(*=============================================================================
  ApoDSReg.PAS
  TApolloTable Registration Code & Property Editors

  Copyright 1999-2002 Vista Software
  Last Updated: 12-22-2002

  - Remember to "touch" this file if you change any conditional defines.
==============================================================================*)

unit Apodsreg;

interface

{MED: 03-10-2001, commented- no need anymore.  Everything is C/S now!}

{$R *.RES}

uses
  TypInfo, Classes, SysUtils, Dialogs, Forms
{$ifdef VER150}  {AC: 08-01-01, Added for D7 support}
  , DesignIntf
  , DesignEditors,
{$endif}

{$ifdef VER140}  {AC: 08-01-01, Added for D6 support}
  , DesignIntf
  , DesignEditors,
{$endif}

{$ifdef VER130}
  , DsgnIntf,
{$endif}

{$ifdef VER125}
  , DsgnIntf,
{$endif}

{$ifdef VER120}
  , DsgnIntf,
{$endif}


  SDE60, Apollo, //\\ [ac] Fts32,
  ApoDSet, ApoEnv, ApCommon, ApConn, ApoQSet,
  ApoServerDLL,           // [MED:] 03-29-2001 changed from ApoStor.pas
  IntSQL
  {$ifdef Apollo70}
  , ScriptBrowse
  , ApoServerScript
  {$endif}
  ;

type
	TDBStringProperty = class(TStringProperty)
	public
		function GetAttributes: TPropertyAttributes; override;
		procedure GetValueList(List: TStrings); virtual; abstract;
		procedure GetValues(Proc: TGetStrProc); override;
	end;

  TMasterFieldsProperty = class(TDBStringProperty)
  public
    procedure GetValueList(List: TStrings); override;
  end;

  TIndexNameProperty = class(TDBStringProperty)
  public
    procedure GetValueList(List: TStrings); override;
  end;

  TTableNameProperty = class(TDBStringProperty)
  public
    procedure GetValueList(List: TStrings); override;
  end;

  TDatabaseNameProperty = class(TDBStringProperty)
  public
    procedure GetValueList(List: TStrings); override;
  end;

{$ifdef Apollo70}

  TServerScriptDatabaseNameProperty = class(TDBStringProperty)
  public
    procedure GetValueList(List: TStrings); override;
  end;

{$endif}

{
  TProcNameProperty = class(TDBStringProperty)
  public
    procedure GetValueList(List: TStrings); override;
  end;
}
  TServerDLLProcNameProperty = class(TDBStringProperty)
  public
    procedure GetValueList(List: TStrings); override;
  end;

  TServerDLLProcDatabaseNameProperty = class(TDBStringProperty)
  public
    procedure GetValueList(List: TStrings); override;
  end;

  TQueryTableNameProperty = class(TDBStringProperty)
  public
    procedure GetValueList(List: TStrings); override;
  end;

  TQueryDatabaseNameProperty = class(TDBStringProperty)
  public

    procedure GetValueList(List: TStrings); override;

  end;
  
  procedure Register;

implementation

(*=============================================================================
  TDBStringProperty
  This is stolen from the TTables implementation of the IndexNames stuff
 =============================================================================*)
(*=============================================================================
  GetAttributes
 =============================================================================*)
function TDBStringProperty.GetAttributes: TPropertyAttributes;
begin
	Result := [paMultiSelect, paValueList, paSortList];
end;

(*=============================================================================
  GetValues
 =============================================================================*)
procedure TDBStringProperty.GetValues(Proc: TGetStrProc);
var
	i: Integer;
	Values: TStringList;
begin
	Values := TStringList.Create;
	try
		GetValueList( Values );
		for i := 0 to Values.Count - 1 do
    	Proc( Values[i] );
	finally
		Values.Free;
	end;
end;

(*=============================================================================
  Retrieves Fields Names From Master Table
 =============================================================================*)
procedure TMasterFieldsProperty.GetValueList(List: TStrings);
var
	bCloseIt: Boolean;
	iFlds : Integer;
  ds: TApolloTable;
begin
	ds := GetComponent(0) as TApolloTable;
  List.Clear;

  if Assigned( ds.MasterSource ) then
  begin
    bCloseIt := False;

    if not ds.MasterSource.DataSet.Active then
    begin
      bCloseIt := True;
      ds.MasterSource.DataSet.DisableControls;
      ds.MasterSource.DataSet.Open;
    end
    else
      sx_Select( ds.WorkArea );

    for iFlds := 0 to ds.MasterSource.DataSet.FieldCount-1 do
      List.Add( ds.MasterSource.DataSet.Fields[iFlds].FieldName );

    if bCloseIt then
    begin
      ds.MasterSource.DataSet.Close;
      ds.MasterSource.DataSet.EnableControls;
    end;
  end;
end;

(*=============================================================================
  Retrieves index tags for the compound index.
 =============================================================================*)
procedure TIndexNameProperty.GetValueList(List: TStrings);
var
  ds: TApolloTable;
  bCloseIt: Boolean;
  iCurOrderNum: Integer;
  sCurIndexString: String;
  sPath, sTblName: String;
	SearchRec : TSearchRec;
  iMore: Integer;
  msPtr : Pointer;
begin
	ds := GetComponent(0) as TApolloTable;
  List.Clear;

  if ds.AccessMethod = amLocal then
  begin
    sTblName := ds.TableName;

    // Make sure a table exists
    if (sTblName > '') then
    begin
      if (ExtractFilePath( sTblName ) = '') then
      begin
  		 	if Assigned( ds.Database ) then
    		  sPath := ds.Database.DataPath
		  	else
  	 	  	sPath := ds.DatabaseName;

  		  // Look in current directy if none was specified
	  		if (sPath = '') then
	    		GetDir( 0, sPath );

  			// Append backslash to end of path, if not there, for directory search
	  		if (sPath[ Length(sPath) ] <> '\') then
		  		sPath := sPath + '\';
  			sTblName := sPath + sTblName;
	  	end;

  		// Append *.DBF to end of file, if no extension there
	  	if (ExtractFileExt( sTblName ) = '') then
		  	sTblName := sTblName + '.dbf';

  		if (FindFirst( sTblName, faAnyFile, SearchRec ) <> 0) then
      	Exit;
  	end;

    bCloseIt := False;

    if not ds.Active then
    begin
      bCloseIt := True;
      ds.DisableControls;
      ds.Open;
    end
    else
      sx_Select( ds.WorkArea );

  	if (ds.TableType = ttSXNTX) then
    begin
		  iMore := FindFirst( sPath + '*.ntx', faAnyFile, SearchRec );
  	  while (iMore = 0) do
    	begin
      	List.Add( ExtractFileName( SearchRec.Name ));
			  iMore := FindNext( SearchRec );
    	end;
	  	FindClose(SearchRec);
    end
    else
    begin
	    iCurOrderNum := 1;
  	  sCurIndexString := sx_TagName( iCurOrderNum );

	    while (sCurIndexString > '') do
    	begin
  	    List.Add( sCurIndexString );
    	  Inc( iCurOrderNum );
      	sCurIndexString := sx_TagName( iCurOrderNum );
  	  end;
	  end;

    if bCloseIt then
    begin
  	  ds.Close;
      ds.EnableControls;
    end;
  end
  else
  begin
    if Assigned(ds.ApolloConnection) then
      if ds.ApolloConnection.Connected then
      begin
      	ds.ApolloConnection.DataBaseName := ds.DatabaseName;
        msPtr := ds.ApolloConnection.GetIndexNames( ds.TableName,
                                                    SmallInt( ds.TableType )+1);
        if msPtr <> nil then
          List.LoadFromStream(TMemoryStream(msPtr) );
      end;
  end;
end;

(*=============================================================================
  Retrieves table names for the specified path or alias
 =============================================================================*)
procedure TTableNameProperty.GetValueList(List: TStrings);
var
	iMore : Integer;
  ds: TApolloTable;
  sPath : String;
	SearchRec : TSearchRec;
  msPtr : Pointer;
begin

	ds := GetComponent(0) as TApolloTable;
  List.Clear;

  if ds.AccessMethod = amLocal then
  begin
   	if Assigned( ds.Database ) then
	    sPath := ds.Database.DataPath
  	else
     	sPath := ds.DatabaseName;

    // Look in current directy if none was specified
	  if (sPath = '') then
  	  GetDir( 0, sPath );

  	// Append backslash to end of path, if not there, for directory search
	  if (sPath[ Length(sPath) ] <> '\') then
		  sPath := sPath + '\';

  	iMore := FindFirst( sPath + '*.dbf', faAnyFile, SearchRec );
    while (iMore = 0) do
    begin
      List.Add( SearchRec.Name );
  		iMore := FindNext( SearchRec );
    end;
  	FindClose(SearchRec);
  end
  else
  begin
    if Assigned(ds.ApolloConnection) then
      if (ds.ApolloConnection.Connected) and (ds.DataBaseName <> '') then
      begin
      	ds.ApolloConnection.DataBaseName := ds.DatabaseName;
        msPtr := ds.ApolloConnection.GetTableNames;
        if msPtr <> nil then
          List.LoadFromStream(TMemoryStream(msPtr) );
     end;
  end;
end;


(*=============================================================================
  Retrieves DatabaseNames (aliases) for all TApolloDatabase components (if any)
 =============================================================================*)
procedure TDatabaseNameProperty.GetValueList(List: TStrings);
var
  ds: TApolloTable;
	i: Integer;
  msPtr: Pointer;
begin
  List.Clear;

	ds := GetComponent(0) as TApolloTable;

  if ds.AccessMethod = amLocal then
  begin
    // use new internal list of databases for the dropdown list in the Obj. Inspector
    {$ifdef DELPHI5_OR_LATER}
    for i:=0 to apDatabaseList.Count - 1 do
      if List.IndexOf( TApolloDatabase(apDatabaseList.Objects[i]).DataPathAlias ) = - 1 then
        List.Add( TApolloDatabase(apDatabaseList.Objects[i]).DataPathAlias ) ;
    {$else}
    for i:=0 to apDatabaseList.Count - 1 do
      if List.IndexOf( TApolloDatabase(apDatabaseList.items[i]).DataPathAlias ) = - 1 then
        List.Add( TApolloDatabase(apDatabaseList.items[i]).DataPathAlias ) ;
    {$endif}
  end
  else
  begin
		if Assigned(ds.ApolloConnection) then
			if ds.ApolloConnection.Connected then
      begin
				msPtr := ds.ApolloConnection.GetDataBaseNames;
        if msPtr <> nil then
          List.LoadFromStream(TMemoryStream(msPtr));
			end;
  end;

end;





{
/////////////////////
procedure TProcNameProperty.GetValueList(List: TStrings);
var
  ds: TApolloConnection;
  msPtr : Pointer;
begin
  List.Clear;
  ds := GetComponent(0) as TApolloConnection;
  if Assigned(ds) then
    if ds.Connected then
    begin
      ds.DataBaseName := '';
      msPtr := ds.GetProcedureNames;
      if msPtr <> nil then
        List.LoadFromStream(TMemoryStream(msPtr));
    end;
end;
}

// for TApolloStoredProc  - We use ApolloConnection property
procedure TServerDLLProcNameProperty.GetValueList(List: TStrings);
var
  ds: TApolloServerDLL;
  msPtr : Pointer;
begin
  List.Clear;
  ds := GetComponent(0) as TApolloServerDLL;
  if Assigned(ds.ApolloConnection) then
    if ds.ApolloConnection.Connected then
    begin
      //ds.ApolloConnection.DataBaseName := ds.DataBaseName;
      msPtr := ds.ApolloConnection.GetDLLProcedureNames(ds.DataBaseName);
      if msPtr <> nil then
        List.LoadFromStream(TMemoryStream(msPtr));
    end;
end;

(*=============================================================================
  Retrieves DatabaseNames (aliases) for all TApolloDatabase components (if any)
 =============================================================================*)
procedure TServerDLLProcDatabaseNameProperty.GetValueList(List: TStrings);
var
  ds: TApolloServerDLL;
  msPtr : Pointer;
begin
  List.Clear;
	ds := TApolloServerDLL(GetComponent(0));
  begin
		if Assigned(ds.ApolloConnection) then
			if ds.ApolloConnection.Connected then
      begin
				msPtr := ds.ApolloConnection.GetDataBaseNames;
				List.LoadFromStream(TMemoryStream(msPtr) );
			end;
  end;
end;


{$ifdef Apollo70}
(*=============================================================================
  Retrieves DatabaseNames (aliases) for all TApolloDatabase components (if any)
 =============================================================================*)
procedure TServerScriptDatabaseNameProperty.GetValueList(List: TStrings);
var
  ds: TApolloServerScript;
  msPtr : Pointer;
begin
  List.Clear;
	ds := TApolloServerScript(GetComponent(0));
  begin
		if Assigned(ds.ApolloConnection) then
			if ds.ApolloConnection.Connected then
      begin
				msPtr := ds.ApolloConnection.GetDataBaseNames;
				List.LoadFromStream(TMemoryStream(msPtr) );
			end;
  end;
end;
{$endif}  // <--- Apollo70

(*=============================================================================
  Retrieves table names for the specified path or alias
 =============================================================================*)
procedure TQueryTableNameProperty.GetValueList(List: TStrings);
var
  msPtr     : Pointer;
  mm        : TApolloTableItem;
  tt        : TApolloTables;
  ds        : TApolloQuery;
  sPath     : String;     {LAS: 05-05-00, Added}
  iMore     : Integer;    {LAS: 05-05-00, Added}
	SearchRec : TSearchRec; {LAS: 05-05-00, Added}
begin
  List.Clear;
  mm := GetComponent(0) as TApolloTableItem;
  tt := TApolloTables(mm.Collection);
  ds := TApolloQuery(tt.GetOwner);
  if ds = nil then
  	exit;

	//-----------------------------------------------( LAS: 05-05-00, Added below)
  if ds.AccessMethod = amLocal then
  begin
   	if Assigned( ds.Database ) then
	    sPath := ds.Database.DataPath
  	else
     	sPath := ds.DatabaseName;

    // Look in current directy if none was specified
	  if (sPath = '') then
  	  GetDir( 0, sPath );

  	// Append backslash to end of path, if not there, for directory search
	  if (sPath[ Length(sPath) ] <> '\') then
		  sPath := sPath + '\';

  	iMore := FindFirst( sPath + '*.dbf', faAnyFile, SearchRec );
    while (iMore = 0) do
    begin
      List.Add( SearchRec.Name );
  		iMore := FindNext( SearchRec );
    end;
  	FindClose(SearchRec);
  end
  else //------------------------------------------( LAS: 05-05-00, Added above)
  begin
    if Assigned(ds.ApolloConnection) then
      if (ds.ApolloConnection.Connected) and (ds.DataBaseName <> '') then
      begin
        ds.ApolloConnection.DataBaseName := ds.DatabaseName;
        msPtr := ds.ApolloConnection.GetTableNames;
        List.LoadFromStream(TMemoryStream(msPtr) );
      end;
  end;
end;

(*=============================================================================
  Retrieves DatabaseNames (aliases) for all TApolloDatabase components (if any)
 =============================================================================*)
procedure TQueryDatabaseNameProperty.GetValueList(List: TStrings);
var
  ds: TApolloQuery;
  msPtr : Pointer;
	i : Integer;     {LAS: 05-05-00, Added}

begin
  List.Clear;
	ds := TApolloQuery(GetComponent(0));

	//-----------------------------------------------( LAS: 05-05-00, Added below)
  if ds.AccessMethod = amLocal then
  begin
    // use new internal list of databases for the dropdown list in the Obj. Inspector
    {$ifdef DELPHI5_OR_LATER}
    for i:=0 to apDatabaseList.Count - 1 do
      if List.IndexOf( TApolloDatabase(apDatabaseList.Objects[i]).DataPathAlias ) = - 1 then
        List.Add( TApolloDatabase(apDatabaseList.Objects[i]).DataPathAlias ) ;
    {$else}
    for i:=0 to apDatabaseList.Count - 1 do
      if List.IndexOf( TApolloDatabase(apDatabaseList.items[i]).DataPathAlias ) = - 1 then
        List.Add( TApolloDatabase(apDatabaseList.items[i]).DataPathAlias ) ;
    {$endif}
  end
  else 	//-----------------------------------------( LAS: 05-05-00, Added above)
  begin
		if Assigned(ds.ApolloConnection) then
			if ds.ApolloConnection.Connected then
      begin
				msPtr := ds.ApolloConnection.GetDataBaseNames;
				List.LoadFromStream(TMemoryStream(msPtr) );
			end;
  end;
end;


procedure Register;
begin
  RegisterComponents( 'Apollo',
                      [
                       TApolloTable,
                       TApolloQuery,
                       TApolloConnection,
                       TApolloEnv,
                       TApolloDatabase,
                       TApolloServerDLL,   // [MED:] 03-29-2001 changed from TApolloStoredProc
                       {$ifdef Apollo70}
                       TApolloServerScript,
                       {$endif}
                       //\\ [ac] TFts,
                       TApolloDataSet,
                       TApollo
                       ] );

	RegisterPropertyEditor( TypeInfo(String), TApolloTable, 'MasterFields', TMasterFieldsProperty);
	RegisterPropertyEditor( TypeInfo(String), TApolloTable, 'IndexName',    TIndexNameProperty);
	RegisterPropertyEditor( TypeInfo(String), TApolloTable, 'TableName',    TTableNameProperty);
	RegisterPropertyEditor( TypeInfo(String), TApolloTable, 'DatabaseName', TDatabaseNameProperty);
{$ifdef Apollo70}
  ////////////////////////////////////////// 11-14-00, added
  // ApolloServercript
  RegisterPropertyEditor( TypeInfo(String), TApolloServerScript, 'DataBaseName',  TServerScriptDataBaseNameProperty);
  RegisterPropertyEditor(TypeInfo(TStringList), TApolloServerScript, 'Scripts', TFuncEditProperty);
  RegisterComponentEditor(TApolloServerScript, TApolloServerScriptEditProperty);

{$endif}

  RegisterPropertyEditor( TypeInfo(String), TApolloTableItem, 'TableName',  TQueryTableNameProperty);
  RegisterPropertyEditor( TypeInfo(String), TApolloQuery, 'DataBaseName',   TQueryDataBaseNameProperty);

  // ApolloServerDLL
  RegisterPropertyEditor( TypeInfo(String), TApolloServerDLL, 'DataBaseName',   TServerDLLProcDataBaseNameProperty);
	RegisterPropertyEditor( TypeInfo(String), TApolloServerDLL, 'ProcName',       TServerDLLProcNameProperty);

end;

initialization

	LoadDLL;  // Always need local .DLLs loaded when in design-mode

	//  apDatabaseList := TStringList.Create; - Removed - RH d.d. 17-12-2002
  // Creation and destruction of apDatabaseList is now handled in ApoEnv.pas - RH d.d. 17-12-2002

finalization

	//  apDatabaseList.Free; - Removed - RH d.d. 17-12-2002
  // Creation and destruction of apDatabaseList is now handled in ApoEnv.pas - RH d.d. 17-12-2002

end.




