(*============================================================================
  ApConn.PAS
  Main TApolloConnection code                          

  Copyright 1999-2002 Vista Software
  Last Updated: 12-15-2002

  NOTE: Tab stops were set at 2 spaces during development of this unit. If you
        are using a tab stop setting other than 2 spaces, the code indents
        may not line up accurately in your editor.

  -10-05-99  TPS methods have been moved to ApTPS.PAS file
  see {$INCLUDE ApTPS.PAS}

  -12-30-99 Added OnServerProgress event feature.  See FillReadStream

  -04-03-01, added compression/encryption to Streams.  Only Stream data
  are affected by this.  The packet header is not crypted.

  -04-04-01, Login parameters are scrambled now.
============================================================================*)

unit ApConn;


interface

{.  $define LICENSE_KEY}      //MED: 03-22-01, added to publish LicenseKey

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, APWin, ExtCtrls,
  db,       // used for TParams
  ApCommon,                         // contains our enumerated types
  IntSQL,
  FileCtrl,
  IntCrypt

 {$ifdef VER150}  {MED: 09-20-02, Added for D7 support}
  , Variants
 {$endif}

 {$ifdef VER140}  {LAS: 05-29-00, Added for D6 support}
  , Variants
 {$endif}

  ;

{$Include ApConst.PAS }


const
  MAXRECLEN    = ((1024)*64)+1;


Type

  TSocketConnectionLoginEvent = procedure(Sender : TObject; LoginAccepted:Boolean; Msg:String) of object;
  TSocketConnectionLogoutEvent= procedure(Sender : TObject) of object;
  TSocketConnectionErrorEvent = procedure(Sender : TObject; Command, Msg: String) of object;
  TSocketConnectionServerMsgEvent = procedure(Sender : TObject; Msg: String) of object;
  TSocketConnectionServerStreamEvent = procedure(Sender : TObject; Stream: TMemoryStream) of object;

  TApolloConnection = class(TApSDEClient)
  private
    FCompression        : Boolean;         // MED: 04-03-01, added compression
    FCompressionPassword: String;
    FCompressionLevel   : Integer;        // MED: 04-05-01, not functional yet.  Set at Level 1 for now

    FTimer        : TTimer;         {MED: 03-17-00, added }
    FTriggersInterval : Integer;    {MED: 01-27-01, added }

    FSQLDataPath  : String;
    FTransList    : TList;      {A list of TApolloTables and current status}
    FTransStream  : TMemoryStream;
    nCurWorkArea  : Word;
    FActiveBuffer : PChar;
    FLogged       : Boolean;
    FVersion      : String;
    FDummyVersion : String;
    FSerial       : String;
    FTimeOut      : LongInt;
    FLicenseKey   : String;       //MED: 03-22-01, added to publish LicenseKey

    FDataBaseName   : String;       {MED: 12-11-99 Added, Public property used quietly among other components}
    FOnLogin        : TSocketConnectionLoginEvent;
    FOnError        : TSocketConnectionErrorEvent;
    FOnLogout       : TSocketConnectionLogoutEvent;
    FOnServerMessage: TSocketConnectionServerMsgEvent;
    FOnServerStream : TSocketConnectionServerStreamEvent;  // triggers only

    FServerName     : String;
    FGroupName      : String;
    FConnType       : Integer;

    FReadStream, FWriteStream : TMemoryStream;
    InBuff        : Array [0..MAXRECLEN] of Char;

    function FillInBuff:PChar;

    procedure SetActive(Value: Boolean); virtual;
    procedure ServiceIncoming(Sender:TObject);    {MED: 03-17-00, added}

    function   GetTriggers:Boolean;
    procedure  SetTriggers(Value:Boolean);
    procedure  SetTriggersInterval(Value:Integer);
    procedure  SetCompressionPassword(Value:String);        // 255 chars

  protected
    { Protected declarations }
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
		function  PathFromAlias: String;   {LAS: 05-09-00, Added}
  public
    { Public declarations }

    // MED: 04-03-2001
    property ConnType : Integer read FConnType write FConnType; // MED: 05-18-2002 one byte-client or server connecting to Relay
    property Compression        : Boolean read FCompression write fCompression;         // MED: 04-03-01, added compression
    property CompressionPassword: String read FCompressionPassword write SetCompressionPassword;

    property TimeOut : Integer read FTimeOut write FTimeOut;  {MED: 03-25-2002, added}

    property SQLDataPath: String read FSQLDataPath write FSQLDataPath; {LAS: 05-09-00, Added}
    property Triggers : Boolean read GetTriggers write SetTriggers;
    property TriggersInterval : Integer read FTriggersInterval write SetTriggersInterval;

    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    function  ap_SetDLLProcName(sDatabaseName, sProcName:String):Pointer;
		procedure CloseDataSets;  {LAS: 10-17-99, Added}
		procedure OpenDataSets;  {LAS: 10-17-99, Added}
    function StartTransaction:integer;   {MED: 03-27-00, Added}
    function Commit:integer;             {MED: 03-27-00, Added}
    function RollBack:integer;           {MED: 03-27-00, Added}


    property ReadStream  : TMemoryStream read FReadStream write FReadStream;
    property WriteStream : TMemoryStream read FWriteStream write FWriteStream;
    property TransStream : TMemoryStream read FTransStream write FTransStream;

    property  ActiveBuffer : PChar read FActiveBuffer;
	  function  Login:Boolean;
	  procedure Logout;

    function  IsServerConnectedToRelay:Boolean;
    function  GetServerListFromRelay:Pointer;     // Relay only

    procedure ClientError(const Command, ErrMsg: String);
		// Connected does not work properly if disconnected by the server for any reason.
    // this is an absolute verification if we still have a connection
    // Note: Connected is faster
    function Ping: Boolean;     {LAS: 11-01-99, Changed from Procedure}

    function  ClientXmit(const scAction:String; sPacket:String; bWaitForResponse:Boolean):PChar;

    //MED: 09-25-2002, made all functions wait mode
    //function  ClientStreamSend(const scAction:String): LongInt;                       // sends a stream

    function  ClientStreamReceive(const scAction:String; sPacket: String ): Pointer;  // sends a packet but receives a stream
    function  FillReadStream(bTriggers:Boolean): Integer;

    {MED: 03-18-00, added to send AND receive streams}
    function  ClientStreamSendReceive(const scAction:String ): TMemoryStream;

    // function GetStationCount:Integer;        // gets # of clients connected to the server
    // function GetTableCount(sTable:String):Integer;           // gets # of clients that have the table open

    function GetConnectionID:String;
	  function GetServerName:String;    //MED 05-18-2002, changed to GetServerName
    // function CopyFile(sSrcDatabaseName, sDesDatabaseName, sFileName : String): Boolean; {MED: 03-23-01}
	  function DeleteFile(sDataBaseName, sFileName: String ): Boolean;
	  function DeleteFileExt(sDataBaseName, sFileName: String ): Boolean;    
	  function FileExists(sDataBaseName, sFileName: String ): Boolean;
	  function FileLength(sDataBaseName,sFileName: String ): Integer;
	  function FileIsInUse(sDataBaseName,sFileName: String ): Boolean;  {MED: 09-30-00, added-tells us if the file is opened by any other process}
	  function FileSend(sDataBaseName,  srcFile, desFile: String): Integer;
	  function FileReceive(sDataBaseName, srcFile, desFile: String): Integer;
    function GetServerDateTime:TDateTime;

	  function  GetTableNames: Pointer;
		function  GetTableNamesList(sDatabaseName: String): TStringList;  {LAS: 05-22-00, Re-added}
	  function  GetDataBaseNames: Pointer;
  	function  GetDatabaseNamesList: TStringList;  {LAS: 05-22-00, Readded}

	  function  GetIndexNames(sTableName:String; iTableType:SmallInt): Pointer;
	  function  GetDLLProcedureNames(sDatabaseName:String): Pointer;

    {MED: 09-08-00, added for 6.0 scripting}
//{$ifdef APOLLO60}
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////
    function  CreateScript(ScriptName:String; ScriptText:TStringList; iLang:Integer; bSave:Boolean):Pointer;  // creates a new script and adds it to the list
    function  StopScript(ScriptName:String):Pointer;        // currently not used
    function  GetScriptsList: Pointer;                      // gets a TStringList of all scripts on the server
    function  DropScript(ScriptName:String): Pointer;       // Deletes it in SCRIPTS.DBF
    function  RunScript(ScriptName:String; InParams:array of Variant):Pointer;
    function  CheckScriptSyntax(ScriptName:String; var CharNo:integer; var LineNo): Pointer;
    function  UpdateScriptText(ScriptName:String; ScriptText:TStrings; bSave:Boolean ):Boolean;   // saves script text
    function  GetScriptText(ScriptName:String):Pointer;     // load script text
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////
//{$endif}

		function  GetDLLProcs( sDLLName:String): Pointer;  {LAS: 01-16-99, Added}
	  function  GetFiles(sSrchStr:String): Pointer;  {MED: 01-16-2000, addeed}
		procedure GetUserNames(uList: TStringList);
    function  GetUserNamesList: TStringList;      {LAS: 05-29-2000, Re-added}
    function  FindDrives:Pointer;
    function  FindDirectories(const ParentDirectory:String):Pointer;
    Function  FindFiles(Directory, Mask:String): Pointer;// not as same as GetFiles

    Function  GetServerLog:Pointer;     //MED: 05-26-02, added

    function ap_ExecDLLProc: Pointer;         // MED: 04-02-2001 changed to TMemoryStream

    property DataBaseName : String read FDataBaseName write FDataBaseName;

    // with a few exceptions the following mimic sx_* methods
    function ap_Alias(iWA:Word): PChar;
    function ap_Append:Integer;                   // 10-11-00, changed to return new record number
    function ap_AppendBlank:Integer;              // 10-11-00, changed to return new record number
    function ap_AppendFrom(cpFileName: PChar; iSourceType: SmallInt;
                            cpScopeExpr: PChar): WordBool;
    function ap_BaseDate: PChar;
    function ap_BaseName: PChar;
    function ap_BlobToFile ( cpFieldName: PChar; cpFileName: PChar;
                             iRecNo: Integer ): WordBool;  {LAS: 06-26-00, Added iRecNo param}
    function ap_Bof: WordBool;
    procedure ap_Close;
    procedure ap_CloseAll;
    procedure ap_CloseIndexes;
    procedure ap_Commit;

    function ap_CommitLevel(iVal:Integer):Integer;

    function ap_CopyFile( sToFileName: PChar ): WordBool;
    function  ap_CopyFileText( sTextFileName:PChar;
                              iFileType : smallInt ) : WordBool;
    function ap_CopyStructure( sFileName : PChar; sAlias:PChar  ) : WordBool;
    function ap_CopyStructureExtended(cpFileName: PChar): WordBool;
    function ap_Count: LongInt;
    function ap_CreateExec: WordBool;
    procedure ap_CreateField(cpName: PChar; cpType: PChar;
                              iLength: SmallInt; iDecimals: SmallInt);
    function ap_CreateFrom(cpFileName: PChar; cpAlias: PChar; cpStruFile: PChar;
                            iRDEType: SmallInt): WordBool;

    function ap_CreateNew(cpDataBaseName, cpFileName: PChar; iRdeType: SmallInt;
                           iNumFields: SmallInt): SmallInt;

    function  ap_DbfDecrypt: WordBool;
    function  ap_DbfEncrypt: WordBool ;
    function  ap_DBFilter: PChar ;
    function  ap_Decrypt( cpBuffer, cpPassword : PChar; iLen :Integer ): PChar;
    procedure ap_Delete;
    function  ap_Deleted:  WordBool ;
    function  ap_Descend(cpKeyString: PChar): PChar ;
    function  ap_Empty( sFieldName: PChar ):  WordBool;
    function  ap_Eof: WordBool ;
    function  ap_Encrypt( cpBuffer, cpPassword : PChar; iLen : Integer ) : PChar;
    function  ap_ErrorLevel(iErrorLevel: Integer): Integer ;
    function  ap_EvalLogical(sExpression: PChar ): WordBool;
    function  ap_EvalNumeric( sExpression: PChar ): Double;
    function  ap_EvalString(sExpression: PChar): PChar ;
    function  ap_EvalTest(sExpression: PChar): SmallInt ;
    function  ap_FieldCount: Word ;
    function  ap_FieldDecimals(cpFieldName: PChar): Word ;
    function  ap_FieldName(uiFieldNum: Word): PChar ;
    function  ap_FieldNum(cpFieldName: PChar): Word ;
    function  ap_FieldOffset(cpFieldName: PChar): Word ;
    function  ap_FieldType(cpFieldName: PChar): PChar ;
    function  ap_FieldWidth(cpFieldName: PChar): Word ;
    function  ap_FilterAlias(cpAliasName: PChar;
                             cpFieldName: PChar): WordBool ;
    function  ap_FilterDlg(hwnd: HWND; cpExpr: PChar; cpCaption: PChar;
                           bHasIndexList: WordBool): PChar;
    function  ap_Flock: WordBool ;
    function  ap_Found: WordBool ;
    function  ap_GetBitMap(cpFieldName: PChar; hwnd: HWND): WordBool ;
    function  ap_GetBlob(cpFieldName: PChar; vpVar: Pointer): LongInt;
    function  ap_GetBlobLength(cpFieldName: PChar): LongInt ;
    function  ap_GetByte(cpFieldName: PChar): PChar ;
    function  ap_GetDateJulian(cpFieldName: PChar): LongInt ;
    function  ap_GetDateString(cpFieldName: PChar): PChar ;
    function  ap_GetDouble(cpFieldName: PChar): Double ;
    function  ap_GetFields:PChar;                  {MED: added to support Fields}
    function  ap_GetInteger(cpFieldName: PChar): SmallInt;
    function  ap_GetLogical(cpFieldName: PChar): WordBool;
    function  ap_GetLong(cpFieldName: PChar): LongInt;
    function  ap_GetMemo( cpFieldName: PChar; uiLineWidth: Word): PChar ;  {LAS: 07-16-00, Added iRecNo param}
    function  ap_GetPage(Count:Integer): PChar;   {MED: 08-11-99, added to support Fields}
    function  ap_GetQueryBit( lRecNum : LongInt ) : WordBool;
    procedure ap_GetRecord(cpRecord: PChar; iRecSize:integer) ;
    function  ap_GetScope(iWhichScope: SmallInt): PChar;
    function  ap_GetString(cpFieldName: PChar): PChar;
    function  ap_GetStructure:Pointer;
    function  ap_GetTrimString(cpFieldName: PChar): PChar ;  {LAS: 06-09-00, Added iRecNo param} {MED:08-29-00, changed back}
    function  ap_Go(lRecNum: Integer): Integer;  {LAS: 06-01-00, Changed to function}
    procedure ap_GoBottom;
    procedure ap_GoTop;
    function  ap_Index(cpFileName: PChar; cpExpr: PChar; iOption: SmallInt;
                       bDescend: WordBool; cpCondition: PChar): SmallInt;
    procedure ap_IndexClose;
    function  ap_IndexCount: Integer;  {LAS: 12-19-99, Added}
    function  ap_IndexCondition: PChar ;
    function  ap_IndexDescending: Integer;   {LAS: 12-19-99, Added}
    function  ap_IndexFlip: WordBool;
    function  ap_IndexKey: PChar ;
    function  ap_IndexKeyField: PChar ;
    function  ap_IndexName(iIndex: SmallInt): PChar ;
    function  ap_IndexOpen(cpFileName: PChar): SmallInt ;
    function  ap_IndexOrd: SmallInt;
    function  ap_IndexTag(cpFileName: PChar; cpTagName: PChar; cpExpr: PChar;
                          iOption: SmallInt; bDescend: WordBool;
                          cpCondition: PChar): SmallInt;
    function  ap_IndexType: SmallInt ;
    function  ap_IsEncrypted(iFileOrRec: SmallInt): WordBool ;
    function  ap_KeyAdd(cpTagname: PChar): WordBool ;
    function  ap_KeyData: PChar ;
    function  ap_KeyDrop(cpTagname: PChar): WordBool ;
    function  ap_Locate (cpExpression: PChar; iDirection: WordBool;
                        bContinue: WordBool): LongInt ;
    function  ap_LockCount: Word ;
    function  ap_Locked(lRecNum: LongInt): WordBool ;
    procedure ap_DBRlockList( ulpArray: PChar );
    function  ap_MemAlloc( lNumberOfBytes: Integer ): PChar ;  //MED: 04-13-01, added
    procedure ap_MemDealloc(vpPtr: Pointer) ;
    function  ap_OpenMode: SmallInt;
    function  ap_OrderPosGet: Double ;
    procedure ap_OrderPosSet( dPosition: Double ) ;
    function  ap_OrderRecNo: LongInt ;
    procedure ap_Pack;
    function  ap_PutBlob(cpFieldName: PChar; vpVar: Pointer;
                          lSize: LongInt): LongInt;
    procedure ap_PutRecord(cpRecord: PChar; iRecSize:Integer);
    function  ap_PutMemo(cpFieldName: PChar; vpData:PChar):Integer;
    function  ap_Query(cpExpression: PChar): LongInt ;
    function  ap_QueryRowCount(cpExpression: PChar; var RowCount:integer): LongInt;    
    function  ap_QueryRecCount: LongInt;
    function  ap_QuerySetExact(bSetQueryExact: LongBool): LongBool ;
    function  ap_QueryTest (cpExpression: PChar): SmallInt ;

    procedure ap_Recall;
    function  ap_RecCount: LongInt ;
    function  ap_RecNo: LongInt ;
    function  ap_RecSize: LongInt ;
    procedure ap_Reindex;
    procedure ap_Replace(cpFieldname: PChar; iDataType: SmallInt;
                          vpData: Pointer) ;
    function ap_Rlock(lRecNum: LongInt): WordBool;
    function ap_RYOFilterActivate(iFilterHandle: SmallInt;
                                   iBoolOperation: SmallInt): WordBool ;
    function  ap_RYOFilterCopy: SmallInt ;
    function  ap_RYOFilterCreate: SmallInt;
    function  ap_RYOFilterDestroy(iFilterHandle: SmallInt): WordBool ;
    function  ap_RYOFilterGetBit(iFilterHandle: SmallInt;
                                 lRecNo: LongInt): WordBool ;
    function  ap_RYOFilterRestore(cpFileName: PChar): WordBool;
    function  ap_RYOFilterSave(iFilterHandle: SmallInt;
                               cpFileName: PChar): WordBool ;
    function  ap_RYOFilterSetBit(iFilterHandle: SmallInt; lRecNo: LongInt;
                                 iOnOrOff: SmallInt): WordBool;
    function  ap_RYOKeyAdd(cpTagname: PChar;
                           cpKey: PChar): WordBool ;
    function  ap_RYOKeyDrop(cpTagname: PChar): WordBool ;
    function  ap_Seek(cpKeyValue: PChar): WordBool;
    function  ap_SeekBin(cpKeyValue: PChar; uiLength: Word): WordBool ;
    function  ap_Select(uiBaseArea: Integer): Integer ;
    procedure ap_SetCentury(uiOnOff: WordBool) ;  {LAS: 11-14-99, Changed param type}
    procedure ap_SetDateFormat(uiDateType: Word) ;
    procedure ap_SetDeleted(uiDeleted: WordBool) ;
    function  ap_SetEpoch( uiBaseYear: Word): Word ;
    function  ap_SetErrorFunc( pFunc, pData : LongInt): LongInt;
    procedure ap_SetExact(uiOnOff: WordBool) ;
    procedure ap_SetFields(FFields:PChar);     {MED: added to support Fields}
    procedure ap_SetFilter(cpExpression: PChar) ;
    procedure ap_SetGaugeHook(hwndGauge: HWND) ;
  //  function  ap_SetHandles(iNumHandles: SmallInt): SmallInt ;   {LAS: 11-19-99, Removed}
    procedure ap_SetLockTimeout(iSeconds: SmallInt) ;
    function ap_SetMemoBlockSize(uiBlockSize: Word):Word ;
    function  ap_SetOrder(iIndex: SmallInt): SmallInt ;
    procedure ap_SetPassword(cpEncodeKey: PChar) ;
    procedure ap_SetQueryBit(lRecNo: LongInt; bValue: WordBool) ;
    procedure ap_SetRelation(uiChildArea: Word; cpKeyExpr: PChar) ;
    function  ap_SetScope(cpLowVal: PChar; cpHighVal: PChar): WordBool ;
    procedure ap_SetSoftSeek(uiOnOff: WordBool) ;
    procedure ap_SetStringType(uiStringType: Word);
    procedure ap_SetTranslate(uiOnOff: WordBool) ;
    procedure ap_SetTurboRead(uiOnOff: WordBool) ;
    procedure ap_Skip(lNumRecs: LongInt) ;
    function  ap_SysProp(uiSysItem: Word; vpData: Pointer): LongInt ;
    function  ap_TagArea(cpTagName: PChar): SmallInt ;
    function  ap_TagDelete(cpTagname: PChar): LongBool;
    function  ap_TagName(iTagArea: SmallInt): PChar ;
    procedure ap_Unlock(lRecNum: LongInt) ;
    function  ap_Use(cpDataBaseName, cpFileName: PChar;  iOpenMode: SmallInt;
                     iRdeType: SmallInt): SmallInt;
    function  ap_UseEx(cpDataBaseName, cpFileName: PChar;  iOpenMode: SmallInt;
                     iRdeType: SmallInt; uiMode : Cardinal  ): SmallInt;

    function  ap_QuickUse(
                   sDataBaseName, sFileName: String;
                   iOpenMode: SmallInt;
                   iRdeType: SmallInt;
                   var FRowCount : Integer;
                   var FFieldCount : Integer;
                   var FRecordSize : Integer;
                   var FIsEncrypted: WordBool;
                   var FPassword   : String;
                   FTurboRead      : WordBool;
                   FIndexName      : String;
                   FEpoch, FDateFormat, FMemoBlockSize : Integer;
                   FDeleted, FCentury, FSoftSeek,
                   FExact,
                   FTranslate,            // MED: 04-07-01
                   FApplyEnvSettings : WordBool;
                   FIndexCondition  : String;
                   FCommitLevel : Integer;
                   var FIndexOrd   : Integer;
                   var FIndexDefs : TIndexDefs;
                   var iError: Integer;
                   var sError:String ): SmallInt;


    function  ap_Version: PChar;
    function  ap_WorkArea(cpAlias: PChar ): Word ;
    procedure ap_Zap;

    {MED: 07-31-01, added}
	  function  ap_GetTableType(sDatabaseName, sTableName:String): SmallInt;

    ///////////////////////////////////////////////////
    //MED: 09-10-01, added
    ///////////////////////////////////////////////////
    procedure ap_SetSystemCollation;
    procedure ap_SetMachineCollation;
    procedure ap_AddDudenCollation;
    procedure ap_AddEtecCollation;


    {MED: 04-18-01, added}

		function ap_GoBofRecord:TMemoryStream;
    function ap_GoEofRecord:TMemoryStream;
		function ap_GoTopRecord:TMemoryStream;
    function ap_GoBottomRecord:TMemoryStream;
		function ap_GoRecord(const lRecNum: Integer):TMemoryStream;
    function ap_GoGetRecord:TMemoryStream;
    // 01-29-02 Renamed from GoRowSkip and changed result to Integer;
		function ap_GoSkip( const lRecNum: Integer; const OffSet: Integer):Integer;
    function ap_GoToRow( const RowNo: Integer ):Integer;
    function ap_GetRowNo:Integer;


    {MED: 02-10-00, added}
    procedure ap_EnvSettings(Epoch, DateFormat, MemoBlockSize:Integer;
                             Deleted, Century, SoftSeek, Exact:WordBool);


    // Additional
    function ap_TraverseTable(
                          FirstCrack, LastCrack:WordBool;
                          FEOFAction:Integer;
                          GetMode: TGetMode;
                          var RecNo : Integer;
                          LSDERecordBuffer:PChar;
                          iRecSize:Integer;
                          FetchCount:Integer): TGetResult;

    // Additional
    function ap_MultiTraverseTable(
                          var ReadCount:Integer;
                          FirstCrack, LastCrack:WordBool;
                          FEOFAction:Integer;
                          GetMode, LastGetMode: TGetMode;
                          LSDERecordBuffer:PChar;
                          iRecSize:Integer;
                          FetchCount:Integer;
                          LastRecNo:Integer;
                          var FFetchRecs: TList): TGetResult;

{
    function ap_SDEFetchRecords(
                          AccessMethod : TAccessMethod;
                          iDirection: TSkipDirection;      // 1 or -1 --- Skip(1) or Skip(-1)
                          var FBOF  : Boolean;
                          var FEOF  : Boolean;
                          FRecSize:Integer;
                          var FRecCount :Integer;             // sx_RecCount
                          var FIndexOrd : Integer;             // sx_IndexOrd
                          FetchCount:Integer;
                          LastRecNo :Integer;
                          lNumRecs : Integer;
                          var FFetchRecs:TList;
                          RecIndex:Integer;
                          StackOrder:TStackOrder;    // -1 for all
                          FSetDeleted:Boolean):Integer;

 }
    // Transaction Processing
    function ap_tps_BeginTransaction:WordBool;
    function ap_tps_Commit: WordBool;
    function ap_tps_RollBack: WordBool;

    // SQL
    {MED: 01-09-2000, modified to support ExecSQL, and OpenSQL}
    function  ap_sql_Prepare(
                              AccessMethod : TAccessMethod;
                              sDataBaseName:String;
                              TableNames:TCollection;
                              Century : WordBool;
                              DateFormat: String;
                              Deleted:WordBool;
                              MemoBlockSize:Integer;
                              FParams:TParams;
                              FReadOnly : Boolean;
                              FDefaultTableType : TApolloTableType;
                              FDefaultPassword : String;
                              FDefaultTranslate: Boolean;
                              FCommitLevel : TApolloCommitLevel;                              
                              var iError: Integer;
                              var sError:String ): Integer;  // returns SQLID

    procedure ap_sql_ExecSQL(
                              AccessMethod:TAccessMethod;
                              WASQL: Word;
                              SQL : String;
                              FParams:TParams;              // Med:04-05-01, added
                              var iError:Integer;
                              var sError:String);           //prepare before calling

    procedure ap_sql_OpenSQL(
                              AccessMethod:TAccessMethod;
                              WASQL: Word;                 //prepare before calling
                              SQL : String;
                              var RowCount:Integer ;
                              var FieldCount:Integer ;
                              var RecordSize: Integer;
                              var iError:Integer;
                              var sError:String);

    procedure ap_sql_CloseSQL(AccessMethod:TAccessMethod;WASQL: Word);

    procedure ap_sql_UnPrepare( AccessMethod:TAccessMethod; WASQL: Word );
    function  ap_sql_RecCount(AccessMethod:TAccessMethod; WASQL:Word): LongInt;
    function  ap_sql_GetStructure(AccessMethod:TAccessMethod; WASQL:Word):Pointer;
    function  ap_sql_GetRecord(AccessMethod:TAccessMethod; WASQL:Word; RecNo, FetchCount, RecordCount:Integer):Pointer;
    function  ap_sql_ReadBlobData(AccessMethod:TAccessMethod; WASQL, RecNo, FieldIndex:Integer):TMemoryStream;
    procedure ap_sql_Refresh( AccessMethod:TAccessMethod; WASQL:Word );

    ///////////////////////////////////////////////////
    //MED: 09-08-01, added
    ///////////////////////////////////////////////////
    procedure ap_sql_SetSystemCollation(AccessMethod:TAccessMethod);
    procedure ap_sql_SetMachineCollation(AccessMethod:TAccessMethod);
    procedure ap_sql_AddDudenCollation(AccessMethod:TAccessMethod);
    procedure ap_sql_AddEtecCollation(AccessMethod:TAccessMethod);


  published
    { Published declarations }
    property Active: Boolean read Connected write SetActive;
    property ServerName: String read FServerName write FServername;
    property GroupName : String read FGroupName write FGroupName;
    property Version: String read FVersion write FDummyVersion;
    property OnLogin: TSocketConnectionLoginEvent read FOnLogin write FOnLogin;
    property OnLogout: TSocketConnectionLogoutEvent read FOnLogout write FOnLogout;
    property OnError: TSocketConnectionErrorEvent read FOnError write FOnError;
    property OnServerMessage: TSocketConnectionServerMsgEvent read FOnServerMessage write FOnServerMessage;

    // MED: 01-27-01, added for triggers
    property OnServerStream: TSocketConnectionServerStreamEvent read FOnServerStream write FOnServerStream;


{$ifdef LICENSE_KEY}  // MED: 03-22-01, added
    property LicenseKey : String read FLicenseKey write FLicenseKey;
{$endif}



  end;


implementation

uses
	ApoDSet, ApoQSet, compStream,
  ApoServerDLL,
  ApoEnv                          // MED: 03-29-2001
  {$ifdef Apollo60}
  , ApoServerScript
  {$endif}  
  ;                 // MED: 03-29-2001


function SubStr(const cString:String; nStart,nEnd: Integer): String;
begin
	Result := Copy(cString, nStart, nEnd);
end;

//////////////////////////////////////////////////////////////////
// Padl()
/////////////////////////////////////////////////////////////////
function Padl ( cStr:String; nRepeat:Integer; cChar:Char): String;
begin
  Result := StringOfChar(cChar, nRepeat - Length(cStr) ) + cStr;
end;

///////////////////////////////////////////////////////////////////////
// Padr()
/////////////////////////////////////////////////////////////////
function Padr ( cStr:String; nRepeat:Integer; cChar:Char): String;
begin
  Result := cStr + StringOfChar(cChar, nRepeat - Length(cStr) );
end;


type
  TVolumeInfo = record
     Name               : String;
     SerialNumber       : DWORD;
     MaxComponentLength : DWORD;
     FileSystemFlags    : DWORD;
     FileSystemName     : String;
  end; // TVolumeInfo

function GetSerialInformation( const Drive : Char; var s:String ) : TVolumeInfo;
var
   lpRootPathName           : PChar; // address of root directory of the file system
   lpVolumeNameBuffer       : PChar; // address of name of the volume
   nVolumeNameSize          : DWORD; // length of lpVolumeNameBuffer
   lpVolumeSerialNumber     : DWORD; // address of volume serial number
   lpMaximumComponentLength : DWORD; // address of system's maximum filename length
   lpFileSystemFlags        : DWORD; // address of file system flags
   lpFileSystemNameBuffer   : PChar; // address of name of file system
   nFileSystemNameSize      : DWORD; // length of lpFileSystemNameBuffer
begin
   GetMem( lpVolumeNameBuffer, MAX_PATH + 1 );
   GetMem( lpFileSystemNameBuffer, MAX_PATH + 1 );
   try
      nVolumeNameSize     := MAX_PATH + 1;
      nFileSystemNameSize := MAX_PATH + 1;

      lpRootPathName := PChar( Drive + ':\' );
      if Windows.GetVolumeInformation( lpRootPathName,
                                       lpVolumeNameBuffer,
                                       nVolumeNameSize,
                                       @lpVolumeSerialNumber,
                                       lpMaximumComponentLength,
                                       lpFileSystemFlags,
                                       lpFileSystemNameBuffer,
                                       nFileSystemNameSize ) then
      begin
         with Result do
         begin
            Name               := lpVolumeNameBuffer;
            SerialNumber       := lpVolumeSerialNumber;
            MaxComponentLength := lpMaximumComponentLength;
            FileSystemFlags    := lpFileSystemFlags;
            FileSystemName     := lpFileSystemNameBuffer;
         end; // with Result
      end // if
      else
      begin
         with Result do
         begin
            Name               := '';
            SerialNumber       := 0;
            MaxComponentLength := 0;
            FileSystemFlags    := 0;
            FileSystemName     := '';
         end; // with Result
      end; // else
      s := Result.Name;
   finally
      FreeMem( lpVolumeNameBuffer );
      FreeMem( lpFileSystemNameBuffer );
   end; // try
end;

function FindDrives:String;
type
  TDriveType = (dtUnknown, dtNoDrive, dtFloppy, dtFixed, dtNetwork, dtCDROM,dtRAM);

  function VolumeID(DriveChar: Char): string;
  var
    OldErrorMode: Integer;
    NotUsed, VolFlags: DWORD;
    Buf: array [0..MAX_PATH] of Char;
  begin
    OldErrorMode := SetErrorMode(SEM_FAILCRITICALERRORS);
    try
      Buf[0] := #$00;
      if GetVolumeInformation(PChar(DriveChar + ':\'), Buf, DWORD(sizeof(Buf)),
        nil, NotUsed, VolFlags, nil, 0) then
        SetString(Result, Buf, StrLen(Buf))
      else Result := '';

      if DriveChar < 'a' then        Result := AnsiUpperCaseFileName(Result)
      else                           Result := AnsiLowerCaseFileName(Result);

      Result := Format('[%s]',[Result]);
    finally
      SetErrorMode(OldErrorMode);
    end;
  end;

  function NetworkVolume(DriveChar: Char): string;
  var
    Buf: Array [0..MAX_PATH] of Char;
    DriveStr: array [0..3] of Char;
    BufferSize: DWORD;
  begin
    BufferSize := sizeof(Buf);
    DriveStr[0] := UpCase(DriveChar);
    DriveStr[1] := ':';
    DriveStr[2] := #0;
    if WNetGetConnection(DriveStr, Buf, BufferSize) = WN_SUCCESS then
    begin
      SetString(Result, Buf, BufferSize);
      if DriveChar < 'a' then        Result := AnsiUpperCaseFileName(Result)
      else                           Result := AnsiLowerCaseFileName(Result);
    end
    else                             Result := VolumeID(DriveChar);
  end;

var
  DriveNum: Integer;
  DriveChar: Char;
  DriveType: TDriveType;
  DriveBits: set of 0..25;
  s : String;
begin
  Result := '';

  Integer(DriveBits) := GetLogicalDrives;

  for DriveNum := 0 to 25 do
  begin
    if not (DriveNum in DriveBits) then      Continue;
    DriveChar := Char(DriveNum + Ord('a'));
    DriveType := TDriveType(GetDriveType(PChar(DriveChar + ':\')));
    case DriveType of
      dtFixed:
      begin
        s := VolumeID(DriveChar);
        if s <> '' then
        begin
          Result := DriveChar;
          exit;
        end;
      end;
     end;
  end;
  // couldn't find it try floppies
  for DriveNum := 0 to 25 do
  begin
    if not (DriveNum in DriveBits) then      Continue;
    DriveChar := Char(DriveNum + Ord('a'));
    DriveType := TDriveType(GetDriveType(PChar(DriveChar + ':\')));
    case DriveType of
      dtFloppy:
      begin
        s := VolumeID(DriveChar);
        if s <> '' then
        begin
          Result := DriveChar;
          exit;
        end;
      end;
     end;
  end;
{  dtNetwork:
   dtCDROM:    s := Format('%s: %s', [DriveChar, VolumeID(DriveChar)]);
   dtRAM:      s := Format('%s: %s', [DriveChar, VolumeID(DriveChar)]);
 }

end;

{MED: 03-11-01, added GetMachinID to freeze maxConnection list on server}
function GetMachineID:String;
var
  s : String;
  vol : TVolumeInfo;
begin
  s   := FindDrives;
  vol := GetSerialInformation(s[1], s);
  s   := inttostr(vol.SerialNumber);
  Result := padr(s, 255, ' ');
end;

function LogicalToYN(YN: Boolean): String;
begin
  if YN then  	Result := 'Y'
  else        	Result := 'N';
end;

function YNToLogical(YN: String): Boolean;
begin
  Result := (YN = 'Y');
end;


{
MED: 04-04-01, added
 Scrambles a word and returns the scrambled version to the calling function

 Example:
   str = "WATER"
        Result = "&@717"
}
Function Scramble ( sStr: string ):String;
var
  i, iLen, ch : Integer;
begin

  iLen  := length(sStr);
  Result:= '';

  for i := 1 to iLen do
  begin
    ch := Ord(substr(sStr,i,1)[1]);

    if      (ch >= 120) AND (ch < 130) then   ch := ch + 125
    else if (ch >= 110) AND (ch < 120) then   ch := ch + 105
    else if (ch >= 100) AND (ch < 110) then   ch := ch + 85
    else if (ch >= 90)  AND (ch < 100) then   ch := ch + 65
    else if (ch >= 80)  AND (ch < 90)  then   ch := ch + 45
    else if (ch >= 70)  AND (ch < 80)  then   ch := ch + 25
    else if (ch >= 60)  AND (ch < 70)  then   ch := ch + 5
    else if (ch >= 50)  AND (ch < 60)  then   ch := ch - 35
    else if (ch >= 40)  AND (ch < 50)  then   ch := ch - 15
    else if (ch >= 30)  AND (ch < 40)  then   ch := ch - 25;
    Result := Result + chr(ch);
  end;
end;

procedure TApolloConnection.ClientError(const Command, ErrMsg: String);
begin
  if Self <> nil then
  begin
    if Assigned(FOnError) then
      FOnError(Self, Command, ErrMsg)
    else
      ShowMessage( '[' + Command + '] ' + ErrMsg );
  //      raise Exception.Create('[Command]'+ErrMsg);
  end;

end;

{MED: 03-17-00, added to support triggers from Server}
{MED: 09-10-00, Put FillReadStream before the IF.  Delphi was shutting down when server was shut down.  fixed}
procedure TApolloConnection.ServiceIncoming(Sender:TObject);
begin
  FillReadStream(TRUE);        // wait for triggers
  if (csDesigning in ComponentState) or (csLoading in ComponentState) or
           (not Connected) then
    exit;
end;

{ TApolloConnection }
constructor TApolloConnection.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  FServerName   := '';
  FGroupName    := '';
  FConnType     := 0;

  FTransList    := TList.Create;

  FSerial       := '0000';
  FTimeOut      := 60000;
  Host          := '127.0.0.1';
  Port          := 8121;
  FVersion      := 'ApolloConnection ' + APOLLOCONNECTION_VER;
  nCurWorkArea  := 0;
  FDatabaseName := '';
  FSQLDataPath  := '';  

  FReadStream   := TMemoryStream.Create;
  FWriteStream  := TMemoryStream.Create;
  FTransStream  := TMemoryStream.Create;  {MED 03-30-00, added}

  FTimer         := TTimer.Create(Self);  {MED 03-18-00, added}
  FTriggersInterval := 1000;              {MED: 01-27-01, added}
  FTimer.Enabled := FALSE;                {Only when Activated-Triggers don't(shouldn't) work in design mode}
  FTimer.Interval:= FTriggersInterval;    {MED 03-18-00, added} {MED: 01-27-01, changed}
  FTimer.OnTimer := ServiceIncoming;      {MED 03-18-00, added}

  FLicenseKey    := '';                   {MED: 03-23-01, added}


  // MED: 04-03-2001
  FCompression        := FALSE;
  FCompressionPassword:= 'Default';
  FCompressionLevel   := 6;

end;

destructor TApolloConnection.Destroy;
begin
	CloseDataSets;
	Active := False;  // Close connection
  FTimer.Enabled := FALSE;  {MED: 02-20-2001, added}
  FTimer.Free;        {MED 03-18-00}
  FReadStream.Free;
  FWriteStream.Free;
  FTransStream.Free;
  FTransList.Free;
  
  inherited Destroy;

end;

{
procedure TApolloConnection.GetActive: Boolean;
begin
  Result := Connected;
end;
}


{==============================================================================
  CloseDataSets  // LAS: 10-17-99, Added
  Closes all TApolloTables connected to this TApolloConnection
 ==============================================================================}
procedure TApolloConnection.CloseDataSets;
var
  Temp: TComponent;
  SearchObj : TApolloSearchObject;
begin
  SearchObj := TApolloSearchObject.Create;
  Try
    Temp      := SearchObj.SearchFirstComponent(TApolloTable);
    while Temp <> nil do
    begin
      if (Temp is TApolloTable) and
           ((Temp as TApolloTable).ApolloConnection = Self) then
          TApolloTable(Temp).Close;
      Temp := SearchObj.SearchNextComponent(TApolloTable);
    end;
  Finally
    SearchObj.Free;
  End;
end;

{==============================================================================
  OpenDataSets  // LAS: 10-17-99, Added
  Opens all TApolloTables connected to this TApolloConnection
 ==============================================================================}
procedure TApolloConnection.OpenDataSets;
var
  Temp: TComponent;
  SearchObj : TApolloSearchObject;
begin
  SearchObj := TApolloSearchObject.Create;
  Try
    Temp      := SearchObj.SearchFirstComponent(TApolloTable);
    while Temp <> nil do
    begin
      if (Temp is TApolloTable) and
           ((Temp as TApolloTable).ApolloConnection = Self) then
      Try
        TApolloTable(Temp).Open;
      Except
        On E: EApolloError do
          MessageDlg(E.Message, mtError, [mbOK], 0);
      End;
      Temp := SearchObj.SearchNextComponent(TApolloTable);
    end;
  Finally
    SearchObj.Free;
  End;

end;
{___________________
|iSize       : 8 bytes   Size of the stream
|iOffSet     : 8 bytes   when 0 then process all. otherwise process table at offset only
|iTableCount : 8 bytes   Use it when iOffset is zero which means process all tables
|_____________________}
function TApolloConnection.StartTransaction: Integer;
var
  Temp: TComponent;
  SearchObj : TApolloSearchObject;
begin
	Result := 0;
  FTransList.Clear;
  FTransStream.Clear;
  FTransStream.Position := 0;
  FTransStream.Write(PChar(IntToHex(0, 8))^, 8);
  FTransStream.Write(PChar(IntToHex(0, 8))^, 8);
  FTransStream.Write(PChar(IntToHex(0, 8))^, 8);
  SearchObj := TApolloSearchObject.Create;
  Try
    Temp   := SearchObj.SearchFirstComponent(TApolloTable);
    while Temp <> nil do
    begin
      if (Temp is TApolloTable) and
         (TApolloTable(Temp).Active) and
         ((Temp as TApolloTable).ApolloConnection = Self) then
      begin
         FTransList.Add(nil);
      end;
      Temp := SearchObj.SearchNextComponent(TApolloTable);
    end;
  Finally
    SearchObj.Free;
  End;
end;

function  TApolloConnection.Commit:Integer;
begin
	Result := 0;
  FTransStream.Position := 8;
  FTransStream.Write(PChar(IntToHex(0, 8))^, 8);

  // on success
  FTransStream.Clear;
end;

function TApolloConnection.RollBack:Integer;
begin
	Result := 0;
end;

function   TApolloConnection.GetTriggers:Boolean;
begin
  Result := FTimer.Enabled;
end;

procedure  TApolloConnection.SetTriggers(Value:Boolean);
begin
  FTimer.Enabled := Value;
end;

procedure  TApolloConnection.SetTriggersInterval(Value:Integer);
var
  bOld : Boolean;
begin
  if (Value<> FTriggersInterval) then
  begin
    bOld              := FTimer.Enabled;
    FTimer.Enabled    := FALSE;
    FTriggersInterval := Value;
    FTimer.Interval   := FTriggersInterval;
    FTimer.Enabled    := bOld;
  end;
end;

{==============================================================================
  MED: 05-21-02 added
  only works while connected to Relay Server
  Note this function only works before LOGIN to server is initiated.
  Call it ONCE before actual login
 ==============================================================================}
function  TApolloConnection.IsServerConnectedToRelay:Boolean;
var
  s : String;
begin
  Result := FALSE;

  if Connected then
    Disconnect;

  Connect;
  if Connected then
  begin
    s := padr(FServerName, 255, ' ')+
         padr(FGroupName, 255, ' ');

    Result := Substr(ClientXmit( sc_IsServerConnectedToRelay, s,  True), 1, 1 ) = 'Y' ;
    // this is necessary to force the thread on the relay to expire
    Disconnect;
  end;
end;

{==============================================================================
  MED: 05-21-02 added
  only works while connected to Relay Server
  Note this function only works before LOGIN to server is initiated.
  Call it ONCE before actual login
 ==============================================================================}
function  TApolloConnection.GetServerListFromRelay:Pointer;
begin
  Result := nil;
  if Connected then
    Disconnect;

  Connect;
  if Connected then
  begin
    try
      ClientStreamReceive( sc_GetServerListFromRelay , FGroupName);
      Result := FReadStream;
    except
      ClientError('ApolloConnection.GetServerListFromRelay', 'GetServerListFromRelay');
    end;
    Disconnect;
  end;
end;


{==============================================================================
  SetActive
 ==============================================================================}
procedure TApolloConnection.SetActive(Value: Boolean);
begin
  if Value <> Connected then
  begin
		if (Value = False) then
    begin
      FTimer.Enabled := FALSE;
      //CloseDataSets;    {LAS: 05-02-00, Uncommented.  Med, did you want this commented still? YES LOREN!}
      if FLogged then
        Logout;
    end;

    if (csDesigning in ComponentState) or (csLoading in ComponentState) then
    begin
      if Value then
      begin
        try
          Connect;
          if (User <> '') and (not Login) then
          begin
            Disconnect;
            ClientError('ApolloConnection.SetActive', 'Login not accepted');
          end;
        except
          ClientError('ApolloConnection.SetActive',
                      'No Apollo Database Server found at ' + Host);
        end;
      end
      else
        Disconnect;
    end
    else if not (csLoading in ComponentState) then
    begin
      if Value then
      begin
        try
          Connect;
          if (User <> '') and (not Login) then
          begin
            Disconnect;
            ClientError('ApolloConnection.SetActive',
                        'Login not accepted. ' +
                        'Maximum user count may have been exceeded');
          end;
          FTimer.Enabled := Connected;
        except
          ClientError('ApolloConnection.SetActive',
                      'No Apollo Database Server found at ' + Host);
        end;
      end
      else
      begin
        Disconnect;
      end;
    end;
  end;
end;

function TApolloConnection.FillInBuff:PChar;
begin
  FReadStream.Position := 0;
  FReadStream.Read(InBuff, FReadStream.Size);
  InBuff[FReadStream.Size] := #0;
  FReadStream.Position := 0;
  Result := InBuff;
end;

{==============================================================================
	Notification   // LAS: 11-14-99, Added
==============================================================================}
procedure TApolloConnection.Notification( AComponent: TComponent;
                                          Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);

  if (Operation = opRemove) then
  begin
    if (AComponent is TApolloTable) and (TApolloTable(AComponent).ApolloConnection = self) then
  		TApolloTable(AComponent).ApolloConnection := nil;

    {MED: 12-10-99 added}
    if (AComponent is TApolloQuery) and (TApolloQuery(AComponent).ApolloConnection = self) then
	  	TApolloQuery(AComponent).ApolloConnection := nil;

    if  (AComponent is TApolloServerDLL) and (TApolloServerDLL(AComponent).ApolloConnection = self) then
	  	TApolloServerDLL(AComponent).ApolloConnection := nil;

    // [MED:] 03-29-2001 added
    if (AComponent is TApolloEnv) and (TApolloEnv(AComponent).ApolloConnection = self) then
	    TApolloEnv(AComponent).ApolloConnection := nil;


  {$ifdef Apollo60}
    // [MED:] 03-29-2001 added
    if (AComponent is TApolloServerScript) and (TApolloServerScript(AComponent).ApolloConnection = self) then
	    TApolloServerScript(AComponent).ApolloConnection := nil;
  {$endif}
  end;
end;

// tries to ping the server. Disconnects upon failure
function TApolloConnection.Ping: Boolean;   {Changed from Procedure}
begin
  try      
    Write('00000000000000000000');    // server expects 20 bytes minimum
		Result := True;
  except
    Active := False;
		Result := False;
  end;
end;

// initial handshake is done here
//  TO  From    Action     Packet
// 0034 0000 CONNECTIONID   SOCKETID
// 0034 0000     0459       0034 <---Socket ID
// this is called ONCE upon successful connection
function TApolloConnection.GetConnectionID: String;  // used by TApolloConnection
begin
  if Connected then
  begin
    try
      Result := ClientXmit(sc_CONNECTIONID, '', True);

      if Result <> '' then
      	FSerial := PadL(Result, 4, '0');
    except
      ClientError('ApolloConnection.GetConnectionID', 'GetConnectionID');
    end;
  end
  else
    Result := '0000';
end;

//--------------------------------------------------
// MED: 04-04-2001- Added
// Initially password is "Default"
//--------------------------------------------------
procedure TApolloConnection.SetCompressionPassword(Value:String);        // 255 chars
var
  s : String;
begin
  if (Value <> FCompressionPassword) then
  begin
    if Connected then
    begin
      s := PadR(Value,255,' ');
      s := Scramble(s);
      if ClientXmit( sc_SetCompressionPassword, s,  True) = 'Y' then
        FCompressionPassword := Value;
    end;
  end;
end;


function TApolloConnection.Login: Boolean;
var
  sMsg, s : String;
// CRVG Hint   OutStrm : TMemoryStream;
// CRVG   iLen : Integer;
begin
  FLogged := False;
  sMsg    := '';
  if Connected then
  begin
    // MED: 04-04-2001- Added scramble
    s :=   PadR(Password,255,' ') +
           PadR(User, 255, ' ')+
           GetMachineID+              // 255 length
           padr(FLicenseKey, 255, ' ')+
           padr(FCompressionPassword, 255, ' ')+
           IntToStr(FCompressionLevel)+                 // one byte
           LogicalTOYN(FCompression)+
           IntToStr(FConnType)+                 // one byte-client or server connecting to Relay
           padr(FServerName, 255, ' ')+
           padr(FGroupName, 255, ' ');

    s := Scramble(s);
    {MED: 03-22-01, added GetMachineID and LicenseKey}



    sMsg := ClientXmit( sc_LOGIN, s,  True);

    if Substr(sMsg, 1, 1 ) = 'Y' then
    begin
      FLogged := TRUE;
      if assigned(FOnLogin) then
        FOnLogin( Self, FLogged, '');
    end
    else
    begin
      Logout;       // MED: 03-23-01, added - logout first before calling the event
      if assigned(FOnLogin) then
        FOnLogin( Self, FLogged, trim(substr(sMsg, 2, length(sMsg)-1)));
    end;
  end;
  Result := FLogged;
end;
                                       
procedure TApolloConnection.Logout;
begin
  if Connected then
    ClientXmit(sc_LOGOUT, PadR(User, 10, ' ') , TRUE);

    // MED:05-26-02, commented.  MachineID is now part of SessionData on the server
    // so there is no need 
    //ClientXmit(sc_LOGOUT, PadR(User, 10, ' ') + GetMachineID, FALSE);

  if assigned(FOnLogout) then
      FOnLogout( Self );

  FLogged := False;
end;

//MED:  Note if DateFormat is set to dfWinDefault then century will always show
procedure TApolloConnection.ap_EnvSettings(Epoch, DateFormat, MemoBlockSize:Integer;
                                           Deleted, Century, SoftSeek, Exact:WordBool);
begin
  try
    ClientXmit( sc_ENVSETTINGS ,
                                IntToHex(Epoch, 4) +
                                IntToHex(DateFormat, 4)+
                                IntToHex(MemoBlockSize, 4)+
                                LogicalTOYN(Deleted)+
                                LogicalToYN(Century)+
                                LogicalToYN(SoftSeek)+
                                LogicalToYN(Exact),
                                True);
  except
    ClientError('ApolloConnection.EnvSettings', 'EnvSettings')
  end;
end;

//=========================================================================================
/////////////////////////////////////////////////////
// this is a bit different than sx_Use
// TableName is passed and no path is allowed here.
// There is no path before filename since the user is not supposed to know where
// the file is. However, the cpAlias contains a path alias that's resolved on
// the SERVER.
//   sx_Use: function(cpFileName: PChar; cpAlias: PChar; iOpenMode: SmallInt;
//                    iRdeType: SmallInt): SmallInt;
// See OpenTable inside the TApolloServer

function  TApolloConnection.ap_Use( cpDataBaseName, cpFileName: PChar;  iOpenMode: SmallInt;
                   iRdeType: SmallInt): SmallInt;
var
  msg, cDataBaseName, cTableName: String;
begin
  Result := 0;
  try
    cDataBaseName := PadR(cpDataBaseName, 255, ' ');
    cTableName := PadR(cpFileName, 255, ' ');
    msg := ClientXmit( sc_USE,
                                 cDataBaseName +
                                 cTableName +
                                 PadR(IntToStr(iOpenMode), 1, ' ')+
                                 PadR(IntToStr(iRdeType), 1, ' ') , True);
    try
      Result := StrToInt('$'+Trim(msg));
    except
      ClientError('ApolloConnection.Use', 'Use');
    end;
  except
    ClientError('ApolloConnection.Use', 'Use');
  end;
end;

function  TApolloConnection.ap_UseEx( cpDataBaseName, cpFileName: PChar;  iOpenMode: SmallInt;
                   iRdeType: SmallInt; uiMode : Cardinal ): SmallInt;
var
  msg, cDataBaseName, cTableName: String;
begin
  Result := 0;
  try
    cDataBaseName := PadR(cpDataBaseName, 255, ' ');
    cTableName := PadR(cpFileName, 255, ' ');
    msg := ClientXmit( sc_UseEX,
                                 cDataBaseName +
                                 cTableName +
                                 PadR(IntToStr(iOpenMode), 1, ' ')+
                                 PadR(IntToStr(iRdeType), 1, ' ')+
                                 PadR(IntToStr(uiMode), 1, ' ' ), True);
    try
      Result := StrToInt('$'+Trim(msg));
    except
      ClientError('ApolloConnection.UseEx', 'UseEx');
    end;
  except
    ClientError('ApolloConnection.Use', 'Use');
  end;
end;

function  TApolloConnection.ap_QuickUse(
                   sDataBaseName, sFileName: String;
                   iOpenMode: SmallInt;
                   iRdeType: SmallInt;
                   var FRowCount : Integer;
                   var FFieldCount : Integer;
                   var FRecordSize : Integer;
                   var FIsEncrypted: WordBool;
                   var FPassword   : String;
                   FTurboRead      : WordBool;
                   FIndexName      : String;
                   FEpoch, FDateFormat, FMemoBlockSize  : Integer;
                   FDeleted, FCentury, FSoftSeek,
                   FExact,
                   FTranslate,            // MED: 04-07-01
                   FApplyEnvSettings: WordBool;
                   FIndexCondition : String;
                   FCommitLevel : Integer;
                   var FIndexOrd  : Integer;
                   var FIndexDefs : TIndexDefs;
                   var iError: Integer;
                   var sError:String ): SmallInt;

var
  s, sRes, cDataBaseName, cTableName: String;
  i, j, iItemCount,  iItemSize : Integer;
  sName, sFields,  sSource, sExpression,
  sxPrimary,  sxUnique,  sxDescending,
  sxCaseIns,  sxExpression,  sxNonMain : String;
begin
  Result          := 0;
  FFieldCount     := 0;
  FRecordSize     := 0;
  FRowCount       := 0;
  iError          := 0;
  FIndexOrd       := 0;
  FIndexCondition := '';
  sError          := '';
  FIndexDefs.Clear;

  try
    cDataBaseName := PadR(sDataBaseName, 255, ' ');
    cTableName := PadR(sFileName, 255, ' ');
    sRes := ClientXmit( sc_QUICKUSE,
                                 cDataBaseName +
                                 cTableName +
                                 PadR(IntToStr(iOpenMode), 1, ' ')+
                                 PadR(IntToStr(iRdeType), 1, ' ') +
                                 PadR(FPassword, 255, ' ') +
                                 LogicalTOYN(FTurboRead)+
                                 PadR(FIndexName, 255, ' ') +
                                 InttoHex(FEpoch, 4)+
                                 IntToHex(FDateFormat, 4)+
                                 IntToHex(FMemoBlockSize, 4)+
                                 LogicalTOYN(FDeleted)+
                                 LogicalTOYN(FCentury)+
                                 LogicalTOYN(FSoftSeek)+
                                 LogicalTOYN(FExact)+
                                 LogicalTOYN(FTranslate)+
                                 LogicalTOYN(FApplyEnvSettings)+
                                 Padr(IntToStr(FCommitLevel),1,' '),
                                 True);
    try
      Result    := StrToInt('$'+Substr(sRes,1,4));
//      if Result > 0 then    {LAS: 06-15-00, Removed.  Handled in ApoDSet}
      begin
      	nCurWorkArea    := Result;
        FRowCount       := StrToInt('$'+Substr(sRes,5,8));
        FFieldCount     := StrToInt('$'+Substr(sRes,13,8));
        FRecordSize     := StrToInt('$'+Substr(sRes,21,8));
        FIsEncrypted    := YNToLogical(Substr(sRes,29,1));
        FIndexCondition := trim(Substr(sRes, 30,255));
        FIndexOrd       := StrToInt('$'+Substr(sRes, 285,4));
        iItemCount      := StrToInt('$'+Substr(sRes,289, 4));
        iItemSize       := StrToInt('$'+Substr(sRes,293, 8));
        iError          := StrToInt('$'+Substr(sRes,301+iItemSize, 8));
        sError          := Substr(sRes, 309+iItemSize, length(sRes)- (308+iItemSize) );

        if iError = 0 then
        begin
          s := Substr(sRes, 301, iItemSize);
          j := 1;
          for i := 0 to iItemCount-1 do
          begin
            sName       := trim(SubStr(s , j, 255));
            sFields     := trim(SubStr(s , j+255, 255));
            sSource     := trim(SubStr(s , j+510, 255));
            sExpression := trim(SubStr(s , j+765, 255));
            sxPrimary   := SubStr(s , j+1020, 1);
            sxUnique    := SubStr(s , j+1021, 1);
            sxDescending:= SubStr(s , j+1022, 1);
            sxCaseIns   := SubStr(s , j+1023, 1);
            sxExpression:= SubStr(s , j+1024, 1);
            sxNonMain   := SubStr(s , j+1025, 1);

            with FIndexDefs do
				    begin
		    			Add( sName, sFields, [] );
          
              Items[i].Source     := sSource;
              if sxExpression = '1' then
              begin
                Items[i].Options    := Items[i].Options + [ixExpression];
                Items[i].Expression := sExpression;
              end;

              if sxPrimary = '1' then
                Items[i].Options := Items[i].Options + [ixPrimary];
              if sxUnique = '1' then
                Items[i].Options := Items[i].Options + [ixUnique];
              if sxDescending = '1' then
                Items[i].Options := Items[i].Options + [ixDescending];
              if sxCaseIns = '1' then
                Items[i].Options := Items[i].Options + [ixCaseInsensitive];

              if sxNonMain = '1' then
                Items[i].Options := Items[i].Options + [ixNonMaintained];

    				end;
            j := j + 1026;  // LAS: 07-18-00, Changed from 1025
          end;
        end
        else
        begin
//		      ClientError('ApolloConnection.Use', 'Error ' + IntToStr(iError) +
//                      ': ' + sError );
        end;
      end;
    except
      ClientError('ApolloConnection.Use', 'Use');
    end;
  except
    ClientError('ApolloConnection.Use', 'Use');
  end;
end;

function TApolloConnection.ap_Alias(iWA:Word): PChar;
begin
  try
    Result := ClientXmit( sc_ALIAS, IntToStr(iWA), True);
  except
    Result := nil;
    ClientError('ApolloConnection.Alias', 'Alias');
  end;
end;

{
MED: 03-18-00
Main Engine for sending/receiving messages

bTrigger is TRUE when called by the timer thread to check for triggers from server.
Otherwise it is FALSE.
if bTrigger is FALSE and we encounter a TRIG_* message, then it's likely the
client was already waiting on some other packets so we need to stay inside untill
we statisfy client's need or timeout.

L_Continue helps us stay within the engine even after an unwanted TRIG message
from the server.

main block purposely indented twice so L_Continue stands out better
}
function TApolloConnection.FillReadStream(bTriggers:Boolean): Integer;
var
  iMin, iMax, iPos:Integer;
  ToRead :Integer;
  FAction, s, t : String;
  InStream : TMemoryStream;
  Buff     : Pointer;             // MED: 04-04-01, added
Label
  L_Continue;
begin
  FTimer.Enabled := FALSE;
  Result := 0;
  ToRead := 0;
  if (Connected) then
  begin

      Try
        if (bTriggers) and (not Readable(1)) then       // no waiting here
        begin
          FTimer.Enabled := Connected;           //('Nothing to read');
          exit;
        end;


  L_Continue:    //<============   sorry there was no other choice!

        FReadStream.Clear;

        s       := Read(20) ; //ReadWaitSec(20, FTimeOut);

//showmessage('s---->'+s);
        Try
          ToRead  := StrToInt('$'+Substr(s, 1, 8));
          if ToRead > 0 then
          begin
            if FCompression then
            begin
              InStream := TMemoryStream.Create;
              try
                ReadToStream( InStream, ToRead );
                if not CryptDLLLoaded then
                  LoadCryptDLL;

                UnCompressStream(InStream, FReadStream);
                ToRead  := FReadStream.Size;
              finally
                InStream.Free;
              end;
            end
            else
              ReadToStream( FReadStream, ToRead );
          end;
        Except
          ToRead := 0;
        End;
        Try
          FAction := substr(s, 17,4);                    {MED: 12-30-99, added}
        Except
        End;

        // ----------------------------------------------------------------
        // Process Action
        // ----------------------------------------------------------------
        if FAction = sc_TRIG_SHUTDOWN then
        begin
          sleep(5000);
          Disconnect;
          if Assigned(FOnServerMessage) then
            OnServerMessage(Self, 'Server shutdown');
          Result := -1;
          exit;
        end
        //--------------------------------------------
        else if FAction = sc_TRIG_STREAM then   // MED: 01-27-01, Added
        begin
          if ToRead > 0 then
          begin
            if Assigned(FOnServerStream) then
              OnServerStream(Self, FReadStream);
          end;
          if not bTriggers then
            goto L_Continue;
        end
        //--------------------------------------------
        else if FAction = sc_TRIG_MESSAGE then   // MED: 09-28-01, text message only.
        begin
          if ToRead > 0 then
          begin
            GetMem(Buff, ToRead+1);
            Try
              FReadStream.Read(PChar(Buff)^, ToRead);
              pchar(Buff)[ToRead] := #0;               // MED: 09-28-01, text message only.
              if Assigned(FOnServerMessage) then
                OnServerMessage(Self, pchar(Buff));
            Finally
              FreeMem(Buff);
            End;
          end;
          if not bTriggers then
            goto L_Continue;
        end
        //--------------------------------------------
        else if FAction = sc_TRIG_PROGRESS then
        begin
          if ToRead > 0 then
          begin
            GetMem(Buff, 24);
            Try
              FReadStream.Read(PChar(Buff)^, 24);
              if Assigned(OnServerProgress) then
              begin
                t    := PChar(Buff);
                iMin := StrToInt('$'+Substr(t,1,8));
                iMax := StrToInt('$'+Substr(t,9,8));
                iPos := StrToInt('$'+Substr(t,17,8));
                OnServerProgress(Self, psXProgress, iMin, iMax, iPos );
              end;
            finally
              FreeMem(Buff);
            end;
          end;
          if not bTriggers then
            goto L_Continue;
        End;

      Except
        Active      := FALSE;   {MED: 12-29-99, Added}
        ClientError('ApolloConnection.FillReadStream', 'Connection lost');
      End;
  end;
  Result := ToRead;
  FTimer.Enabled := Connected;
  
end;

//--------------------------------------------------------------------
// bytes to read does not include self
//     4bytes    4bytes    4bytes
// [   0001<-scramble byte  ] [   Serial  ] [action.]  [..packet.......could be empty]
// however when packets are receive they also contain cLen = 8 tyes in front
// Byte 12 is flipped to 1 when FCompression is TRUE
//--------------------------------------------------------------------
function TApolloConnection.ClientXmit( const scAction:String; sPacket: String;
                                       bWaitForResponse: Boolean): PChar;
var
  iLen :Integer;
  OutStr : String;
  InStrm, OutStrm : TMemoryStream;
begin
  FTimer.Enabled := FALSE;
  Result         := '';
  iLen           := Length(sPacket);

  if (Connected) then
  begin
    try
      if (FCompression) AND (iLen > 0) then
      begin
        if not CryptDLLLoaded then        LoadCryptDLL;

        InStrm  := TMemoryStream.Create;
        OutStrm := TMemoryStream.Create;
        Try
          InStrm.Write(pchar(sPacket)^, iLen);
          CompressStream(InStrm, OutStrm);
          OutStr := IntToHex(OutStrm.Size, 8) + '0001' + FSerial + scAction;
          WriteToStream2(outStr, OutStrm);
        Finally
          InStrm.Free;
          OutStrm.Free;
        End;
      end
      else
      begin
        Write(IntToHex(iLen, 8) + '0000' + FSerial + scAction + sPacket);
      end;

    except
      Active := FALSE;            {MED: 12-29-99 , Added}
      ClientError('ApolloConnection.ClientXmit', 'Connection lost');
      exit;
    end;

    if (bWaitForResponse) and (FillReadStream(FALSE) > 0) then
    begin
      try
        Result := FillInBuff;
      except
        ClientError('ApolloConection.ClientXmit', 'Connection lost');
      end;
    end;
  end
  else
    ClientError('ApolloConection.ClientXmit('+scAction+')', 'Not connected');

  FTimer.Enabled := Connected;
end;

(*
//--------------------------------------------------------------------
// the first write is 20 bytes and contains BTR+ '0001' (Fcompress)+Action
// the second write is the stream itself
// bytes to read (BTR) does not include self
// New improved (faster)
//--------------------------------------------------------------------
function TApolloConnection.ClientStreamSend(const scAction:String): Integer;    // sends a stream
var
  s : String;
  OutStream : TMemoryStream;
begin
  Result    := 0;
  try
    if (Connected) and (FWriteStream.Size > 0) then   // can't send empty ones!
    begin
      if FCompression then
      begin
        OutStream := TMemoryStream.Create;
        Try
          CompressStream(FWriteStream, OutStream);
          s      := IntToHex(OutStream.Size, 8) + '0001' + FSerial + scAction;
          Result := WriteToStream2(s, OutStream);
        Finally
          OutStream.Free;
        End;
      end
      else
      begin
        s      := IntToHex(FWriteStream.Size, 8) + '0000' + FSerial + scAction;
        Result := WriteToStream2(s, FWriteStream);
      end;
    end;
  except
    ClientError('ApolloConnection.ClientStreamSend', 'Connection lost');
  end;
end;
*)

//------------------------------------------------------------------------
// bytes to read (BTR) does not include self NOR 0000+FSerial+scAction
//
// pointer starts at ACTION+PACKET
// sends a single packet but receives a stream
//-------------------------------------------------------------------------
function TApolloConnection.ClientStreamReceive( const scAction: String;
                                                sPacket: String ): Pointer;
var
  iLen : Integer;
  s : String;
  InStrm, OutStrm : TMemoryStream;
begin
  FTimer.Enabled := FALSE;
  Result         := nil;
  iLen           := Length(sPacket);      // minus the scAction
  FReadStream.Clear;
  if (Connected) then
  begin
		try
      if (FCompression) AND (iLen > 0) then
      begin
        if not CryptDLLLoaded then        LoadCryptDLL;

        InStrm  := TMemoryStream.Create;
        OutStrm := TMemoryStream.Create;
        Try
          InStrm.Write(pchar(sPacket)^, iLen);
          CompressStream(InStrm, OutStrm);
          s := IntToHex(OutStrm.Size, 8) + '0001' + FSerial + scAction;
          WriteToStream2(s, OutStrm);
        Finally
          InStrm.Free;
          OutStrm.Free;
        End;
      end
      else
      begin
        s := IntToHex(iLen, 8) + '0000' + FSerial + scAction + sPacket;
        Write(s);
      end;

		Except
      Active := FALSE;                    {MED: 12-29-99 , Added}
      ClientError('ApolloConnection.ClientStreamReceive', 'Connection lost');
			exit;
		end;
		iLen := FillReadStream(FALSE);
    if  iLen > 0 then
		begin
			FReadStream.Position := 0;
			Result := FReadStream;
		end;
  end;
  FTimer.Enabled := Connected;
end;

{-----------------------------------------------------------------------------
  MED: 03-18-00 Added-This is the holy grail of stream transmission!
  Using this we can now send a stream and receive a stream back.
  // the first write is 20 bytes and contains BTR+'0000'+TO+ACTION
  // the second write is the stream itself
  // bytes to read (BTR) does not include self
  Sends FWriteStream and then waits for another stream (FReadStream) to come back from server
-----------------------------------------------------------------------------}
function TApolloConnection.ClientStreamSendReceive(const scAction:String):TMemoryStream;
var
  iOutSize, iInSize, iSent : Integer;
  s : String;
  OutStream : TMemoryStream;
begin
  FTimer.Enabled := FALSE;
  Result         := nil;
  iSent          := 0;
  iOutSize       := 0;
  try
    if (Connected) and (FWriteStream.Size > 0) then   // can't send empty ones!
    begin
      if FCompression then
      begin
        OutStream := TMemoryStream.Create;
        Try
          CompressStream(FWriteStream, OutStream);
          iOutSize  := OutStream.Size;
          s         := IntToHex(iOutSize, 8) + '0001' + FSerial + scAction;
          iSent     := WriteToStream2(s, OutStream);
        Finally
          OutStream.Free;
        End;
      end
      else
      begin
        iOutSize  := FWriteStream.Size;
        s         := IntToHex(iOutSize, 8) + '0000' + FSerial + scAction;
        iSent     := WriteToStream2(s, FWriteStream);
      end;
    end;
    //==========================================================
    // now wait for stream to come from server
    if (Connected) and (iSent=iOutSize) then
    begin
		  FReadStream.Clear;
		  iInSize := FillReadStream(FALSE);
      if  iInSize > 0 then
		  begin
			  FReadStream.Position := 0;
  			Result := FReadStream;
	  	end;
    end;
  except
    ClientError('ApolloConnection.ClientStreamSendReceive', 'Connection lost');
  end;
  FTimer.Enabled := Connected;
end;

function TApolloConnection.GetServerName : String;  // used by TApolloConnection
begin
  try
    Result := ClientXmit(sc_SERVERNAME, '', True);
  except
    ClientError('ApolloConnection.ServerName', 'ServerName');
  end;
end;

Function TApolloConnection.ap_Append : Integer;
begin
  try
    Result := StrToInt('$'+ClientXmit( sc_APPEND , '', TRUE));
  except
    ClientError('ApolloConnection.Append', 'Append');
    Result := 0;
  end;
end;

Function TApolloConnection.ap_AppendBlank: Integer;
begin
  try
    Result := StrToInt('$'+ClientXmit( sc_APPENDBLANK, '', TRUE));
  except
    ClientError('ApolloConnection.AppendBlank', 'AppendBlank');
    Result := 0;
  end;
end;

function TApolloConnection.ap_AppendFrom( cpFileName: PChar;
                         iSourceType: SmallInt; cpScopeExpr: PChar ): WordBool;
begin
  try
    Result := ClientXmit( sc_APPENDFROM,
                          PadR( cpFileName, 255, ' ') +
                          PadR( IntToStr( iSourceType ), 10, ' ') +
                          PadR( cpScopeExpr, 255, ' '),
                          True ) = 'Y';
  except
	  Result := False;
  end;
end;

function TApolloConnection.ap_BaseDate: PChar;
begin
  Result := '';
  try
    Result :=  PChar(ClientXmit( sc_BASEDATE ,'',  True));
  except
    ClientError('BaseDate', 'BaseDate');
  end;
end;

function TApolloConnection.ap_BaseName: PChar;
begin
  Result := '';
  try
    Result := ClientXmit( sc_BASENAME , '', True);
  except
    ClientError('ApolloConnection.BaseName', 'BaseName');
  end;
end;

{MED: 03-07-2000, added}
function  TApolloConnection.ap_BlobToFile( cpFieldName, cpFileName: PChar;
                                          iRecNo: Integer ): WordBool;  {LAS: 06-26-00, Added iRecNo param}
begin
  Result := False;
  try
    Result := ClientXmit( sc_BLOBTOFILE,
                          PadR( cpFieldName, 10, ' ')+
                          PadR(cpFileName, 255, ' ') +
                          IntToHex(iRecNo, 8),   {LAS: 06-26-00, Added iRecNo param}
                          True ) = 'Y';

  except
    ClientError('ApolloConnection.BlobToFile', 'BlobToFile');
  end;
end;

function  TApolloConnection.ap_Bof: WordBool;
begin
  Result := False;
  try
    Result := ClientXmit( sc_BOF, '', True) = 'Y';
  except
    ClientError('ApolloConnection.Bof', 'Bof');
  end;
end;

procedure TApolloConnection.ap_Close;
begin
  try
    ClientXmit( sc_CLOSETABLE , '', TRUE);
  except
    ClientError('ApolloConnection.Close', 'Close');
  end;
end;

procedure TApolloConnection.ap_CloseAll;
begin
  try
    ClientXmit( sc_CLOSEALL, '', TRUE);
  except
    ClientError('ApolloConnection.CloseAll', 'CloseAll');
  end;
end;


procedure TApolloConnection.ap_CloseIndexes;
begin
  try
    ClientXmit( sc_CLOSEINDEXES, '', TRUE);
  except
    ClientError('CloseIndexes', 'CloseIndexes');
  end;
end;

procedure TApolloConnection.ap_Commit;
begin
  try
    ClientXmit( sc_COMMIT, '', TRUE);
  except
    ClientError('ApolloConnection.Commit', 'Commit');
  end;
end;

function TApolloConnection.ap_CommitLevel(iVal:Integer):Integer;
begin
  Result := 0;
  try
    Result := StrToInt(ClientXmit( sc_COMMITLEVEL  , IntToStr(iVal), True));
  except
    ClientError('ApolloConnection.CommitLevel', 'CommitLevel');
  end;
end;

function TApolloConnection.ap_CopyFile( sToFileName: PChar ): WordBool;
begin
  try
    Result := ClientXmit( sc_COPYFILE,
                          PadR( sToFileName, 255, ' '),
                          True ) = 'Y';
  except
    Result := False;
    ClientError('ApolloConnection.CopyFile', 'CopyFile');
  end;
end;

function  TApolloConnection.ap_CopyFileText( sTextFileName:PChar; iFileType : smallInt ) : WordBool;
begin
  try
    Result := ClientXmit( sc_COPYFILETEXT,
                          PadR( sTextFileName, 255, ' ')+
                          IntToHex(iFileType, 4),
                          True ) = 'Y';
  except
    Result := False;
    ClientError('ApolloConnection.CopyFileText', 'CopyFileText');
  end;
end;

function TApolloConnection.ap_CopyStructure( sFileName : PChar; sAlias:PChar  ) : WordBool;
begin
  try
    Result := ClientXmit( sc_COPYSTRUCTURE,
                          PadR( sFileName, 255, ' ')+
                          PadR( sAlias, 255, ' '),
                          True ) = 'Y';
  except
    Result := False;
    ClientError('ApolloConnection.CopyStructure', 'CopyStructure');
  end;
end;

function TApolloConnection.ap_CopyStructureExtended(cpFileName: PChar): WordBool;
begin
  try
    Result := ClientXmit( sc_COPYSTRUCTUREEXTENDED,
                          PadR( cpFileName, 255, ' '),
                          True ) = 'Y';
  except
    Result := False;
    ClientError('ApolloConnection.CopyStructureExtended', 'CopyStructureExtended');
  end;
end;


function TApolloConnection.ap_Count: LongInt;
begin
  Result := 0;
  try
    Result := StrToInt(ClientXmit( sc_COUNT  , '', True));
  except
    ClientError('ApolloConnection.Count', 'Count')
  end;
end;

function  TApolloConnection.ap_CreateExec: WordBool ;
begin
  Result := False;
  try
    if ClientXmit( sc_CREATEEXEC , '', True) = 'Y' then
			Result := True;
  except
    ClientError('ApolloConnection.CreateExec', 'CreateExec');
  end;
end;

procedure TApolloConnection.ap_CreateField(cpName, cpType: PChar;
                            iLength: SmallInt; iDecimals: SmallInt);
begin
//   Result := False;
  try
    ClientXmit( sc_CREATEFIELD,
                     PadR(cpName, 10, ' ')+
                     PadR(cpType, 10, ' ')+
                     PadR(IntToStr(iLength), 10, ' ')+
                     PadR(IntToStr(iDecimals), 10, ' '),
                     True);
  except
    ClientError('ApolloConnection.CreateField', 'CreateField')
  end;
end;

function  TApolloConnection.ap_CreateFrom( cpFileName,  cpAlias,  cpStruFile: PChar;
                          iRDEType: SmallInt): WordBool;
begin
  Result := False;
  try
    if ClientXmit( sc_CREATEFROM,
                     PadR(cpFileName, 255, ' ')+
                     PadR(cpAlias, 255, ' ')+
                     PadR(cpStruFile, 255, ' ')+
                     PadR(IntToStr(iRDEType), 10, ' '),
                     True) = 'Y' then
    Result := True;
  except
  end;
end;

{MED: 03-20-00, added cpDataBaseName}
function TApolloConnection.ap_CreateNew( cpDataBaseName, cpFileName: PChar;
                    iRdeType: SmallInt; iNumFields: SmallInt): SmallInt;
var
  msg : String;
begin
  Result := 0;
  try
    msg := ClientXmit( sc_CREATENEW ,
                     PadR(cpDataBaseName, 255, ' ')+
                     PadR(cpFileName, 255, ' ')+
                     IntToHex(iRDEType, 1)+
                     IntToHex(iNumFields, 8), True);

    Result := StrToInt(Trim(msg));
  except
    ClientError('ApolloConnection.CreateNew', 'CreateNew');
  end;
end;

{MED: 03-24-00, re-wrote to use stream to make nulls compatible}
function TApolloConnection.ap_Decrypt( cpBuffer, cpPassword : PChar;   iLen : Integer  ) : PChar;
var
  pMem : TMemoryStream;
begin
  Result := nil;
  Try
    FWriteStream.Clear;
    FWriteStream.Position := 0;
    FWriteStream.Write(pchar(IntToHex(iLen, 8))^, 8);
    FWriteStream.Write(pchar(cpPassword)^, 255);
    FWriteStream.Write(pchar(cpBuffer)^, iLen);
    pMem         := ClientStreamSendReceive(sc_DECRYPT);
    Result       := pMem.Memory;
    Result[iLen] := #0;
  except
    ClientError('ApolloConnection.Decrypt', 'Decrypt');
  end;
end;

{MED: 03-24-00, re-wrote to use stream to make nulls compatible}
function TApolloConnection.ap_Encrypt( cpBuffer, cpPassword : PChar;   iLen : Integer ) : PChar;
var
  pMem : TMemoryStream;
begin
  Result := Nil;
  Try
    FWriteStream.Clear;
    FWriteStream.Position := 0;
    FWriteStream.Write(pchar(IntToHex(iLen, 8))^, 8);
    FWriteStream.Write(pchar(cpPassword)^, 255);
    FWriteStream.Write(pchar(cpBuffer)^, iLen);
    pMem         := ClientStreamSendReceive(sc_ENCRYPT);
    Result       := pMem.Memory;
    Result[iLen] := #0;
  except
    ClientError('ApolloConnection.Encrypt', 'Encrypt');
  end;
end;


function  TApolloConnection.ap_DbfDecrypt: WordBool;
begin
  Result := False;
  try
   Result := ClientXmit( sc_DBFDECRYPT , '', True) = 'Y';
  except
    ClientError('ApolloConnection.DbfDecrypt', 'DbfDecrypt');
  end;
end;

function TApolloConnection.ap_DbfEncrypt: WordBool;
begin
  Result := False;
  try
    Result := ClientXmit( sc_DBFENCRYPT , '', True) = 'Y';
  except
    ClientError('ApolloConnection.DbfEncrypt', 'DbfEncrypt');
  end;
end;

function TApolloConnection.ap_DBFilter  :PChar;
begin
  Result := '';
  try
    Result := ClientXmit( sc_DBFILTER , '', True);
  except
    ClientError('ApolloConnection.DBFilter', 'DBFilter');
  end;
end;


function TApolloConnection.ap_Descend( cpKeyString: PChar ): PChar;
begin
  Result := '';
  try
    Result := ClientXmit( sc_DESCEND , cpKeyString, True);
  except
    ClientError('ApolloConnection.Descend', 'Descend')
  end;
end;

procedure TApolloConnection.ap_Delete;
begin
  try
    ClientXmit( sc_DELETE , '', TRUE);
  except
    ClientError('ApolloConnection.Delete', 'Delete')
  end;
end;

function  TApolloConnection.ap_Deleted: WordBool;
begin
  Result := False;
  try
    Result := ClientXmit( sc_DELETED, '', True) = 'Y';
  except
    ClientError('ApolloConnection.Deleted', 'Deleted');
  end;
end;


function TApolloConnection.ap_Eof: WordBool;
begin
  try
   Result := ClientXmit( sc_EOF, '', True ) = 'Y';
  except
    Result := False;
  end;
end;


function TApolloConnection.ap_Empty( sFieldName: PChar ): WordBool;
begin
  try
    Result := YNToLogical(ClientXmit( sc_EMPTY , sFieldName, True));
  except
    Result := False;
  end;
end;
                  
                                        
// NOTE: sx_SetErrorFunc should be used to trap server-side errors to Log file
function  TApolloConnection.ap_ErrorLevel(iErrorLevel: Integer ): Integer;
begin
	// Server-side ErrorLevel is ALWAYS set to 0;
	Result := 0;
end;
{ // [ac] 12-13-2002 changed to below
function  TApolloConnection.ap_EvalLogical(sExpression: PChar ): WordBool;
begin
  try
    Result := YNToLogical(ClientXmit( sc_EVALLOGICAL, sExpression, True));
  except
	  Result := False;
    ClientError('ApolloConnection.EvalLogical', 'EvalLogical');
  end;
end;
}
function  TApolloConnection.ap_EvalLogical(sExpression: PChar ): WordBool;
var 
	sResult: string;
begin
	try
		sResult := ClientXmit( sc_EVALLOGICAL, sExpression, True);
		if sResult = '' then
			result := False
		else if sResult = #1 then
			result := True
		else
			Result := YNToLogical(sResult);
  except
   	Result := False;
    ClientError('ApolloConnection.EvalLogical', 'EvalLogical');
  end;
end;

function TApolloConnection.ap_EvalNumeric( sExpression: PChar ): Double;
begin
  try
    Result := StrToFloat(ClientXmit( sc_EVALNUMERIC, sExpression, True));
  except
	  Result := 0.00;
    ClientError('ApolloConnection.EvalNumeric', 'EvalNumeric');
  end;
end;

function TApolloConnection.ap_EvalString (sExpression: PChar ): PChar;
begin
  try
    Result := ClientXmit( sc_EVALSTRING, sExpression, True);
  except
	  Result := nil;
    ClientError('ApolloConnection.EvalString', 'EvalString');
  end;
end;

function  TApolloConnection.ap_EvalTest (sExpression: PChar ): SmallInt;
begin
  Result := 0;
  try
    Result := StrToInt(ClientXmit( sc_EVALTEST, Trim(sExpression), True) );
  except
    ClientError('ApolloConnection.EvalTest', 'EvalTest');
  end;
end;



function  TApolloConnection.ap_FieldCount: Word;
begin
  Result := 0;
  try
    Result := StrToInt(ClientXmit( sc_FIELDCOUNT , '', True));
  except
    ClientError('ApolloConnection.FieldCount', 'FieldCount');
  end;
end;


function  TApolloConnection.ap_FieldDecimals ( cpFieldName: PChar ): Word;
begin
  Result := 0;
  try
    Result := StrToInt(ClientXmit( sc_FIELDDECIMALS,   cpFieldName, True));
  except
    ClientError('ApolloConnection.FieldDecimals', 'FieldDecimals');
  end;
end;

function  TApolloConnection.ap_FieldName( uiFieldNum: Word ): PChar;
begin
   Result := '';
   try
      Result := PChar(ClientXmit( sc_FIELDNAME ,  IntToStr(uiFieldNum) , True));
   except
      ClientError('ApolloConnection.FieldName', 'FieldName');
   end;
end;

function  TApolloConnection.ap_FieldNum( cpFieldName: PChar ): Word;
begin
  Result := 0;
  try
    Result := StrToInt(ClientXmit( sc_FieldNum , cpFieldName, True));
  except
    ClientError('ApolloConnection.FieldNum', 'FieldNum');
  end;
end;


function  TApolloConnection.ap_FieldOffset( cpFieldName: PChar ): Word;
begin
   Result := 0;
   try
      Result := StrToInt(ClientXmit( sc_FieldOffset , cpFieldName, True));
   except
      ClientError('ApolloConnection.FieldOffset', 'FieldOffset');
   end;
end;


function  TApolloConnection.ap_FieldType( cpFieldName: PChar ): PChar;
begin
  Result := '';
  try
    Result := PChar(ClientXmit( sc_FieldType, cpFieldName, True));
  except
    ClientError('ApolloConnection.FieldType', 'FieldType');
  end;
end;

function  TApolloConnection.ap_FieldWidth( cpFieldName: PChar ): Word;
begin
  Result := 0;
  try
     Result := StrToInt(ClientXmit( sc_FieldWidth , cpFieldName, True));
  except
    ClientError('ApolloConnection.FieldWidth', 'FieldWidth');
  end;
end;

{MED: 03-07-2000, this function should not be called}
function TApolloConnection.ap_FilterAlias(cpAliasName: PChar; cpFieldName: PChar): WordBool;
begin
  Result := False;
end;

{MED: 03-07-2000, this function should not be called}
function TApolloConnection.ap_FilterDlg(hwnd: HWND; cpExpr: PChar; cpCaption: PChar;
                         bHasIndexList: WordBool): PChar;
begin
  Result := nil;
end;


function  TApolloConnection.ap_FLock: WordBool;
begin
  try
    Result := YNToLogical(ClientXmit( sc_FLOCK , '', True ));
  except
    Result := False;
  end;
end;


function  TApolloConnection.ap_Found: WordBool;
begin
  Result := False;
  try
    if ClientXmit( sc_FOUND, '', True ) = 'Y' then     Result := True;
  except
    ClientError('ApolloConnection.Found', 'Found');
  end;
end;

function TApolloConnection.FileLength(sDataBaseName, sFileName : String): LongInt;
var
  sFile,s  : String;
begin
  Result := 0;
  if (sDataBaseName='') or (sFileName='') then
    exit;

  sFile := ExtractFileName(sFileName);
  try
    s := (ClientXmit( sc_FILELength ,
                                 PadR(sDataBaseName , 255, ' ')+
                                 PadR(sFile, 255, ' ') , True));
    Result := StrToInt('$'+s);
  except
    ClientError('ApolloConnection.FileLength', 'FileLength');
  end;
end;

{MED: 09-30-00, Added}
function TApolloConnection.FileIsInUse(sDataBaseName, sFileName : String): Boolean;
var
  sFile,s  : String;
begin
  Result := FALSE;
  if (sDataBaseName='') or (sFileName='') then
    exit;

  sFile := ExtractFileName(sFileName);
  try
    s := (ClientXmit( sc_FileIsInUse ,
                                 PadR(sDataBaseName , 255, ' ')+
                                 PadR(sFile, 255, ' ') , True));
    Result := YNToLogical(s);
  except
    ClientError('ApolloConnection.FileIsInUse', 'FileIsInUse');
  end;
end;


//====================================================
// CAUTION: Overwrites existing file
// Syntax : FileReceive('DATA', 'PHONE.DBF', '\TMP\PHONELIST.DBF');
//====================================================
function TApolloConnection.FileReceive(sDataBaseName, srcFile, desFile: String): LongInt;
var
  sFile     : String;
  FS        : TFileStream;
begin
  Result    := 0;
  if (sDataBaseName='') or (srcFile='') then
    exit;

  if desFile='' then
    desFile  := ExtractFileName(srcFile);

  sFile := ExtractFileName(srcFile);  // get rid of path on source if any-server doesn't see path
  FS    := TFileStream.Create(desFile, fmCreate);
  try
    try
      FReadStream.Clear;
      ClientStreamReceive( sc_FILERECEIVE , PadR(sDataBaseName, 255, ' ')+ sFile );
      FReadStream.Position := 0;
      if FReadStream.Size > 0 then
        FS.CopyFrom (FReadStream, FReadStream.Size);
      Result := FS.Size;
    except
      ClientError('ApolloConnection.FileReceive', 'Error in stream');
    end;
  finally
    FS.Free;
  end;
end;


// Client must have create/write rights on the server directory
function TApolloConnection.FileSend(sDataBaseName, srcFile, desFile: String): Integer;
var
  sFile : String;
  FS    : TFileStream;
begin
	FS        := nil;
  Result    := 0;

  if (sDataBaseName='') or (srcFile='') then
    exit;

  if desFile='' then
    desFile  := srcFile;

  sFile  := ExtractFileName(desFile);  // get rid of path if any  - server doesn't see path
  if (not SysUtils.FileExists(srcFile)) then
  begin
    ClientError('ApolloConnection.FileSend', 'No such file or directory');
    exit;
  end;

  try
    try
      FS := TFileStream.Create(srcFile, fmOpenRead);
    except
      ClientError('ApolloServer.FileSend', 'Opening source:'+srcFile);
      exit;
    end;

    Try
      FWriteStream.Clear;
      FWriteStream.Position := 0;
      FWriteStream.Write(PChar(PadR(sDataBaseName, 255, ' '))^, 255);
      FWriteStream.Write(PChar(PadR(sFile, 255, ' '))^, 255);
      FWriteStream.CopyFrom(FS, FS.Size);
      FWriteStream.Position := 0;    // this is a must have
      Result := FWriteStream.Size - 510 ;

      {MED: 09-25-2002. Changed from Send to make it wait}
      //Result := ClientStreamSend(sc_FILESEND);
      ClientStreamSendReceive(sc_FILESEND);

      {if Result >= 510 then
        Result := Result - 510
      else
        Result := 0;
      }


      FWriteStream.Clear;
      FWriteStream.Position := 0;
    except
      ClientError('ApolloServer.FileSend', 'Copying from source');
      Result := 0;
    end;
  Finally
    FS.free;
  End;
end;

{MED: 01-08-00
NOTE: this function is pretty handy.  Make sure you synch client's time with server.
Observe that your server could be in a different time zone than your clients.  For example: a client on
the east coast of U.S. connecting to a server on the west coast of U.S.
will have a different time than the server.  Suppose a record is time stamped by a client on the east coast and half
an hour later a second client connects from the mid-west of the U.S.  He/she might see records with time stamps
that have not even been reached in his time-zone.  This can create a lot of confusion.  The best thing to do is
synch everyone with the server. Call this function prior to stamping any records that use time&date }
function TApolloConnection.GetServerDateTime: TDateTime;
begin
  try
    Result := StrToFloat(ClientXmit( sc_GETSERVERDATETIME , '', True ));
  except
    Result := 0;
  end;
end;



// not needed anymore
function TApolloConnection.ap_GetBitMap (cpFieldName: PChar; hwnd: HWND): WordBool;
begin
  Result := False;
  // do nothing
end;

// user MUST Make sure there is enough room for the blob.
// Call ApolloConnection.GetBlobLength first
function  TApolloConnection.ap_GetBlob( cpFieldName: PChar; vpVar: Pointer ): LongInt;
var
  pP : Pointer;
begin
  Result := 0;
  try
    pP := ClientStreamReceive( sc_GETBLOB,  cpFieldName );
    CopyMemory(vpVar, TMemoryStream(pP).Memory, TMemoryStream(pP).Size);
    Result := TMemoryStream(pP).Size;
  except
    ClientError('ApolloConnection.GetBlob', 'GetBlob');
  end;
end;

function  TApolloConnection.ap_GetBlobLength( cpFieldName: PChar ): LongInt;
begin
  Result := 0;
  try
    Result := StrToInt(ClientXmit( sc_GETBLOBLength , cpFieldName , True));
  except
    ClientError('ApolloConnection.GetBlobLength', 'GetBlobLength');
  end;
end;

function TApolloConnection.ap_GetByte( cpFieldName: PChar ): PChar;  {LAS: 11-07-99, Coded}
begin
  Result := '';
  try
    Result := ClientXmit( sc_GETBYTE , cpFieldName , True);
  except
    ClientError('ApolloConnection.GetByte', 'GetByte');
  end;
end;

function  TApolloConnection.ap_GetDateJulian( cpFieldName: PChar ): LongInt;
begin
  Result := 0;
  try
    Result := StrToInt('$'+ClientXmit( sc_GETDATEJULIAN, cpFieldName, True));
	except
    ClientError('ApolloConnection.GetDateJulian', 'GetDateJulian');
  end;
end;


function  TApolloConnection.ap_GetDateString( cpFieldName: PChar): PChar;
begin
  Result := '';
  try
    Result := ClientXmit( sc_GETDATESTRING, cpFieldName, True);
	except
    ClientError('ApolloConnection.GetDateString', 'GetDateString');
  end;
end;


function TApolloConnection.ap_GetDouble(cpFieldName: PChar): Double;
begin
  Result := 0.00;
  try
    Result := StrToFloat(ClientXmit( sc_GETDOUBLE, cpFieldName, True));
  except
    ClientError('ApolloConnection.GetDouble', 'GetDouble');
  end;
end;


function  TApolloConnection.ap_GetInteger(cpFieldName: PChar): SmallInt;
begin
  Result := 0;
  try
    Result := StrToInt(ClientXmit( sc_GETINTEGER, cpFieldName, True));
  except
    ClientError('ApolloConnection.GetInteger', 'GetInteger');
  end;
end;

function  TApolloConnection.ap_GetLogical(cpFieldName: PChar): WordBool;
begin
  Result := False;
  try
    Result := YNToLogical(ClientXmit( sc_GETLOGICAL, cpFieldName, True));
  except
    ClientError('ApolloConnection.GetLogical', 'GetLogical');
  end;
end;

function  TApolloConnection.ap_GetLong(cpFieldName: PChar): LongInt;
begin
   Result := 0;
   try
      Result := StrToInt(ClientXmit( sc_GETLONG , cpFieldName , True));
   except
      ClientError('ApolloConnection.GetLong', 'GetLong');
   end;
end;

function  TApolloConnection.ap_GetMemo( cpFieldName: PChar; uiLineWidth: Word  ): PChar;   {LAS: 07-16-00, Added iRecNo param}
begin
  Result := nil;
  try
    Result := PChar(ClientXmit( sc_GETMEMO,
											IntToHex(0, 8) +     {LAS: 07-16-00, Added iRecNo param}
                     	IntToHex(uiLineWidth, 8)+
                     	cpFieldName, True));
  except
    ClientError('ApolloConnection.GetMemo', 'GetMemo');
  end;
end;


function  TApolloConnection.ap_PutMemo( cpFieldName: PChar; vpData:PChar):Integer;
begin
  Result := 0;
  try
    Result := StrToInt(ClientXmit( sc_PUTMEMO ,
                     PadR(cpFieldName,10, ' ') +
                     PChar(vpData), True));
  except
    ClientError( 'ApolloConnection.PutMemo', 'PutMemo');
  end;
end;

function  TApolloConnection.ap_GetString(cpFieldName: PChar): PChar;
begin
  Result := '';
  try
    Result := ClientXmit( sc_GETSTRING, cpFieldName, True);
  except
    ClientError('ApolloConnection.GetString', 'GetString');
  end;
end;
////////////////////////////////
function  TApolloConnection.ap_GetStructure:Pointer;
begin
  Result := nil;
  try
    ClientStreamReceive( sc_GETSTRUCTURE, '');
    Result := FReadStream;
  except
    ClientError('ApolloConnection.GetStructure', 'GetStructure');
  end;
end;

function  TApolloConnection.ap_GetTrimString(cpFieldName: PChar ): PChar;  {LAS: 06-09-00, Added iRecNo param}{MED:08-29-00, changed back}
begin
	Result := '';
	try
		Result := PChar(ClientXmit( sc_GETTRIMSTRING,  cpFieldName, True));  {LAS: 06-09-00, Added iRecNo param}{MED:08-29-00, changed back}
	except
		ClientError('ApolloConnection.GetTrimString', 'GetTrimString');  {LAS: 06-09-00, Changed from .GetString}{MED:08-29-00, changed back}
  end;
end;

function TApolloConnection.ap_GetQueryBit( lRecNum : LongInt ) : WordBool;
begin
  Result := False;
  try
    Result := ClientXmit( sc_GETQUERYBIT, IntToStr(lRecNum), True) = 'Y';
  except
    ClientError('ApolloConnection.GetQueryBit', 'GetQueryBit');
  end;
end;


function  TApolloConnection.ap_GetScope( iWhichScope: SmallInt ): PChar;
begin
  Result := nil;
  try
    Result := ClientXmit( sc_GETSCOPE, IntToStr(iWhichScope), True);
  except
    ClientError('ApolloConnection.GetScope', 'GetScope');
  end;
end;

{LAS: 06-01-00, Changed to return order position}
function TApolloConnection.ap_Go( lRecNum: Integer): Integer;
begin
	Result := 0;
  try
    Result := StrToInt( '$' + ClientXmit( sc_GO, IntToHex(lRecNum,8), True ));
  except
    ClientError('ApolloConnection.Go', 'Go');
  end;
end;

procedure TApolloConnection.ap_GoBottom;
begin
  try
    ClientXmit( sc_GOBOTTOM , '', TRUE);
  except
    ClientError('ApolloConnection.GoBottom', 'GoBottom');
  end;
end;

procedure TApolloConnection.ap_GoTop;
begin
  try
    ClientXmit( sc_GOTOP, '', TRUE );
  except
    ClientError('ApolloConnection.GoTop', 'GoTop');
  end;
end;

function TApolloConnection.ap_GetFields:PChar;
begin
  Result := nil;
  try
    Result := ClientXmit(sc_READFIELDS, '', True);
  except
    ClientError('ApolloConnection.GetFields', 'GetFields');
  end;
end;

procedure TApolloConnection.ap_SetFields(FFields:PChar);
begin
  try
    ClientXmit(sc_SETFIELDS , FFields, TRUE);
  except
    ClientError('ApolloConnection.SetFields', 'SetFields');
  end;
end;

function TApolloConnection.ap_GetPage(Count:Integer):PChar;
begin
	Result := nil;
	try
		Result := ClientXmit(sc_GETPAGE, IntToHex(Count, 4), True);
  except
		ClientError('ApolloConnection.GetPage', 'GetPage');
	end;
end;


function TApolloConnection.ap_TraverseTable(
                          FirstCrack, LastCrack:WordBool;
                          FEOFAction:Integer;
                          GetMode: TGetMode;
                          var RecNo : Integer;
                          LSDERecordBuffer:PChar;
                          iRecSize:Integer;
                          FetchCount:Integer): TGetResult;
var
  pMem : TMemoryStream;
  Buff : Array [0..8] of char;
begin
	Result := TGetResult(0);
  try
    pMem := ClientStreamReceive(sc_TRAVERSETABLE,
                                  LogicalToYN(FirstCrack)+
                                  LogicalToYN(LastCrack)+
                                  IntToStr(FEOFAction)+
                                  IntToStr(Ord(GetMode))+
                                  IntToHex(FetchCount, 8));
    if pMem <> nil then
    begin
      pMem.Read( (@buff[0])^, 1);      buff[1] := #0;   // Result
      Result := TGetResult(StrToInt(Buff));

      pMem.Read( pchar(@buff)^, 8);      buff[8] := #0; // Record Number
      RecNo         := StrToInt('$'+buff);

      pMem.Read( pchar(LSDERecordBuffer)^, iRecSize);   // Active Buffer itself
    end;
  except
    ClientError( 'ApolloConnection.TraverseTable',
                 'Make sure DatabaseName is set to a valid alias.' );
  end;
end;

function TApolloConnection.ap_MultiTraverseTable(
                          var ReadCount:Integer;
                          FirstCrack, LastCrack:WordBool;
                          FEOFAction:Integer;
                          GetMode, LastGetMode: TGetMode;
                          LSDERecordBuffer:PChar;
                          iRecSize:Integer;
                          FetchCount:Integer;
                          LastRecNo :Integer;
                          var FFetchRecs:TList): TGetResult;
var
  i, RecOrdinal, RecNo: Integer;
  pMem : TMemoryStream;
  Buff : Array [0..8] of char;
begin
	Result := TGetResult(0);
  try
    pMem := ClientStreamReceive(sc_MULTITRAVERSETABLE,
                                  LogicalToYN(FirstCrack)+
                                  LogicalToYN(LastCrack)+
                                  IntToStr(FEOFAction)+
                                  IntToStr(Ord(GetMode))+
                                  IntToStr(Ord(LastGetMode))+
                                  IntToHex(FetchCount, 8)+
                                  IntToHex(LastRecNo, 8));
    if pMem <> nil then
    begin
      pMem.Read( pchar(@buff)^, 8);      buff[8] := #0;   // Total Read Count
      ReadCount := StrToInt('$'+Buff);
      for i:=0 to ReadCount-1 do
      begin
        pMem.Read( (@buff[0])^, 1);      buff[1] := #0;   // Result
        Result := TGetResult(StrToInt(Buff));

        pMem.Read( pchar(@buff)^, 8);      buff[8] := #0; // Record Number
        RecNo         := StrToInt('$'+buff);

        pMem.Read( pchar(@buff)^, 8);      buff[8] := #0; // Ordinal Record Number
        RecOrdinal     := StrToInt('$'+buff);

        pMem.Read( pchar(LSDERecordBuffer)^, iRecSize);   // Active Buffer itself
        CopyMemory(PFetchList(FFetchRecs.Items[i])^.Data, LSDERecordBuffer, iRecSize);
        PFetchList(FFetchRecs.Items[i])^.RecNum     := RecNo;
        PFetchList(FFetchRecs.Items[i])^.RecOrdinal := RecOrdinal;
        PFetchList(FFetchRecs.Items[i])^.GetResult  := Result;
      end;
    end;
  except
    ClientError( 'ApolloConnection.TraverseTable',
                 'Make sure DatabaseName is set to a valid alias.' );
  end;
end;

//=============================================================
//

function TApolloConnection.ap_GoRecord(const lRecNum: Integer):TMemoryStream;
begin
  Result := Nil;
  try
    Result := ClientStreamReceive(sc_GORecord,  IntToHex(lRecNum, 8));
  except
    ClientError( 'ApolloConnection.GoRecord', '' );
  end;
end;

// lRecNum = sx_Go(lRecNum)
// OffSet = sx_Skip(OffSet)
function TApolloConnection.ap_GoSkip( const lRecNum: Integer; const OffSet:Integer):Integer;
var
// CRVG   sPack : String;
  pMem : TMemoryStream;
  pBuff : PChar;
begin
  Result := 0;
  //sPack := IntToHex(lRecNum, 8) + IntToHex(OffSet, 8);
  try
    FWriteStream.Clear;
    FWriteStream.Position := 0;
    FWriteStream.Write(pchar(IntToHex(lRecNum, 8))^, 8);
    FWriteStream.Write(pchar(IntToHex(OffSet, 8))^, 8);

    pMem        := ClientStreamSendReceive(sc_GoSkip);
    if pMem <> nil then
    begin
      pBuff := NIL;
      Try
        pBuff := pchar(AllocMem(pMem.Size+1));
        pMem.Read( pchar(@pbuff[0])^, 8);pbuff[8] := #0;         // Result
        Result := StrToInt('$'+pbuff);
      Finally
        FreeMem(pBuff);
      End;
    end;

{
    Result := StrToInt('$'+ClientXmit(sc_GOSKIP, IntToHex(lRecNum, 8)+  IntToHex(OffSet, 8), TRUE));
 }
  except
    ClientError( 'ApolloConnection.GoSkip', '' );
  end;
end;

function TApolloConnection.ap_GetRowNo:Integer;
var
  s : String; // dummy
  pMem : TMemoryStream;
  pBuff : PChar;
begin
  Result := 0;
  s := '1';
  //showmessage('GetRowNo');
  try
    FWriteStream.Clear;
    FWriteStream.Position := 0;
    FWriteStream.Write(pchar(s)^, 1);
    pMem        := ClientStreamSendReceive(sc_GetRowNo);
    if pMem <> nil then
    begin
      pBuff := NIL;
      Try
        pBuff := pchar(AllocMem(pMem.Size+1));
        pMem.Read( pchar(@pbuff[0])^, 8);pbuff[8] := #0;         // Result
        Result := StrToInt('$'+pbuff);
      Finally
        FreeMem(pBuff);
      End;
    end;
  except
    ClientError( 'ApolloConnection.GetRowNo', '' );
  end;
end;

//MED: 01-29-02, added
function TApolloConnection.ap_GoToRow( const RowNo: Integer ):Integer;
var
  pMem : TMemoryStream;
  pBuff : PChar;
begin
  Result := 0;
  //sPack := IntToHex(lRecNum, 8) + IntToHex(OffSet, 8);
//showmessage('GoSkip');
  try
    FWriteStream.Clear;
    FWriteStream.Position := 0;
    FWriteStream.Write(pchar(IntToHex(RowNo, 8))^, 8);
    pMem        := ClientStreamSendReceive(sc_GoToRow);
    if pMem <> nil then
    begin
      pBuff := NIL;
      Try
        pBuff := pchar(AllocMem(pMem.Size+1));
        pMem.Read( pchar(@pbuff[0])^, 8);pbuff[8] := #0;         // Result
        Result := StrToInt('$'+pbuff);
      Finally
        FreeMem(pBuff);
      End;
    end;

  except
    ClientError( 'ApolloConnection.GoToRow', '' );
  end;

exit;

  try
    Result := StrToInt('$'+ClientXmit(sc_GOSKIP,  IntToHex(RowNo, 8), TRUE));
  except
    ClientError( 'ApolloConnection.GoToRow', '' );
  end;
end;


function TApolloConnection.ap_GoGetRecord:TMemoryStream;
begin
  Result := nil;
  try
    Result := ClientStreamReceive(sc_GoGetRecord, '');
  except
    ClientError( 'ApolloConnection.GoGetRecord',   '' );
  end;
end;

function TApolloConnection.ap_GoBofRecord:TMemoryStream;
begin
  Result := nil;
  try
    Result := ClientStreamReceive(sc_GoBofRecord, '');
  except
    ClientError( 'ApolloConnection.GoBofRecord', '');
	end;
end;

function TApolloConnection.ap_GoEofRecord:TMemoryStream;
begin
  Result := nil;
  try
    Result := ClientStreamReceive(sc_GoEofRecord, '');
  except
    ClientError( 'ApolloConnection.GoEofRecord',   '' );
  end;
end;

function TApolloConnection.ap_GoBottomRecord:TMemoryStream;
begin
  Result := nil;
  try
    Result := ClientStreamReceive(sc_GoBottomRecord, '');
  except
    ClientError( 'ApolloConnection.GoBottomRecord', '');
	end;
end;

function TApolloConnection.ap_GoTopRecord:TMemoryStream;
begin
  Result := nil;
  try
    Result := ClientStreamReceive(sc_GoTopRecord, '');
  except
    ClientError( 'ApolloConnection.GoTopRecord',   '' );
  end;
end;


// See GetRows
{
 _____________________   Stream's first 24 bytes-Always present even when empty
|iSize           : 8 bytes   Size of the stream
|iCount          : 8 bytes   (hex) # of records in the stream for this table
|____________________________________________________= 16 minimum bytes each stream header
|FRecCount       : 8 Bytes   (hex) # of records sx_RecCount
|RecLength       : 8 Bytes length starts @ DATA(includes all blobs)
|____________________________________________________= 32 each stream header
|PhysicalRec No  : 8 bytes            // -1 bof, -2 eof  otherwise good record
|OrdinalRec No   : 8 bytes
|BlobCount       : 4 Bytes  0 if none
|____________________________________= 8+8 +4 = 20 bytes each record header
|DATA            : FRecSize bytes
 _________________________________________________________
  |
  |BlobField       : 20 Bytes  -- only if BlobCount > 0
  |BlobLength      : 8 Bytes
  |BlobData        : Variable bytes
  |................
  |BlobField       : 20 Bytes
  |BlobLength      : 8 Bytes
  |BlobData        : Variable bytes
  |..............
  |___________________

//////////////////////////
}{
function TApolloConnection.ap_SDEFetchRecords(
                          AccessMethod : TAccessMethod;
                          iDirection: TSkipDirection;    // --- Skip(1) or Skip(-1)
                          var FBOF  : Boolean;
                          var FEOF  : Boolean;
                          FRecSize:Integer;
                          var FRecCount :Integer;             // sx_RecCount
                          var FIndexOrd : Integer;             // sx_IndexOrd
                          FetchCount:Integer;
                          LastRecNo :Integer;
                          lNumRecs : Integer;             // skip this many before start reading
                          var FFetchRecs:TList;
                          RecIndex : Integer;
                          StackOrder:TStackOrder;  // -1 for all
                          FSetDeleted:Boolean):Integer;
var
  iSize, iCount, i, j,
  FRecNum,
  iBlobCount,
  iRecLen,
  iBlobReadCount,
  FFieldNo : Integer;
  pMem : TMemoryStream;
  Buff : Array [0..20] of char;
  ActBuffer : PChar;
  BlobName : String;
  MemStream : TMemoryStream;
begin
	Result    := 0;
  FEOF      := FALSE;
  FBOF      := FALSE;
  MemStream := TMemoryStream.Create;
  Try
    try
      if AccessMethod = amLocal then
      begin
        InternalSDEFetchRecords(MemStream, FetchCount, LastRecNo, lNumRecs, iDirection, FSetDeleted);
        pMem := MemStream;
      end
      else
        pMem := ClientStreamReceive(sc_SDEFetchRecords,
                                    IntToHex(FetchCount, 8)+
                                    IntToHex(LastRecNo, 8) +
                                    IntToHex(lNumRecs, 8) +
                                    IntToHex(Ord(iDirection), 1));

      if pMem <> nil then
      begin
        //_____________________________________________________                 Size of the stream
        pMem.Read( pchar(@buff)^, 8); buff[8] := #0;
        //StreamSize := StrToInt('$'+Buff);

        //_____________________________________________________                 Total Read Count
        pMem.Read( pchar(@buff)^, 8); buff[8] := #0; iCount     := StrToInt('$'+Buff);
        //_______________________________________________________               FRecCount
        pMem.Read( pchar(@buff)^, 8); buff[8] := #0; FRecCount  := StrToInt('$'+Buff);

        //________________________________________________________               FRecLength this we don't set currently
        pMem.Read( pchar(@buff)^, 8); buff[8] := #0;
        //FRecLength := StrToInt('$'+Buff);

        //_____________________________________________
        pMem.Read( pchar(@buff)^, 4); buff[4] := #0; FIndexOrd := StrToInt('$'+Buff);

        //if FRecCount < 1 then exit;
        //if iCount < 1 then exit;
        IF (iCount>0) and (FRecCount>0) then
        BEGIN
          ActBuffer   := AllocMem(FRecSize);
          Try
            iRecLen     := 0;

            if (StackOrder = soSingle) then
              //RecIndex <> -1 then            // a single record was requested
              j := RecIndex-1
            else if (StackOrder = soReverse ) then
              j := iCount -1
            else
              j := 0;

            FOR i := 0 to iCount-1 do
            begin
              //iBlobReadCount := 0;

              pMem.Read( (@buff[0])^, 1);      buff[1] := #0;
              FBOF  := YNToLogical(Buff);
              PSkipList(FFetchRecs.Items[j])^.Bof :=  YNToLogical(Buff);

              pMem.Read( (@buff[0])^, 1);      buff[1] := #0;
              //FEOF  := YNToLogical(Buff);
              PSkipList(FFetchRecs.Items[j])^.Eof    := YNToLogical(Buff);

              pMem.Read( pchar(@buff)^, 8);      buff[8] := #0; // Physical Record Number
              FRecNum := StrToInt('$'+buff);

              PSkipList(FFetchRecs.Items[j])^.RecNum     := FRecNum;

              pMem.Read( pchar(@buff)^, 8);      buff[8] := #0; // Ordinal Record Number
              PSkipList(FFetchRecs.Items[j])^.RecOrdinal := StrToInt('$'+buff); //RecOrdinal;
              pMem.Read( pchar(@buff)^, 4);      buff[4] := #0;
              iBlobCount   := StrToInt('$'+buff);  //any blobs??

              pMem.Read( PChar(PSkipList(FFetchRecs.Items[j])^.Data)^, FRecSize);   // Active Buffer itself

              //CopyMemory(PChar(PSkipList(FFetchRecs.Items[j])^.Data), pchar('CCCCC'), 5);
              //pMem.Position := pMem.Position + FRecSize;

              iRecLen := iRecLen + 54 + FRecSize;
              // see if there are any blobs
              if (iBlobCount> 0) then
              begin
                iBlobReadCount := iBlobCount;
                while (iBlobReadCount > 0) do
                begin
                  pMem.Read( pchar(@buff)^, 20);      buff[20] := #0;
                  BlobName := UpperCase(Trim(Buff));
                  pMem.Read( pchar(@buff)^, 8);      buff[8] := #0;
                  iSize    := StrToInt('$'+buff);
                  pMem.Read( pchar(@buff)^, 4);      buff[4] := #0;
                  FFieldNo    := StrToInt('$'+buff);
                  if Assigned(PSkipList(FFetchRecs.Items[j])^.Blobs[FFieldNo]) then
                    FreeMem(PSkipList(FFetchRecs.Items[j])^.Blobs[FFieldNo]);
                  PSkipList( FFetchRecs.Items[j])^.Blobs[FFieldNo] := AllocMem(iSize+1);  //<--watch the extra one - blobs may not like this
                  pMem.Read( PChar(PSkipList(FFetchRecs.Items[j])^.Blobs[FFieldNo])^, iSize);   // Active Buffer itself
                  iRecLen := iRecLen + 32 + iSize;
                  dec(iBlobReadCount);
                end;
              end;
              inc(Result);
              if (StackOrder = soSingle) then                break
              else if (StackOrder = soReverse ) then         dec(j)
              else                                           inc(j);
            End;  // FOR
          Finally
            FreeMem(ActBuffer);
          End;
        END;  // IF
      End;
    except
      ClientError( 'ApolloConnection.SDEFetchRecords','Error in reading stream.' );
    end;
  Finally
    MemStream.Free;
  End;
end;
}

procedure TApolloConnection.ap_GetRecord(cpRecord:PChar; iRecSize:Integer);
var
  pMem : TMemoryStream;
begin
  try
    pMem := ClientStreamReceive(sc_READRECORD, '');
    if pMem <> nil then
    begin
      pMem.Read( PChar(cpRecord)^, iRecSize);         // Active Buffer itself
    end;
  except
    ClientError('ApolloConnection.GetRecord', 'GetRecord');
  end;
end;


(*

// 10-12-99 special function written to speed up GetRecord
// see ApoDSet .GetRecord
// some strange reasons TEOFAction is not recognized by ApoDSET so we just pass the ORD()
// MED: 03-21-00, changed to make one call instead
// added RecNo and LSDERecordBuffer.  The function still returns TGetResult
// but the structure of return data has been changed.
// first byte = GetResult
// 8     bytes= record number
// *     bytes= data
function TApolloConnection.ap_TraverseTable(
                          FirstCrack, LastCrack:WordBool;
                          FEOFAction:Integer;
                          GetMode: TGetMode;
                          var RecNo : Integer;
                          LSDERecordBuffer:PChar;
                          iRecSize:Integer): TGetResult;
var
  sGetRes : SmallInt;
  pRec : PChar;
begin
	Result := TGetResult(0);
  try

    pRec := ClientXmit( sc_TRAVERSETABLE,
                LogicalTOYN(FirstCrack)+
                LogicalTOYN(LastCrack)+
                IntToStr(FEOFAction)+
                IntToStr(Ord(GetMode))
                , True);


//    Result        := TGetResult(StrToInt(pRec[0]));
//    RecNo         := StrToInt('$'+Substr(pRec, 2, 8));
    FActiveBuffer := @pRec[0]; //ClientXmit(sc_READRECORD, '', True);
    CopyMemory(LSDERecordBuffer, FActiveBuffer, iRecSize); {MED:03-22-00, changed StrLen(FActiveBuffer) );}

//////////////////

                 {
    Result := TGetResult( StrToInt(ClientXmit( sc_TRAVERSETABLE,
                LogicalTOYN(FirstCrack)+
                LogicalTOYN(LastCrack)+
                IntToStr(FEOFAction)+
                IntToStr(Ord(GetMode))
                , True)));
                }
  except
    ClientError( 'ApolloConnection.TraverseTable',
                 'Make sure DatabaseName is set to a valid alias.' );
  end;
end;



// same as sx_GetRecord
procedure TApolloConnection.ap_GetRecord(cpRecord:PChar; iRecSize:Integer);
begin
  try
    FActiveBuffer := ClientXmit(sc_READRECORD, '', True);
    CopyMemory(cpRecord, FActiveBuffer, iRecSize); {MED: 03-22-00, Changed StrLen(FActiveBuffer) )}
  except
    ClientError('ApolloConnection.GetRecord', 'GetRecord');
  end;
end;
*)
function  TApolloConnection.ap_Select( uiBaseArea: Integer ): Integer;

begin
  Result := nCurWorkArea;
  try
    if nCurWorkArea <> uiBaseArea then
    begin
      ClientXmit( sc_SELECT , IntToHex(uiBaseArea, 4), True);
      nCurWorkArea := uiBaseArea;
    end;
  except
    Result := 0;
    ClientError('ApolloConnection.Select', 'Select');
  end;
end;


// we convert to hex to keep the server ReadStream.Size the same
procedure TApolloConnection.ap_Skip(lNumRecs:LongInt);
begin
	try
		ClientXmit( sc_SKIP, IntToHex(lNumRecs, 8), TRUE);
	except
		ClientError('ApolloConnection.Skip', 'Skip');
	end;
end;

function  TApolloConnection.ap_Index(cpFileName: PChar; cpExpr: PChar; iOption: SmallInt;
                     bDescend: WordBool; cpCondition: PChar): SmallInt;
begin
	try
		Result :=  StrToInt(ClientXmit( sc_INDEX,
                     PadR(cpFileName, 255, ' ')+
                     PadR(cpExpr, 255, ' ')+
                     PadR(IntToStr(iOption), 10, ' ')+
                     LogicalTOYN(bDescend)+
                     PadR(cpCondition, 255, ' '),
                     True));
	except
		Result := 0;
	end;
end;


function  TApolloConnection.ap_IndexCondition: PChar;
begin
  try
   	Result := ClientXmit( sc_INDEXCONDITION , '', True);
  except
    Result := nil;
  end;
end;

function  TApolloConnection.ap_IndexFlip : WordBool;
begin
  try
   	Result := YNToLogical(ClientXmit( sc_INDEXFLIP , '', True));
  except
		Result := False;
  end;
end;

// 05-03-99 sPath taken out to make method compatible
// Assumes file is opened for EXCLUSIVE use
// cpFileName should NOT contain any path just the table name
function  TApolloConnection.ap_IndexTag (cpFileName: PChar; cpTagName: PChar; cpExpr: PChar;
                        iOption: SmallInt; bDescend: WordBool;
                        cpCondition: PChar): SmallInt;
begin
	try
		Result := StrToInt(PChar(ClientXmit( sc_INDEXTAG,
                     PadR(cpFileName, 255, ' ')+
                     PadR(cpTagName, 255, ' ')+
                     PadR(cpExpr, 255, ' ')+
                     PadR(IntToStr(iOption), 10, ' ')+
                     LogicalTOYN(bDescend)+
                     PadR(cpCondition, 255, ' '),
                     True) ));
	except
		Result := 0;
	end;
end;


procedure TApolloConnection.ap_IndexClose;
begin
	try
		ClientXmit( sc_INDEXCLOSE , '', True);
	except
		ClientError('ApolloConnection.IndexClose', 'IndexClose');
	end;
end;

function  TApolloConnection.ap_IndexCount: Integer;
begin
	Result := 0;
	try
		Result := StrToInt(ClientXmit( sc_INDEXCOUNT, '', True ));
	except
		ClientError('ApolloConnection.IndexCount', 'IndexCount');
	end;
end;

function  TApolloConnection.ap_IndexDescending:Integer;
begin
	Result := 0;
	try
		Result := StrToInt(Trim(ClientXmit( sc_INDEXDESCEND, '', True)));
	except
		ClientError('ApolloConnection.IndexDescending', 'Error in IndexDescending')
	end;
end;


function  TApolloConnection.ap_IndexKey: PChar;
begin
  Result := '';
  try
    Result := ClientXmit( sc_INDEXKEY , '', True);
  except
    ClientError('ApolloConnection.IndexKey', 'IndexKey');
  end;
end;

function  TApolloConnection.ap_IndexKeyField: PChar;
begin
	Result := '';
	try
		Result := ClientXmit( sc_INDEXKEYFIELD, '', True);
	except
		ClientError('ApolloConnection.IndexKeyField', 'IndexKeyField');
	end;
end;

function TApolloConnection.ap_IndexName( iIndex: SmallInt ): PChar;
begin
  Result := '';
  try
    Result := ClientXmit( sc_INDEXNAME, IntToStr(iIndex) , True);
  except
    ClientError('ApolloConnection.IndexName', 'IndexName')
  end;
end;

// get a list of indicies from the server direcotry
function TApolloConnection.GetIndexNames(sTableName:String; iTableType:SmallInt): Pointer;
//var NewList : TStrings;
begin
  Result := nil;
  try
    Result := ClientStreamReceive( sc_GETIndexNames ,
                                   PadR(FDataBaseName,255, ' ')+
                                   PadR(sTableName, 255, ' ')+IntToStr(iTableType) );
  except
    ClientError('ApolloConnection.GetIndexNames', 'GetIndexNames');
  end;
end;

{LAS: 11-28-99, Implemented}
function TApolloConnection.ap_IndexOpen(cpFileName: PChar): SmallInt;
begin
  Result := 0;
  try
    Result := StrToInt(Trim(ClientXmit( sc_INDEXOPEN,
                                 PadR( FDataBaseName, 255, ' ') +
                                 PadR( String(cpFileName), 255, ' '), True)));
  except
    ClientError('ApolloConnection.IndexOpen', 'IndexOpen');
  end;
end;

function  TApolloConnection.ap_IndexOrd: SmallInt;
begin
  Result := 0;
  try
    Result := StrToInt(Trim(ClientXmit( sc_INDEXORD , '', True)));
  except
    ClientError('ApolloConnection.IndexOrd', 'IndexOrd');
  end;
end;


function  TApolloConnection.ap_IndexType: SmallInt;
begin
  Result := 0;
  try
    Result := StrToInt(Trim(ClientXmit( sc_INDEXTYPE , '', True)));
  except
    ClientError('ApolloConnection.IndexType', 'IndexType');
  end;
end;


function  TApolloConnection.ap_IsEncrypted( iFileOrRec: SmallInt ): WordBool;
begin
  Result := False;
  try
    Result := ClientXmit( sc_ISENCRYPTED, IntToStr(iFileOrRec), True) = 'Y';
  except
    ClientError('ApolloConnection.IsEncrypted', 'IsEncrypted');
  end;
end;


function TApolloConnection.ap_KeyAdd(cpTagname: PChar): WordBool;
begin
  Result := False;
  try
    Result := ClientXmit( sc_KEYADD, Trim(cpTagName), True) = 'Y';
  except
    ClientError('ApolloConnection.KeyAdd', 'KeyAdd');
  end;
end;


{LAS: 11-19-99, implemented}
function  TApolloConnection.ap_KeyData: PChar;
begin
   Result := '';
   try
      Result := ClientXmit( sc_KEYDATA, '', True);
   except
      ClientError('ApolloConnection.KeyData', 'KeyData');
   end;
end;

function  TApolloConnection.ap_KeyDrop(cpTagName : PChar) : WordBool;
begin
  Result := False;
  try
    Result := ClientXmit( sc_KEYDROP, Trim(cpTagName), True) = 'Y';
  except
    ClientError('ApolloConnection.KeyDrop', 'Error in KeyDrop');
  end;
end;
function  TApolloConnection.ap_Locate(cpExpression: PChar; iDirection: WordBool;
                      bContinue: WordBool): LongInt;
var
  cBack, cCont, msg: String;
begin
  Result := 0;
  cBack := 'N';
  cCont := 'N';

  if iDirection then
  	cBack := 'Y';
  if bContinue then
  	cCont := 'Y';

  try
    msg := ClientXmit( sc_LOCATE , cBack + cCont + cpExpression, True);
    if msg[1] <> '0' then
      Result := StrToInt(Trim(substr(msg, 1, 10)));
  except
    ClientError('ApolloConnection.Locate', 'Locate');
  end;
end;

function TApolloConnection.ap_LockCount: Word;
begin
  Result := 0;
  try
    Result := StrToInt(Trim(ClientXmit( sc_LOCKCOUNT , ''  , True)));
  except
    ClientError('ApolloConnection.LockCount', 'LockCount');
  end;
end;


procedure TApolloConnection.ap_DBRlockList( ulpArray: PChar );
var
  pP : Pointer;
begin
  try                                                     
    pP := ClientStreamReceive( sc_DBRLOCKLIST,  '' );
    if not (pP = nil) then   
	    TMemoryStream(pP).Read((@ulpArray[0])^, TMemoryStream(pP).Size); //, ulpArray, TMemoryStream(pP).Size);
  except
    ClientError('ApolloConnection.DBRLockList', 'DBRLockList');
  end;
end;

function TApolloConnection.ap_Locked( lRecNum: LongInt ): WordBool;
begin
  Result := False;
  try
    Result := ClientXmit( sc_LOCKED, IntToStr(lRecNum)  , True) = 'Y';
  except
    ClientError('ApolloConnection.Locked', 'Locked ');
  end;
end;


function TApolloConnection.ap_MemAlloc( lNumberOfBytes: Integer ): PChar ;
begin
  //  no action required here
  Result := NIL;
end;

procedure TApolloConnection.ap_MemDealloc(vpPtr: Pointer);
begin
//  no action required here
end;

function TApolloConnection.ap_OpenMode: SmallInt;
begin
  Result := 0;
  try
    Result := StrtoInt(ClientXmit( sc_OPENMODE, ''  , True) );
  except
    ClientError('ApolloConnection.OpenMode', 'OpenMode ');
  end;
end;

function  TApolloConnection.ap_OrderPosGet: Double;
begin
  Result := 0.00;
  try
    Result := StrToFloat(ClientXmit( sc_ORDERPOSGET , '', True));
  except
    ClientError('ApolloConnection.OrderPosGet', 'OrderPosGet');
  end;
end;

procedure TApolloConnection.ap_OrderPosSet(dPosition: Double);
begin
  try
    ClientXmit( sc_ORDERPOSSET , FloatToStr(dPosition), TRUE);
  except
    ClientError('ApolloConnection.OrderPosSet','OrderPosSet');
  end;
end;

function TApolloConnection.ap_OrderRecNo: LongInt;
begin
  Result := 0;
  try
    Result := StrToInt(ClientXmit( sc_ORDERRECNO  , '', True));
  except
    ClientError('ApolloConnection.OrderRecNo', 'OrderRecNo');
  end;
end;


procedure TApolloConnection.ap_Pack;
begin
  try
    ClientXmit( sc_PACK , '', True);
  except
    ClientError('ApolloConnection.Pack', 'Pack');
  end;
end;

function TApolloConnection.ap_PutBlob( cpFieldName: PChar; vpVar: Pointer; lSize: LongInt ): LongInt;
var
  sSize : String;
begin
  sSize := IntToHex(lSize, 8);
  Result := 0;
  try
    FWriteStream.Clear;
    FWriteStream.Position := 0;
    FWriteStream.Write(PChar(PadR(cpFieldName, 10, ' '))^, 10);
    FWriteStream.Write(PChar(sSize)^, 8);
    FWriteStream.Write(vpVar^, lSize);
    FWriteStream.Position := 0;     // this is a must have
    ClientStreamSendReceive(sc_PUTBLOB);
    FWriteStream.Position := 0;    // this is a must have

    Result := StrToInt('$'+ClientXmit( sc_PUTBLOBBYTES, '', True));
  except
    ClientError('ApolloConnection.PutBlob', 'PutBlob');
  end;
end;

//function ap_Query(cpExp:PChar):Integer;
function TApolloConnection.ap_Query(cpExpression: PChar): LongInt;
begin
  Result := 0;
  try
    Result := StrToInt(Trim(ClientXmit( sc_QUERY, cpExpression , True)));
  except
    ClientError('ApolloConnection.Query', 'Query');
  end;
end;


//MED: 03-05-2002, added to speed up the query method
function TApolloConnection.ap_QueryRowCount(cpExpression: PChar; var RowCount:integer): LongInt;
var
  s : String;
begin
  Result := 0;
  RowCount:= 0;
  try
    s       := Trim(ClientXmit( sc_QUERYRowCount, cpExpression , True));
    Result  := StrToInt('$'+substr(s, 1, 8));
    RowCount:= StrToInt('$'+substr(s, 9, 8));
  except
    ClientError('ApolloConnection.QueryRowCount', 'QueryRowCount');
  end;

end;

function TApolloConnection.ap_QueryRecCount: LongInt;
begin
  Result := 0;
  try
    Result := StrToInt(ClientXmit( sc_QUERYRECCOUNT , '', True));
  except
    ClientError('ApolloConnection.QueryRecCount', 'QueryRecCount');
  end;
end;

// MED: 04-12-01, added
function  TApolloConnection.ap_QuerySetExact(bSetQueryExact: LongBool): LongBool ;
begin
  Result := FALSE;
  try
    Result := YNToLogical(ClientXmit( sc_QuerySETEXACT, LogicalTOYN(bSetQueryExact), TRUE));
  except
    ClientError('ApolloConnection.QuerySetExact', 'QuerySetExact');
  end;
end;

function  TApolloConnection.ap_QueryTest( cpExpression: PChar ):SmallInt;
begin
  Result := 0;
  try
    Result := StrToInt(Trim(ClientXmit( sc_QUERYTEST , cpExpression , True)));
  except
    ClientError('ApolloConnection.QueryTest', 'QueryTest');
  end;
end;

procedure TApolloConnection.ap_Recall;
begin
  try
    ClientXmit( sc_RECALL , '', TRUE);
  except
    ClientError('ApolloConnection.Recall', 'Recall');
  end;
end;


function  TApolloConnection.ap_RecCount: LongInt;
begin
  Result := 0;
  try
    Result := StrToInt(Trim(ClientXmit( sc_RECCOUNT , '', True)));
  except
    ClientError('ApolloConnection.RecCount', 'RecCount');
  end;
end;

function  TApolloConnection.ap_RecNo: LongInt;
begin
  Result := 0;
  try
    Result := StrToInt(ClientXmit( sc_RECNO , '', True));
  except
    ClientError('ApolloConnection.RecNo', 'RecNo');
  end
end;

function TApolloConnection.ap_RecSize: LongInt;
begin
  Result := 0;
  try
    Result := StrToInt(Trim(ClientXmit( sc_RECSIZE , '', True)));
  except
    ClientError('ApolloConnection.RecSize', 'RecSize');
  end;
end;


procedure TApolloConnection.ap_Reindex;
begin
  try
    ClientXmit( sc_REINDEX  , '', TRUE);
  except
    ClientError('ApolloConnection.Reindex', 'Reindex');
  end;
end;


{MED: 09-25-02, Modified.  Made it modal}
procedure TApolloConnection.ap_Replace( cpFieldname: PChar; iDataType: SmallInt; vpData: Pointer );
var
  iSize : Integer;
  cDataType : String;
begin
	iSize := 0;
  case iDataType of
    R_CHAR, R_CHAREX: iSize := Length(PChar(vpData));
    R_INTEGER:        iSize := SizeOf(Integer);
    R_LONG:           iSize := SizeOf(LongInt);
    R_JULIAN:         iSize := SizeOf(LongInt);
    R_DOUBLE:         iSize := SizeOf(Double);
    R_LOGICAL:        iSize := SizeOf(Boolean);
  	R_DATESTR:        iSize := Length(PChar(vpData));
  	R_MEMO:           iSize := Length(PChar(vpData));
  end;

  cDataType := IntToHex(iDataType, 4);

  begin
    FWriteStream.Clear;
    FWriteStream.Position := 0;
    FWriteStream.Write(PChar(PadR(cpFieldName, 10, ' '))^, 10);
    FWriteStream.Write(PChar(cDataType)^, 4);
    FWriteStream.Write(vpData^, iSize);
    FWriteStream.Position := 0;     // this is a must have
    ClientStreamSendReceive(sc_REPLACE);
    FWriteStream.Position := 0;    // this is a must have
  end;
end;

function TApolloConnection.ap_RLock( lRecNum: LongInt ): WordBool;
begin
  Result := False;
  try
    Result := YNToLogical(ClientXmit( sc_RLOCK, IntToStr(lRecNum), True));
  except
    ClientError('ApolloConnection.Rlock', 'Rlock');
  end;
end;

{LAS: 11-19-99, implemented.}
function  TApolloConnection.ap_RYOFilterActivate( iFilterHandle: SmallInt;
                                iBoolOperation: SmallInt): WordBool;
begin
  Result := False;
  try
    Result := YNToLogical( ClientXmit( sc_RYOFILTERACTIVATE,
                           IntToHex(iFilterHandle, 4) +
                           IntToHex(iBoolOperation, 4),
                           True));
  except
    ClientError('ApolloConnection.RYOFilterActivate', 'RYOFilterActivate');
  end;
end;


{LAS: 11-19-99, implemented.}
function  TApolloConnection.ap_RYOFilterCopy: SmallInt;
begin
  Result := 0;
  try
    Result := StrToInt(Trim(ClientXmit( sc_RYOFILTERCOPY, '', True)));
  except
    ClientError('ApolloConnection.RYOFilterCopy', 'RYOFilterCopy');
  end;
end;


{LAS: 11-19-99, implemented.}
function TApolloConnection.ap_RYOFilterCreate: SmallInt;
begin
  Result := 0;
  try
    Result := StrToInt(Trim(ClientXmit( sc_RYOFILTERCREATE, '', True)));
  except
    ClientError('ApolloConnection.RYOFilterCreate', 'RYOFilterCreate');
  end;
end;

{LAS: 11-19-99, implemented.}
function TApolloConnection.ap_RYOFilterDestroy(iFilterHandle: SmallInt): WordBool;
begin
	Result := False;
	try
		Result := YNToLogical(ClientXmit( sc_RYOFILTERDESTROY,
                          IntToHex(iFilterHandle, 4), True));
  except
		ClientError('ApolloConnection.RYOFilterDestroy', 'RYOFilterDestroy');
   end;
end;

{LAS: 11-19-99, implemented.}
function  TApolloConnection.ap_RYOFilterGetBit(iFilterHandle: SmallInt; lRecNo: LongInt): WordBool;
begin
  Result := False;
  try
    Result := YNToLogical( ClientXmit( sc_RYOFILTERGETBIT,
                           IntToHex(iFilterHandle, 4) +
                           IntToHex(lRecNo, 4), True));
  except
    ClientError('ApolloConnection.RYOFilterGetBit', 'RYOFilterGetBit');
  end;
end;

{LAS: 11-19-99, implemented.}
function  TApolloConnection.ap_RYOFilterRestore(cpFileName: PChar): WordBool;
begin
  Result := False;
  try
    Result := YNToLogical(ClientXmit( sc_RYOFILTERRESTORE, cpFileName, True));
  except
    ClientError('ApolloConnection.RYOFilterRestore', 'RYOFilterRestore');
  end;
end;

{LAS: 11-19-99, implemented}
function  TApolloConnection.ap_RYOFilterSave(iFilterHandle: SmallInt; cpFileName: PChar): WordBool;
var
  msg : String;
begin
  Result := False;
  try
    msg := ClientXmit( sc_RYOFILTERSAVE,
                              IntToHex(iFilterHandle, 4) +
                              cpFileName, True);
    if msg[1] = 'Y' then
    	Result := True;
  except
    ClientError('ApolloConnection.RYOFilterSave', 'RYOFilterSave');
  end;
end;


{LAS: 11-19-99, implemented}
function  TApolloConnection.ap_RYOFilterSetBit( iFilterHandle: SmallInt; lRecNo: LongInt;
                              iOnOrOff: SmallInt): WordBool;
var
  msg : String;
begin
  Result := False;
  try
    msg := ClientXmit( sc_RYOFILTERSETBIT,
                              IntToHex(iFilterHandle, 4) +
                              IntToHex(lRecNo, 8) +
//                              IntToHex(iOnOrOff, 4),  {LAS: 12-09-99, Removed}
                              LogicalToYN(Boolean(iOnOrOff)),    {LAS: 12-09-99, Added}
                              True);
    if msg[1] = 'Y' then
    	Result := True;
  except
    ClientError('ApolloConnection.RYOFilterSetBit', 'RYOFilterSetBit');
  end;
end;

{LAS: 11-19-99 implemented, Need to test!!!}
function  TApolloConnection.ap_RYOKeyAdd( cpTagname: PChar; cpKey: PChar ): WordBool;
begin
  Result := False;
  try
    Result := YNToLogical(ClientXmit( sc_RYOKEYADD,
                                             PadR( cpTagName, 10, ' ' ) +
                                             cpKey, True));
  except
    ClientError('ApolloConnection.RYOKeyAdd', 'RYOKeyAdd');
  end;
end;

{LAS: 11-19-99, implemented.  Need to test!!!}
function  TApolloConnection.ap_RYOKeyDrop(cpTagname: PChar): WordBool;
begin
  Result := False;
  try
    Result := YNToLogical(ClientXmit( sc_RYOKEYDROP, cpTagName, True));
  except
    ClientError('ApolloConnection.RYOKeyDrop', 'RYOKeyDrop');
  end;
end;

function TApolloConnection.ap_Seek( cpKeyValue: PChar): WordBool;
var
  msg: String;
begin
  Result := False;
  try
    msg := ClientXmit( sc_SEEK, cpKeyValue, True);
    if msg[1] = 'Y' then
    	Result := True;
  except
    // ClientError....
  end;
end;


function  TApolloConnection.ap_SeekBin(cpKeyValue: PChar; uiLength: Word): WordBool;
var
  msg : String;
begin
  Result := False;
  try
    msg := ClientXmit( sc_SEEKBIN ,
                              IntToHex( uiLength, 8 )+
                              cpKeyValue, True);
    if msg[1] = 'Y' then
    	Result := True;
  except
    ClientError('ApolloConnection.SeekBin', 'SeekBin');
  end;
end;


// select area not necessary  {LAS: 11-14-99, Modified}
procedure TApolloConnection.ap_SetCentury( uiOnOff: WordBool );
begin
  if Self = nil then
  	exit;
  try
    With Self do
    begin
      if Connected then        // this is to prevent design time access
      begin
        try
          ClientXmit( sc_SETCENTURY, LogicalToYN(uiOnOff), TRUE);
        except
          ClientError('ApolloConnection.SetCentury', 'SetCentury');
        end;
      end;
    end;
  except
  end;
end;

// select area not necessary
procedure TApolloConnection.ap_SetDateFormat(uiDateType: Word);
begin
  try
    ClientXmit( sc_SETDATEFORMAT,  IntToHex(uiDateType, 8), TRUE );
  except
    ClientError('ApolloConnection.SetDateFormat', 'SetDateFormat');
  end;
end;

// select area not necessary
procedure TApolloConnection.ap_SetExact(uiOnOff: WordBool);
begin
  if Self = nil then
  	exit;
  try
    With Self do
    begin
      if Connected then                    // this is to prevent design time access
      begin
        try
          ClientXmit( sc_SETEXACT, LogicalToYN(uiOnOff), TRUE);
        except
          ClientError('ApolloConnection.SetExact', 'SetExact');
        end;
      end;
    end;
  except
    ClientError('ApolloConnection.SetExact', 'SetExact');
  end;
end;

procedure TApolloConnection.ap_SetSoftSeek(uiOnOff: WordBool);
begin
  if Self = nil then
  	exit;
  try
    if not Connected then
    	exit;
    try
      ClientXmit( sc_SETSOFTSEEK , LogicalTOYN(uiOnOff), TRUE);
    except
    end;
  except
    ClientError('ApolloConnection.SetSoftSeek', 'SetSoftSeek');
  end;
end;

// returns previous order
function  TApolloConnection.ap_SetOrder( iIndex: SmallInt ):SmallInt;
var
  msg : String;
begin
  Result := 0;
  try
    msg := ClientXmit( sc_SETORDER, IntToHex(iIndex,4), True);
    Result := StrToInt(Trim(msg));
  except
    ClientError('ApolloConnection.SetOrder', 'SetOrder');
  end;
end;

procedure TApolloConnection.ap_SetRelation(uiChildArea: Word; cpKeyExpr: PChar);
begin
  try
    ClientXmit( sc_SETRELATION , IntToHex(uiChildArea, 8)+ cpKeyExpr, TRUE);
  except
    ClientError('ApolloConnection.SetRelation', 'SetRelation');
  end;
end;

{LAS: 11-19-99.  This function should not be called remotely.}
procedure TApolloConnection.ap_SetStringType(uiStringType: Word);
begin
	// LAS: This should be set to 1 at Server start-up.
end;

procedure TApolloConnection.ap_SetPassword( cpEncodeKey: PChar );
begin
  try
    ClientXmit( sc_SETPASSWORD, cpEncodeKey, False);
  except
    ClientError('ApolloConnection.SetPassword', 'SetPassword');
  end;
end;

procedure TApolloConnection.ap_SetQueryBit(lRecNo: LongInt; bValue: WordBool);
var
  msg : String;
begin
  msg := 'N';
  if bValue then
  	msg := 'Y';
  try
    ClientXmit( sc_SETQUERYBIT, IntToHex(lRecNo, 8) + msg, False);
  except
    ClientError('ApolloConnection.SetQueryBit', 'SetQueryBit');
  end;
end;

function  TApolloConnection.ap_SetScope(cpLowVal, cpHighVal: PChar): WordBool;
begin
  Result := False;
	try
    Result := YNTOLOGICAL(ClientXmit( sc_SETSCOPE,
                                    IntToHex( Length( cpLowVal ),4 ) +   {LAS: 06-21-00, Added}
                                    IntToHex( Length( cpHighVal ),4 ) +  {LAS: 06-21-00, Added}
                                    PadR(cpLowVal, 255, ' ') +
                                    PadR(cpHighVal, 255, ' '), True));
  except
    ClientError('ApolloConnection.SetScope', 'SetScope');
  end;
end;

{LAS: 11-19-99, implemented.  Need to test!!!}
procedure TApolloConnection.ap_SetFilter( cpExpression: PChar );
begin
  try
    ClientXmit( sc_SETFILTER, cpExpression, False );
  except
    ClientError('ApolloConnection.SetFilter', 'SetFilter');
  end;
end;


procedure TApolloConnection.ap_SetDeleted(uiDeleted: WordBool);
var
	msg : String;
begin
  if self = nil then
  	exit;
  if uiDeleted then
  	msg := 'Y'
  else
    msg := 'N';
  try
    With Self do
    begin
      if Connected then               // this is to prevent design time access
      begin
        try
          ClientXmit( sc_SETDELETED, msg, TRUE);
        except
          ClientError('ApolloConnection.SetDeleted', 'SetDeleted');
        end;
      end;
    end;
  except
  end;
end;


{LAS: 11-19-99, implemented.  Need to test!!!}
function  TApolloConnection.ap_SetEpoch(uiBaseYear: Word): Word;
var
  cSDE, msg: String;
begin
  Result := 0;
  try
    if not Connected then
    	exit;

    cSDE := IntToHex(uiBaseYear, 8);
    try
      if self <> nil then
      begin
        try
          msg := ClientXmit( sc_SETEPOCH, cSDE, True );
          Result := StrToInt(Trim(msg));
        except
          ClientError('ApolloConnection.SetEpoch', 'SetEpoch');
        end;
      end;
    except
      ClientError('ApolloConnection.SetEpoch', 'SetEpoch');
    end;
  except
  end;
end;

{LAS: 11-19-99.  This function should not be called remotely.}
function  TApolloConnection.ap_SetErrorFunc( pFunc, pData : LongInt): LongInt;
begin
  Result := 0;
end;

{LAS: 11-19-99.  This function should not be called remotely.}
procedure TApolloConnection.ap_SetGaugeHook(hwndGauge: HWND);
begin
  //showmessage('SetGaugeHook not implemented');
end;

{LAS: 11-19-99, Implimented.  Need to test!!!}
//MED: 03-29-01, made it into a function
function TApolloConnection.ap_SetMemoBlockSize(uiBlockSize: Word):Word;
var
  cSDE: String;
begin
  Result := 0;

  if not Connected then
    exit;

  cSDE := PadR(IntToStr(uiBlockSize), 10, ' ');
  try
    if Self <> nil then
    begin
      try
        Result := StrToInt('$'+ClientXmit( sc_SETMEMOBLOCKSIZE, cSDE, TRUE));
      except
        ClientError('ApolloConnection.SetMemoBlockSize', 'SetMemoBlockSize');
      end;
    end;
  except
    ClientError('ApolloConnection.SetMemoBlockSize', 'SetMemoBlockSize');
  end;
end;

procedure TApolloConnection.ap_SetLockTimeout(iSeconds: SmallInt);
begin
  try
    ClientXmit( sc_SetLockTimeOut, IntToStr(iSeconds), False);
  except
    ClientError('ApolloConnection.SetLockTimeOut', 'SetLockTimeOut');
  end;
end;

procedure TApolloConnection.ap_SetTranslate(uiOnOff: WordBool);
begin
  try
    ClientXmit( sc_SETTRANSLATE , LogicalToYN(uiOnOff), False);
  except
    ClientError('ApolloConnection.SetTranslate', 'SetTranslate');
  end;
end;

function  TApolloConnection.ap_SysProp(uiSysItem: Word; vpData: Pointer): LongInt;
var
  cSDE, msg,
  sData  : String;
begin
  Result := 0;
  try
    if not Connected then
    	exit;

    cSDE  := IntToHex(uiSysItem, 8);
    sData := '';
    case uiSysItem of                      // MED: 04-02-2001-some of SysProp methods still need additional work
      SDE_SP_GETSOFTSEEK,
	    SDE_SP_GETEXACT,
	    SDE_SP_GETDELETED,
	    SDE_SP_GETOBUFFER,
	    SDE_SP_GETSTRINGTYPE,
	    SDE_SP_SETSTRINGTYPE,
	    SDE_SP_GETDISABLEAUTO,
	    SDE_SP_SETDISABLEAUTO,
	    SDE_SP_SETOEMCOLLATE,
	    SDE_SP_GETDUDENCOLLATE,
	    SDE_SP_SETDUDENCOLLATE,
	    SDE_SP_GETLIMITCASECONV,
	    SDE_SP_GETADDQUERY,
	    SDE_SP_SETADDQUERY,
	    SDE_SP_GETINDEXCOUNT:
        sData := '';

	    SDE_SP_SETSOFTSEEK,
	    SDE_SP_SETEXACT,
	    SDE_SP_SETOBUFFER,
	    SDE_SP_SETUSECONDITIONAL:
        if Boolean(vpData) then sData := 'Y' else sData := 'N';

	    SDE_SP_SETLIMITCASECONV,
	    SDE_SP_SETWRITEBLOBHDR:if Integer(vpData) = 1 then sData := '1'
                             else                        sData := '0';

	    SDE_SP_GETUSECONDITIONAL:  sData := '0';

      SDE_SP_GETDRIVER,
	    SDE_SP_GETDESCENDING,
	    SDE_SP_GETEMPTY:
        sData := IntToHex(Integer(vpData), 8);
    end;

    if Self <> nil then
    begin
      try
        msg := ClientXmit( sc_SYSPROP, cSDE + sData, True);
        Result := StrToInt('$'+Trim(msg));
      except
        ClientError('ApolloConnection.SysProp', 'Converting data');
      end;
    end;
  except
    ClientError('ApolloConnection.SysProp', 'SysProp');
  end;
end;

function  TApolloConnection.ap_TagArea( cpTagName: PChar ): SmallInt;
begin
  Result := 0;
  try
    Result := StrToInt(Trim(ClientXmit( sc_TAGAREA, cpTagName , True)));
  except
    ClientError('ApolloConnection.TagArea', 'TagArea');
  end;
end;

function  TApolloConnection.ap_TagDelete(cpTagname: PChar): LongBool;
begin
  Result := FALSE;
  try
    Result := YNToLogical(Trim(ClientXmit( sc_TAGDELETE, cpTagName , TRUE)));
  except
    ClientError('ApolloConnection.TagDelete', 'TagDelete');
  end;
end;

function TApolloConnection.ap_TagName( iTagArea: SmallInt ): PChar;
begin
  Result := '';
  try
    Result := ClientXmit( sc_TAGNAME, IntToHex( iTagArea, 4 ), True );
  except
    ClientError('ApolloConnection.TagName', 'TagName');
  end;
end;

procedure TApolloConnection.ap_SetTurboRead(uiOnOff: WordBool);
var msg : String;
begin
  try
    if not Connected then
    	exit;

    if uiOnOff then    	msg := 'Y'
    else               	msg := 'N';
    ClientXmit( sc_SETTURBOREAD, msg, TRUE);
  except
    ClientError('ApolloConnection.SetTurboRead', 'SetTurboRead');
  end;
end;


procedure TApolloConnection.ap_Unlock( lRecNum: LongInt );
begin
  try
    ClientXmit( sc_UNLOCK, IntToHex(lRecNum, 8), TRUE );
  except
    ClientError('ApolloConnection.Unlock', 'Unlock');
  end;
end;

function TApolloConnection.ap_WorkArea(cpAlias: PChar ): Word ;
begin
	Result := 0;
  if self = nil then
  	exit;
  try
    Result := StrToInt(Trim(ClientXmit( sc_WORKAREA, cpAlias, True)));
  except
    ClientError('ApolloConnection.WorkArea', 'WorkArea');
  end;
end;


function TApolloConnection.ap_Version:PChar;
begin
  Result := nil;
  try
    Result := ClientXmit( sc_VERSION, '' , True);
  except
    ClientError('ApolloConnection.Version', 'Version')
  end;
end;

procedure TApolloConnection.ap_Zap;
begin
  try
    ClientXmit( sc_ZAP, '', True );
  except
    ClientError('ApolloConnection.Zap', 'Zap');
  end;
end;

{MED: 07-31-01, added}
function  TApolloConnection.ap_GetTableType(sDatabaseName, sTableName:String): SmallInt;
begin
  Result := 0;
  try
    Result := StrToInt(Trim(ClientXmit( sc_GETTABLETYPE,
                                        PadR(sDatabaseName, 255, ' ') +
                                    PadR(sTableName, 255, ' '), True)));

  except
    ClientError('ApolloConnection.GetTableType', 'GetTableType');
  end;

end;

procedure TApolloConnection.ap_PutRecord( cpRecord: PChar; iRecSize:Integer ); // same as sx_PutRecord
begin
  try
    FWriteStream.Clear;
    FWriteStream.Position := 0;
    FWriteStream.Write(PChar(cpRecord)^, iRecSize); //StrLen(cpRecord));
    FWriteStream.Position := 0;    // this is a must have
    ClientStreamSendReceive(sc_WRITERECORD);
    FWriteStream.Clear;
  except
    ClientError('ApolloConnection.PutRecord', 'PutRecord');
  end;
end;
(*
procedure TApolloConnection.ap_PutRecord( cpRecord: PChar ); // same as sx_PutRecord
begin
  try
//    ClientXmit(sc_WRITERECORD, cpRecord, True);
  except
    ClientError('ApolloConnection.PutRecord', 'PutRecord');
  end;
end;
*)

function TApolloConnection.GetTableNames: Pointer;
begin
  Result := nil;
  try
    ClientStreamReceive( sc_GETTableNames, FDataBaseName );
    Result := FReadStream;
  except
    ClientError('ApolloConnection.GetTableNames', 'GetTableNames');
  end;
end;

{LAS: 05-22-00, Readded}
//------------------------------------------------------------------------------
function TApolloConnection.GetTableNamesList(sDatabaseName: String): TStringList;
var
  msPtr : Pointer;
  sOldAlias: String;
begin
  Result := TStringList.Create;
  try
    if Connected then
    begin
      sOldAlias := FDatabaseName;
      FDatabaseName := sDatabaseName;
      msPtr := GetTableNames;
      FDatabaseName := sOldAlias;
      if ( msPtr <> nil ) then
        Result.LoadFromStream( TMemoryStream( msPtr ));
    end;
  except
    ClientError('ApolloConnection.GetTableNamesList', 'GetTableNamesList');
  end;
end;



function TApolloConnection.GetDLLProcs( sDLLName:String): Pointer;
begin
  Result := nil;
  try
    ClientStreamReceive( sc_GETDLLProcs,
                         PadR(FDataBaseName,255, ' ') +
                         PadR(sDLLName, 255, ' '));
    Result := FReadStream;
  except
    ClientError('ApolloConnection.GetDLLProcs', 'GetDLLProcs');
  end;
end;

function TApolloConnection.GetFiles(sSrchStr:String): Pointer;
begin
  Result := nil;
  try
    ClientStreamReceive( sc_GETFILES,
                                   PadR(FDataBaseName,255, ' ')+
                                   PadR(sSrchStr, 255, ' ') );
    Result := FReadStream;
  except
    ClientError('ApolloConnection.GetFiles', 'GetFiles');
  end;
end;

{
 DriveChar= 1
 DriveType= 1
 VolName  = 255

 first 4 bytes is count
 returns Count
}
function  TApolloConnection.FindDrives:Pointer;
begin
  Result := nil;
  try
    ClientStreamReceive(sc_FINDDRIVES, '' );
    Result := FReadStream;
  except
    ClientError('ApolloConnection.FindDrives', 'FindDrives');
  end;
end;

function TApolloConnection.FindDirectories(const ParentDirectory:String): Pointer;
begin
  Result := nil;
  try
    ClientStreamReceive( sc_FINDDIRECTORIES,
                          PadR(ParentDirectory, 255, ' ') );
    Result := FReadStream;
  except
    ClientError('ApolloConnection.FindDirectories', 'FindDirectories');
  end;
end;

// DOES NOT USE FDATABASENAME - see GetFiles
function TApolloConnection.FindFiles(Directory, Mask:String): Pointer;
begin
  Result := nil;
  try
    ClientStreamReceive( sc_FINDFILES, PadR(Directory, 255, ' ')+
                                      PadR(Mask, 255, ' ')  );
    Result := FReadStream;
  except
    ClientError('ApolloConnection.FindFiles', 'FindFiles');
  end;
end;

function TApolloConnection.GetServerLog: Pointer;
begin
  Result := nil;
  try
    Result := ClientStreamReceive( sc_GETSERVERLOG, '' );
  except
    ClientError('ApolloConnection.GetServerLog', 'GetServerLog');
  end;
end;

// set FReadStream.Position := 0 if reading another way
function TApolloConnection.GetDataBaseNames: Pointer;
begin
  Result := nil;
  try
    Result := ClientStreamReceive( sc_GetDataBaseNames , '');
  except
    ClientError('ApolloConnection.GetDataBaseNames', 'GetDataBaseNames');
  end;
end;

{LAS: 05-22-00, Readded}
//------------------------------------------------------------------------------
function TApolloConnection.GetDatabaseNamesList: TStringList;
var
  msPtr : Pointer;
begin
  Result := TStringList.Create;
  try
    if Connected then
    begin
      msPtr := GetDataBaseNames;
      if ( msPtr <> nil ) then
        Result.LoadFromStream( TMemoryStream( msPtr ));
    end;
  except
    ClientError('ApolloConnection.GetDatabaseNamesList', 'GetDatabaseNamesList');
  end;
end;


procedure TApolloConnection.GetUserNames(uList: TStringList);
var
  msPtr : pointer;
begin
  try
    msPtr := ClientStreamReceive( sc_GetUserNames , '');
    if (msPtr <> nil) then
    begin
      uList.Clear;
      uList.LoadFromStream(TMemoryStream(msPtr) );
    end;
  except
    ClientError('ApolloConnection.GetUserNames', 'GetUserNames');
  end;
end;


{LAS: 05-29-00, Readded.  Changed from GetUserNames procedure}
function TApolloConnection.GetUserNamesList: TStringList;
var
  msPtr : pointer;
begin                                  
  Result := TStringList.Create;
  try
    if Connected then
    begin
	    msPtr := ClientStreamReceive( sc_GetUserNames , '');
  	  if (msPtr <> nil) then
    	  Result.LoadFromStream(TMemoryStream(msPtr) );
    end;
  except
    ClientError('ApolloConnection.GetUserNamesList', 'GetUserNamesList');
  end;
end;
{
//-----------------------------------------------------------------
// MED: 03-23-01, added
//-----------------------------------------------------------------
function TApolloConnection.CopyFile(sSrcDatabaseName, sDesDatabaseName, sFileName : String): Boolean;
var
  sFile : String;
begin
  Result := False;

  if (sDesDatabaseName='') or (sFileName='') or (sSrcDatabaseName= '') then
    exit;

  sFile := ExtractFileName(sFileName);
  try
    Result := YNToLogical(ClientXmit( sc_COPYFILE_EX ,
                                 PadR(sSrcDatabaseName , 255, ' ')+
                                 PadR(sDesDatabaseName , 255, ' ')+
                                 PadR(sFile, 255, ' ') , TRUE));
  except
    ClientError('ApolloConnection.CopyFile', 'CopyFile');
  end;
end;
}

//IIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIII
function TApolloConnection.DeleteFile(sDataBaseName, sFileName : String): Boolean;
var
  sFile : String;
begin
  Result := False;

  if (sDataBaseName='') or (sFileName='') then
    exit;

  sFile := ExtractFileName(sFileName);
  try
    Result := YNToLogical(ClientXmit( sc_DELETEFILE ,
                                 PadR(sDataBaseName , 255, ' ')+
                                 PadR(sFile, 255, ' ') , True));
  except
    ClientError('ApolloConnection.DeleteFile', 'DeleteFile');
  end;
end;

function TApolloConnection.DeleteFileExt(sDataBaseName, sFileName : String): Boolean;
var
  sFile : String;
begin
  Result := False;

  if (sDataBaseName='') or (sFileName='') then
    exit;

  sFile := ExtractFileName(sFileName);
  try
    Result := YNToLogical(ClientXmit( sc_DELETEFILE ,
                                 PadR(sDataBaseName , 255, ' ')+
                                 PadR(sFile, 255, ' ') , True));
  except
    ClientError('ApolloConnection.DeleteFile', 'DeleteFile');
  end;
end;


//ppppppppppppppppppppppp
function TApolloConnection.FileExists(sDataBaseName, sFileName : String): Boolean;
var
  sFile : String;
begin
  Result := False;
  if (sDataBaseName='') or (sFileName='') then
    exit;

  sFile := ExtractFileName(sFileName);
  try
    Result := YNToLogical(ClientXmit( sc_FILEEXISTS,
                                 PadR(sDataBaseName, 255, ' ')+
                                 PadR(sFile, 255, ' ') , True));
  except
    ClientError('ApolloConnection.FileExists', 'FileExists');
  end;
end;


/////////////////////////////////////////////////////////////////////////////////////
//_________________________________________________________________
//
//  Script routines
//    function  CreateScript(ScriptName:String; ScriptText:TStrings; bSave:Boolean):Pointer;  // creates a new script and adds it to the list
//    function  GetScriptText(ScriptName:String):Pointer;     // load script text
//_________________________________________________________________
/////////////////////////////////////////////////////////////////////////////////////////////

function TApolloConnection.GetScriptsList: Pointer;
begin
  Result := nil;
  try
    Result := ClientStreamReceive( sc_GetScriptsList , '');
  except
    ClientError('ApolloConnection.GetScriptsList', 'GetScriptsList');
  end;
end;

//creates the script and adds it to the list (use Remove to remove it from the list)
// if bSave is TRUE then the script is also saved in SCRIPTS.DBF otherwise will be gone next time you reboot server
// or if LoadAllScript is called
// it returns a new list of scripts if successfull or nil on failure
function TApolloConnection.CreateScript(ScriptName:String; ScriptText:TStringList; iLang:Integer; bSave:Boolean):Pointer;  // creates a new script and adds it to the list
begin
  Result := nil;
  try
    FWriteStream.Clear;
    FWriteStream.Position := 0;
    FWriteStream.Write(pchar(padr(ScriptName, 32, ' '))^, 32);
    FWriteStream.Write(pchar(Inttostr(iLang))^, 1);
    FWriteStream.Write(pchar(LogicalToYN(bSave))^, 1);
    ScriptText.SaveToStream(FWriteStream);
    Result := ClientStreamSendReceive(sc_CreateScript);
    FWriteStream.Position := 0;    // this is a must have
  except
    ClientError('ApolloConnection.CreateScript', 'CreateScript');
  end;
end;

function  TApolloConnection.StopScript(ScriptName:String):Pointer;    // stops executiong
begin
  Result := nil;
  try
    Result := ClientStreamReceive( sc_StopScript, padr(ScriptName, 32, ' '));
  except
    ClientError('ApolloConnection.StopScript', 'StopScript');
  end;
end;



function TApolloConnection.DropScript(ScriptName:String): Pointer;    // Deletes it in SCRIPTS.DBF
begin
  Result := nil;
  try
    Result := ClientStreamReceive( sc_DropScript, padr(ScriptName, 32, ' '));
  except
    ClientError('ApolloConnection.DropScript', 'DropScript');
  end;
end;

// if bSave also saves it to disk
// true most of the time
function  TApolloConnection.UpdateScriptText(ScriptName:String; ScriptText:TStrings; bSave:Boolean ):Boolean;   // saves script text
begin
  Result := FALSE;
  try
    FWriteStream.Clear;
    FWriteStream.Position := 0;
    FWriteStream.Write(pchar(padr(ScriptName, 32, ' '))^, 32);
    FWriteStream.Write(pchar(LogicalToYN(bSave))^, 1);
    ScriptText.SaveToStream(FWriteStream);
    ClientStreamSendReceive(sc_UpdateScriptText);
    FWriteStream.Position := 0;    // this is a must have
    Result := TRUE;
  except
    ClientError('ApolloConnection.UpdateScript', 'UpdateScript');
  end;
end;

function  TApolloConnection.GetScriptText(ScriptName:String):Pointer;     // load script text
begin
  Result := nil;
  try
    Result := ClientStreamReceive( sc_GetScriptText, padr(ScriptName, 32, ' '));
  except
    ClientError('ApolloConnection.GetScriptText', 'GetScriptText');
  end;
end;


{
Receiving bytes
8 = error code
8 = line number
8 = char number
254 = descriptiong
rest is variant
}
Function TApolloConnection.RunScript(ScriptName:String; InParams : array of Variant):Pointer;
Var
  s : String;
begin
  Result := nil;
  FWriteStream.Clear;
  FWriteStream.Position := 0;
  FWriteStream.Write(pchar(padr(ScriptName, 32, ' '))^, 32);
  if varType(InParams[0]) = varEmpty then       // 0-fullrun 2-No params
  { if  High(InParams) < 1 then      // 0-fullrun 2-No params
  }
  begin
    s := '2';
    FWriteStream.Write(pchar(s)^, 1);
  end
  else
  begin
    s := '0';
    FWriteStream.Write(pchar(s)^, 1);
    VariantToStream(InParams, FWriteStream);
  end;
  FWriteStream.Position := 0;
  try
    Result := ClientStreamSendReceive(sc_RunScript);
  except
    ClientError('ApolloConnection.RunScript', 'RunScript');
  end;
end;

function TApolloConnection.CheckScriptSyntax(ScriptName:String; var CharNo:integer; var LineNo):Pointer;
var
  s : String;
begin
  Result := nil;
  FWriteStream.Clear;
  FWriteStream.Position := 0;
  FWriteStream.Write(pchar(padr(ScriptName, 32, ' '))^, 32);
  s := '1';                        // syntax only
  FWriteStream.Write(pchar(s)^, 1);
  FWriteStream.Position := 0;
  try
    Result := ClientStreamSendReceive(sc_CheckScriptSyntax);
  except
    ClientError('ApolloConnection.CheckScriptSyntax', 'CheckScriptSyntax');
  end;
end;

////////////////////////////////////////////////////////////////////////////////
//_________________________________________________________________
//
//  Procedures routines
//_________________________________________________________________
////////////////////////////////////////////////////////////////////////////////

function TApolloConnection.GetDLLProcedureNames(sDatabaseName:String): Pointer;
begin
  Result := nil;
  try
    Result := ClientStreamReceive( sc_GetProcedureNames , padr(sDataBaseName, 255, ' '));
  except
    ClientError('ApolloConnection.GetProcedureNames', 'GetProcedureNames');
  end;
end;

///////////////////////////////////////////////////////////
function TApolloConnection.ap_SetDLLProcName(sDatabaseName, sProcName:String):Pointer;
begin
  Result := nil;
  if (sProcName <> '') AND (Connected) then
  begin                                         
    try
      Result := ClientStreamReceive(
                             sc_GetParams,
                             padr(sDataBaseName, 255, ' ') +
                             padr(sProcName, 255 , ' ') );

    except
      ClientError('ApolloConnection.SetDLLProcName', 'SetDLLProcName');
    end;
  end;
end;


function TApolloConnection.ap_ExecDLLProc: Pointer;
begin
  Result := NIL;
  if Connected then
  begin
    try
      Result := ClientStreamSendReceive(sc_EXECPROC);
    except
      ClientError('ApolloConnection.ExecProc', 'ExecProc');
    end;
  end
  else
    ClientError('ApolloConnection.ExecProc', 'ApolloConnection not connected');
end;


(*============================================================================
  ApTPS.PAS
  Apollo Client/Server Transaction Processing
  Copyright 1999 Vista Software
  Last Updated: 10-17-99

  NOTE: Tab stops were set at 2 spaces during development of this unit. If you
        are using a tab stop setting other than 2 spaces, the code indents
        may not line up accurately in your editor.
============================================================================*)

{=======================================================
 Transaction Table is

 TableName      Status        Method      Record
 __254 bytes     _1 bytes     _ 1 Byte   ______Record Length byte

-Status is initially set at 0
-1 is committed
-2 is rolled back

Methods are
-AppendBlank
-Delete
-Replace
-Recall
-PutRecord
=======================================================}

function TApolloConnection.ap_tps_BeginTransaction: WordBool;
begin
  Result := FALSE;
  Try
    Result := ClientXmit( sc_tps_BeginTransaction , '', TRUE) = 'Y';
  Except
    ClientError('ApolloConnection.tps_BeginTransaction', 'tps_BeginTransaction');
  End;
end;


function TApolloConnection.ap_tps_Commit: WordBool;
begin
  Result := FALSE;
  Try
    Result := ClientXmit( sc_tps_COMMIT , '', TRUE) = 'Y';
  Except
    ClientError('ApolloConnection.tps_Commit', 'tps_Commit');
  End;
end;


function TApolloConnection.ap_tps_RollBack: WordBool;
begin
  Result := FALSE;
  Try
    Result := ClientXmit( sc_tps_ROLLBACK , '', TRUE) = 'Y';
  Except
    ClientError('ApolloConnection.tps_RollBack', 'tps_RollBack');
  End;
end;

{LAS: 05-09-00, Added}
{==============================================================================
  PathFromAlias - If TApolloQuery Alias is used, gets physical path (LOCAL ONLY)
 ==============================================================================}
function TApolloConnection.PathFromAlias: String;
begin
	Result := DatabaseName;  // Default

  if ((Pos( '\', Result ) > 0) or    // Check for characters that would
      (Pos( ':', Result ) > 0) or    //   make it clear that this is a physical
      (Pos( '.', Result ) > 0)) and  //    file path and not an alias.
			not(Result = '') then   // If NOT empty, bail now
  	exit;

	if FSQLDataPath > '' then
    Result := FSQLDataPath;

  // If no path was specified, use .DBF current directory
  if (Result = '') then
    GetDir( 0, Result );

  if (Length(Result) > 0) and (Result[ Length(Result) ] <> '\') then
    Result := Result + '\';
end;

{//////////////////////////////////////////////////////////////////////////////////
////////                            SQL
//////////////////////////////////////////////////////////////////////////////////}

// transmits the SQL Statement to the server
// server will parse the statement and try to create a Result Set
// returns # of Rows (if 0 no result set)
// When sql_RecCount > 0 it means the SQL statement was successful and a result set
// is waiting for us on the server.
// before we attempt to fetch any record we must call ap_sql_structure
// to obtain data structure of the Result Set.
// returns SQLID
{MED: 01-09-2000, modified to support ExecSQL, and OpenSQL}
// returns SQL ID

function TApolloConnection.ap_sql_Prepare(
                              AccessMethod: TAccessMethod;   {MED: 03-05-2000, added}
                              sDataBaseName:String;
                              TableNames:TCollection;
                              Century : WordBool;
                              DateFormat : String;
                              Deleted:WordBool;
                              MemoBlockSize:Integer;
                              FParams: TParams;
                              FReadOnly : Boolean;
                              FDefaultTableType : TApolloTableType;
                              FDefaultPassword : String;
                              FDefaultTranslate: Boolean;
                              FCommitLevel : TApolloCommitLevel;
                              var iError: Integer;
                              var sError:String ): Integer;  // returns SQLID
var
  s,
  sPath, sTables  : String;
  SQLID, iCount,
  i, iSize : Integer;
  bFound : WordBool;
  BuffPtr : pointer;
  pBuff   : Pchar;
  pMem    : TMemoryStream;
  FParamsStrm : TMemoryStream;      // MED: 09-05-01, added to support D6
begin
  Result  := 0;
  iCount  := 0;
  sTables := '';
  sError  := '';
  iError  := 0;
  SQLID   := 0;
  for i := 0 to TableNames.Count -1 do
  begin
    if trim(TApolloTables(TableNames).Items[i].TableName) <> '' then  //skip empty ones
    begin
      sTables := sTables + (padr(Trim(TApolloTables(TableNames).Items[i].TableName), 64, ' '));
      sTables := sTables + (padr(Trim(TApolloTables(TableNames).Items[i].Alias), 64, ' '));
      sTables := sTables + IntToHex(Ord(TApolloTables(TableNames).Items[i].TableType), 1);
			sTables := sTables + (padr(Trim(TApolloTables(TableNames).Items[i].Password), 8, ' '));
			sTables := sTables + LogicalToYN(TApolloTables(TableNames).Items[i].OEMTranslate); {LAS: 05-09-00, Added}
      {MED: 02-07-00, Added}
      s       := Trim(TApolloTables(TableNames).Items[i].ExtraIndexes.Text);
      s       := IntToHex(Length(s), 8)+s;
			sTables := sTables + s;
      sTables := sTables + IntToHex(Ord(TApolloTables(TableNames).Items[i].CommitLevel), 1);

      inc(iCount);
    end;
  end;

  //////////////////////////////////////////////////
  // Data packaged.  Now call the engine
  //////////////////////////////////////////////////
  Try
    if (AccessMethod = amServer) then
    begin
      FWriteStream.Clear;
      FWriteStream.Position := 0;
      FWriteStream.Write(pchar(padr(sDataBaseName, 255, ' '))^, 255);
      FWriteStream.Write(pchar(LogicalToYN(Century))^, 1);
      FWriteStream.Write(pchar(padr(DateFormat,10, ' '))^, 10);
      FWriteStream.Write(pchar(LogicalToYN(Deleted))^, 1);
      FWriteStream.Write(pchar(LogicalToYN(FReadOnly))^, 1);
      FWriteStream.Write(pchar(IntToHex(MemoBlockSize,8))^, 8);
      FWriteStream.Write(pchar(IntToHex(Integer(FDefaultTableType),1))^, 1);
      FWriteStream.Write(pchar(Padr(FDefaultPassword, 255, ' '))^, 255);
      FWriteStream.Write(pchar(LogicalToYN(FDefaultTranslate))^, 1);
      FWriteStream.Write(pchar(IntToHex(iCount,8))^, 8);
      FWriteStream.Write(pchar(IntToHex(Integer(FCommitLevel),1))^, 1);// CRVG Added CommitLevel
      FWriteStream.Write(pchar(IntToHex(length(sTables),8))^, 8);
      FWriteStream.Write(pchar(sTables)^, length(sTables));
      BuffPtr := NIL;
      for i:= 0 to FParams.Count -1 do
      begin
        iSize := FParams.Items[i].GetDataSize;
        try
          try
            BuffPtr := AllocMem(iSize);
          except
            ClientError('ApolloConnection.sql_Prepare', 'Allocating memory of '+inttostr(iSize)+' size');
          end;
          FParams.Items[i].GetData( BuffPtr);
          FWriteStream.Write(PChar(IntToHex(iSize, 8))^, 8);
          FWriteStream.Write(PChar(IntToHex(ord(FParams.Items[i].DataType), 2))^, 2);
          FWriteStream.Write(PChar(IntToHex(ord(FParams.Items[i].ParamType), 2))^, 2);
          FWriteStream.Write(PChar(PadR(FParams.Items[i].name, 20, ' '))^, 20);
          FWriteStream.Write(PChar(BuffPtr)^, iSize);
        finally
          FreeMem(BuffPtr);
        end;
      end;
      pMem := ClientStreamSendReceive(sc_SQL_PREPARE);
      if pMem <> nil then
      begin
        pBuff := NIL;
        Try
          pBuff := pchar(AllocMem(pMem.Size+1));   //add one in case we get no ErrorMsg
          pMem.Read( pchar(@pbuff[0])^, 8);pbuff[8] := #0;         // Result
          Result := StrToInt('$'+pbuff);
          pMem.Read( pchar(@pbuff[0])^, 8); pbuff[8] := #0;       // Record Number
          iError := StrToInt('$'+pbuff);
          pMem.Read( pchar(@pbuff[0])^, pMem.Size-16); pbuff[pMem.Size-16] := #0;       // Record Number
          sError := pbuff;
        Finally
          FreeMem(pBuff);
        End;
      end;
    end
    else
    //______________________________________________________________________________
    // LOCAL SQL
    //_______________________________________________________________________________
    begin
	  	LoadSQLDLL;                      {LAS: 05-09-00, Added}

  	  sDatabaseName := PathFromAlias;  {LAS: 05-09-00, Added}

      sPath := Trim(sDataBaseName);
      if Length(sPath) > 0 then
        if (sPath[ Length(sPath) ] <> '\') then
          sPath := sPath + '\';

      bFound        := DirectoryExists(sPath);
      if not bFound then
      begin
        //sError := IntToHex(0, 8) + IntToHex(1, 8)+'No such DataBaseName: '+sPath;
        sError := 'No such DataBaseName: '+sPath;
        exit;
      end;
      FSQLDataPath  := sPath;

      FParamsStrm := TMemoryStream.Create;         // MED: 09-06-01, added for D6
      Try
        ParamsToStream(FParams, FParamsStrm);      // MED: 09-06-01, added for D6

        sError := dll_sql_Prepare(
                        Self,
                        SQLID,
                        OnServerProgress,
                        Ord(FDefaultTableType),
                        PChar(FDefaultPassword),
                        FDefaultTranslate,
                        iCount,
                        pchar(sTables),
                        pchar(sPath),
                        Deleted, Century,
                        MemoBlockSize, pchar(DateFormat),
                        FParamsStrm,
                        FReadOnly,  {CRVG : 2001-06-03, Added}
                        FCommitLevel,
                        iError);

        Result := SQLID;
      finally
        FParamsStrm.Free;
      end;
    end;
  except
    ClientError('ApolloConnection.sql_Prepare', 'sql_Prepare')
  end;
end;


{MED: 01-09-2000, added}
procedure TApolloConnection.ap_sql_ExecSQL(
                                            AccessMethod:TAccessMethod;
                                            WASQL: Word;
                                            SQL : String;
                                            FParams:TParams;
                                            var iError: integer;
                                            var sError:String);
var
  iSQLlen, i, iSize : Integer;
  BuffPtr : pointer;
  pMem    : TMemoryStream;
  pBuff   : Pchar;
  FParamsStrm : TMemoryStream;    // MED: 09-05-01 added
begin
  iError     := 0;
  sError     := '';
  try
    if (AccessMethod = amServer) then
    begin
      // MED: 04-05-01, added Streaming to ExecSQL
      FWriteStream.Clear;
      FWriteStream.Position := 0;
      iSQLLen               := length(trim(SQL));

      FWriteStream.Write(pchar(IntToHex(WASQL,8))^, 8);
      FWriteStream.Write(pchar(IntToHex(iSQLLen,8))^, 8);
      FWriteStream.Write(pchar(SQL)^, iSQLLen);

      BuffPtr := NIL;
      for i:= 0 to FParams.Count -1 do
      begin

        iSize := FParams.Items[i].GetDataSize;
        try
          try
            BuffPtr := AllocMem(iSize);
          except
            ClientError('ApolloConnection.sql_ExecSQL', 'Allocating memory of '+inttostr(iSize)+' size');
          end;
          FParams.Items[i].GetData( BuffPtr);
          FWriteStream.Write(PChar(IntToHex(iSize, 8))^, 8);
          FWriteStream.Write(PChar(IntToHex(ord(FParams.Items[i].DataType), 2))^, 2);
          FWriteStream.Write(PChar(IntToHex(ord(FParams.Items[i].ParamType), 2))^, 2);
          FWriteStream.Write(PChar(PadR(FParams.Items[i].name, 20, ' '))^, 20);
          FWriteStream.Write(PChar(BuffPtr)^, iSize);
        finally
          FreeMem(BuffPtr);
        end;
      end;
      pMem := ClientStreamSendReceive(sc_SQL_EXECSQL);
      if pMem <> nil then
      begin
        pBuff := NIL;
        Try
          pBuff := pchar(AllocMem(pMem.Size+1));  //add one in case we get no ErrorMsg
          pMem.Read( pchar(@pbuff[0])^, 8);
          pbuff[8] := #0;       // ErrorNo
          iError := StrToInt('$'+pbuff);
          if pMem.Size >8 then
          begin
            pMem.Read( pchar(@pbuff[0])^, pMem.Size-8);
            pbuff[pMem.Size-8] := #0;       // ErrorMsg
            sError := pbuff;
          end;
        Finally
          FreeMem(pBuff);
        End;
      end;
{      sRes    := ClientXmit( sc_SQL_EXECSQL, IntToHex(WASQL, 8)+   trim(SQL),  TRUE);
      iError  := StrToInt('$'+Substr(sRes,1,8));
      sError  := Substr(sRes, 9, length(sRes)- 8);
      }
    end
    else
    begin
      FParamsStrm := TMemoryStream.Create;
      Try
        ParamsToStream(FParams, FParamsStrm);      // MED: 09-06-01, added for D6
        sError := dll_sql_ExecSQL(WASQL, pchar(SQL), FParamsStrm, iError);
      Finally
        FParamsStrm.Free;
      End;
    end;
  except
    ClientError('ApolloConnection.sql_ExecSQL', 'sql_ExecSQL')
  end;
end;

{MED: 01-09-2000, added}
procedure TApolloConnection.ap_sql_OpenSQL(
                          AccessMethod:TAccessMethod;
                          WASQL: Word;                 //prepare before calling
                          SQL : String;
                          var RowCount:Integer ;
                          var FieldCount:Integer ;
                          var RecordSize: Integer;
                          var iError: Integer;
                          var sError:String);
var
  sRes : String;
//  vOldCursor  : TCursor; {MED:09-07-00, added}
begin
  RowCount  := 0;
  RecordSize:= 0;
  FieldCount:= 0;
  iError    := 0;
  sError    := '';
  //vOldCursor := Screen.Cursor;        {MED: 09-07-00, Added}
  try
    //Screen.Cursor := crSQLWait;

    if AccessMethod = amServer then
    begin
      sRes      := ClientXmit( sc_SQL_OPENSQL, IntToHex(WASQL, 8)+trim(SQL), TRUE);
      RowCount  := StrToInt('$'+Substr(sRes,1,8));
      FieldCount:= StrToInt('$'+Substr(sRes,9,8));
      RecordSize:= StrToInt('$'+Substr(sRes,17,8));
      iError    := StrToInt('$'+Substr(sRes,25,8));
      sError    := Substr(sRes, 33, length(sRes)- 32);
    end
    else
    begin
      //      FOptIndex := 0;   {LAS: 01-08-00, Added}
      //TheSQL    := FindSQL(WASQL);
      sError := dll_sql_OpenSQL(WASQL, pchar(SQL), RowCount,  FieldCount, RecordSize, iError);
    end;
    //Screen.Cursor := vOldCursor;
  except
    //Screen.Cursor := vOldCursor;
    ClientError('ApolloConnection.sql_OpenSQL', 'sql_OpenSQL')
  end;
end;


{MED: 01-09-2000, added}
procedure TApolloConnection.ap_sql_CloseSQL(AccessMethod:TAccessMethod;
                                            WASQL: Word);
var
  iError :Integer;
begin
  iError := 0;
  try
    if (AccessMethod = amServer) then
      ClientXmit( sc_SQL_CLOSESQL, IntToHex(WASQL, 8), TRUE)
    else
      dll_sql_CloseSQL(WASQL, iError);
  except
    ClientError('ApolloConnection.sql_CloseSQL', 'sql_CloseSQL')
  end;
end;


procedure  TApolloConnection.ap_sql_UnPrepare(
                                            AccessMethod:TAccessMethod;
                                            WASQL:Word );
var
  iError : Integer;
begin
  iError := 0;
  try
    if (AccessMethod = amServer) then
      ClientXmit( sc_SQL_UNPREPARE , IntToHex(WASQL, 8), TRUE)
    else
    begin
        dll_sql_UnPrepareSQL(WASQL, iError);
        //FreeAndNil(TheSQL.Query);
//        FSQLSession.Remove(WASQL);
//        FSQLSession.Pack;
    end;
  except
    ClientError('ApolloConnection.sql_UnPrepare', 'sql_UnPrepare')
  end;
end;

///////////////////////////////
function  TApolloConnection.ap_sql_GetStructure(
                                            AccessMethod:TAccessMethod;
                                            WASQL:Word):Pointer;
var
  iSize : Integer;
  sError : String;
  p : PChar;
begin
  iSize := 0;
  sError := '';
  Result := nil;
  try
    if AccessMethod = amServer then
    begin
      ClientStreamReceive( sc_SQL_GETSTRUCTURE, IntToHex(WASQL,8) );
    end
    else
    begin
      FReadStream.Clear;
      sError := dll_sql_GetStructure(WASQL,  p, iSize);    // set iSize to -1 if error
      if iSize > -1 then
          FReadStream.Write(pchar(p)^, iSize);

    end;
    FReadStream.Position := 0;
    Result := FReadStream;
  except
    ClientError('ApolloConnection.sql_GetStructure', 'sql_GetStructure');
  end;
end;

function TApolloConnection.ap_sql_GetRecord(
                                           AccessMethod:TAccessMethod;
                                           WASQL:Word;
                                           RecNo, FetchCount,
                                           RecordCount:Integer):Pointer;
var
  iSize, iRecRead  : Integer;
  sRec, sError : String;
   p : PChar;
begin
  iSize := 0;
  sError := '';
  Result := nil;

  if FetchCount > (RecordCount - RecNo) then
  	FetchCount := (RecordCount - RecNo) + 1;

  try
    if AccessMethod = amServer then
    begin
      ClientStreamReceive( sc_SQL_READRECORD,
                                  IntToHex(WASQL, 8)+
                                  IntToHex(RecNo, 8)+
                                  IntToHex(FetchCount,8));
    end
    else
    begin
      iRecRead    := 0;             // total records actually read
      sRec        := IntToHex(iRecRead, 8);             // initially then change
      FReadStream.Clear;
      sError := dll_sql_GetRecord(WASQL, p, RecNo, FetchCount, iRecRead, iSize);
      if iSize > -1 then
        FReadStream.Write(pchar(p)^, iSize);
    end;
    FReadStream.Position := 0;
    Result := FReadStream;
  except
    ClientError('ApolloConnection.sql_GetRecord', 'sql_GetRecord');
  end;

end;

///////////////////////////////
function  TApolloConnection.ap_sql_ReadBlobData(
                                              AccessMethod:TAccessMethod;
                                              WASQL, RecNo, FieldIndex:Integer):TMemoryStream;
var
  iSize : Integer;
  sError  : String;
  p : PChar;
begin
  iSize := 0;
  sError := '';
  Result := nil;
  Try
    if AccessMethod = amServer then
    begin
      ClientStreamReceive( sc_SQL_READBLOBDATA,
                                  IntToHex(WASQL, 8)+
                                  IntToHex(RecNo,8)+
                                  IntToHex(FieldIndex, 8) );
    end
    else
    begin
      FReadStream.Clear;
      sError := dll_sql_ReadBlobData(WASQL, p, RecNo, FieldIndex, iSize);
      if iSize > -1 then
        FReadStream.Write(pchar(p)^, iSize);
    end;
    FReadStream.Position := 0;
    Result := FReadStream;
  Except
    ClientError('ApolloConnection.sql_ReadBlobData', 'sql_ReadBlobData');
  End;
end;

function  TApolloConnection.ap_sql_RecCount(AccessMethod:TAccessMethod;
                                            WASQL:Word): LongInt;
Var
  iError : Integer;
  iRecCount : Integer;
  sError  : String;
begin
  Result := 0;
  try
    if AccessMethod = amServer then
      Result := StrToInt('$'+Trim(ClientXmit( sc_SQL_RECCOUNT , IntToHex(WASQL, 8), True)))
    else
    begin
      sError := dll_sql_RecCount(WASQL, iRecCount, iError);
      Result := iRecCount;
    end;
  except
    ClientError('ApolloConnection.sql_RecCount', 'sql_RecCount');
  end;
end;

procedure  TApolloConnection.ap_sql_Refresh(AccessMethod:TAccessMethod; WASQL:Word );
Var
  iError : Integer;
  sError  : String;
begin
  try
    if AccessMethod = amServer then
      ClientXmit( sc_SQL_REFRESH , IntToHex(WASQL, 8), TRUE)
    else
    begin
      sError := dll_sql_Refresh(WASQL, iError);
    end;
  except
    ClientError('ApolloConnection.sql_Refresh', 'sql_Refresh')
  end;
end;


///////////////////////////////////////////////////
//MED: 09-09-01, added
///////////////////////////////////////////////////
procedure TApolloConnection.ap_SetSystemCollation;
begin
  try
    ClientXmit( sc_SetSystemCollation , '', True );
  except
    ClientError('ApolloConnection.SetSystemCollation', 'SetSystemCollation');
  end;
end;

procedure TApolloConnection.ap_SetMachineCollation;
begin
  try
    ClientXmit( sc_SetMachineCollation , '', True );
  except
    ClientError('ApolloConnection.SetMachineCollation', 'SetMachineCollation');
  end;
end;

procedure TApolloConnection.ap_AddDudenCollation;
begin
  try
    ClientXmit( sc_AddDudenCollation , '', True );
  except
    ClientError('ApolloConnection.AddDudenCollation', 'AddDudenCollation');
  end;
end;

procedure TApolloConnection.ap_AddEtecCollation;
begin
  try
    ClientXmit( sc_AddEtecCollation , '', True );
  except
    ClientError('ApolloConnection.AddEtecCollation', 'AddEtecCollation');
  end;
end;



///////////////////////////////////////////////////
//MED: 09-08-01, added
///////////////////////////////////////////////////
procedure TApolloConnection.ap_sql_SetSystemCollation(AccessMethod:TAccessMethod);
begin
  try
    if AccessMethod = amServer then
      ClientXmit( sc_SQL_SetSystemCollation , '', True )
    else
      dll_sql_SetSystemCollation;
  except
    ClientError('ApolloConnection.sql_SetSystemCollation', 'sql_SetSystemCollation');
  end;
end;

procedure TApolloConnection.ap_sql_SetMachineCollation(AccessMethod:TAccessMethod);
begin
  try
    if AccessMethod = amServer then
      ClientXmit( sc_SQL_SetMachineCollation , '', True )           
    else
      dll_sql_SetMachineCollation;
  except                                             
    ClientError('ApolloConnection.sql_SetMachineCollation', 'sql_SetMachineCollation');
  end;
end;

procedure TApolloConnection.ap_sql_AddDudenCollation(AccessMethod:TAccessMethod);
begin
 try
    if AccessMethod = amServer then
      ClientXmit( sc_SQL_AddDudenCollation , '', True )
    else
      dll_sql_AddDudenCollation;
  except
    ClientError('ApolloConnection.sql_AddDudenCollation', 'sql_AddDudenCollation');
  end;
end;

procedure TApolloConnection.ap_sql_AddEtecCollation(AccessMethod:TAccessMethod);
begin
 try
    if AccessMethod = amServer then
      ClientXmit( sc_SQL_AddEtecCollation , '', True )
    else
      dll_sql_AddEtecCollation;
  except
    ClientError('ApolloConnection.sql_AddEtecCollation', 'sql_AddEtecCollation');
  end;
end;


end.
