{Updated: 03-11-01}

unit ApWin;


//{$R *.dcr}        {MED: 08-10-99, commented}

interface

uses
  Classes,
  SysUtils,
  Windows,
  ApGlobal,
  Winsock,
  ApCommon,   //MED: 03-25-01, added
  Dialogs;         // remove before production


var

  sStackDescription: String = '';
  iMaxSockets: Integer = 0;
  iMaxUDPSize: Integer = 0;

Const
  TCPIP_BUFFERCHUNK = 8192;       {MED: 02-20-01, added}
    
type

  TProgressEvent = procedure(Sender: TComponent; const plWork, plWorkSize: LongInt) of object;

  TReadProgressEvent  = procedure(Sender: TComponent; Min, Max, Position: Integer) of object;
  TWriteProgressEvent = procedure(Sender: TComponent; Min, Max, Position: Integer) of object;
//  TServerProgressEvent = procedure(Sender: TComponent; Min, Max, Position: Integer) of object;

  ExceptionClass = class of Exception;
  EApSDEException = class(Exception);
  TClassApSDEException = class of EApSDEException;

  TSunB = packed record
    s_b1, s_b2, s_b3, s_b4: u_char;
  end;

  EApSDESocketError = class(EApSDEexception)
  private
    FiLastError: Integer;
  public
    constructor CreateError(const sMsg: string; const iErr, iVoid: Integer); virtual;
    property LastError: Integer read FiLastError;
  end;

  EApSDEResponseError = class(EApSDEException);
  EApSDEClosedSocket = class(EApSDEException);
  EApSDEInvalidSocket = class(EApSDEException);
  EApSDEConnClosedGraceful = class(EApSDEException);
  EApSDEConnectRefused = class(EApSDEException);
  EApSDEGetMessage = class(EApSDEException);
  EApSDEInterrupted = class(EApSDEException);
  EApSDETimedOut = class(EApSDEInterrupted);
  EApSDENotWaiting = class(EApSDEException);
  EApSDENotListening = class(EApSDEexception);
  EApSDEAlreadyListening = class(EApSDEException);
  EApSDEAlreadyConnected = class(EApSDEException);
  EApSDEAlreadyConnecting = class(EApSDEException);

const {THERE ARE MORE, BUT THESE ARE THE BASIC PROTOCOLS: [OZZ]}
  WSPORT_ECHO   = 7;
  WSPORT_DISCARD= 9;
  WSPORT_SYSTAT = 11;
  WSPORT_DAYTIME= 13;
  WSPORT_NETSTAT= 15;
  WSPORT_QOTD   = 17;
  WSPORT_CHARGEN= 19; {UDP Server!}
  WSPORT_FTP    = 21;
  WSPORT_TELNET = 23;
  WSPORT_SMTP   = 25;
  WSPORT_TIME   = 37;
  WSPORT_WHOIS  = 43;
  WSPORT_DOMAIN = 53;
  WSPORT_TFTP   = 69;
  WSPORT_FINGER = 79;
  WSPORT_HTTP   = 80;
  WSPORT_POP2   = 109;
  WSPORT_POP3   = 110;
  WSPORT_AUTH   = 113;
  WSPORT_NNTP   = 119;
  WSPORT_SSL    = 443;

  wsOk = 1;
  wsErr = 0;


type

  // Temp hack for Kelp parsing
  TDummy = class
  end;

  TApSDE = class(TComponent)
  private
    fiPort: Integer;
    fsBoundIP: string;
    fslstLocalAddresses: TStrings;
    //
    function GetLocalAddress: string;
    function GetLocalAddresses: TStrings;
    function GetLocalName: string;
    // Cannot be static method - CBuilder doesnt handle it correctly for a prop accessor
    //function GetVersion: string;  {MED: 08-08-99, Added}
    //    procedure SetVersion(const Value: string);
    procedure PopulateLocalAddresses;
  protected
    fHandle: TSocket;
    FHexHandle : String;    
    //
    procedure AllocateSocket(const piSocketType: Integer); dynamic;
    procedure Bind;
    procedure Listen;
    procedure ListenNonDefault(const piQueueCount: integer);
  public
    procedure CloseSocket; virtual;
    function Connected: Boolean; Virtual;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Disconnect; Virtual;
    class procedure Raise_SocketError(const iErr: integer);
    class function ResolveHost(const psHost: string; var psIP: string): u_long;
    // Resolves host passed in sHost. sHost may be an IP or a HostName.
    // sIP returns string version of the IP
    class function ResolveIP(const psIP: string): string;

    class function CheckForSocketError(const iResult: integer): boolean;
    class function CheckForSocketError2(const iResult: integer;
     const aiIgnore: array of integer): boolean;
    // Pass 0 to ignore none
    class function SockErrMsg(const iErr: integer): string;
    class function TInAddrToString(prIP: TInAddr): string;
    //
    property Handle: TSocket read FHandle;
    property HexHandle: String read FHexHandle write FHexHandle;  {Med:09-05-00, added}    
    property LocalAddress: string read GetLocalAddress;
    property LocalAddresses: TStrings read GetLocalAddresses;
    property LocalName: string read GetLocalName;

    property BoundIP: string read fsBoundIP write fsBoundIP; {MED: 08-08-99, moved from published}
  published

    property Port: Integer read FiPort write FiPort;

    //    property Version: string read GetVersion write SetVersion stored false;    {MED: 08-08-99, Added}
  end;

  TApSDESocket = class(TApSDE)
  private
    //    FConnectionBusy : Boolean ;  {02-20-01, added}
    FUser : String ;          {MED: 01-13-2000, added}
    FPassword : String ;          {MED: 10-15-2001, added}
    fbClosedGracefully, fbUseNagle: Boolean;
    fnLogFile, fsEOLTerminator: string;
    FbASCIIFilter: Boolean;
    FnResultNo: SmallInt;
    FlLastWork, FiBufferChunk, fiReadBufferPos: integer;
    lWork, lWorkSize: LongInt;
    strmLog: TFileStream;
    fOnStatus, fOnCaptureLine: TStringEvent;
    fOnWork: TProgressEvent;
    FOnReadProgress: TReadProgressEvent;        {MED: 12-25-99, added}
    FOnWriteProgress: TWriteProgressEvent;      {MED: 12-25-99, added}
    FOnServerProgress: TxProgressEvent;       {MED: 03-25-01, changed}

  protected
    fsPeerAddress, fsCommandResult, fsReadBuffer: string;
    FslstCommandResultLong: TStrings;

    procedure SRaise(exception: TClassApSDEException);
    property OnCaptureLine: TStringEvent read FOnCaptureLine write FOnCaptureLine;
/////////////////////
    procedure BeginWork(const lSize: LongInt);
    function Capture(objc: TObject; const asDelim: array of string): integer;
    procedure CaptureHeader(pslstHeaders: TStrings; const psDelim: string);
    procedure CheckForDisconnect;
    function Command(const sOut: string; const nResponse: SmallInt; const nLevel: byte;
     const sMsg: string): SmallInt; virtual;
    procedure ConvertHeadersToValues(pslstHeaders: TStrings);
    procedure DoWork(lNewWork: LongInt); Virtual;
    function ExtractXBytesFromBuffer(const piPos: Integer): string;
    procedure DoStatus(const sMsg: string); dynamic;
    procedure EndWork;
    function GetResponse: SmallInt;
    function Readable(const piMSec: Integer): boolean;
    procedure ReadToBuffer(var pBuffer; const piByteCount: Integer);
    function ReadBuffer: String;
    // ReadBuffer WILL wait for data if none exists
    function ReadBufferSize: Integer;
    function ReadByte: Byte;
    procedure ReadFromWinsock;
    function ReadShortInt: ShortInt;

    function ReadLnAndEcho(const psMask: string): string;
    function ReadLnWait: string;
    procedure SetBufferChunk(const iValue: Integer);
    function WriteFromBuffer(const pBuffer; const piByteCount: Integer):Integer; {MED: 02-20-01, changed to Function}

    procedure WriteHeaders(pslstHeaders: TStrings);
    procedure WriteTStrings(pslst: TStrings);
    //
    property CommandResult: string read FsCommandResult;
    property CommandResultLong: Tstrings read FslstCommandResultLong;
    property EOLTerminator: string read fsEOLTerminator write fsEOLTerminator;

    property ResultNo: SmallInt read FnResultNo;

    {MED: 08-08-99, Moved to protected}
    property ASCIIFilter: boolean read FbASCIIFilter write FbASCIIFilter default False;
    property BufferChunk: Integer read FiBufferChunk write SetBufferChunk;
    property LogFile: string read FnLogFile write FnLogFile;
    property UseNagle: Boolean read FbUseNagle write FbUseNagle default TRUE;   {MED: 12-24-99, changed to TRUE}
    property ClosedGracefully: Boolean read fbClosedGracefully write fbClosedGracefully;

    //////////////
  public
    //property  ConnectionBusy : Boolean read FConnectionBusy write FConnectionBusy;
    constructor Create(AOwner: TComponent); override;     {MED: 03-26-01, moved from protected}
    destructor Destroy; override;                         {MED: 03-26-01, moved from protected}

    property User       : String read FUser write FUser;   {MED: 01-13-2000}
    property Password   : String read FPassword write FPassword;   {MED: 10-15-2001}
    property PeerAddress: string read FsPeerAddress;

    procedure WriteLn(const sOut: string);
    function  ReadLn: string;
    function  Read(const piLen: Integer): string; virtual; {MED: 01-20-2000, moved from private}

    // MED: 05-25-02, added for later implemenation in ApConn.FillReadStream
    function  ReadWaitSec(const piLen: Integer; piMS:Integer): string; virtual; 

    function  Write(psOut: string):Integer; {MED: 02-20-01, changed to function}
    procedure ReadToStream(strm: TStream; const iCount: Integer);
    function  WriteToStream(strm: TStream; const bAll: Boolean):Integer;    {MED: 08-12-99, changed from WriteStream}
    function  WriteToStream2(sHeader:String; strm: TStream):Integer;    {MED: 04-03-01, Added}

  published

    property OnStatus: TStringEvent read FOnStatus write FOnStatus;

//    property OnWork: TProgressEvent read FOnWork write FOnWork;  {MED: 08-10-99, commented}

    property OnReadProgress: TReadProgressEvent read FOnReadProgress write FOnReadProgress;  {MED: 12-25-99, Added}
    property OnWriteProgress: TWriteProgressEvent read FOnWriteProgress write FOnWriteProgress;  {MED: 12-25-99, Added}
    property OnServerProgress: TxProgressEvent read FOnServerProgress write FOnServerProgress;  {MED: 12-29-99, Added}
  end;

  TApSDEClient = class(TApSDESocket)
  private
    fsUser : String;                   {MED: 09-04-00, added}
    fsHost, fsPassword, fsUserID : string;
  protected
    property UserID: string read FsUserID write FsUserID;
  	procedure SetUser( Value: String );  {LAS: 04-20-00, Added}   {LOREN, This will suffice you don't need it in the server class above--MED}
  public
    procedure Connect; virtual;
    destructor Destroy; override;
    procedure Disconnect; override;
  published
    property Host: string read FsHost write FsHost;
  	property User: string read fsUser write SetUser;  {LAS: 04-20-00, Changed to SetUser}
    property Password: string read FsPassword write FsPassword;  {09-04-00, moved from protected}

  end;

  function GmtOffsetStrToDateTime(s: string) : TDateTime;
  function DateTimeToGmtOffSetStr(dttm: TDateTime; bSubGMT: Boolean) : string;
  //  function UnixDateStrToDateTime(s: string) : TDateTime;          {MED: 12-29-99, commented.  Doesn't seem it is used}

const
  MultiPartBoundary = '=_NextPart_2rfksadvnqw3nerasdf';

implementation

{MED: 08-08-99, Added}
{
procedure Register;
begin
end;
}

procedure TApSDESocket.Sraise;
begin
  raise Exception.Create(FsCommandResult);
end;

function TApSDE.Connected;
begin
  Result := Handle <> INVALID_SOCKET;
end;

procedure TApSDE.Disconnect;
begin
  if Connected then
  begin
    CloseSocket;
  end;
end;

class function TApSDE.ResolveHost;
var
  pa: PChar;
  sa: TInAddr;
  aHost: PHostEnt;
begin
  psIP := psHost;
  // Sometimes 95 forgets who localhost is
  if CompareText(psHost, 'LOCALHOST') = 0 then begin
    sa.S_un_b.s_b1 := #127;
    sa.S_un_b.s_b2 := #0;
    sa.S_un_b.s_b3 := #0;
    sa.S_un_b.s_b4 := #1;
    psIP := '127.0.0.1';
    Result := sa.s_addr;
  end else begin
    // Done if is tranlated (ie There were numbers}
    Result := inet_addr(PChar(psHost));
    // If no translation, see if it resolves}
    if Result = u_long(INADDR_NONE) then begin
      aHost := gethostbyname(PChar(psHost));
      if aHost = nil then
      begin
      	CheckForSocketError(SOCKET_ERROR);
      end
      else
      begin
        pa := aHost^.h_addr_list^;
        sa.S_un_b.s_b1 := pa[0];
        sa.S_un_b.s_b2 := pa[1];
        sa.S_un_b.s_b3 := pa[2];
        sa.S_un_b.s_b4 := pa[3];
        psIP := TInAddrToString(sa);
      end;
      Result := sa.s_addr;
    end;
  end;
end;

constructor TApSDESocket.Create;
begin
  inherited Create(AOwner);

  fHexHandle      := '0000';         {MED:09-05-00, added}
  fsEOLTerminator := #13#10;
  fiReadBufferPos := 1;
  FiBufferChunk   := TCPIP_BUFFERCHUNK;
  FbUseNagle      := TRUE;         
  FslstCommandResultLong := TstringList.Create;
end;

procedure TApSDESocket.SetBufferChunk;
begin
  if Connected or (csDesigning in ComponentState) or (csLoading in ComponentState) then
    exit;

  fiBufferChunk := iValue;
  if not (csDesigning in ComponentState) then
    SetLength(fsReadBuffer, fiBufferChunk);
end;

destructor TApSDESocket.Destroy;
begin
  FslstCommandResultLong.Free;
  inherited Destroy;
end;


class function TApSDE.CheckForSocketError(const iResult: integer): boolean;
begin
  result := TApSDE.CheckForSocketError2(iResult, [0]);
end;

// Returns True if error was ignored (Matches iIgnore), false if no error occurred
class function TApSDE.CheckForSocketError2(const iResult: integer;
    const aiIgnore: array of integer): boolean;
var
  i, iErr: integer;
begin
  Result := false;
  if iResult = SOCKET_ERROR then
  begin

    iErr := WSAGetLastError;

    for i := Low(aiIgnore) to High(aiIgnore) do
    begin
      if iErr = aiIgnore[i] then
      begin
        Result := true;
        exit;
      end;
    end;
    Raise_SocketError(iErr);
  end;
end;

class function TApSDE.SockErrMsg;
begin
  Result := '';
  case iErr of
    WSAEINTR: Result           := 'Interrupted system call.';
    WSAEBADF: Result           := 'Bad file number.';
    WSAEACCES: Result          := 'Access denied.';
    WSAEFAULT: Result          := 'Bad address.';
    WSAEINVAL: Result          := 'Invalid argument.';
    WSAEMFILE: Result          := 'Too many open files.';

    WSAEWOULDBLOCK: Result     := 'Operation would block. ';
    WSAEINPROGRESS: Result     := 'Operation now in progress.';
    WSAEALREADY: Result        := 'Operation already in progress.';
    WSAENOTSOCK: Result        := 'Socket operation on non-socket.';
    WSAEDESTADDRREQ: Result    := 'Destination address required.';
    WSAEMSGSIZE: Result        := 'Message too long.';
    WSAEPROTOTYPE: Result      := 'Protocol wrong type for socket.';
    WSAENOPROTOOPT: Result     := 'Bad protocol option.';
    WSAEPROTONOSUPPORT: Result := 'Protocol not supported.';
    WSAESOCKTNOSUPPORT: Result := 'Socket type not supported.';
    WSAEOPNOTSUPP: Result      := 'Operation not supported on socket.';
    WSAEPFNOSUPPORT: Result    := 'Protocol family not supported.';
    WSAEAFNOSUPPORT: Result    := 'Address family not supported by protocol family.';
    WSAEADDRINUSE: Result      := 'Address already in use.';
    WSAEADDRNOTAVAIL: Result   := 'Can''t assign requested address.';
    WSAENETDOWN: Result        := 'Network is down.';
    WSAENETUNREACH: Result     := 'Network is unreachable.';
    WSAENETRESET: Result       := 'Net dropped connection or reset.';
    WSAECONNABORTED: Result    := 'Software caused connection abort.';
    WSAECONNRESET: Result      := 'Connection reset by peer.';
    WSAENOBUFS: Result         := 'No buffer space available.';
    WSAEISCONN: Result         := 'Socket is already connected.';
    WSAENOTCONN: Result        := 'Socket is not connected.';
    WSAESHUTDOWN: Result       := 'Cannot send or receive after socket is closed.';
    WSAETOOMANYREFS: Result    := 'Too many references, can''t splice.';
    WSAETIMEDOUT: Result       := 'Connection timed out.';
    WSAECONNREFUSED: Result    := 'Connection refused.';
    WSAELOOP: Result           := 'Too many levels of symbolic links.';
    WSAENAMETOOLONG: Result    := 'File name too long.';
    WSAEHOSTDOWN: Result       := 'Host is down.';
    WSAEHOSTUNREACH: Result    := 'No route to host.';
    WSAENOTEMPTY: Result       := 'Directory not empty';
    WSAEPROCLIM: Result        := 'Too many processes.';
    WSAEUSERS: Result          := 'Too many users.';
    WSAEDQUOT: Result          := 'Disk Quota Exceeded.';
    WSAESTALE: Result          := 'Stale NFS file handle.';
    WSAEREMOTE: Result         := 'Too many levels of remote in path.';

    WSASYSNOTREADY: Result     := 'Network subsystem is unavailable.';
    WSAVERNOTSUPPORTED: Result := 'WINSOCK DLL Version out of range.';
    WSANOTINITIALISED: Result  := 'Winsock not loaded yet.';

    WSAHOST_NOT_FOUND: Result  := 'Host not found.';
    WSATRY_AGAIN: Result       := 'Non-authoritative response (try again or check DNS setup).';
    WSANO_RECOVERY: Result     := 'Non-recoverable errors: FORMERR, REFUSED, NOTIMP.';
    WSANO_DATA: Result         := 'Valid name, no data record (check DNS setup).';
  end;
  Result := 'Socket Error #' + IntToStr(iErr) + EOL + Result;
end;

function TApSDESocket.Command;
begin
  if sOut <> CHAR0 then
  begin
    if strmLog <> nil then
      StreamWriteLn(strmLog, 'O: ' + sOut);
    WriteLn(sOut);
  end;

  GetResponse;
  if (nResponse > -1) and (ResultNo <> nResponse) then
    raise EApSDEResponseError.Create(CommandResult);

  if sMsg = '*' then
    DoStatus(stringOfChar(' ', nLevel * 2) + CommandResult)
  else if length(sMsg) > 0 then
    DoStatus(stringOfChar(' ', nLevel * 2) + sMsg);

  Result := ResultNo;
end;

function TApSDESocket.GetResponse;
var
  s, t: string;
begin
  FslstCommandResultLong.Clear;
  s := ReadLnWait;
  if strmLog <> nil then
    StreamWriteLn(strmLog, 'I: ' + s + EOL);
  if length(s) > 3 then
  begin
    if s[4] = '-' then
    begin {Multi line response coming}
      t := Copy(s, 1, 3) + CHAR32;
      FslstCommandResultLong.Add(s);
      repeat
        s := ReadLnWait;
        if strmLog <> nil then
          StreamWriteLn(strmLog, 'I: ' + s + EOL);
        FslstCommandResultLong.Add(s);
      until Copy(s, 1, 4) = t;
    end;
  end;
  FsCommandResult := s;

  if Copy(FsCommandResult, 1, 3) = '+OK' then
    FnResultno := wsOK
  else if Copy(FsCommandResult, 1, 4) = '-ERR' then
    FnResultno := wsErr
  else
    FnResultno := StrToIntDef(Copy(FsCommandResult, 1, 3), 0);

  Result := ResultNo;
end;

function TApSDESocket.ReadLn;
var
  i: Integer;
begin
  repeat
    {TODO make searches more efficient}
    i := Pos(EOLTerminator, Copy(fsReadBuffer, 1, fiReadBufferPos - 1));
    // ReadFromWinsock blocks - dont call unless we need to
    if i = 0 then
      ReadFromWinsock;
  until i > 0;
  Result := ExtractXBytesFromBuffer(i + Length(EOLTerminator) - 1);
  SetLength(Result, i - 1);
end;

procedure TApSDESocket.ReadToStream(strm: TStream; const iCount: Integer);
var
  iChunk, iTotal: integer;
begin
  //strm.SetSize(iCount);
  strm.Position := 0;                       // MED: 04-03-2001

  iTotal:= iCount;

  if Assigned(FOnReadProgress) then                   {MED:12-25-99, added}
    FOnReadProgress(self, 0, iTotal, 0);

  while iTotal > 0 do
  begin
    if iTotal > FiBufferChunk then
      iChunk:= FiBufferChunk
    else
      iChunk:= iTotal;

    strm.WriteBuffer(Read(iChunk)[1], iChunk);
    dec(iTotal, iChunk);
    if Assigned(FOnReadProgress) then                   {MED:12-25-99, added}
        FOnReadProgress(self, 0, 0, strm.Position);
  end;
  if Assigned(FOnReadProgress) then                   {MED:01-08-00, added}
    FOnReadProgress(self, 0, 0, 0);

  strm.Position := 0;               // MED: 04-03-2001
end;

function TApSDESocket.ReadByte;
begin
  Result := Ord(Read(1)[1]);
end;

function TApSDESocket.ReadShortInt;
begin
  Result := ReadByte;
  Result := ReadByte * 256 + Result;
end;

procedure TApSDESocket.ReadFromWinsock;
var
  i, j: Integer;
begin
  if (Length(fsReadBuffer) - fiReadBufferPos) < (fiReadBufferPos div 2) then
    SetLength(fsReadBuffer, Length(fsReadBuffer) + FiBufferChunk)
  else if ((Length(fsReadBuffer) - fiReadBufferPos) > (FiBufferChunk * 2))
   and ((Length(fsReadBuffer) - FiBufferChunk) > fiBufferChunk) then
    SetLength(fsReadBuffer, Length(fsReadBuffer) - FiBufferChunk);

  UniqueString(fsReadBuffer);
  i := recv(Handle, fsReadBuffer[fiReadBufferPos], Length(fsReadBuffer) - fiReadBufferPos, 0);

  fbClosedGracefully := i = 0;
  if not fbClosedGracefully then
  begin
    //showmessage(inttostr(    WSAGetLastError));
    if CheckForSocketError2(i, [10054]) then      //MED:09-10-2002, added to handle sudden disconnect from server
    begin
      Disconnect;
      Raise_SocketError(10054);
    end;

    if CheckForSocketError2(i, [10058]) then
    begin
      Disconnect;
      Raise_SocketError(10058);
    end;

    DoWork(lWork + i);
    if fbASCIIFilter then
    begin
      for j := fiReadBufferPos to fiReadBufferPos + i - 1 do
        fsReadBuffer[i] := Chr(Ord(fsReadBuffer[i]) and $7F);
    end;
    fiReadBufferPos := fiReadBufferPos + i;
  end;
  CheckForDisconnect;
  {TODO Put all DoWorks here and in write - remove all others}
  // DoWork(lWork + piLen);
end;

function TApSDESocket.ExtractXBytesFromBuffer(const piPos: Integer): string;
begin
  if piPos >= fiReadBufferPos then
    raise Exception.Create('Not enough data in buffer.');

  result := Copy(fsReadBuffer, 1, piPos);
  fiReadBufferPos := fiReadBufferPos - piPos;
  Uniquestring(fsReadBuffer);
  MoveMemory(PChar(fsReadBuffer), PChar(fsReadBuffer) + piPos, fiReadBufferPos - 1);
end;

function TApSDESocket.Read;
begin
  result := '';
  // Called by ReadBuffer - May be 0
  if piLen = 0 then begin
    ReadFromWinsock;
    exit;
  end;

  while fiReadBufferPos <= piLen do
    ReadFromWinsock;

  Result := ExtractXBytesFromBuffer(piLen);
end;

{MED: 05-26-2002, added}
function  TApSDESocket.ReadWaitSec(const piLen: Integer; piMS:Integer): string; 
var
  rClock : TTickClock;
begin
  Result := '';
  rClock := TTickClock.Create(FALSE);
  Try
    // Called by ReadBuffer - May be 0
    if piLen = 0 then
    begin
      ReadFromWinsock;
    end
    else
    begin
      While (fiReadBufferPos <= piLen) AND (not rClock.Delay(piMs)) do
        ReadFromWinsock;

      Result := ExtractXBytesFromBuffer(piLen);
    end;
  finally
    rClock.Free;
  end;
end;

function TApSDESocket.ReadLnWait;
begin
  Result := '';
  while length(Result) = 0 do
    Result := Trim(ReadLn);
end;

function TApSDE.GetLocalName;
var
  s: String;
begin
  SetLength(s, MAX_PATH);
  CheckForSocketError(GetHostName(PChar(s), Length(s)));
  Result := String(PChar(s));
end;

function TApSDE.GetLocalAddress;
begin
	// Prevents "List out of bounds" error when no addresses listed
	Result := '127.0.0.1';              {LAS: 07-27-00, Added}
	if LocalAddresses.Count > 0 then    {LAS: 07-27-00, Added}
	  Result := LocalAddresses[0];
end;

procedure TApSDE.PopulateLocalAddresses;
type
  TaPInAddr = Array[0..250] of PInAddr;
  PaPInAddr = ^TaPInAddr;
var
  i: integer;
  aHost: PHostEnt;
  padrptr: PaPInAddr;
begin
  fslstLocalAddresses.Clear;

  aHost := GetHostByName(PChar(LocalName));
  if aHost = nil then
  begin
    CheckForSocketError(SOCKET_ERROR)
  end
  else
  begin
    padrptr := PaPInAddr(aHost^.h_addr_list);
    i := 0;
    while padrptr^[i] <> nil do
    begin
      fslstLocalAddresses.Add(INet_ntoa(padrptr^[I]^));
      Inc(I);
    end;
  end;
end;

procedure TApSDESocket.DoWork;
begin
  lWork := lNewWork;
  if (lWorkSize <> -2) then begin
    if (lWork = -1) or (lNewWork div 1024 > flLastWork) then begin
      flLastWork := lNewWork div 1024;
      if assigned(FOnWork) then
        FOnWork(Self, lWork, lWorkSize);
    end;
  end;
end;

procedure TApSDESocket.DoStatus;
begin
  if assigned(FOnStatus) then
  begin
    try
      FOnStatus(Self, sMsg);
    except
      on E: Exception do
        raise ExceptionClass(E.ClassType).Create(
          E.Message + ' occurred in user''s OnStatus Event.');
    end;
  end;
end;

procedure TApSDESocket.WriteLn;
begin
  Write(sOut + EOLTerminator);
end;

{MED: 08-08-99, Added}
{
}


procedure TApSDESocket.beginWork;
begin
  lWork := 0;
  lWorkSize := lSize;
end;

procedure TApSDESocket.endWork;
begin
  lWorkSize := -1;
  DoWork(-1);
  lWorkSize := -2;
end;

class procedure TApSDE.Raise_SocketError;
begin
  raise EApSDESocketError.CreateError(SockErrMsg(iErr), iErr, 0);
end;

constructor EApSDESocketError.CreateError(const sMsg: string; const iErr, iVoid: Integer); 
begin
  inherited Create(sMsg);
  FiLastError := iErr;
end;


function GmtOffsetStrToDateTime(s: string) : TDateTime;
begin       {-Typical string: -0500  or  +0200  or  -0130}
  Result := 0.0;
  S := Copy(Trim(s), 1, 5);

  if Length(s) > 0 then
  begin
    if s[1] in ['-', '+'] then
    begin
      try
        Result := EncodeTime(StrToInt(Copy(s, 2, 2)), StrToInt(Copy(s, 4, 2)), 0, 0);
        if s[1] = '-' then
          Result := -Result;
      except Result := 0.0; end;
    end;
  end;
end;

function DateTimeToGmtOffSetStr(dttm: TDateTime; bSubGMT: Boolean) : string;
var
  wHour, wMin, wSec, wMSec: Word;

  function ZeroPad(s: string) : string;
  begin
    if Length(s) < 2 then
      Result := '0' + s
    else
      Result := s;
  end;

begin
  if (dttm = 0.0) and bSubGMT then begin
    Result := 'GMT';
    Exit;
  end;

  DecodeTime(dttm, wHour, wMin, wSec, wMSec);
  Result := CHAR32 + ZeroPad(IntToStr(wHour)) + ZeroPad(IntToStr(wMin));
  if dttm < 0.0 then
    Result[1] := '-'
  else
    Result[1] := '+';
end;


function  TApSDESocket.Write(psOut: string):Integer; {MED: 02-20-01, changed to function}
//function TApSDESocket.Write:Integer;
begin
  Result := WriteFromBuffer(psOut[1], Length(psOut));
end;

function TApSDESocket.WriteFromBuffer(const pBuffer; const piByteCount: Integer):integer;
var
  iPos, i: Integer;
  BufferPtr: PChar;
begin
  Result := 0;

  if piByteCount = 0 then
    exit;
  if Not Connected then
    raise EApSDEClosedSocket.Create('Socket is closed, cannot write');

  BufferPtr := @pBuffer;
  iPos := 0;
  Repeat
    i := Send(Handle, BufferPtr[iPos], piByteCount - iPos, 0);

    fbClosedGracefully := i = 0;
    
    CheckForDisconnect;

    if CheckForSocketError2(i, [10054]) then      //MED:09-10-2002, added to handle sudden disconnect from server
    begin
      Disconnect;
      Raise_SocketError(10054);
    end;

    if CheckForSocketError2(i, [10058]) then
    begin
      Disconnect;
      Raise_SocketError(10058);
    end;


    iPos := iPos + i
  until iPos = piByteCount;
  Result := i;
end;

procedure TApSDEClient.Disconnect;
begin
  try
    DoStatus('Disconnecting');
    strmLog.Free;
    fsReadBuffer := '';
    fiReadBufferPos := 1;
    FsPeerAddress := '';
    inherited;
    DoStatus('Disconnected');
  except end;

end;

procedure TApSDEClient.Connect;
var
  sao: TSockAddr;
begin
  if Connected then
    raise EApSDEAlreadyConnected.Create('Already connected.');

  DoStatus('Connecting to ' + FsHost);
  SetBufferChunk(BufferChunk);
  AllocateSocket(SOCK_STREAM);
  try

    // 10-12-99 commented since UseNagle is TRUE now
    // if not UseNagle then      CheckForSocketError(SetSockOpt(FHandle, IPPROTO_TCP, TCP_NODELAY, '1111', 4));

    sao.sin_family := PF_INET;

    //  if FiPort > 65535	then
    //    sao.sin_port := htonl(FiPort)
    //  else
    sao.sin_port := htons(FiPort);  //MED: 09-21-2002 : Might have to change from htons (Range Check error)?

    sao.sin_addr.S_addr := ResolveHost(fsHost, fsPeerAddress);;

    CheckForSocketError(Winsock.Connect(Handle, sao, SizeOf(sao)));
    if length(FnLogFile) > 0 then
    begin
      strmLog := TFileStream.Create(FnLogFile, fmCreate);
    end;
    DoStatus('Socket Connected');

  except
    On E: Exception do
    begin
      Disconnect;
      raise;
    end;
  end;
  fbClosedGracefully := False;
end;

{LAS: 04-20-00, Added}
procedure TApSDEClient.SetUser( Value: String );
begin
	FsUser := UpperCase( Value );
end;


{MED: 04-03-01 Added}
function TApSDESocket.WriteToStream2(sHeader:String; strm: TStream ):Integer;
var
  sChunk: string;
  iSize: Integer;
begin
  Result := 0;
  if Write(sHeader) > 0 then
  begin
    strm.Position := 0;
    BeginWork(strm.Size);
    try
      if Assigned(FOnWriteProgress) then                   {MED:12-25-99, added}
        FOnWriteProgress(self, 0, strm.size, 0);
      while true do
      begin
        iSize := strm.Size - strm.Position;
        if iSize > TCPIP_BUFFERCHUNK then
          iSize := TCPIP_BUFFERCHUNK
        else if iSize = 0 then
          break;
        SetLength(sChunk, iSize);
        strm.ReadBuffer(sChunk[1], iSize);
        Write(sChunk);
        if Assigned(FOnWriteProgress) then                   {MED:12-25-99, added}
          FOnWriteProgress(self, 0,0, strm.Position);
      end;
      Result := strm.Position;

      if Assigned(FOnWriteProgress) then                   {MED:01-08-00, added}
        FOnWriteProgress(self, 0, 0, 0);
    Finally
      EndWork;
    end;
  end;
end;

{MED: 08-12-99 changed from WriteStream}
function TApSDESocket.WriteToStream(strm: TStream; const bAll: Boolean):Integer;    {MED: 08-12-99, changed from WriteStream}
var
  sChunk: string;
  iSize: Integer;
begin

  //FConnectionBusy := TRUE;    {MED: 02-20-01, added}
  if bAll then
    strm.Position := 0;
  BeginWork(strm.Size);
  try
    if Assigned(FOnWriteProgress) then                   {MED:12-25-99, added}
      FOnWriteProgress(self, 0, strm.size, 0);
    while true do
    begin
      iSize := strm.Size - strm.Position;
      if iSize > TCPIP_BUFFERCHUNK then
        iSize := TCPIP_BUFFERCHUNK
      else if iSize = 0 then
        break;
      SetLength(sChunk, iSize);
      strm.ReadBuffer(sChunk[1], iSize);
      Write(sChunk);
      if Assigned(FOnWriteProgress) then                   {MED:12-25-99, added}
        FOnWriteProgress(self, 0,0, strm.Position);
    end;
    Result := strm.Position;

    if Assigned(FOnWriteProgress) then                   {MED:01-08-00, added}
      FOnWriteProgress(self, 0, 0, 0);
  Finally
    EndWork;
    //FConnectionBusy := FALSE;    {MED: 02-20-01, added}
  end;

end;

function TApSDESocket.Readable(const piMSec: Integer): boolean;
var
  tmTo: TTimeVal;
  FDRead: TFDSet;
begin
  result := ReadBufferSize > 0;
  if result then
    exit;

  tmTo.tv_sec := piMSec div 1000;
  tmTo.tv_usec := piMSec mod 1000;

  FDRead.fd_count := 1;
  FDRead.fd_array[0] := Handle;

  if piMSec = -1 then
    Result := Select(0, @FDRead, nil, nil, nil) = 1
  else
    Result := Select(0, @FDRead, nil, nil, @tmTO) = 1;
end;

destructor TApSDEClient.Destroy;
begin                           
  if Connected then
    Disconnect;
  FOnStatus := nil;
  inherited Destroy;
end;



procedure TApSDE.Bind;
var
  m_addr: TSockAddrIn;
begin
  // Setup local address for receiving socket
  with m_addr do                                                         
  begin
    sin_family := PF_INET;

    // if FiPort > 65535	then
    //    sin_port := htonl(FiPort)  //MED: 09-21-2002 : Might have to change from htons (Range Check error)?
    // else
    sin_port := htons(FiPort);

    sin_addr.s_addr := INADDR_ANY;
    if length(BoundIP) > 0 then
      sin_addr.s_addr := inet_addr(PChar(BoundIP));
  end;

  if CheckForSocketError2(winsock.bind(FHandle, m_addr, sizeof(m_addr))
   , [WSAEADDRINUSE]) then
    raise EApSDEException.Create(
     'Could not bind socket. Address and port are already in use.');
end;

procedure TApSDE.ListenNonDefault;
begin
  CheckForSocketError(winsock.listen(FHandle, piQueueCount));
end;

procedure TApSDE.AllocateSocket;
begin
  FHandle := Socket(PF_INET, piSocketType, IPPROTO_IP);
  if FHandle = INVALID_SOCKET then
    raise EApSDEInvalidSocket.Create('Cannot allocate socket.');
end;

procedure TApSDE.Listen;
begin
  ListenNonDefault(5);
end;

procedure Main;
var
  sData: TWSAData;
begin
  if WSAStartup($101, sData) = SOCKET_ERROR then
    raise EApSDEException.Create('Winsock Initialization Error.');
  sStackDescription := StrPas(sData.szDescription);
  iMaxUDPSize := sData.iMaxUdpDg;
  iMaxSockets := sData.iMaxSockets;
end;

function TApSDESocket.Capture;
var
  s: string;
  i: Integer;
begin
  result := 0;
  beginWork(-1); try
    repeat

      // Retrieve next line
      s := ReadLn;

      // End of message?
      for i := 0 to High(asDelim) do begin
        if s = asDelim[i] then begin
          result := i + 1;
          exit;
        end;
      end;

      // For RFC 822 retrieves
      if (asDelim[0] = '.') and (Copy(s, 1, 2) = '..') then
        Delete(s, 1, 1);

      // Write to output
      if objc is TStrings then begin
        TStrings(objc).Add(s);
      end else if objc is TStream then begin
        StreamWriteLn(TStream(objc), s);
      end else if objc = nil then begin
        if assigned(fOnCaptureLine) then
          fOnCaptureLine(Self, s);
      end else begin
        raise Exception.Create('objc type not supported.');
      end;
    until false;
  finally EndWork; end;
end;

procedure TApSDESocket.CaptureHeader(pslstHeaders: TStrings; const psDelim: string);
begin
  pslstHeaders.Clear;
  Capture(pslstHeaders, psDelim);
  ConvertHeadersToValues(pslstHeaders);
end;

procedure TApSDESocket.WriteTStrings(pslst: TStrings);
var
  i: integer;
begin
  for i := 0 to Pred(pslst.Count) do
    WriteLn(pslst[i]);
end;

procedure TApSDESocket.ConvertHeadersToValues(pslstHeaders: TStrings);
var
  i: Integer;
  s: string;
  slst: TStringList;
begin
  slst := TStringList.Create; try
    for i := 0 to Pred(pslstHeaders.Count) do begin
      s := pslstHeaders[i];
      // Length will never be 0 - this would have terminated the header
      if s[1] in [#9, ' '] then begin
        s := ' ' + Trim(StringReplace(s, #9, ' ', [rfReplaceAll]));
        slst[pred(slst.Count)] := slst[pred(slst.Count)] + s;
      end else begin
        // No ReplaceAll flag - we only want to replace the first one
        slst.add(StringReplace(s, ': ', '=', []));
      end;
    end;
    pslstHeaders.Assign(slst);
  finally slst.free; end;
end;

procedure TApSDESocket.WriteHeaders(pslstHeaders: TStrings);
var
  i: Integer;
begin
  for i := 0 to Pred(pslstHeaders.Count) do
    // No ReplaceAll flag - we only want to replace the first one
    WriteLn(StringReplace(pslstHeaders[i], '=', ': ', []));
  WriteLn('');
end;

procedure TApSDESocket.CheckForDisconnect;
begin
  //showmessage('CheckForDisconnect');
  // ************************************************************* //
  // An exception may occur here. This only happens in the IDE and this is normal.
  // ************************************************************* //
  if fbClosedGracefully then
  begin
    Disconnect;
    raise EApSDEConnClosedGraceful.Create('Connection Closed Gracefully.');
  end;
end;

function TApSDESocket.ReadBuffer: String;
begin
  Result := '';
  if Readable(1) then
    ReadFromWinsock; // Must call first to adjust readbuffersize
  if ReadBufferSize > 0 then
    Result := Read(ReadBufferSize);
end;

function TApSDESocket.ReadBufferSize: Integer;
begin
  Result := fiReadBufferPos - 1;
end;

{
procedure TApSDE.SetVersion(const Value: string);
begin
  // Void
end;
}

function TApSDESocket.ReadLnAndEcho;
var
  c: Char;
begin
  result := '';
  Repeat
    c := Read(1)[1];
    if c = #9 then begin
      Write(#9#32#9);
      if length(Result) > 0 then
        SetLength(Result, length(result) - 1);
    end else if not (c in [#13, #10]) then begin
      Result := Result + c;
      if psMask = '' then
        Write(c)
      else
        Write(psMask);
    end;
  until c = #13;
  Write(EOL);
end;

class function TApSDE.TInAddrToString;
begin
  {TODO change to inet_ntoa()}
  result := IntToStr(Ord(prIP.S_un_b.s_b1)) + '.' + IntToStr(Ord(prIP.S_un_b.s_b2)) + '.'
   + IntToStr(Ord(prIP.S_un_b.s_b3)) + '.' + IntToStr(Ord(prIP.S_un_b.s_b4));
end;

constructor TApSDE.Create(AOwner: TComponent);
begin
  inherited;
  fHandle := INVALID_SOCKET;
end;

destructor TApSDE.Destroy;
begin
  fslstLocalAddresses.Free;
  inherited;
end;

class function TApSDE.ResolveIP(const psIP: string): string;
var
  i: Integer;
  P: PHostEnt;
begin
  result := '';
  if CompareText(psIP, '127.0.0.1') = 0 then begin
    result := 'LOCALHOST';
  end else begin
    i := inet_addr(PChar(psIP));
    P := GetHostByAddr(@i, 4, PF_INET);
    if P <> nil then
      result := P.h_name
    else
      CheckForSocketError2(SOCKET_ERROR, [WSANO_DATA]);
  end;
end;

function TApSDE.GetLocalAddresses: TStrings;
begin
  if fslstLocalAddresses = nil then begin
    fslstLocalAddresses := TStringList.Create;
    PopulateLocalAddresses;
  end;
  result := fslstLocalAddresses;
end;

procedure TApSDE.CloseSocket;
var
  OldHandle: THandle;
begin
  // Must be first, closing socket will trigger some errors, and they
  // may then check (in other threads) Connected, which checks this.
  OldHandle   := Handle;
  fHandle     := INVALID_SOCKET;
  FHexHandle  := '0000';          {med:09-05-00, ADDED}
  Winsock.CloseSocket(OldHandle);
end;

procedure TApSDESocket.ReadToBuffer(var pBuffer; const piByteCount: Integer);
var
  sBuffer: String;
begin
  sBuffer := Read(piByteCount);
  System.Move(sBuffer[1], pBuffer, Length(sBuffer));
end;

initialization
  Main;
finalization
  WSACleanup;
end.


