(*============================================================================
  ApoEnv.PAS
  Main TApolloEnv and TApolloDatabase code

  Copyright 1999-2002 Vista Software
  Last Updated: 12-22-2002

  NOTE: Tab stops were set at 2 spaces during development of this unit. If you
        are using a tab stop setting other than 2 spaces, the code indents
        may not line up accurately in your editor.
============================================================================*)
unit ApoEnv;                     

interface

uses
  SDE60, DB, Classes, SysUtils, DBCommon, Windows, dialogs, ApConn, ApCommon;

  {03-12-01, commented -- no need }
// {$I CS_BUILD.PAS}  // Modify this file to toggle between C/S build and std

const
  // Epoch Default
  EPOCH_DEFAULT = 1980;

  // Date types
  AMERICAN    = 0;
  ANSI        = 1;
  BRITISH     = 2;
  FRENCH      = 3;
  GERMAN      = 4;
  ITALIAN     = 5;
  SPANISH     = 6;
  WIN_DEFAULT = 99;

type
  {============================================================================
		TApolloEnv - Handles all Global, application-wide, settings
  ============================================================================}

  {MED: 10-28-00, implement this as published property for 7.0- see ApCommon}
  //TApolloErrorLevel = (ERRLEVEL_NONE, ERRLEVEL_FATAL, ERRLEVEL_STANDARD); {MED: 10-28-00, added}

	TApolloApplyType = (atLocal, atServer);
	TApolloApplyTo = set of TApolloApplyType;

  TApolloEnv = class(TComponent)
	private
  	FApolloApplyTo: TApolloApplyTo;
    FApolloConnection: TApolloConnection;
		FAutoRefresh: Boolean;
    FCentury : WordBool;
    FDateFormat : TApolloDateFrmt;
    FDeleted : WordBool;
    FLockTimeOut: Integer;
    FMemoBlockSize: Word;  {LAS: 06-23-00, Changed type from Integer}
    FOptimisticBuffer: Boolean;
    FSoftSeek: Boolean;
    FExact: Boolean;
  	FEpoch: Integer;
    FErrorLevel : Integer;
  	FCountNeedsRefresh: WordBool;
	  FErrorFunc: Integer;

	protected
		function  GetCentury: WordBool;
    function  GetDateFormat: TApolloDateFrmt;
    function  GetLockTimeout: Integer;
		function  GetDeleted: WordBool;
		function  GetEpoch: Integer;
    function  GetExact: WordBool;
		function  GetOptimisticBuffer: WordBool;
    function  GetMemoBlockSize: Word;   {LAS: 06-23-00, Changed type from Integer}
    function  GetSoftSeek: WordBool;
    procedure Loaded; override;

  public
    constructor Create(Owner: TComponent); override;
    destructor Destroy; override;

    function  ErrorLevel( const Value: Integer ): Integer;
    function  GetErrorLevel:Integer;

  	procedure CloseDataSets;
    procedure OpenDataSets;
		procedure RefreshDataSets;
    procedure SetCentury( const Value: WordBool );
		procedure SetDateFormat( const Value: TApolloDateFrmt );
    procedure SetDeleted( const Value: WordBool );
    procedure SetEpoch( const Value: Integer );
		function  SetErrorFunc( const pFunc, pData: Integer ): Integer;
    procedure SetExact( const Value: WordBool );
		procedure SetLockTimeOut( const Value: Integer );
		procedure SetOptimisticBuffer( const Value: WordBool );
    procedure SetMemoBlockSize( const Value: Word );  {LAS: 06-23-00, Changed type from Integer}
    procedure SetSoftSeek( const Value: WordBool );
		function SysProp( uiSysItem: Word; vpData: Pointer ): Integer;
  published
  	property ApolloConnection: TApolloConnection read FApolloConnection write FApolloConnection;
    property ApplyTo: TApolloApplyTo read FApolloApplyTo write FApolloApplyTo;
		property AutoRefresh: Boolean read FAutoRefresh write FAutoRefresh default True;
    property Century: WordBool read GetCentury write SetCentury default True;
		property DateFormat: TApolloDateFrmt read GetDateFormat write SetDateFormat
                                            default dfWinDefault;
    property Deleted: WordBool read GetDeleted write SetDeleted default False;
		property Epoch: Integer read GetEpoch write SetEpoch default EPOCH_DEFAULT;
    property Exact: WordBool read GetExact write SetExact default False;
    property ErrorFunc : integer read FErrorFunc;
    property CountNeedsRefresh: WordBool read	FCountNeedsRefresh;
		property LockTimeOut: Integer read GetLockTimeOut write SetLockTimeOut
                                                       default 1;
    property MemoBlockSize: Word read GetMemoBlockSize     {LAS: 06-23-00, Changed type from Integer}
                                    write SetMemoBlockSize default 64;
		property OptimisticBuffer: WordBool read GetOptimisticBuffer
                                    write SetOptimisticBuffer default False;
    property SoftSeek: WordBool read GetSoftSeek write SetSoftSeek
                                                 default False;
  end;


  {============================================================================
		TApolloDatabase - Handles database path aliasing
  ============================================================================}
  TApolloDatabase = class(TComponent)
  private
  	FDataPath: String;
    FDataPathAlias: String;
    //\\ new implementation for D6 bug (datamodule)
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    constructor Create(Owner: TComponent); override;
    destructor Destroy; override;
  	procedure CloseDataSets;
    procedure EnumTableNames(List: TStrings; const Ext: string = 'dbf'); // RH d.d. 17-12-2002
    procedure EnumDataSets(List: TList); // RH d.d. 17-12-2002
    procedure OpenDataSets;
		procedure RefreshDataSets;
  published
    property DataPath: String read FDataPath write FDataPath;
    property DataPathAlias: String read FDataPathAlias write FDataPathAlias;
  end;

  // Changed to a function, which will create the Object as needed - RH d.d. 17-12-2002
  function apDatabaseList: TList;

  procedure apAddDatabase( oComp: TComponent);
  procedure apRemoveDatabase( oComp: TComponent);

implementation

uses ApoDSet;

var
  _apDatabaseList: TList;
type
{ -------------------------------------------------------------------------
   - RH d.d. 17-12-2002

  The object TapdbStringList will function as a normal TStringList, but
  will prevent a user from being able to free the Stringlistobject.

  The object is internally Freed - By InternalDestroy
  (Free cals Destroy - wich will raise an error)
  ------------------------------------------------------------------------- }

  TapdbStringList = class(TList)
  public
    Destructor Destroy; override;
    Destructor InternalDestroy;
  end;

{ TapdbStringList }

destructor TapdbStringList.Destroy;
begin
  Raise Exception.CreateFmt('Cannot manually destroy a %s', [Self.ClassName]);
end;

destructor TapdbStringList.InternalDestroy;
begin
  if _apDatabaseList = Self then
    _apDatabaseList := nil;
  inherited Destroy;
end;

// End of TapdbStringList declaration - RH d.d. 17-12-2002

function apDatabaseList: TList;
begin
  if not Assigned(_apDatabaseList) then
    _apDatabaseList := TapdbStringList.Create;
  Result := _apDatabaseList;
end;

//------------------------------------------------------------------------------
procedure apAddDatabase( oComp: TComponent);
begin
  apDatabaseList.Add(oComp);
end;
//------------------------------------------------------------------------------
procedure apRemoveDatabase( oComp: TComponent);
var
  i : integer;
begin
  i := apDatabaseList.IndexOf( oComp );
  apDatabaseList.Delete(i);
end;

{==============================================================================
  TApolloEnv
 ==============================================================================}
{==============================================================================
  Create
 ==============================================================================}
constructor TApolloEnv.Create(Owner: TComponent);
begin
  inherited;
  FApolloApplyTo := [atLocal];  // Default to True for Local

	// Initialize global (ENV) variables
	FDateFormat := dfWinDefault;
  FLockTimeOut := 1;
	FMemoBlockSize := 64;
  FCentury := True;
	FEpoch := EPOCH_DEFAULT;
  FErrorFunc := -1;
  FOptimisticBuffer := False;
  FAutoRefresh := True;

end;

{==============================================================================
  Destroy
 ==============================================================================}
destructor TApolloEnv.Destroy;
begin
  inherited;
end;

{==============================================================================
 Loaded
==============================================================================}
procedure TApolloEnv.Loaded;
begin
  inherited;
	if (FApolloApplyTo = [atLocal]) then
    LoadDLL;
end;

{LAS: 04-20-00, Rewritten}
{==============================================================================
  CloseDataSets
  Closes all TApolloTables in entire application
 ==============================================================================}
procedure TApolloEnv.CloseDataSets;
var
  Temp: TComponent;
  SearchObj : TApolloSearchObject;
begin
	SearchObj := TApolloSearchObject.Create;
  Try
    Temp      := SearchObj.SearchFirstComponent(TApolloTable);
    while Temp <> nil do
    begin
      if (Temp is TApolloTable) then
          TApolloTable(Temp).Close;
      Temp := SearchObj.SearchNextComponent(TApolloTable);
    end;
  Finally
    SearchObj.Free;
  End;
end;

{LAS: 04-20-00, Rewritten}
{==============================================================================
  OpenDataSets
  Opens all TApolloTables in entire application
 ==============================================================================}
procedure TApolloEnv.OpenDataSets;
var
  Temp: TComponent;
  SearchObj : TApolloSearchObject;
begin
	SearchObj := TApolloSearchObject.Create;
  Try
    Temp      := SearchObj.SearchFirstComponent(TApolloTable);
    while Temp <> nil do
    begin
      if (Temp is TApolloTable) then
      Try
        TApolloTable(Temp).Open;
      Except
        On E: EApolloError do
          MessageDlg(E.Message, mtError, [mbOK], 0);
      End;

      Temp := SearchObj.SearchNextComponent(TApolloTable);
    end;
  Finally
    SearchObj.Free;
  End;
end;

{LAS: 04-20-00, Rewritten}
{==============================================================================
  RefreshDataSets
  Refreshes all TApolloTables in entire application
 ==============================================================================}
procedure TApolloEnv.RefreshDataSets;
var
  Temp: TComponent;
  SearchObj : TApolloSearchObject;
begin
  if not FAutoRefresh then Exit;      (*Add this line to see how it should work*)

	SearchObj := TApolloSearchObject.Create;
  Try
    Temp := SearchObj.SearchFirstComponent(TApolloTable);
    while Temp <> nil do
    begin
      if (Temp is TApolloTable) and (TApolloTable(Temp).Active) then
          TApolloTable(Temp).Refresh;
      Temp := SearchObj.SearchNextComponent(TApolloTable);
    end;
  Finally
    SearchObj.Free;
  End;
end;


{==============================================================================
  ErrorLevel
 ==============================================================================}
function TApolloEnv.ErrorLevel( const Value : Integer ) : Integer;
begin
	Result := 0;
  if (atLocal in FApolloApplyTo) then
  begin
  	LoadDLL;
    FErrorLevel := Value;
	  Result      := sx_ErrorLevel( Value );
	end;
end;

function TApolloEnv.GetErrorLevel:Integer;
begin
	Result := FErrorLevel;
end;

{==============================================================================
  SetLockTimeOut
 ==============================================================================}
procedure TApolloEnv.SetLockTimeOut( const Value : Integer );
begin
	if FApolloApplyTo = [] then
  	exit;

  if (Value <> FLockTimeOut) then
  begin
		// Global, so no need to make sure that a table is already open
		if (atServer in FApolloApplyTo) and Assigned(FApolloConnection) and
  	                                   (FApolloConnection.Connected) then
			FApolloConnection.ap_SetLockTimeout( Value );
		if (atLocal in FApolloApplyTo) then
    begin
    	LoadDLL;
		  sx_SetLockTimeOut( Value );
    end;

	  FLockTimeOut := Value;
	end;
end;

{==============================================================================
  SetMemoBlockSize
 ==============================================================================}
procedure TApolloEnv.SetMemoBlockSize( const Value : Word );  {LAS: 06-23-00, Changed param type from Integer}
begin
	if FApolloApplyTo = [] then
  	exit;

  if (Value <> FMemoBlockSize) then
  begin
		// Global, so no need to make sure that a table is already open
		if (atServer in FApolloApplyTo) and Assigned(FApolloConnection) and
  	                                   (FApolloConnection.Connected) then
			FApolloConnection.ap_SetMemoBlockSize( Value );
		if (atLocal in FApolloApplyTo) then
    begin
    	LoadDLL;
	  	sx_SetMemoBlockSize( Value );
    end;
	  FMemoBlockSize := Value;
	end;
end;

{==============================================================================
  GetMemoBlockSize
 ==============================================================================}
function TApolloEnv.GetMemoBlockSize: Word;  {LAS: 06-23-00, Changed type from Integer}
begin
	Result := FMemoBlockSize;
end;

{==============================================================================
  GetOptimisticBuffer
 ==============================================================================}
function TApolloEnv.GetOptimisticBuffer: WordBool;
begin
	if not (atLocal in FApolloApplyTo) then
  	Result := False
	else
		Result := FOptimisticBuffer;
end;

{==============================================================================
  SetOptimisticBuffer
 ==============================================================================}
procedure TApolloEnv.SetOptimisticBuffer( const Value : WordBool );
begin
	if FApolloApplyTo = [] then
  	exit;

  // Global, so no need to make sure that a table is already open
  // Local only.  Optimistic buffer is always OFF on remote server
  if (atLocal in FApolloApplyTo) then
  begin
    LoadDLL;
    sx_SysProp( SDE_SP_SETOBUFFER, Pointer( Value ));
    FOptimisticBuffer := (sx_SysProp( SDE_SP_GETOBUFFER, nil ) <> 0);
  end;
end;

{==============================================================================
  GetSoftSeek
 ==============================================================================}
function TApolloEnv.GetSoftSeek: WordBool;
begin
	Result := FSoftSeek;
end;

{==============================================================================
  SetSoftSeek
 ==============================================================================}
procedure TApolloEnv.SetSoftSeek( const Value: WordBool);
begin
	if FApolloApplyTo = [] then
  	exit;

	if FSoftSeek <> Value then
  begin
		FSoftSeek := Value;

    // Global, so no need to make sure that a table is already open
    if (atServer in FApolloApplyTo) and Assigned(FApolloConnection) and
                                       (FApolloConnection.Connected) then
      FApolloConnection.ap_SetSoftSeek( Value );
    if (atLocal in FApolloApplyTo) then
    begin
    	LoadDLL;
      sx_SetSoftSeek( Value );
    end;
  end;
end;


{==============================================================================
  GetDateFormat
 ==============================================================================}
function TApolloEnv.GetDateFormat: TApolloDateFrmt;
begin
	Result := FDateFormat;
end;

{==============================================================================
  GetLockTimeout
 ==============================================================================}
function TApolloEnv.GetLockTimeout: Integer;
begin
	Result := FLockTimeout;
end;

{==============================================================================
  GetCentury
 ==============================================================================}
function TApolloEnv.GetCentury: WordBool;
begin
	Result := FCentury;
end;

{==============================================================================
  SetCentury
 ==============================================================================}
procedure TApolloEnv.SetCentury( const Value: WordBool );
var
	tmpDate: String;
	iPos, iPos2, iLen: Integer;
begin
	if FApolloApplyTo = [] then
  	exit;
	if (atServer in FApolloApplyTo) and Assigned(FApolloConnection) and
                                     (FApolloConnection.Connected) then
		FApolloConnection.ap_SetCentury( Value );

	if (atLocal in FApolloApplyTo) then
  begin
  	LoadDLL;
	  sx_SetCentury(Value);
  end;
	FCentury := Value;

	tmpDate := ShortDateFormat;  // ShortDateFormat is a Delphi Global Variable
	iPos := Pos( 'yy', tmpDate );
	iPos2 := Pos( 'yyyy', tmpDate );
	iLen := Length( tmpDate );

	if FCentury then
  begin
		if (iPos2 = 0) and (iPos > 0) then
			ShortDateFormat := Copy( tmpDate, 1, iPos-1 ) + 'yyyy' +
                         Copy( tmpDate, iPos+2, iLen-iPos-1 );
	end
  else
  begin
		if iPos2 > 0 then
			ShortDateFormat := Copy( tmpDate, 1, iPos2+1 ) +
                         Copy( tmpDate, iPos2 + 4, iLen - 3 - iPos2 );
	end;
	// Refresh all open workareas, if any
	RefreshDataSets;
end;

{==============================================================================
  SetDateFormat
 ==============================================================================}
procedure TApolloEnv.SetDateFormat( const Value: TApolloDateFrmt );
var
	wValue : Word;
	DefaultLCID : LCID;
begin
	if FApolloApplyTo = [] then
	 	exit;

	if Value = dfWinDefault then
  	wValue := WIN_DEFAULT
	else
		wValue := Word( Value );

	if (atServer in FApolloApplyTo) and Assigned(FApolloConnection) and
                                     (FApolloConnection.Connected) then
		FApolloConnection.ap_SetDateFormat( wValue );

	if (atLocal in FApolloApplyTo) then
  begin
  	LoadDLL;
	  sx_SetDateFormat(wValue);
  end;

	FDateFormat := Value;

	// Set up the Delphi Globals
	case Value of
		dfAmerican :
			begin
				DateSeparator   := '/';
				ShortDateFormat := 'mm/dd/yy';
			end;

		dfAnsi :
			begin
				DateSeparator   := '.';
				ShortDateFormat := 'yy.mm.dd';
			end;

		dfBritish, dfFrench :
			begin
				DateSeparator   := '/';
				ShortDateFormat := 'dd/mm/yy';
			end;

		dfGerman :
			begin
				DateSeparator   := '.';
				ShortDateFormat := 'dd.mm.yy';
			end;

		dfItalian, dfSpanish :
    	begin
				DateSeparator   := '-';
				ShortDateFormat := 'dd-mm-yy';
      end;
	else
		begin
			// Goes back to Windows default
			DefaultLCID := GetSystemDefaultLCID;
			ShortDateFormat := GetLocaleStr(DefaultLCID, LOCALE_SSHORTDATE, 'm/d/yy');
			DateSeparator := GetLocaleChar(DefaultLCID, LOCALE_SDATE, '/');
		end;
	end;

	// If Century is set, call SetCentury to reset the date formats
	if FCentury then
		SetCentury( FCentury )
	else
    // Refresh all open workareas, if any
    RefreshDataSets;
end;

{==============================================================================
  GetDeleted
 ==============================================================================}
function TApolloEnv.GetDeleted: WordBool;
begin
	Result := FDeleted;
end;

{==============================================================================
  SetDeleted
  Global, so no need to make sure that a table is already open
 ==============================================================================}
procedure TApolloEnv.SetDeleted( const Value: WordBool );
begin
	if FApolloApplyTo = [] then
  	exit;

	FDeleted := Value;

	if (atServer in FApolloApplyTo) and Assigned(FApolloConnection) and
                                     (FApolloConnection.Connected) then
		FApolloConnection.ap_SetDeleted( Value );

	if (atLocal in FApolloApplyTo) then
  begin
  	LoadDLL;
		sx_SetDeleted( Value );
  end;

	// Refresh all open workareas, if any
	RefreshDataSets;
end;

{==============================================================================
  GetEpoch
 ==============================================================================}
function TApolloEnv.GetEpoch: Integer;
begin
	Result := FEpoch;
end;

{==============================================================================
  SetEpoch
 ==============================================================================}
procedure TApolloEnv.SetEpoch( const Value : Integer );
{$ifdef NOT_D3_OR_CB3}
var
	Present: TDateTime;
	Year, Month, Day: Word;
{$endif}
begin
	if FApolloApplyTo = [] then
  	exit;

	if Value = 0 then
		FEpoch := 1900
	else if Value < 100 then
     FEpoch := 100
	else if Value > 2999 then
		FEpoch := 2999
	else
		FEpoch := Value;

	// Global, so no need to make sure that a table is already open
	if (atServer in FApolloApplyTo) and Assigned(FApolloConnection) and
                                     (FApolloConnection.Connected) then
		FApolloConnection.ap_SetEpoch( FEpoch );

	if (atLocal in FApolloApplyTo) then
  begin
  	LoadDLL;
	 	sx_SetEpoch( FEpoch );
  end;

{$ifdef NOT_D3_OR_CB3}
	// Set system TwoDigitCenturyWindow variable
  Present := Now;
  DecodeDate( Present, Year, Month, Day );
	TwoDigitYearCenturyWindow := Abs(Year - FEpoch) mod 100;
{$endif}
end;

{==============================================================================
  SetErrorFunc
 ==============================================================================}
function TApolloEnv.SetErrorFunc( const pFunc, pData : Integer ) : Integer;
begin
	Result := -1;

	// Global, so no need to make sure that a table is already open
  // Only applies to local side.  Remote errors are handled by server.
	if (atLocal in FApolloApplyTo) then
  begin
  	LoadDLL;
	 	Result := sx_SetErrorFunc( Integer(pFunc), Integer(pData) );
		if pFunc = 0 then
  		FErrorFunc := -1
	  else
		  FErrorFunc := Result;
	end;
end;

{==============================================================================
  GetExact
 ==============================================================================}
function TApolloEnv.GetExact: WordBool;
begin
	Result := FExact;
end;

{==============================================================================
  SetExact
 ==============================================================================}
procedure TApolloEnv.SetExact( const Value: WordBool );
begin
	if FApolloApplyTo = [] then
  	exit;

	if FExact <> Value then
  begin
		FExact := Value;

    // Global, so no need to make sure that a table is already open
    if (atServer in FApolloApplyTo) and Assigned(FApolloConnection) and
       (FApolloConnection.Connected) then
      FApolloConnection.ap_SetExact( Value );

    if (atLocal in FApolloApplyTo) then
    begin
    	LoadDLL;
      sx_SetExact( Value );
    end;
  end;
end;


{==============================================================================
  SysProp
 ==============================================================================}
function TApolloEnv.SysProp( uiSysItem: Word; vpData: Pointer ): Integer;
begin
	Result := 0;

	if (atServer in FApolloApplyTo) and Assigned(FApolloConnection) and
                               (FApolloConnection.Connected) then
	  Result := FApolloConnection.ap_SysProp( uiSysItem, vpData );

	if (atLocal in FApolloApplyTo) then
  begin
  	LoadDLL;
	  Result := sx_SysProp( uiSysItem, vpData );
  end;
end;


{==============================================================================
  TApolloDatabase
 ==============================================================================}
{==============================================================================
  Create
  Make sure we're added to the list of databases
 ==============================================================================}
constructor TApolloDatabase.Create(Owner: TComponent);
begin
  inherited;
end;

{==============================================================================
  Destroy
 ==============================================================================}
destructor TApolloDatabase.Destroy;
begin
  inherited;
end;

{LAS: 04-20-00, Rewritten - And again 2002-12-18 RH/CA }
{==============================================================================
  CloseDataSets
  Closes all TApolloTables using EnumDatasets
 ==============================================================================}
procedure TApolloDatabase.CloseDataSets;
var
  n: integer;
  aList: TList;
begin
  // Removed previous Enum routine and replaced it with the internal EnumDatasets-method
  // RH/CA 2002-12-18
  aList:= TList.Create;
  try
    EnumDataSets(aList); // The list will return only TApolloTables linked to this Database
    for n:=0 to aList.Count-1 do
      TDataSet(aList[n]).Close; // No need to cast to TApolloTable
  Finally
    aList.Free;
  End;
end;

{LAS: 04-20-00, Rewritten - And again 2002-12-18 RH/CA }
{==============================================================================
  OpenDataSets
  Opens all TApolloTables using EnumDatasets
 ==============================================================================}
procedure TApolloDatabase.OpenDataSets;
var
  n: integer;
  aList: TList;
begin
  // Removed previous Enum routine and replaced it with the internal EnumDatasets-method
  // RH/CA 2002-12-18
  aList:= TList.Create;
  try
    EnumDataSets(aList); // The list will return only TApolloTables linked to this Database
    for n:=0 to aList.Count-1 do
    Try
      TDataSet(aList[n]).Open; // No need to cast to TApolloTable
    Except
      On E: EApolloError do
        MessageDlg(E.Message, mtError, [mbOK], 0);
    End;
  Finally
    aList.Free;
  End;
end;

{LAS: 04-20-00, Rewritten - And again 2002-12-18 RH/CA }
{==============================================================================
  RefreshDataSets
  Refreshes all TApolloTables using EnumDatasets
 ==============================================================================}
procedure TApolloDatabase.RefreshDataSets;
var
  n: integer;
  aList: TList;
begin
  // Removed previous Enum routine and replaced it with the internal EnumDatasets-method
  // RH/CA 2002-12-18
  aList:= TList.Create;
  try
    EnumDataSets(aList); // The list will return only TApolloTables linked to this Database
    for n:=0 to aList.Count-1 do
      if TDataSet(aList[n]).Active then  // No need to cast to TApolloTable
        TDataSet(aList[n]).Refresh;
  Finally
    aList.Free;
  End;
end;

{==============================================================================
	Notification
==============================================================================}

procedure TApolloDatabase.Notification( AComponent: TComponent;
                                        Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);

  // Need to manage the list of TApolloDatabase components for use by
  // Apollo table and query (ApoDSReg.pas)

  if (Operation = opRemove) and (AComponent.ClassType = TApolloDatabase) then
  begin
    //showmessage('removing');
    apRemoveDatabase( acomponent );
  end;

  if (Operation = opInsert) and (AComponent.ClassType = TApolloDatabase) then
  begin
    //showmessage('adding');
    apAddDatabase( acomponent );
  end;
end;

{ -------------------------------------------------------------------------
  EnumDataSets

  Will enumerate all evailable TApolloDatasets linked to the TApolloDatabase
  based on the DatapathAlias

  Most of the routine is copied from the VistaSoftware code which was
  available in CloseDatasets, OpenDatasets and RefreshDatasets methods.

  RH/CA 2002-12-18
  ------------------------------------------------------------------------- }

procedure TApolloDatabase.EnumDataSets(List: TList);
var
  Temp: TComponent;
  SearchObj : TApolloSearchObject;
begin
  if Assigned(List) then
  begin
    List.Clear;

    SearchObj := TApolloSearchObject.Create;
    Try
      Temp      := SearchObj.SearchFirstComponent(TApolloTable);
      while Temp <> nil do
      begin
        if (Temp is TApolloTable) and (TApolloTable(Temp).DatabaseName = FDataPathAlias) then
          List.Add(Temp);

        Temp := SearchObj.SearchNextComponent(TApolloTable);
      end;
    Finally
      SearchObj.Free;
    End;
  end;
end;

{ -------------------------------------------------------------------------
  EnumDataSets

  Will enumerate all evailable tablenames (fysicle) in the
  TApolloDatabase.DataPath

  Default to extention "dbf", but different extentions can be used

  Most of the routine is copied from the VistaSoftware code from then
  property editers

  RH/CA 2002-12-18
  ------------------------------------------------------------------------- }

procedure TApolloDatabase.EnumTableNames(List: TStrings;
  const Ext: string);
var
  iMore : Integer;
  sPath : String;
  SearchRec : TSearchRec;
begin
  if Assigned(List) then
  begin
    List.Clear;
    sPath := DataPath;

    // Look in current directy if none was specified
    if (sPath = '') then
      GetDir( 0, sPath );

    // Append backslash to end of path, if not there, for directory search
    if (sPath[ Length(sPath) ] <> '\') then
      sPath := sPath + '\';

    iMore := FindFirst( sPath + '*.' + Ext, faAnyFile, SearchRec );
    try
      while (iMore = 0) do
      begin
        List.Add( SearchRec.Name );
        iMore := FindNext( SearchRec );
      end;
    finally
      SysUtils.FindClose(SearchRec);
    end;
  end;
end;

initialization

  // Initilize local variable - RH d.d. 17-12-2002
  _apDatabaseList := nil;

finalization

  // Free/Destroy Stringlistobject if created - RH d.d. 17-12-2002
  if Assigned(_apDatabaseList) then
    TapdbStringList(_apDatabaseList).InternalDestroy;

end.
