(*============================================================================
  ApoQSet.PAS
  Main TApolloQuery code

  Copyright 1999-2002 Vista Software
  Last Updated: 10-22-2002

  NOTE: Tab stops were set at 2 spaces during development of this unit. If you
        are using a tab stop setting other than 2 spaces, the code indents
        may not line up accurately in your editor.
============================================================================*)

unit ApoQSet;

{$ifndef VER100} // CBuilder only
{$ObjExportAll On}
{$ASSERTIONS ON}
{$endif}

interface


uses
  DB, TypInfo, Classes, SysUtils, Forms, Controls,
  ApConn, ApoDSet, ApoEnv,
  ApCommon, IntSQL
  , Dialogs
{$ifdef EVAL}
  ,  Registry
{$endif}

{$ifdef VER120}  {LAS: 05-30-00, Added for D6 support}
  , ApoCompat
{$endif}

{$ifdef VER125}  {LAS: 05-30-00, Added for D6 support}
  , ApoCompat
{$endif}
{$ifdef VER140}  // D6,CB6
  , Variants
{$endif}
{$ifdef VER150}   // D7,CB7?
  , Variants
{$endif}

  ;
                                         
const
	// RDE types
  SDENTX = 1;
  SDEFOX = 2;
  SDENSX = 3;

  // Epoch Default
  EPOCH_DEFAULT = 1980;

  // Date types
  AMERICAN    = 0;
  ANSI        = 1;
  BRITISH     = 2;
  FRENCH      = 3;
  GERMAN      = 4;
  ITALIAN     = 5;
  SPANISH     = 6;
  WIN_DEFAULT = 99;



	ERR_NO_APOLLOCONNECTION_SPECIFIED =
  	'No ApolloConnection specified';
	ERR_NOT_CONNECTED =
  	'Not connected';
  SCircularReference =
  	'Circular reference not allowed';
  SSQLIsEmpty =
  	'No SQL statement available';

type
  TServerRecNo = array of Integer;

  TApolloQuery = Class;

  TApolloQueryBlobStream = class(TMemoryStream)
  private
    FAccessMethod : TAccessMethod;
    FIndex : Integer;
    FRecNo : Integer;
    iWA    : Integer;
    pConn  : TApolloConnection;
    procedure ReadBlobData;
  public
    constructor Create( amMode: TAccessMethod;
                        FApolloConnection: TApolloConnection; iRecNO: Integer;
                        Field: TBlobField; Mode: TBlobStreamMode);
  end;

  EApolloQueryError = class(Exception);

  TFieldInfo = class
    SDEOffset: Integer;
    NativeOffset: Integer;
    Width: Integer;
    Decimals: Integer;
  end;

  TRecInfo = record
    Bookmark: Integer;
    RecordNo: Integer;
    BookmarkFlag: TBookmarkFlag;
  end;
  PRecInfo=^TRecInfo;

  TApolloQuerySaveFlag = (mtfSaveData, mtfSaveCalculated, mtfSaveLookup,mtfSaveNonVisible);
  TApolloQuerySaveFlags = set of TApolloQuerySaveFlag;

  TApolloTableItem = class(TCollectionItem)
  private
    FItemCommitLevel : TApolloCommitLevel;
    FTableName: String;
    FAlias : String;
    FExtraIndexes: TStringList;
    FPassword: String;
		FTranslate: Boolean;
    FTableType : TApolloTableType;
    procedure SetExtraIndexes( Value : TStringList );
    procedure SetTableName(const Value: String);
  protected
    function GetDisplayName: String; override;
    function GetCaption : string;
  public
		constructor Create(Collection: TCollection); override;
		destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
  published
    property Alias : String read FAlias write FAlias;
		property ExtraIndexes: TStringList read FExtraIndexes write SetExtraIndexes;
		property OEMTranslate: Boolean read FTranslate write FTranslate;  {LAS: 05-09-00, Added}
    property TableName: String read FTableName write SetTableName;
    property TableType: TApolloTableType read FTableType write FTableType;
    property Password: String read FPassword write FPassword;
    property CommitLevel : TApolloCommitLevel Read FItemCommitLevel Write FItemCommitLevel default clNormal;
  end;

  TApolloTables = class(TCollection)
  private
    FStatusBar: TApolloQuery;
    function GetItem(Index: Integer): TApolloTableItem;
    procedure SetItem(Index: Integer; Value: TApolloTableItem);
  protected
  public
    constructor Create(StatusBar: TApolloQuery);
    function GetOwner: TPersistent; override;
    function Add: TApolloTableItem;
    property Items[Index: Integer]: TApolloTableItem read GetItem write SetItem; default;
  end;

  TApolloQuery = class(TDataSet)
  private
    //FRegSerial        : String;            // MED: 08-20-01, added

    FCommitLevel      : TApolloCommitLevel;     
    FErrorMsg         : String;     {MED: 11-20-00, added for trapping error messages on ApolloASP}

		FDummyNum         : Integer;  // To get RecCount info to show up in Object Inspector
    FCentury          : WordBool;
    FDataPath         : String;            {LAS: 05-09,00, Added}
    FDatabase         : TApolloDatabase;   {LAS: 05-05-00, Added}
    //FDateFormat       : TApolloDateFrmt;
    FDateFormat       : String;            {MED: 10-16-00, Format  m/d/yyyy(month, day, year) }
    FDeleted          : WordBool;
    FMemoBlockSize    : Integer;
    LOpeningTable     : Boolean;
    FRowCount         : Integer;
    FRecordSize       : Integer;
    //    FFieldCount       : Integer;
    iSQLID            : Integer;            // this is our SQL ID on the server
    iExecSQLID        : Integer;            // our ExecSQL has its own ID
    FApolloTables     : TApolloTables;
    FFieldOfs         : Array[0..1024] of Integer;  // MED: 02-15-01, changed from 255
    FDummyString      : String;
    FSQL              : TStrings;
    FFetchCount       : Integer;
    FDatabaseName     : String;
    FPassword         : String;
    FTranslate        : Boolean;
    FWaitCursor       : TCursor;          // MED: 04-10-01, added
    FOldCursor        : TCursor;          // gets set in SetCursor, UnSetCursor uses it

    FApolloConnection : TApolloConnection;
    FInternalConnection:TApolloConnection; {MED: 03-29-01, added to create a
                                            Conn on the fly in case one was not assigned
                                            !!! works only in amLocal mode  !!!  }
    FCursorOpen       : Boolean;
    FRecNo            : Integer;
    FFilterBuffer     : PChar;
    FRecords          : TList;
    FORecords         : TList;             // Our Original Records
    FServerRecNo      : TServerRecNo;  // dynamic array for server physical record #
    FBufferSize,
    FStartRecInfo,
    FStartCalculated  : Integer;
    FReadOnly         : Boolean;
    FParams           : TParams;
    FParamCheck       : Boolean;
    FSQLCounter       : Integer;
    FAccessMethod     : TAccessMethod;
    FTableType        : TApolloTableType;
    FAutoDeleteConn   : Boolean;            //MED: 03-16-2002, added

    FCalcFieldsOfs: Integer;  // CRVG : 05-14-2002

    procedure RaiseApolloQuery(msg:String);{MED:08-30-01, added}

    procedure SetCursor;                                //MED : 04-10-01, added
    procedure UnSetCursor;

    procedure OpenTable;
    procedure SetDatabaseName(Value: String);
		procedure SetTableType( Value: TApolloTableType );
    procedure SetPanels(Value: TApolloTables);
    function  GetPanels:TApolloTables;
    procedure SetApolloConnection(Val:TApolloConnection);
		procedure SetAccessMethod( Value: TAccessMethod );
    procedure SetSQL( Value: TStrings);
    function  BuildSqlTable: ByteBool;

    //    function  GetActiveRecordBuffer: PChar;   {LAS: 05-05-00, Removed}
    function  FilterRecord(Buffer: PChar): Boolean;
    procedure _InternalAdd(Buffer:Pointer);
    procedure _InternalDelete(Pos:Integer);
    procedure _InternalInsert(Pos:Integer; Buffer: Pointer);
    procedure _InternalEmpty;
    procedure _InternalFirst;
    procedure _InternalLast;
    function  _InternalNext: Boolean;
    function  _InternalPrior: Boolean;
    procedure SetFetchCount( Val: Integer );
    function  SetToApolloEnv:TApolloEnv;
  protected
    {Protected Procedure}
    function  FetchRecord( iRecNo: Integer ): Boolean; // MED: 11-02-01 moved here from private
    procedure Loaded; override;
    function  GetDatabase: TApolloDatabase;  {LAS: 05-05-00, Added}
	  function  GetVersion: String;
    procedure InternalOpen; override;
    procedure InternalClose; override;
    procedure InternalFirst;override;
    procedure InternalLast;override;
    procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); override;
    procedure InternalDelete; override;
    procedure InternalInitRecord(Buffer: PChar); override;
    procedure InternalPost; override;
    procedure InternalInitFieldDefs; override;
    procedure InternalSetToRecord(Buffer: PChar); override;
    procedure InternalRefresh; override;
    function  IsCursorOpen: Boolean; override;
    function  GetCanModify: Boolean; override;
		function  PathFromAlias: String;   {LAS: 05-09-00, Added}
    function  GetRecordSize: Word; override;
    function  GetRecordCount: Integer; override;
    function  AllocRecordBuffer: PChar; override;
    procedure FreeRecordBuffer(var Buffer: PChar); override;
    procedure SetFieldData(Field: TField; Buffer: Pointer); override;
    function  GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult; override;
    function  FindRecord(Restart, GoForward: Boolean): Boolean; override;
    function  GetRecNo: Integer; override;
    procedure SetRecNo(Value: Integer); override;
    function  GetRecCount: Integer;
    function  GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; override;
    procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override;
    procedure GetBookmarkData(Buffer: PChar; Data: Pointer); override;
    procedure SetBookmarkData(Buffer: PChar; Data: Pointer); override;
    procedure InternalGotoBookmark(Bookmark: Pointer); override;
    procedure InternalHandleException; override;
    procedure SetCommaText(AString: String);
    function  GetCommaText: String;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    property  Filter;
    property  Filtered;
    procedure RefreshInternalCalcFields(Buffer: PChar); override; // CRVG : 05-14-2002
    procedure ClearCalcFields(Buffer: PChar); override;           // CRVG : 05-14-2002
  public
    FFieldName     : array of String;  {MED: 11-06-00, added the following}
    FFieldWidth    : array of Integer;
    FFieldDecimals : array of Integer;
    FFieldOffset   : array of Integer;
    FFieldType     : array of String;
    FFieldCount    : integer;

    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    property AutoDeleteConn : Boolean read FAutoDeleteConn write FAutoDeleteConn;
    property  ErrorMsg: String read FErrorMsg write FErrorMsg;     {MED: 11-20-00, added for ApolloASP}

    property  Records : TList read FRecords;                {MED: 11-06-00, added}
    property  SQLID: Integer read iSQLID;                   {MED: 11-06-00, added}
    property  Database: TApolloDatabase read GetDatabase;   {LAS: 05-05-00, Added}
		property  DataPath: String read FDataPath;              {LAS: 05-09-00, Added}
    property  CommaText:string read GetCommaText write SetCommaText;
    property  RecSize : Integer read FRecordSize;           {MED: 11-06-00, added}
    property  ServerRecNo: TServerRecNo read FServerRecNo;

    procedure GotoBookmark(Bookmark: Pointer);  {MED:08-23-00, added}
    function  BookmarkValid(Bookmark: TBookmark): Boolean; override;
    function  CompareBookmarks( Bookmark1, Bookmark2: TBookmark ): Integer; override;
    function  GetFieldData(Field: TField; Buffer: Pointer): Boolean; override;
    function  CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream; override;
    function  Prepare: Integer;  // used only with  iExecSQLID & ExecSQL
    procedure UnPrepare;         // used only with  iExecSQLID & ExecSQL
    procedure ExecSQL;           // used only with  iExecSQLID
    procedure LoadFromFile(const FileName: string);
    procedure LoadFromStream(Stream: TStream);
    procedure SaveToFile(const FileName: string; flags:TApolloQuerySaveFlags);
    procedure SaveToStream(Stream: TStream; flags:TApolloQuerySaveFlags);
    procedure EmptyTable;
    procedure FetchAll;
    function ParamByName(const Value: string): TParam;
		{LAS: 04-19-00, Added}
    function Locate( const KeyFields: String; const KeyValues: Variant;
                              Options: TLocateOptions): Boolean; override;

    // MED: 09-08-01, added
    procedure SetSystemCollation;
    procedure SetMachineCollation;
    procedure AddDudenCollation;
    procedure AddEtecCollation;

  published
    {Published Declarations}
    property AccessMethod: TAccessMethod read FAccessMethod   write SetAccessMethod;
    property Active;
    property DatabaseName: String read FDatabaseName write SetDatabaseName;
    property TableNames: TApolloTables read GetPanels write SetPanels;
    property ApolloConnection: TApolloConnection read FApolloConnection write SetApolloConnection;

    // CRVG: 2001-11-02 Added
    property CommitLevel : TApolloCommitLevel Read FCommitLevel Write FCommitLevel default clNormal;

    // MED: 11-14-00 added the following two properties
    property DateFormat : String read FDateFormat write FDateFormat;
    property SetDeleted: WordBool read FDeleted write FDeleted default TRUE;

    // MED: 04-10-01, added
    property WaitCursor : TCursor read FWaitCursor write FWaitCursor;

    property FetchCount : Integer read FFetchCount write SetFetchCount default 100;  {LAS: 05-26-00, Added Default}
    property Params: TParams read FParams write FParams;
    property ParamCheck: Boolean read FParamCheck write FParamCheck default True;
    property Password: String read FPassword write FPassword;
    property RecNo: Integer read GetRecNo write SetRecNo;
    property RecCount: Integer read GetRecCount write FDummyNum;
    property TableType: TApolloTableType read FTableType write SetTableType default ttSXFOX;
		property OEMTranslate: Boolean read FTranslate write FTranslate;  {LAS: 05-09-00, Added}
    property SQL: TStrings read FSQL write SetSQL;
    property Version: String read GetVersion write FDummyString stored False;  {LAS: 04-11-00, Added stored}
    property ReadOnly: Boolean read FReadOnly write FReadOnly default False;


    property BeforeOpen;
    property AfterOpen;
    property BeforeClose;
    property AfterClose;
    property BeforeInsert;
    property AfterInsert;
    property BeforeEdit;
    property AfterEdit;
    property BeforePost;
    property AfterPost;
    property BeforeCancel;
    property AfterCancel;
    property BeforeDelete;
    property AfterDelete;
    property BeforeScroll;
    property AfterScroll;
    property OnCalcFields;
    property OnDeleteError;
    property OnEditError;
    property OnNewRecord;
    property OnPostError;
  end;


{$ifdef EVAL}
	{$I EvalDefs.PAS}    {LAS: 05-05-00, Added EVAL support}
{$endif}

implementation
                                                  
uses
  Windows,  IniFiles;

{$ifdef EVAL}
	{$I EvalTimeOut.PAS}    {LAS: 05-05-00, Added EVAL support}
{$endif}

{MED:08-30-01}
procedure TApolloQuery.RaiseApolloQuery(msg:String);{MED:08-30-01, added}
begin
  FErrorMsg := msg;
//  if not IsApolloASPRunning then
    raise EApolloQueryError.Create( msg );
end;

function SubStr(const cString:String; nStart,nEnd: Integer): String;
begin
	Result := Copy(cString, nStart, nEnd);
end;

function StripStr(var Str:String; Token:Char; var iPos:Integer):String;
var
  NewStr : String;
begin
  NewStr := '';
  Result := '';
  iPos := Pos(Token, Str);

  if iPos > 0 then
  begin
    Result := Substr(Str, 1, iPos-1);
    NewStr := Substr(Str, iPos+1, length(Str)-iPos);
  end;
  Str := NewStr;
end;

{==============================================================================
 Loaded
==============================================================================}
procedure TApolloQuery.Loaded;
begin
  inherited;

	if Assigned( FApolloConnection ) then
   	FApolloConnection.DataBaseName := FDatabaseName;
end;

//------------------------------------------------------------------------------
constructor TApolloQuery.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  //FRegSerial    := '';     // MED:08-20-01, added

  FCommitLevel  := clNormal;
  FErrorMsg     := '';

  iSQLID        := 0;              // No active work area
  FRecords      := TList.Create;
  FORecords     := TList.Create;
  FSQL          := TStringList.Create;
  FApolloTables := TApolloTables.Create(Self);

  {MED: 03-29-01, added to create a Conn on the fly in case one was not assigned
  !!! works only in amLocal mode  !!!  }
  // MED: 08-16-01, moved this to Prepare and InternalOpen.
  // No need to create on construct.

  //FInternalConnection := TApolloConnection.Create(Self);
  //FInternalConnection.Name := 'InternalConnection';

  FFetchCount   := 100;    {LAS: 05-26-00, Changed Default from 25 to 100}
  FCursorOpen   := False;  // Table cursor open flag
  FDataPath      := '';    // Physical path of table (DatabaseName location)  {LAS: 05-09-00, Added}

	FDateFormat     := 'm/d/yyyy';          {MED: 10-16-00, made it a property}
	// Initialize global (ENV) variables
	FMemoBlockSize  := 64;
  FCentury        := FALSE;
  FDeleted        := TRUE;               {MED: 11-16-00, changed to TRUE }

  {MED: 03-03-2000, added LOCAL support}
  FSQLCounter     := 0;
  FAccessMethod   := amLocal;    // MED: 04-02-2001- changed to amLocal

  FTableType      := ttSXFOX;
  FParams         := TParams.Create(self); {MED: 03-28-00, added}
  FOldCursor      := crDefault;
end;

//------------------------------------------------------------------------------
destructor TApolloQuery.Destroy;
begin
  FParams.Free;
  FApolloTables.Free;
  FreeAndNil(FInternalConnection);

  inherited Destroy;

  // Delete allocated records.
  _InternalEmpty;
  FRecords.free;
  FRecords := NIL;

  FORecords.Free;
  FORecords := NIL;

  FSQL.Free;

  {MED: 03-03-2000, added for local SQL only}
//  FreeSQLs;

end;

{==============================================================================
  SetAccessMethod
 ==============================================================================}
procedure TApolloQuery.SetAccessMethod( Value: TAccessMethod );
begin
	if (Value <> FAccessMethod) then
  begin
    if Active then
    	RaiseApolloQuery( ERR_MUST_CLOSE_TABLE_TO_CHANGE_DATA_ACCESS );
		FAccessMethod := Value;
  end;
end;

{==============================================================================
  SetTableType
 ==============================================================================}
procedure TApolloQuery.SetTableType( Value: TApolloTableType );
begin
	if (Value <> FTableType) then
  begin
		FTableType := Value;
  end;
end;


//------------------------------------------------------------------------------
procedure TApolloQuery.Notification( AComponent: TComponent;
                                     Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = FApolloConnection) then
    FApolloConnection := nil;
end;

//------------------------------------------------------------------------------
procedure TApolloQuery._InternalAdd(Buffer:Pointer);
begin
  FRecords.Add(Buffer);
end;

//------------------------------------------------------------------------------
procedure TApolloQuery._InternalInsert(Pos:Integer; Buffer:Pointer);
var
  i:Integer;
  b:PChar;
begin
  if Pos <= 0 then
    Pos := 1;
  if (Pos = FRecords.Count+1) or (Pos = -1) then
    FRecords.Add(Buffer)
  else
    FRecords.Insert(Pos,Buffer);

  for i:=Pos to FRecords.Count do
  begin
    b := FRecords.Items[i-1];
    FServerRecNo[i-1] := 0;
    Inc(PRecInfo(b+FStartRecInfo)^.RecordNo);
  end;
end;

//------------------------------------------------------------------------------
procedure TApolloQuery._InternalDelete(Pos:Integer);
var
	i: Integer;
	b: PChar;
begin
  if Pos > 1 then
  begin
    FServerRecNo[Pos-1] := 0;
    FreeMem(FRecords.Items[Pos-1]);
    FRecords.Delete(Pos-1);
    for i:=Pos to FRecords.Count do
    begin
      b := FRecords.Items[i-1];
      Dec(PRecInfo(b+FStartRecInfo)^.RecordNo);
    end;
  end;
end;

// Purge all records.
//------------------------------------------------------------------------------
procedure TApolloQuery._InternalEmpty;
var
  i : Integer;
begin
  for i:=0 to FRecords.Count-1 do
  begin
    FServerRecNo[i] := 0;
    FreeMem(FRecords[i]);
  end;
  FRecords.Clear;
end;


{==============================================================================
  OpenTable
  This opens the query table and makes sure it was opened correctly.  It also
  creates a valid SQL ID.  This is a private method used by InternalOpen and
  InternalCreateFieldDefs.
 ==============================================================================}
procedure TApolloQuery.OpenTable;
var
  iError : Integer;
  sError : String;
  // sDLLFullPath : String;
begin

{$ifdef EVAL}
  	if not NewCheckTimeLimit(REGKEY,
                             REGKEY_KILL_NAME,
                             REGKEY_CHECK_NAME,
                             'Apollo SQL 6.1', TRUE) then
    Exit;

{$endif}


  {LAS: 05-31-00, Removed.  TableNames are dynamically created as required now}
  //  if TableNames.Count = 0 then
  //    raise EApolloQueryError.Create( 'No TableNames have been defined' );
	if (AccessMethod = amServer) and (FDatabaseName = '') then         {LAS: 06-09-00, Added. 07-04-00, Added amServer check}
		RaiseApolloQuery( ERR_MISSING_DATABASENAME_PROPERTY );  {LAS: 06-09-00, Added}

  if FAccessMethod = amLocal then
  begin
    LoadSQLDLL;

    FDatabase := GetDatabase;
    FDataPath := PathFromAlias;
    if Assigned( FApolloConnection ) then
      FApolloConnection.SQLDataPath := FDataPath;
  end;

  SetToApolloEnv;

  SetCursor;

  iSQLID := FApolloConnection.ap_sql_Prepare(
                              FAccessMethod,
                              FDataBaseName,
                              TableNames,
                              FCentury,
                              FDateFormat,
                              FDeleted,
                              FMemoBlockSize,
                              FParams,
                              FReadOnly,
                              FTableType,
                              FPassword,
                              FTranslate,
                              FCommitLevel,
                              iError, sError );  // returns a SQL ID

  FErrorMsg := sError;

  // Unable to open file  -- this is the SQL ID NOT the work area!
  if (iSQLID < 1) then
  begin
    UnSetCursor;
    RaiseApolloQuery( sError);
  end;

  {Preparing ain't enough!  We need to tell the server to actually "Active"
  the query and execute the SQL statement to give us a few parameters
  we need:  FRowCount, FFieldCount, FRecordSize }
  FApolloConnection.ap_sql_OpenSQL(FAccessMethod, iSQLID,
                                SQL.TEXT, FRowCount, FFieldCount,
                                FRecordSize, iError, sError);  // iWA > 0 at this point

  UnSetCursor;

  FErrorMsg := sError;
  if iError > 0 then
  begin
  	if sError = '' then
    	FErrorMsg := 'Unable to execute SQL statement:' + #13#13 + SQL.Text;
    RaiseApolloQuery( FErrorMsg );
  end;

  try
    SetLength(FServerRecNo, FRowCount);
  except
    FErrorMsg := 'Out of memory while expanding dynamic array';
    RaiseApolloQuery(FErrorMsg);
  end;
end;

{MED:04-10-01, added}
procedure TApolloQuery.SetCursor;
begin
  if not IsApolloASPRunning then
  begin
    FOldCursor := Screen.Cursor;        {MED: 10-17-00, Added}
    Screen.Cursor := FWaitCursor;
  end;
end;

procedure TApolloQuery.UnSetCursor;
begin
  if not IsApolloASPRunning then
  begin
    Screen.Cursor := FOldCursor;
  end;
end;

{==============================================================================
  InternalOpen
 ==============================================================================}
procedure TApolloQuery.InternalOpen;
var  // CRVG : 05-14-2002
  i, iFldCount: Integer; // CRVG : 05-14-2002
  CurField: TField;  // CRVG : 05-14-2002
begin
  FRowCount   := 0;
  FFieldCount := 0;
  FRecordSize := 0;
	iSQLID      := 0;

  SetLength(FServerRecNo, 0);
  if Active then
  	Close;

	if (AccessMethod = amServer) and (FDatabaseName = '') then         {LAS: 06-09-00, Added. 07-04-00, Added amServer check}
		RaiseApolloQuery( ERR_MISSING_DATABASENAME_PROPERTY );  {LAS: 06-09-00, Added}

  if Length(FSQL.Text) = 0 then
     RaiseApolloQuery(SSQLIsEmpty);

  {MED: 03-29-01, added to create a Conn on the fly in case one was not assigned
  !!! works only in amLocal mode  !!!  }
  if (FApolloConnection = nil) then
  begin
    if (FAccessMethod = amLocal) then
    begin
          // MED: 08-15-01, moved here from Constructor
      if FInternalConnection = nil then
      begin
        FInternalConnection := TApolloConnection.Create(Self);
        FInternalConnection.Name := 'InternalConnection';
      end;

      FApolloConnection := FInternalConnection;
      FApolloConnection.DataBaseName := FDatabaseName;
    end
    else
      RaiseApolloQuery( ERR_NO_APOLLOCONNECTION_SPECIFIED )
  end;


  if (FAccessMethod = amServer) and (not FApolloConnection.Connected) then
  begin
    iSQLID:= 0;
    Active:= False;
    RaiseApolloQuery( ERR_NOT_CONNECTED );
  end;

  LOpeningTable := True;
  try

    InternalInitFieldDefs;

    if (iSQLID = 0) then   {MED: 08-12-00, Added FRowCount test}
    begin                  {MED:10-28-00, commented the FRowCount.  QuickReport crashes if no records are found}
      exit;
    end;

    if DefaultFields then
      CreateFields;

    BindFields(True);    // Bind the fields

    FCursorOpen     := True;

    FRecNo          := -1;

    if (FRowCount < 1) then   {MED: 08-12-00, Added FRowCount test}
    begin                     {MED:10-28-00, commented the FRowCount.  QuickReport crashes if no records are found}
      exit;                   {MED:03-17-01, moved it here for QuickReport Still crashed when combined with iSQLID=0 above}
    end;

    // CRVG : 05-14-2002
    FCalcFieldsOfs := 0;
    iFldCount := FieldCount;
    for i := 0 to iFldCount-1 do
    begin
      CurField := Fields[i];
      if CurField.FieldNo > 0 then
      begin

        if CurField.DataSize = 0 then
          // Inc( FCalcFieldsOfs, 1 )  // Leave enough room for NULL indicator
          Inc( FCalcFieldsOfs )  // Leave enough room for NULL indicator
        else
          Inc( FCalcFieldsOfs, CurField.DataSize );
          // Inc( FCalcFieldsOfs, CurField.DataSize+1 );
      end;
    end;

    BookmarkSize    := sizeof(longint);
    FStartRecInfo   := FRecordSize;
    FStartCalculated:= FStartRecInfo+SizeOf(TRecInfo);
    FBufferSize     := FRecordSize+Sizeof(TRecInfo)+CalcFieldsSize;

    BuildSqlTable;
  finally
    LOpeningTable := False;
  end;

end;

//------------------------------------------------------------------------------
procedure TApolloQuery.InternalClose;
begin
  if iSQLID > 0 then
  begin
    FApolloConnection.ap_sql_CloseSQL(FAccessMethod, iSQLID);
    FApolloConnection.ap_sql_UnPrepare(FAccessMethod, iSQLID);
  end;
  iSQLID := 0;
  _InternalEmpty;
  FCursorOpen := False;
  BindFields(False);

  // Destroy the fields if we created them
  if (DefaultFields) then
  	DestroyFields;
end;

{==============================================================================
  InternalInitFieldDefs
  This creates the TFieldDefs based on the table structure.
  We have to open the table to get the field defs.
 ==============================================================================}
procedure TApolloQuery.InternalInitFieldDefs;
var
  FieldOffSet, i,  j : Integer;
  FldName : String;
  FldType : TFieldType;
  FldSize : Integer;
  hdr: TStrings;
  msPtr: Pointer;
begin
	if (iSQLID = 0) then
		OpenTable;

  if (iSQLID = 0) then
  	exit;

	FieldDefs.Clear;

  if FFieldCount < 1 then
  	RaiseApolloQuery( 'FFieldCount < 1');

  {MED: 11-05-00, added}
  SetLength(FFieldName, FFieldCount+1);  // 1- based array
  SetLength(FFieldType, FFieldCount+1);  // 1- based array
  SetLength(FFieldWidth, FFieldCount+1);  // 1- based array
  SetLength(FFieldDecimals, FFieldCount+1);  // 1- based array
  SetLength(FFieldOffSet, FFieldCount+1);  // 1- based array

  hdr := TStringList.Create;
  try
    hdr.Clear;
    msPtr := FApolloConnection.ap_sql_GetStructure(AccessMethod, iSQLID);
  
    hdr.LoadFromStream(TMemoryStream(msPtr) );

    j := 0;
    for i := 1 to FFieldCount do
    begin
  	  FldName     := Trim( hdr.strings[j] );

      FldType     := TFieldType(StrToInt('$'+Trim( hdr.Strings[j+1] )));
      FldSize     := StrToInt( '$' + hdr.strings[j+2] );
      FieldOffSet := StrToInt( '$' + hdr.strings[j+3] );
      Inc( j, 5 );

      FieldDefs.Add(fldName, fldType, fldSize, FALSE);
      FFieldOfs[i-1] := FieldOffset;

      {MED: 11-06-00, added}
      FFieldName[i]    := UpperCase(fldName);

      Case FldType of
        ftString :      FFieldType[i]    := 'C';
        ftFloat  :      FFieldType[i]    := 'N';
        ftInteger:      FFieldType[i]    := 'N';
        ftBoolean:      FFieldType[i]    := 'L';
        ftDateTime:     FFieldType[i]    := 'D';
        ftDate:         FFieldType[i]    := 'D';  // MED: 08-28-01, added
        ftMemo:         FFieldType[i]    := 'M';
        ftSmallInt:     FFieldType[i]    := 'N';
      End;

      FFieldWidth[i]   := FldSize;              // only valid for strings
      FFieldDecimals[i]:= 0;
      FFieldOffSet[i]  := FieldOffset;
    end;
  finally
    hdr.Free;
  end;

  // If we're just doing an InfoQuery and the table's not already open,
  // then we need to close the table on the way out.
  if (not FCursorOpen) and (not LOpeningTable)  then
    iSQLID := 0;
end;

function TApolloQuery.BuildSqlTable: ByteBool;
var
  i:Integer;
  b: Pointer;
begin
  // Allocate room for buffer in list.
  b := nil;
  try
    GetMem(b,FBufferSize);

    LOpeningTable := TRUE;
    FRecords.Clear;
    FORecords.Clear;
    if FieldCount > 0 then
    begin
      for i:=1 to FRowCount do
      begin
        InternalAddRecord(b, TRUE);
        FORecords.Add(FRecords[i-1]);
      end;
    end;
  finally
    FreeMem(b);
  end;
  LOpeningTable := FALSE;
  Result := True;
end;

function TApolloQuery.FetchRecord(iRecNo:Integer): Boolean;
var
  pP, Buff : Pointer;
  i, nRead, iRecToRead: Integer;
  ServerRecordNum : Integer;
begin
  Result := False;
  if (iRecNo < 1 ) or (iRecNo > FRecords.Count) then
  	exit;
  if FRecords.Count < 1 then
  	exit;

  if FServerRecNo[iRecNo-1] = 0 then
  begin
    pP := FApolloConnection.ap_sql_GetRecord(FAccessMethod, iSQLID, iRecNo, FFetchCount, FRecords.Count);

    if TMemoryStream(pP).Size = 0 then
      RaiseApolloQuery('NIL stream');

    TMemoryStream(pP).Position := 0;
    TMemoryStream(pP).Seek(0, sofrombeginning);
    Buff := nil;
    try
      GetMem(Buff, FRecordSize);
      nRead :=  TMemoryStream(pP).Read( PChar(buff)^, 8);  //# of records we received in the stream
      if nRead > 0 then
      begin
        iRecToRead := StrToInt('$'+SubStr(PChar(buff),1 ,8));
        if iRecToRead > 0 then
        begin
          for i:=1 to iRecToRead do
          begin
            // get server record number
            nRead := TMemoryStream(pP).Read( PChar(buff)^, 8);
            if (nRead < 8 ) then
              RaiseApolloQuery('Error while reading stream');

            ServerRecordNum := StrToInt('$'+SubStr(PChar(buff),1 ,8));
///////////////////
            nRead := TMemoryStream(pP).Read( PChar(buff)^, FRecordSize);

            if (nRead < FRecordSize ) then
              RaiseApolloQuery('Error while reading stream');
            Move(pchar(Buff)^, FRecords.Items[iRecNo-1]^, FRecordSize);
            FServerRecNo[iRecNo-1] := ServerRecordNum;
            Inc(iRecNo);
          end;
        end;
      end;
    finally
      FreeMem(Buff);
    end;
    Result := True;
  end;
end;

// Result is data in the buffer and a boolean return (true=not null, false=is null).
//------------------------------------------------------------------------------
function TApolloQuery.GetFieldData(Field: TField; Buffer: Pointer): Boolean;
var
  DataBuffer: PChar;
  FieldData: PChar;
  OffSet : Integer;
  vData      : TDateTimeRec;    {MED: 09-04-00, added to make 0/0/00 appear blank}
begin
  Result := FALSE;  {MED: 08-16-00, added }
  // CRVG : 05-14-2002 Removed to support calcfields -> if (Field.FieldNo < 1) then    Exit;  {MED: 08-16-00, added }
  if ActiveBuffer = nil then    Exit;  {MED: 08-16-00, added }

  if (not FCursorOpen) then
    RaiseApolloQuery( ERR_MUST_OPEN_TABLE );

  if Field.FieldNo > 0 then
    OffSet := FFieldOfs[Field.FieldNo-1]
  else
    OffSet := FStartCalculated + Field.OffSet;

  if State = dsCalcFields then // CRVG : 05-14-2002
  	DataBuffer := CalcBuffer
  else
  	DataBuffer := ActiveBuffer;

  FieldData := DataBuffer+Offset;

	if Assigned(Buffer) then
  begin
    case Field.DataType of
      ftDate:
      begin
        Move(FieldData^, vData, sizeof(TDateTimeRec)); {MED:09-04-00, added to support blank dates}
        if vData.Date = 0 then
        begin
          pchar(Buffer^) := #0;
          Result := FALSE;
        end
        else
        begin
          Move(vData, Buffer^, SizeOf(TDateTimeRec));
          Result := True;
        end;
      end
      else
      begin
        Move(FieldData^, Buffer^, Field.DataSize);
        Result := True;
      end;
    end;
  end
  else
  begin  {MED:10-27-00, added}  // If FieldData is not blank then return TRUE
    If (FieldData <>'') then
      Result:=True;
  end;
end;

(* LAS: 05-05-00, Removed
//------------------------------------------------------------------------------
function TApolloQuery.GetActiveRecordBuffer: PChar;
begin
	case State of
		dsBrowse:
    	if IsEmpty then
      	Result := nil
			else
				Result := ActiveBuffer;
		dsCalcFields:
    	Result := CalcBuffer;
		dsFilter:
    	Result:=FFilterBuffer;
		dsEdit,dsInsert:
    	Result:=ActiveBuffer;
	else
		Result := nil;
	end;
end;
*)

//------------------------------------------------------------------------------
procedure TApolloQuery.SetFieldData(Field: TField; Buffer: Pointer);
var
  OffSet:  Integer;
  DataBuffer: PChar;
  FieldData: PChar;
begin

  if State = dsInactive then
  	Exit;

  // ShowMessage(Field.Name);

  if Field.FieldNo > 0 then
  begin
   	Field.Validate(Buffer);
    OffSet := FFieldOfs[Field.FieldNo-1];
	end
  else  // Calculated fields have Field.FieldNo values of -1
   OffSet := FStartCalculated + Field.OffSet;

  if State = dsCalcFields then
    DataBuffer := CalcBuffer
  else
    DataBuffer := ActiveBuffer;

  FieldData := DataBuffer+Offset;

  if (Assigned(Buffer)) then
  begin
    // Not Null
    // FieldData[0] := FLD_NOTNULL;
    // Inc( FieldData );
    Move(Buffer^, FieldData^, Field.DataSize);
  end
  else
  begin
    // Null data
    // FieldData[0] := FLD_NULL;
    // Inc( FieldData );
    FillChar(FieldData^, Field.DataSize, 0);
  end;
end;

//------------------------------------------------------------------------------
function TApolloQuery.IsCursorOpen: Boolean;
begin
	Result := FCursorOpen;
end;

//------------------------------------------------------------------------------
function TApolloQuery.GetCanModify: Boolean;
begin
  Result := not FReadOnly;
end;

//------------------------------------------------------------------------------
function TApolloQuery.GetRecordSize: Word;
begin
  Result := FRecordSize;
end;

//------------------------------------------------------------------------------
function TApolloQuery.AllocRecordBuffer: PChar;
begin
	GetMem(Result,FBufferSize);
	FillChar(Result^,FBufferSize,0);
end;

//------------------------------------------------------------------------------
procedure TApolloQuery.FreeRecordBuffer(var Buffer: PChar);
begin
	FreeMem(Buffer);
end;

//------------------------------------------------------------------------------
procedure TApolloQuery.InternalFirst;
begin
  _InternalFirst;
end;

//------------------------------------------------------------------------------
procedure TApolloQuery.InternalLast;
begin
  _InternalLast;
end;

//------------------------------------------------------------------------------
procedure TApolloQuery._InternalFirst;
begin
  FRecNo := -1;
end;

//------------------------------------------------------------------------------
procedure TApolloQuery._InternalLast;
begin
  FRecNo := FRecords.Count + 1;
end;

//------------------------------------------------------------------------------
procedure TApolloQuery.FetchAll;
var
	i: Integer;
begin
  for i:=1 to FRecords.Count do
  begin
    FetchRecord(i);
  end;
  InternalFirst;
end;

//------------------------------------------------------------------------------
procedure TApolloQuery.InternalRefresh;
begin
	//
end;

//------------------------------------------------------------------------------
function TApolloQuery._InternalNext:boolean;
begin
  if FRecNo = -1 then
  	FRecNo := 0;

  if FRecNo < FRecords.Count then
  begin
    Inc(FRecNo);
    FetchRecord(FRecNo);
    Result := True;
  end
  else
  	Result := False;
end;

//------------------------------------------------------------------------------
function TApolloQuery._InternalPrior: Boolean;
begin
  if FRecNo > 1 then
  begin
    Dec(FRecNo);
    FetchRecord(FRecNo);
    Result := True;
  end
  else
  	Result := False;
end;

//------------------------------------------------------------------------------
procedure TApolloQuery.ClearCalcFields(Buffer: PChar);
begin
  FillChar( Buffer[FCalcFieldsOfs], CalcFieldsSize, 0 );
end;

function TApolloQuery.GetRecord( Buffer: PChar; GetMode: TGetMode;
                                 DoCheck: Boolean): TGetResult;
var
  Acceptable: Boolean;
begin
  Result := grOK;
  if LOpeningTable then
    exit;

  Acceptable := False;
  repeat
    case GetMode of
      gmCurrent:
      begin
        if (FRecNo > FRecords.Count) then      {MED: 09-01-00, fixed bug by changing from >= }
          Result := grEOF
        else if (FRecNo < 1) then
        begin
        	Result := grBOF
        end
        else
        	Result := grOk;
      end;

      gmNext:
      begin
        if _InternalNext then
          Result := grOK
        else
          Result := grEOF;
      end;

      gmPrior:
      begin
        if _InternalPrior then
        begin
          Result := grOK;
        end
        else
        begin
          Result := grBOF;
        end;
      end;
    end;

    if (Result = grOk) and (FRecNo > -1) then
    begin
      ClearCalcFields(Buffer);
      //fill TARrecord part of buffer
      Move(FRecords.Items[FRecNo-1]^,Buffer^,FBufferSize);

      //fill information part of buffer
      with PRecInfo(Buffer+FStartRecInfo)^ do
      begin
        BookMark := FRecNo;
        RecordNo := FRecNo;

        if Result = grEof then
        begin
//  				FLastCrack := True;
          BookmarkFlag := bfEof;
        end
        else
	        BookmarkFlag := bfCurrent;
      end;

      //fill calc fields part of buffer
      // CRVG : 05-14-2002
      GetCalcFields(Buffer);
      Acceptable := FilterRecord(Buffer);
      if (GetMode = gmCurrent) and not Acceptable then
      	Result := grError;
    end;

  until (Result<>grOk) or Acceptable;
end;

//------------------------------------------------------------------------------
function TApolloQuery.FindRecord(Restart, GoForward: Boolean): Boolean;
var
  Status: Boolean;
begin
  CheckBrowseMode;
	if not ControlsDisabled then  {LAS: 06-01-00, Added}
	  DoBeforeScroll;
  SetFound(False);
  UpdateCursorPos;
  CursorPosChanged;

  if GoForward then
  begin
    if Restart then
    	_InternalFirst;
    Status := _InternalNext;
  end
  else
  begin
    if Restart then
    	_InternalLast;
    Status := _InternalPrior;
  end;

  if Status then
  begin
    Resync([rmExact, rmCenter]);
    SetFound(True);
  end;
  Result := Found;
  if Result and (not ControlsDisabled) then  {LAS: 06-01-00, Added ControlsDisabled check}
  	DoAfterScroll;
end;

//------------------------------------------------------------------------------
function TApolloQuery.FilterRecord(Buffer: PChar): Boolean;
var
  SaveState: TDatasetState;
begin
  Result:=True;
  if not Filtered or not Assigned(OnFilterRecord) then
  	Exit;
  SaveState:=SetTempState(dsFilter);
  FFilterBuffer:=Buffer;
  OnFilterRecord(self,Result);
  RestoreState(SaveState);
end;

//------------------------------------------------------------------------------
procedure TApolloQuery.InternalSetToRecord(Buffer: PChar);
begin
  FRecNo := PRecInfo(Buffer+FStartRecInfo).RecordNo;
end;

//------------------------------------------------------------------------------
function TApolloQuery.GetRecordCount: Integer;
var
  SaveState: TDataSetState;
  SavePosition: Integer;
  TempBuffer: PChar;
begin
  if not Filtered then
  	Result := FRecords.Count
  else
  begin
    Result := 0;
    SaveState := SetTempState(dsBrowse);
    SavePosition := FRecNo;
    try
      TempBuffer:=AllocRecordBuffer;
      InternalFirst;
	    while GetRecord(TempBuffer,gmNext,True) = grOk do
      	Inc(Result);
    finally
    	RestoreState(SaveState);
     	FRecNo := SavePosition;
     	FreeRecordBuffer(TempBuffer);
    end;
  end;
end;

//------------------------------------------------------------------------------
function TApolloQuery.GetRecCount: Integer;
begin
	Result := RecordCount;
end;

//------------------------------------------------------------------------------
function TApolloQuery.GetRecNo: Integer;
var
  SaveState: TDataSetState;
  SavePosition: Integer;
  TempBuffer: PChar;
begin
	if not Active then
  begin
  	Result := 0;
    exit;
  end;

 	UpdateCursorPos;
  if not Filtered then
  	Result :=	FRecNo
  else
  begin
    Result := 0;
    SaveState := SetTempState(dsBrowse);
    SavePosition := FRecNo;
    try
      TempBuffer := AllocRecordBuffer;
      InternalFirst;
      repeat
				if GetRecord(TempBuffer,gmNext,True) = grOk then
        	Inc(Result);
      until PRecInfo(TempBuffer+FStartRecInfo).RecordNo = SavePosition
    finally
      RestoreState(SaveState);
      FRecNo := SavePosition;
      FreeRecordBuffer(TempBuffer);
    end;
  end;
end;

//------------------------------------------------------------------------------
procedure TApolloQuery.SetRecNo(Value: Integer);
var
  SaveState: TDataSetState;
  SavePosition: Integer;
  TempBuffer: PChar;
begin
	if not Active  then
  	Exit;

  if not Filtered then
  	FRecNo := Value
  else
  begin
    SaveState := SetTempState(dsBrowse);
    SavePosition := FRecNo;
    try
      TempBuffer := AllocRecordBuffer;
      InternalFirst;
      repeat
        begin
          if GetRecord(TempBuffer, gmNext, True) = grOk then
          	Dec(Value)
          else
          begin
            FRecNo := SavePosition;
            break;
          end;
        end;
      until Value = 0;
    finally
      RestoreState(SaveState);
      FreeRecordBuffer(TempBuffer);
    end;
  end;
  with PRecInfo(ActiveBuffer + FStartRecInfo)^ do
	begin
		BookmarkFlag := bfCurrent;
		Bookmark     := FRecNo;
	end;
  Resync([]);
end;

//------------------------------------------------------------------------------
procedure TApolloQuery.InternalAddRecord(Buffer: Pointer; Append: Boolean);
var
   b: Pointer;
begin
  // Allocate room for buffer in list.
  GetMem(b,FBufferSize);
  Move(Buffer^, b^, FBufferSize);
  if Append then
    _InternalAdd(b)
  else
    _InternalInsert(FRecNo,b);
end;

//------------------------------------------------------------------------------
procedure TApolloQuery.InternalDelete;
begin
	_InternalDelete(FRecNo);
end;

//------------------------------------------------------------------------------
procedure TApolloQuery.InternalInitRecord(Buffer: PChar);
begin
  FillChar( Buffer^, FBufferSize, 0 );
  PRecInfo( Buffer+FStartRecInfo)^.RecordNo := FRecNo;
end;

{==============================================================================
  RefreshInternalCalcFields
 ==============================================================================}
// CRVG : 05-14-2002
procedure TApolloQuery.RefreshInternalCalcFields(Buffer: PChar);
begin
  inherited;
//  CalculateFields( Buffer );
end;


//------------------------------------------------------------------------------
procedure TApolloQuery.InternalPost;
var
  b: Pointer;
  n: Integer;
begin
  n := PRecInfo(ActiveBuffer+FStartRecInfo)^.RecordNo;
  if State = dsEdit then
    Move(ActiveBuffer^, FRecords.Items[n-1]^, FBufferSize)
  else
  begin
    GetMem( b, FBufferSize );
    Move(ActiveBuffer^, b^, FBufferSize);
    if GetBookmarkFlag(b) = bfEOF then
			_InternalAdd(b)
    else
			_InternalInsert(n,b);
  end;
end;

{==============================================================================
  CompareBookmarks
  This is used by the grid's multi-select to know if a record should be
  highlighted or not.
 ==============================================================================}
function TApolloQuery.CompareBookmarks( Bookmark1, Bookmark2: TBookmark): Integer;
begin
 if (Bookmark1 = nil) or (Bookmark2 = nil) then
   Result := 2
 else if (Integer(Bookmark1^) = Integer(Bookmark2^)) then
   Result := 0
 else if (Integer(Bookmark1^) < Integer(Bookmark2^)) then
   Result := -1
 else
   Result := 1;
end;

//------------------------------------------------------------------------------
function TApolloQuery.BookmarkValid(Bookmark: TBookmark): Boolean;
begin
  Result := not(Bookmark = nil);
end;

//------------------------------------------------------------------------------
procedure TApolloQuery.SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag);
begin
	PRecInfo(Buffer + FStartRecInfo).BookmarkFlag := Value;
end;

//------------------------------------------------------------------------------
function TApolloQuery.GetBookmarkFlag(Buffer: PChar): TBookmarkFlag;
begin
	Result := PRecInfo(Buffer+FStartRecInfo).BookmarkFlag;
end;

//------------------------------------------------------------------------------
procedure TApolloQuery.GetBookmarkData(Buffer: PChar; Data: Pointer);
begin
	PInteger(Data)^ := PRecInfo(Buffer + FStartRecInfo).Bookmark;
end;

//------------------------------------------------------------------------------
procedure TApolloQuery.SetBookmarkData(Buffer: PChar; Data: Pointer);
begin
	PRecInfo(Buffer + FStartRecInfo).Bookmark := PInteger(Data)^;
end;

//------------------------------------------------------------------------------
// MED: 08-22-00, added
procedure TApolloQuery.GotoBookmark(Bookmark: Pointer);
begin
  InternalGotoBookmark(Bookmark);
end;

procedure TApolloQuery.InternalGotoBookmark(Bookmark: Pointer);
var
  ReqBookmark: Integer;
begin
  ReqBookmark := Integer(Bookmark^);   {LAS: 08-13-00, Added pointer reference to BookMark param}
  if (ReqBookmark >= 0) and (ReqBookmark <= RecordCount) then {MED:08-25-00, changed to <= from <}
  begin
    FRecNo := ReqBookmark;
    FetchRecord( FRecNo );
    Resync([]);
  end;
end;

//------------------------------------------------------------------------------
procedure TApolloQuery.InternalHandleException;
begin
	Application.HandleException(Self);
end;

//------------------------------------------------------------------------------
procedure TApolloQuery.SaveToFile(const FileName: string; flags:TApolloQuerySaveFlags);
var
  Stream: TStream;
begin
  Stream := TFileStream.Create(FileName, fmCreate);
  try
    SaveToStream(Stream,flags);
  finally
    Stream.Free;
  end;
end;

//------------------------------------------------------------------------------
procedure TApolloQuery.SaveToStream(Stream: TStream; flags:TApolloQuerySaveFlags);
var
  i: Integer;
  bm: TBookmark;
  nf: Integer;
  s,a: String;
  l: Integer;
  fset,f: ^Boolean;
  Ods,Oms: Char;
begin
  // Setup standard layout for data.
  Ods := DateSeparator;
  Oms := DecimalSeparator;
  DateSeparator := '/';
  DecimalSeparator := '.';

  bm := GetBookmark;
  fset := nil;
  try
    DisableControls;

    // Setup flags for fields to save.
    nf := Fieldcount;
    GetMem(fset,nf * SizeOf(Boolean));
    f := fset;
    for i:=0 to nf-1 do
    begin
      f^:=false;
      case Fields[i].FieldKind of
        fkData:
        	if (mtfSaveData in flags) then
          	f^ := True;
        fkCalculated:
        	if (mtfSaveCalculated in flags) then
          	f^ := True;
        fkLookup:
        	if (mtfSaveLookup in flags) then
          	f^ := True;
        else
        	f^ := True;
      end;
      if not (Fields[i].Visible or (mtfSaveNonVisible in flags)) then
      	f^ := False;
      Inc(f);
    end;

    // Write all field display names in CSV format.
    s := '';
    a := '';
    f := fset;
    for i := 0 to (nf-1) do
		begin
			if f^ then
			begin
				s := s + a + AnsiQuotedStr(PChar(Fields[i].DisplayName),'"');
				a := ',';
			end;
			Inc(f);
		end;
		s := s + #13 + #10;
		l := length(s);
		Stream.Write(Pointer(s)^, l);

		// Write all records in CSV format.
		First;
		while not EOF do
		begin
			// Write current record.
			s := '';
			a := '';
			f := fset;
      for i := 0 to nf-1 do
			begin
				if f^ then
				begin
					if (Fields[i].IsNull) then
          	s := s + a
					else
          	s := s + a + AnsiQuotedStr(PChar(Fields[i].AsString),'"');
          a := ',';
				end;
				Inc(f);
      end;
			s := s + #13 + #10;
			l := length(s);
			Stream.WriteBuffer(Pointer(s)^, l);

      // Next record.
      Next;
		end;
  finally
		GotoBookmark(bm);
		EnableControls;
		FreeBookmark(bm);
		if fset <> nil then
    	FreeMem(fset);
		DateSeparator := Ods;
    DecimalSeparator := Oms;
  end;
end;

//------------------------------------------------------------------------------
procedure TApolloQuery.LoadFromFile(const FileName: string);
var
  Stream: TStream;
begin
  Stream := TFileStream.Create(FileName, fmOpenRead);
  try
    LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;

//------------------------------------------------------------------------------
procedure TApolloQuery.LoadFromStream(Stream: TStream);
const
	BUFSIZE = 8192;
var
  i: Integer;
  bm: TBookmark;
  nf: Integer;
  s: String;
  buf,ptr: PChar;
  buflen: Integer;
  Line: String;
  lptr,elptr: PChar;
  null: Boolean;
  Ods,Oms: Char;

  //-----------------------------------------------
  function GetLine: Boolean;
  var
    Start: PChar;
    sz: Integer;
  begin
    // If less than 1024 bytes left in buffer, fill up the buffer.
    // Notice: This means that if a line is longer than 1024 bytes it could fail.
    sz := BUFSIZE-(ptr-buf);
    if (sz < 1024) then
    begin
      // Move the rest of data to the start of the buffer.
      if (sz>0) then
      	Move(ptr,buf,sz);
      ptr := buf + sz;

      // Fill up the buffer.
      buflen := BUFSIZE;
      if Stream.Size < buflen then
      	buflen := Stream.Size;
      buflen := Stream.Read(Pointer(buf)^,buflen-sz)+sz;
      ptr := buf;
    end;

    // Check if finished.
    if ((ptr-buf) = buflen) then
    begin
      Result := False;
      exit;
    end;

    // Cut out a line.
    Start := ptr;
    while not (ptr^ in [#0, #10, #13]) do
    	Inc(ptr);
    SetString(Line, Start, ptr - Start);
    lptr := PChar(Line);
    elptr := PChar(Line)+Length(Line)-1;
    if ptr^ = #13 then
    	Inc(ptr);
    if ptr^ = #10 then
    	Inc(ptr);
    Result := True;
  end;

  //-----------------------------------------------
  function GetWord(var null:boolean): String;
  label
	  L_exit;
  begin

    // Cut out next word.
    Result := '';

    // Look for starting " or ,.
    while (lptr^ <> '"') and (lptr^ <> ',') and (lptr<elptr) do
      Inc(lptr);
    if (lptr>=elptr) then
      exit;
    if (lptr^ = ',') then
    begin
      null := True;
      Inc(lptr);
      exit;
    end
    else
      null := False;
    Inc(lptr);

    while True do
    begin
      // Look for ending ".
      while not (lptr^ = '"') do
      begin
        if (lptr>=elptr) then
          goto L_exit;
        Result := Result+lptr^;
        Inc(lptr);
      end;
      Inc(lptr);

      // Is it a double "" or end of word ?.
      if (lptr^ = '"') then
      begin
        Result:=Result+'"';
        Inc(lptr);
        continue;
      end;

L_exit:
      // Found end, remove comma's if any.
      while (lptr<elptr) and (lptr^ = ',') do
        Inc(lptr);

      break;
     end;
  end;

begin
  // Setup standard layout for data.
  Ods := DateSeparator;
  Oms := DecimalSeparator;
  DateSeparator := '/';
  DecimalSeparator := '.';

  bm := GetBookmark;

  try
    // Allocate space for a buffer.
    GetMem(buf,BUFSIZE);

    // Place pointer at end of buffer to notify getword to read a chunk of streamdata.
    ptr := buf + BUFSIZE;

    // Read data from stream.
    nf := Fieldcount;

    // Read headerline and skip it.
    GetLine;

    DisableControls;

    // Read all lines in CSV format.
    while GetLine do
    begin
			Append;
			i := 0;
			while (lptr<elptr) and (i<nf) do
			begin
        s := GetWord(null);
        if null then
        	Fields[i].Clear
        else
        	Fields[i].AsString := s;
        Inc(i);
      end;
			Post;
    end;
  finally
    FreeMem(buf);
    GotoBookmark(bm);
    EnableControls;
    FreeBookmark(bm);
    DateSeparator:=Ods;
    DecimalSeparator:=Oms;
  end;
end;

//------------------------------------------------------------------------------
procedure TApolloQuery.EmptyTable;
begin
  _InternalEmpty;
end;

//------------------------------------------------------------------------------
procedure TApolloQuery.SetCommaText(AString: String);
var
  stream: TMemoryStream;
begin
  EmptyTable;
  stream := TMemoryStream.Create;
  try
    stream.Write(Pointer(AString)^,length(AString));
    stream.Seek(0,soFromBeginning);
    LoadFromStream(stream);
  finally
    stream.free;
  end;
end;

//------------------------------------------------------------------------------
function TApolloQuery.GetCommaText: String;
var
  stream: TMemoryStream;
  sz: Integer;
  p: PChar;
begin
  Result := '';
  stream := TMemoryStream.Create;
  try
    SaveToStream(stream,[mtfSaveData]);
    stream.Seek(0,soFromBeginning);
    sz := stream.Size;
    p := stream.Memory;
    SetString(Result,p,sz);
  finally
    stream.free;
  end;
end;

// returns SQLID from server
function  TApolloQuery.Prepare: Integer;  // returns a SQL ID
var
  iError : Integer;
  sError : String;
begin
  Result     := 0;
  iExecSQLID := 0;

	if (AccessMethod = amServer) and (FDatabaseName = '') then         {LAS: 06-09-00, Added. 07-04-00, Added amServer check}
		RaiseApolloQuery( ERR_MISSING_DATABASENAME_PROPERTY );  {LAS: 06-09-00, Added}

  {MED: 03-29-01, added to create a Conn on the fly in case one was not assigned
  !!! works only in amLocal mode  !!!  }
  if (FApolloConnection = nil) then
  begin
    if (FAccessMethod = amLocal) then
    begin
      // MED: 08-15-01, moved here from Constructor
      if FInternalConnection = nil then
      begin
        FInternalConnection := TApolloConnection.Create(Self);
        FInternalConnection.Name := 'InternalConnection';
      end;

      FApolloConnection := FInternalConnection;
      FApolloConnection.DataBaseName := FDatabaseName;
    end
    else
      RaiseApolloQuery( ERR_NO_APOLLOCONNECTION_SPECIFIED )
  end;

  if (FAccessMethod = amServer) and (not FApolloConnection.Connected) then
  begin
    iSQLID:= 0;
    Active:= False;
    RaiseApolloQuery( ERR_NOT_CONNECTED );
  end;

  {LAS: 06-24-00, Added.  Must resolve DataPath here just like in OpwnSQL}
  if FAccessMethod = amLocal then
  begin
    LoadSQLDLL;
    FDatabase := GetDatabase;
    FDataPath := PathFromAlias;

    if Assigned( FApolloConnection ) then
      FApolloConnection.SQLDataPath := FDataPath;
  end;

  if (Assigned(FApolloConnection)) then  {MED: 05-28-00, took out Connected}
  begin
    SetToApolloEnv;

    SetCursor;

    iExecSQLID := FApolloConnection.ap_sql_Prepare(
                                           FAccessMethod,
                                           FDataBaseName, TableNames,
                                           FCentury,
                                           FDateFormat,
                                           FDeleted,
                                           FMemoBlockSize,
                                           FParams,
                                           FReadOnly,
                                           FTableType,
                                           FPassword,
                                           FTranslate,
                                           FCommitLeveL,
                                           iError, sError );

    UnSetCursor;

    FErrorMsg := sError;
    // Unable to open file  -- this is the SQL ID NOT the work area!
    if (iExecSQLID < 1) then
      RaiseApolloQuery( sError);

    Result := iExecSQLID;
  end;
end;

procedure TApolloQuery.ExecSQL;
var
  iError         : Integer;
  sError         : String;
  bLocalPrepared : WordBool;
begin

{LAS: 04-25-00, Added EVAL Support}
{$ifdef EVAL}
  	if not NewCheckTimeLimit(REGKEY,
                             REGKEY_KILL_NAME,
                             REGKEY_CHECK_NAME,
                             'Apollo SQL 6.1', TRUE) then
    Exit;
{$endif}


	if (AccessMethod = amServer) and (FDatabaseName = '') then         {LAS: 06-09-00, Added. 07-04-00, Added amServer check}
		RaiseApolloQuery( ERR_MISSING_DATABASENAME_PROPERTY );  {LAS: 06-09-00, Added}

	if FAccessMethod = amLocal then    {LAS: 05-05-00, Added}
  	LoadSQLDLL;                      {LAS: 05-05-00, Added}

  bLocalPrepared := FALSE;
  if iExecSQLID < 1 then
  begin
    bLocalPrepared := TRUE;
    iExecSQLID     := Prepare;
  end;
  if iExecSQLID > 0 then
  begin
    with FApolloConnection do
    begin
      SetCursor;

      ap_sql_ExecSQL(FAccessMethod,
                     iExecSQLID,
                     SQL.Text,
                     FParams,
                     iError, sError);

      UnSetCursor;

      FErrorMsg := sError;
      if bLocalPrepared then
      begin
        ap_sql_UnPrepare(FAccessMethod, iExecSQLID);
        iExecSQLID := 0;
      end;
      if iError > 0 then
        RaiseApolloQuery( FErrorMsg );
    end;
  end;
end;

procedure TApolloQuery.UnPrepare;
begin
  if iExecSQLID > 0 then
    FApolloConnection.ap_sql_UnPrepare(FAccessMethod, iExecSQLID);
  iExecSQLID := 0;
end;


{==============================================================================
  SetSQL
==============================================================================}
procedure TApolloQuery.SetSQL( Value: TStrings );
begin
  if Active = True then
  	Close;
  if (Value <> FSQL) then
    FSQL.Assign( Value );
end;


{LAS: 04-25-00, Updated for EVAL}
{==============================================================================
  GetVersion - Returns Version Info
 ==============================================================================}
function TApolloQuery.GetVersion: String;
const
{$ifdef EVAL}
	EVAL = '[EVAL] ';
{$else}
	EVAL = '';
{$endif}
begin
	Result := 'ApolloQuery ' + EVAL + ApolloSQL_VER + ', ' +
                        DateToStr( EncodeDate( VER_YEAR, VER_MONTH, VER_DAY ));
end;

{==============================================================================
  SetApolloConnection for Client/Server
 ==============================================================================}
procedure TApolloQuery.SetApolloConnection(Val:TApolloConnection);
begin
  // CRVG, 2001-11-15 : Changed back
	if FApolloConnection <> Val then
		FApolloConnection := Val;

{ [ac] 11-05-2001 Suggested :
	if FApolloConnection <> Val then
	begin
		FApolloConnection := Val;
    if csDesigning in ComponentState then
      if FApolloConnection = nil then
        FAccessMethod := amLocal
      else
        FAccessMethod := amServer;
	end;}

end;

{==============================================================================
  SetFetchCount
 ==============================================================================}
procedure TApolloQuery.SetFetchCount(Val:Integer);
begin
	if FFetchCount <> Val then
	begin
    if Val < 1 then
    	Val := 1;
		FFetchCount := Val;
	end;
end;

//==============================================================================
procedure TApolloQuery.SetPanels(Value: TApolloTables);
begin
  FApolloTables.Assign(Value);
end;

function TApolloQuery.GetPanels: TApolloTables;
begin
  Result := FApolloTables;
end;

///////////////////////////////////////////////////
//MED: 09-08-01, added
///////////////////////////////////////////////////
procedure TApolloQuery.SetSystemCollation;
begin
  FApolloConnection.ap_sql_SetSystemCollation(FAccessMethod);
end;

procedure TApolloQuery.SetMachineCollation;
begin
  FApolloConnection.ap_sql_SetMachineCollation(FAccessMethod);
end;

procedure TApolloQuery.AddDudenCollation;
begin
  FApolloConnection.ap_sql_AddDudenCollation(FAccessMethod);
end;

procedure TApolloQuery.AddEtecCollation;
begin
  FApolloConnection.ap_sql_AddEtecCollation(FAccessMethod);
end;

////////////////////////////////////

constructor TApolloTableItem.Create(Collection: TCollection);
begin
  inherited Create(Collection);
  FItemCommitLevel := clNormal;
  FExtraIndexes := TStringList.Create;
  FTableType    := ttSXFOX;
end;

destructor TApolloTableItem.Destroy;
begin
  FExtraIndexes.Free;
	inherited Destroy;
end;

procedure TApolloTableItem.Assign(Source: TPersistent);
begin
  if Source is TApolloTableItem then
  begin
    Alias        := TApolloTableItem(Source).Alias;
    TableName    := TApolloTableItem(Source).TableName;
    TableType    := TApolloTableItem(Source).TableType;
    Password     := TApolloTableItem(Source).Password;
    OEMTranslate := TApolloTableItem(Source).OEMTranslate;  {LAS: 05-09-00, Added}
    ExtraIndexes := TApolloTableItem(Source).ExtraIndexes;
  end
  else
  	inherited Assign(Source);
end;

function TApolloTableItem.GetDisplayName: String;
begin
  Result := GetCaption;
  if Result='' then
    Result := inherited GetDisplayName;
end;

procedure TApolloTableItem.SetExtraIndexes(Value: TStringList);
begin
  if (Value <> FExtraIndexes) then
		FExtraIndexes.Assign( Value );
end;

procedure TApolloTableItem.SetTableName(const Value: String);
var
  iPos : Integer;
  sAlias : String;
begin
  if FTableName <> Value then
  begin
    FTableName := Value;
    sAlias := Value;
    if (FAlias = '') and (Value <> '') then
    begin
      FAlias := StripStr(sAlias, '.', iPos);
    end;

    Changed(False);
  end;
end;

function TApolloTableItem.GetCaption: String;
begin
  if FTableName <> '' then
    Result := FTableName + ' ('+FAlias+')'
  else
    Result := 'Undefined';
end;


{ TApolloTables }
constructor TApolloTables.Create(StatusBar: TApolloQuery);
begin
  inherited Create(TApolloTableItem);
  FStatusBar := StatusBar;
end;

function TApolloTables.Add: TApolloTableItem;
begin
  Result := TApolloTableItem(inherited Add);
end;

function TApolloTables.GetItem(Index: Integer): TApolloTableItem;
begin
  Result := TApolloTableItem(inherited GetItem(Index));
end;

function TApolloTables.GetOwner: TPersistent;
begin
  Result := FStatusBar;
end;

procedure TApolloTables.SetItem(Index: Integer; Value: TApolloTableItem);
begin
  inherited SetItem(Index, Value);
end;


{LAS: 05-09-00, Added}
{==============================================================================
  PathFromAlias - If TApolloQuery Alias is used, gets physical path (LOCAL ONLY)
 ==============================================================================}
function TApolloQuery.PathFromAlias: String;
begin
	Result := DatabaseName;  // Default

  if ((Pos( '\', Result ) > 0) or    // Check for characters that would
      (Pos( ':', Result ) > 0) or    //   make it clear that this is a physical
      (Pos( '.', Result ) > 0)) and  //    file path and not an alias.
			not(Result = '') then   // If NOT empty, bail now
  	exit;

	if Assigned( Database ) then
		Result := Database.DataPath;

  // If no path was specified, use .DBF current directory
  if (Result = '') then
    GetDir( 0, Result );

  if (Length(Result) > 0) and (Result[ Length(Result) ] <> '\') then
    Result := Result + '\';
end;

{==============================================================================
  SetDatabaseName  // LAS: 05-09-00, Rewritten 100% for local support
 ==============================================================================}
procedure TApolloQuery.SetDatabaseName( Value: String );
begin
  if (Value <> FDatabaseName) then
  begin
		FDatabase := nil;   {LAS: 06-03-00, Added}
    FDatabaseName := Value;

    // If local physical path, change AccessMethod to amLocal {LAS: 06-03-00, Added}

    // MED: 11-09-00, commented.  If user wants it, he/she will change it.
    //if (Pos(':', FDatabaseName) > 0) or (Pos('\', FDatabaseName) > 0) then
    //	FAccessMethod := amLocal;

		if Assigned( FApolloConnection ) then
    	FApolloConnection.DataBaseName := Value;

		if FAccessMethod = amLocal then
    begin
      FDatabase := GetDatabase;
      FDataPath := PathFromAlias;

			if Assigned( FApolloConnection ) then
        FApolloConnection.SQLDataPath := FDataPath;
		end;
	end;
end;

{==============================================================================
  GetDatabase  // LAS: 05-05-00, Added
 ==============================================================================}
function TApolloQuery.GetDatabase: TApolloDatabase;
var
//  bFound : Boolean;
//  sPath : String;
//  SearchRec : TSearchRec;
  Temp: TComponent;
  SearchObj : TApolloSearchObject;
begin
	Result := FDatabase;
//  bFound := False;

  if Assigned(FDataBase) then
    exit;

	SearchObj := TApolloSearchObject.Create;
  try
    Temp := SearchObj.SearchFirstComponent(TApolloDataBase);
    while Temp <> nil do
    begin
      if (Temp is TApolloDataBase) then
      begin
        if (UpperCase(TApolloDatabase(temp).DataPathAlias) = UpperCase(DatabaseName)) then
        begin
          FDatabase := TApolloDataBase(temp);
//          bFound    := TRUE;
          break;
        end;
      end;
      Temp := SearchObj.SearchNextComponent(TApolloDataBase);
    end;
  finally
    SearchObj.Free;
  end;
(* ----------------------------------{ LAS: 05-11-00, Removed.  Not needed. }
  if not bFound then
  begin
    // we didn't find it
		sPath := DatabaseName;

		// Strip backslash from end of path
		if (Length(sPath) > 0) and (sPath[ Length(sPath) ] = '\') then
		  sPath := Copy( sPath, 1, Length( sPath )-1 );

    if SysUtils.FindFirst( sPath, faDirectory, SearchRec ) <> 0 then
    begin
      // Only raise exception if DatabaseName is not empty but also not valid
      if (Length(DatabaseName) > 0) then
        RaiseApolloQuery( ERR_INVALID_DATABASENAME +' : ' + DatabaseName );
		end;
    SysUtils.FindClose( SearchRec );
  end;
*)
	Result := FDatabase;
end;

//------------------------------------------------------------------------------
function TApolloQuery.CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream;
begin
	UpdateCursorPos;
  Result := TApolloQueryBlobStream.Create( FAccessMethod, FApolloConnection,
                                           FRecNo, Field as TBlobField, Mode);
end;


{ TTApolloQueryBlobStream }
constructor TApolloQueryBlobStream.Create( amMode: TAccessMethod;
                                           FApolloConnection: TApolloConnection;
                                           iRecNo: Integer; Field: TBlobField;
                                           Mode: TBlobStreamMode);
begin
  inherited Create;

  FAccessMethod := amMode;
  FIndex        := Field.Index;
  iWA           := TApolloQuery(Field.DataSet).iSQLID;
  FRecNo        := iRecNo;
  pConn         := FApolloConnection;
  ReadBlobData;
end;

procedure TApolloQueryBlobStream.ReadBlobData;
var
  ptr: Pointer;
begin
  ptr := TApolloConnection(pConn).ap_sql_ReadBlobData(FAccessMethod, iWA, FRecNo, FIndex);
  if ptr <> nil then
    Self.CopyFrom(TMemoryStream(ptr), TMemoryStream(ptr).Size);

  Self.Position:= 0;
end;

{==============================================================================
  SetToApolloEnv
 ==============================================================================}
function TApolloQuery.SetToApolloEnv: TApolloEnv;
var
  Temp: TComponent;
  SearchObj : TApolloSearchObject;
begin
  Result := Nil;
  SearchObj := TApolloSearchObject.Create;
  try
    Temp := SearchObj.SearchFirstComponent(TApolloEnv);
  finally
    SearchObj.Free;
  end;

  if Temp <> nil then
  begin
    {MED: 03-05-2000, changed to support both modes}
    //if (Temp is TApolloEnv) and ((Temp as TApolloEnv).ApplyTo = [atServer]) then
    if (Temp is TApolloEnv) then
    begin
      Result         := TApolloEnv(Temp);
      FCentury       := Result.Century;
      //FDateFormat    := Result.DateFormat;      {MED: 10-17-00, made it a property}
      FDeleted       := Result.Deleted;
      FMemoBlockSize := Result.MemoBlockSize;
    end;
  end;
end;

{MED: 03-28-00, added same as ApolloQuery1.Params.ParamByName}
function TApolloQuery.ParamByName(const Value: string): TParam;
begin
	Result:= FParams.ParamByName(Value);
end;

{LAS: 04-19-00, Added}
{==============================================================================
  Locate
 ==============================================================================}
function TApolloQuery.Locate(const KeyFields: String; const KeyValues: Variant;
                             Options: TLocateOptions): Boolean;
var
	i, iCurRec, iFldCount: Integer;
	tmpFields: TList;
  A : Variant;
	sFldVal, sKeyVal: String;
  bCont: Boolean;
begin
	Result := False;

  // Preserve current record position, in case of failed Locate
  iCurRec := RecNo;

  bCont := ControlsDisabled;
  DisableControls;

	tmpFields := TList.Create;

	try
    // Parse the field names into a list of fields
    GetFieldList(tmpFields, KeyFields);

    iFldCount := tmpFields.Count;
    A := VarArrayCreate([0, iFldCount-1], varVariant);
		A := KeyValues;

	  First;
    while not Eof and (Result = False) do
    begin
      Result := True;                               {Start with True}
      for i := 0 to iFldCount-1 do                  {Cycle through all fields}
      begin
        { Handle Case-sensitive and Partial match options for strings }
				if (TField(tmpFields[i]).DataType = ftString) then
        begin
          sFldVal := TField(tmpFields[i]).Value;
          if iFldCount = 1 then  {Only a single value passed in}
            sKeyVal := A
          else
            sKeyVal := A[i];

        	if not (loPartialKey in Options) and
                 (Length( sFldVal ) <> Length( sKeyVal )) then
					begin
          	Result := False;
						break;
          end;

        	if (loCaseInsensitive in Options) then
          begin
						sFldVal := UpperCase( sFldVal );
          	sKeyVal := UpperCase( sKeyVal );
					end;

          if (Copy( sFldVal, 1, Length( sKeyVal )) <> sKeyVal ) then
					begin
          	Result := False;
						break;
          end;

        end
        else
        begin
          if VarType(A) < varArray then
          begin
            if (TField(tmpFields[i]).Value <> A) then
						begin
          		Result := False;
							break;
  	        end;
          end
					else
          begin
            if (TField(tmpFields[i]).Value <> A[i]) then
						begin
          		Result := False;
							break;
  	        end;
	        end;
				end;
			end;

      if Result then                          {If match was found...}
      	break;                                {..., stop looking}

      Next;                                   {Otherwise, move to next record}
    end;

	finally
		tmpFields.Free;
	end;

  if not bCont then
  	EnableControls;

  {If a match was not found, restore original record position}
  if not Result then
    if (not EOF) and (not BOF) then  // 2001-10-10 CRVG Added checking for an empty table 
    RecNo := iCurRec;

end;





end.

