(*============================================================================
  ApoServerDLL.PAS
  Main TApolloServerDLL code

  Copyright 1999-2002 Vista Software
  Last Updated: 10-22-2002
  
  NOTE: Tab stops were set at 2 spaces during development of this unit. If you
        are using a tab stop setting other than 2 spaces, the code indents
        may not line up accurately in your editor.
============================================================================*)

unit ApoServerDLL;

interface

uses
  DB, TypInfo, Classes, SysUtils, Dialogs, Forms, ApConn, ApoDSet, ApCommon;

const
	ERR_EMPTY_PROCNAME = 'ProcName is empty!';

type

  // MED: 04-02-2001
  TExecDLLErrorEvent = procedure(Sender : TObject; ErrorMsg:String; ErrorNo: Integer) of object;

  EApolloServerDLL = class(Exception);

  TApolloServerDLL = class(TComponent)
  private
    FDummyVersion : String;
    FApolloConnection : TApolloConnection;
    FDatabaseName     : String;
    FVersion          : String;
    FProcName         : String;
    FProcParams       : TParams;

    FErrorMsg         : String;    // MED: 04-02-2001
    FErrorNo          : Integer;
    FOnError          : TExecDLLErrorEvent;

    procedure SetApolloConnection(Val:TApolloConnection);
    procedure SetDatabaseName(Value: String);
    procedure SetDLLProcName(Value : String);   //  also calls GetProcedureNames;

    // MED: 04-02-2001   moved here from ApConn.pas
    procedure InternalGetParams;
    procedure InternalSetParams;

  protected
    { Protected declarations }
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure Loaded; override;   // MED: 04-02-2001

  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    // MED: 04-02-2001
    property ErrorMsg : String read FErrorMsg write FErrorMsg;
    property ErrorNo  : Integer read FErrorNo write FErrorNo;

  published

    property ApolloConnection: TApolloConnection read FApolloConnection
                                                 write SetApolloConnection;

    property DatabaseName: String read FDatabaseName write SetDatabaseName;
    property ProcName: String read FProcName write SetDLLProcName;  
    property ProcParams : TParams read FProcParams write FProcParams;
    property OnError: TExecDLLErrorEvent read FOnError write FOnError;
    property Version: String read FVersion write FDummyVersion;
    
    Function ExecProc : Boolean;
  end;

implementation

function SubStr(const cString:String; nStart,nEnd: Integer): String;
begin
	Result := Copy(cString, nStart, nEnd);
end;

function Padr ( cStr:String; nRepeat:Integer; cChar:Char): String;
begin
  Result := cStr + StringOfChar(cChar, nRepeat - Length(cStr) );
end;

procedure TApolloServerDLL.Loaded;
begin
  inherited;
	if Assigned( FApolloConnection ) then
    FApolloConnection.DataBaseName := FDatabaseName;
end;

{==============================================================================}
procedure TApolloServerDLL.SetApolloConnection(Val:TApolloConnection);
begin
	if FApolloConnection <> Val then
		FApolloConnection := Val;
end;


constructor TApolloServerDLL.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FProcParams := TParams.Create(self);
  FVersion := ApolloServerDLL_VER;
  FErrorNo := 0;
  FErrorMsg:= '';
end;

destructor TApolloServerDLL.Destroy;
begin
  FProcParams.Free;
  inherited Destroy;
end;


{==============================================================================
	Notification
==============================================================================}
procedure TApolloServerDLL.Notification( AComponent: TComponent;
                                          Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = FApolloConnection) then
    FApolloConnection := nil;
end;

{==============================================================================
  SetDatabaseName
 ==============================================================================}
procedure TApolloServerDLL.SetDatabaseName( Value: String );
begin
  if (Value <> FDatabaseName) then
  begin
    FDatabaseName := Value;
  end;
end;

// InternalGetParam is called only when ProcName is changed.
// anytime you change the ProcName the parameter list is fetched from the server
procedure TApolloServerDLL.SetDLLProcName(Value:String);
var
  msPtr : Pointer;
begin
  if FProcName <> Value then
  begin
    FProcParams.Clear;
    FProcName := Value;

    if FApolloConnection = nil then       // MED: it is nil when loading form
      exit;

    if (not Assigned(FApolloConnection)) then
      raise EApolloServerDLL.Create( ERR_NO_APOLLOCONNECTION_SPECIFIED );

    if not FApolloConnection.Connected then
      raise EApolloServerDLL.Create( ERR_NOT_CONNECTED );

    msPtr := FApolloConnection.ap_SetDLLProcName(FDatabaseName, FProcName);
    if msPtr <> nil then
      InternalGetParams;
  end;

end;

// InternalGetParam is called only when ProcName is changed.
// therefore for internal DLL procedures set ProcName = '' at design time
function TApolloServerDLL.ExecProc:Boolean;
begin
  Result := FALSE;
  if (not Assigned(FApolloConnection)) then
    raise EApolloServerDLL.Create( ERR_NO_APOLLOCONNECTION_SPECIFIED );
  if not FApolloConnection.Connected then
    raise EApolloServerDLL.Create( ERR_NOT_CONNECTED );
  if (FProcName = '') then
    raise EApolloServerDLL.Create( ERR_EMPTY_PROCNAME );

  InternalSetParams;
  if (FApolloConnection.ap_ExecDLLProc <> nil) then
  begin
    InternalGetParams;
    if ErrorNo <> 0 then
    begin
      if (Assigned(FOnError)) then
        FOnError(Self, FErrorMsg, FErrorNo)
      else
        raise EApolloServerDLL.Create( ErrorMsg );
    end;
    Result := TRUE;
  end;
end;


procedure TApolloServerDLL.InternalGetParams;
var
  buff : Pointer;
  dType, pType,
  nRead, ToRead: Integer;
  sName : String;
begin
  Buff := nil;
  FProcParams.Clear;
  With FApolloConnection do
  begin

    if ReadStream.Size = 0 then
      exit;

    ReadStream.Position := 0;

    try
      GetMem(Buff, ReadStream.Size);
      nRead     := 0;
      nRead     := nRead + ReadStream.Read( PChar(buff)^, 4);
      FErrorNo  := StrToInt('$'+SubStr(PChar(buff), 1, 4));
      nRead     := nRead + ReadStream.Read( PChar(buff)^, 255);
      FErrorMsg := SubStr(PChar(buff),1 , 255);
      //showmessage('ErrorMsg--->'+FErrorMsg);

      ToRead    := 0;
      Repeat
        nRead := nRead + ReadStream.Read( PChar(buff)^, 8);
        if nRead > 259 then
          ToRead := StrToInt('$'+SubStr(PChar(buff),1 ,8));
        if ToRead > 0 then
        begin
          nRead := nRead + ReadStream.Read( PChar(buff)^, 2);   // datatype
          dType := StrToInt('$'+SubStr(PChar(buff),1,2));
          nRead := nRead + ReadStream.Read( PChar(buff)^, 2);   // paramtype
          pType := StrToInt('$'+SubStr(PChar(buff),1,2));
          nRead := nRead + ReadStream.Read( PChar(buff)^,20 ); // name
          sName := Trim(SubStr(PChar(buff), 1, 20));
          nRead := nRead + ReadStream.Read( PChar(buff)^, ToRead ); // Data
          FProcParams.CreateParam(TFieldType(dType), sName, TParamType(pType));
          FProcParams.ParamByName(sName).SetData(buff);
        end;
      Until ((nRead = ReadStream.Size)); // and (ToRead = 0) );
    finally
      FreeMem(Buff);
    end;
  end;
end;

// 4-7-99 Parameter name is restricted to 20 chars
// send current parameters to the server and tell the server to use these parameters
procedure TApolloServerDLL.InternalSetParams;
var
  i : Integer;
  buffPtr : Pointer;
  size : Integer;
  sName, sType, cSize : String;
begin
  With FApolloConnection do
  begin
    WriteStream.Clear;
    WriteStream.Position := 0;
    WriteStream.Write(PChar(PadR(FDataBaseName, 255, ' '))^, 255);
    WriteStream.Write(PChar(PadR(FProcName, 255, ' '))^, 255);  // write the procedure name to stream
    BuffPtr   := Nil;
    if (FProcParams <> nil) and (FProcParams.Count > 0) then
    begin
      for i:= 0 to FProcParams.Count -1 do
      begin
        size := FProcParams.Items[i].GetDataSize;
        try
          try
            BuffPtr := AllocMem(size);
          except
            FErrorMsg :=  'ApolloServerDLL.InternalSetParams' ;
            if (Assigned(FOnError)) then
              FOnError(Self, FErrorMsg,  -1)
            else
              raise EApolloServerDLL.Create(ErrorMsg);

          end;
          FProcParams.Items[i].GetData( BuffPtr);

          cSize := IntToHex(Size, 8);
          WriteStream.Write(PChar(cSize)^, 8);

          sType := IntToHex(ord(FProcParams.Items[i].DataType), 2);
          WriteStream.Write(PChar(sType)^, 2);

          sType := IntToHex(ord(FProcParams.Items[i].ParamType), 2);
          WriteStream.Write(PChar(sType)^, 2);

          sName := PadR(FProcParams.Items[i].name, 20, ' ');
          WriteStream.Write(PChar(sName)^, 20);
          WriteStream.Write(PChar(BuffPtr)^, size);
        finally
          FreeMem(BuffPtr);
        end;
      end;
    end;
    WriteStream.Position := 0;     // this is a must have
  end;
end;


end.
