{******************************************************************************}
{                                                                              }
{           ReportBuilder Data Access Developement Environment (DADE)          }
{                                                                              }
{             Copyright (c) 1996-1999 Digital Metaphors Corporation            }
{                                                                              }
{******************************************************************************}
{      Originally modified to be used with Apollo by:                          }
{                          Tony Wood, Turnkey Computer Technology Ltd          }
{                                     Glasgow, Scotland                        }
{      No guarantees are given as to the functionality of the modifications.   }
{      You should be prepared to validate any application using this code.     }
{                                                                              }
{******************************************************************************}


//-----------------------------------------------------------------------------
// This code was contributed by Tony Wood. http://www.tctl.co.uk/
// (c)2002 Vista Software. All rights reserved.
// www.vistasoftware.com
//-----------------------------------------------------------------------------


unit daApollo;

interface

uses Classes, SysUtils, Forms, ExtCtrls,  DB,  Dialogs,
     ppComm, ppClass, ppDBPipe, ppDB, ppClasUt, ppTypes,
     daDB, daQueryDataView, daDataView, daPreviewDataDlg,
     {these is the unit containing the Apollo objects}
     ApoEnv, ApoDSet, ApoQSet, APwin, ApConn, ApCommon, daQClass;

type
  {Apollo DataView Classes:
     1.  Apollo TDataSet descendants
           - TDataSets that can be children of a DataView.
           - Override the HasParent method of TComponent to return True
           - Must be registerd with the Delphi IDE using the RegisterNoIcon
             procedure
       a. TdaChildApolloQuery - TApolloQuery descendant that can be a child
                                of a DataView
       b. TdaChildApolloTable - TApolloTable descendant that can be a child
                                of a DataView
     2.  TdaApolloSession
           - descendant of TppSession
           - implements GetDatabaseNames, GetTableNames, etc.
     3.  TdaApolloDataSet
          - descendant of TppDataSet
          - implements GetFieldNames for SQL
     4.  TdaApolloQueryDataView
          - descendant of TppQueryDataView
          - uses the above classes to create the required
            Query -> DataSource -> Pipeline -> Report connection
          - uses the TdaSQL object built by the QueryWizard to assign
            SQL to the TApolloQuery etc.
      }

  { TdaChildApolloQuery }
  TdaChildApolloQuery = class(TApolloQuery)
    private
      FConnection: TApolloConnection;
    public
      constructor Create(aComponent: TComponent); override;
      destructor Destroy; override;
      function HasParent: Boolean; override;

    end;  {class, TdaChildApolloQuery}

  { TdaChildDBISMTable }
  TdaChildApolloTable = class(TApolloTable)
    public
      function HasParent: Boolean; override;

    end;  {class, TdaChildApolloTable}


  { TdaApolloSession }
  TdaApolloSession = class(TdaSession)
    private
      procedure AddDatabase(aDatabase: TComponent);

    protected
      procedure SetDataOwner(aDataOwner: TComponent); override;

    public
      class function ClassDescription: String; override;
      class function DataSetClass: TdaDataSetClass; override;
      class function DatabaseClass: TComponentClass; override;

      function  DefaultSQLType(aDatabaseType: TppDatabaseType): TppSQLType; override;
      procedure GetDatabaseNames(aList: TStrings); override;
      function  GetDatabaseType(const aDatabaseName: String): TppDatabaseType; override;
      procedure GetTableNames(const aDatabaseName: String; aList: TStrings); override;
      function ValidDatabaseTypes: TppDatabaseTypes; override;
      function GetSearchCriteriaDateFormat(aDatabaseType:
                TppDatabaseType; const aDatabaseName: String): String; override;

  end; {class, TdaApolloSession}



  { TdaApolloDataSet }
  TdaApolloDataSet = class(TdaDataSet)
    private
      FDataSet: TApolloTable;

      function GetDataSet: TDataSet;

    protected
      procedure BuildFieldList; override;
      function  GetActive: Boolean; override;
      procedure SetActive(Value: Boolean); override;
      procedure SetDatabaseName(const aDatabaseName: String); override;
      procedure SetDataName(const aDataName: String); override;

      property DataSet: TDataSet read GetDataSet;

    public
      constructor Create(aOwner: TComponent); override;
      destructor Destroy; override;

      class function ClassDescription: String; override;

      procedure GetFieldNamesForSQL(aList: TStrings; aSQL: TStrings); override;
      procedure GetFieldsForSQL(aList: TList; aSQL: TStrings); override;

  end; {class, TdaApolloDataSet}


  { TdaApolloQueryDataView }
  TdaApolloQueryDataView = class(TdaQueryDataView)
    private
      FDataSource: TppChildDataSource;
      FQuery: TdaChildApolloQuery;

    protected
      procedure SQLChanged; override;

    public
      constructor Create(aOwner: TComponent); override;
      destructor Destroy; override;

      class function PreviewFormClass: TFormClass; override;
      class function SessionClass: TClass; override;

      procedure Init; override;
      procedure ConnectPipelinesToData; override;

    published
      property DataSource: TppChildDataSource read FDataSource;

  end; {class, TdaApolloQueryDataView}


  {global functions to access default Apollo session and database}

  {utility routines}

  {global functions to access default Apollo connection}
  function daGetDefaultApolloConnection: TApolloDatabase;

  {utility routines}
  procedure daGetApolloConnectionNames(aList: TStrings);
  function daGetApolloConnectionForName(aDatabaseName: String): TApolloDatabase;

  function daGetApolloConnectionList: TppComponentList;

  {Delphi design time registration}
  procedure Register;

implementation

const
  cDefaultDatabase = 'DefaultApolloDatabase';

var
  FApolloDatabase: TApolloDatabase;
  FApolloConnectionList: TppComponentList;

function daAddBackSlash(aSource: string): string;
var
  lChar: string[1];

begin

  Result := aSource;

  lChar := copy(aSource, Length(aSource), 1);

  if (lChar <> ':') and (lChar <> '\') then
    Result := Result + '\';

end; {function, daAddBackSlash}


{******************************************************************************
 *
 ** R E G I S T E R
 *
{******************************************************************************}

procedure Register;
begin
  {Apollo DataAccess Components}
  RegisterNoIcon([TdaChildApolloQuery, TdaChildApolloTable]);

  {Apollo DataViews}
  RegisterNoIcon([TdaApolloQueryDataView]);
end;

{******************************************************************************
 *
 ** C H I L D   Apollo  D A T A   A C C E S S   C O M P O N  E N T S
 *
{******************************************************************************}

{------------------------------------------------------------------------------}
constructor TdaChildApolloQuery.Create(aComponent: TComponent);
begin
  inherited Create(aComponent);
  FConnection := TApolloConnection.Create(Self);
  ApolloConnection:=FConnection;
  AccessMethod:=amLocal;
end;
destructor TdaChildApolloQuery.Destroy;
begin
  FConnection.Free;
  inherited Destroy;
end;
{ TdaChildApolloQuery.HasParent }
function TdaChildApolloQuery.HasParent: Boolean;
begin
  Result := True;
end; {function, HasParent}

{------------------------------------------------------------------------------}
{ TdaChildApolloTable.HasParent }
function TdaChildApolloTable.HasParent: Boolean;
begin
  Result := True;
end; {function, HasParent}


{******************************************************************************
 *
 ** Apollo   S E S S I O N
 *
{******************************************************************************}

{------------------------------------------------------------------------------}
{ TdaApolloSession.ClassDescription }

class function TdaApolloSession.ClassDescription: String;
begin
  Result := 'ApolloSession';
end; {class function, ClassDescription}

{------------------------------------------------------------------------------}
{ TdaApolloSession.DataSetClass }

class function TdaApolloSession.DataSetClass: TdaDataSetClass;
begin
  Result := TdaApolloDataSet;
end; {class function, DataSetClass}

{------------------------------------------------------------------------------}
{ TdaApolloSession.DatabaseClass }

class function TdaApolloSession.DatabaseClass: TComponentClass;
begin
  Result := TApolloDatabase;
end; {class function, DatabaseClass}

{------------------------------------------------------------------------------}
{ TdaApolloSession.GetTableNames }

procedure TdaApolloSession.GetTableNames(const aDatabaseName: String; aList: TStrings);
var
  lDatabase: TApolloDatabase;
  lsPath: String;
  liFind: Integer;
  liIndex: integer;
  lSearchRec: TSearchRec;
begin
  {get the database}
  lDatabase := daGetApolloConnectionForName(aDatabaseName);

  aList.Clear;
  lsPath:= daAddBackSlash(lDatabase.DataPath);
  lsPath := lsPath+'*.DBF';
  {build list of the files in the directory}
  liFind := FindFirst(lsPath, 0, lSearchRec);
  while (liFind = 0) do
    begin
      liIndex := pos('.', lSearchRec.name);
      if liIndex >0 then
        aList.Append(Copy(lSearchRec.name, 1, liIndex-1) )
      else
        aList.Append(lSearchRec.name);
     liFind := FindNext(lSearchRec);
    end;

end; {procedure, GetTableNames}

{------------------------------------------------------------------------------}
{ TdaApolloSession.SetDataOwner }

procedure TdaApolloSession.SetDataOwner(aDataOwner: TComponent);
var
  lList: TStringList;
begin

  inherited SetDataOwner(aDataOwner);

  lList := TStringList.Create;

  {refresh database names}
  GetDatabaseNames(lList);

  lList.Free;

end; {procedure, SetDataOwner}

{------------------------------------------------------------------------------}
{ TdaApolloSession.GetDatabaseType }

function TdaApolloSession.GetDatabaseType(const aDatabaseName: String): TppDatabaseType;
begin
  Result := dtOther;
end; {function, GetDatabaseType}

{------------------------------------------------------------------------------}
{ TdaApolloSession.GetDatabaseNames }

procedure TdaApolloSession.GetDatabaseNames(aList: TStrings);
var
  liIndex: Integer;

begin

  {call utility routine to get list of database names}
  daGetApolloConnectionNames(aList);

  {get additional names from Database objects residing on Owner}
  daGetDatabaseObjectsFromOwner(TdaSessionClass(Self.ClassType), aList, DataOwner);

  {build list of connection objects}
  for liIndex := 0 to aList.Count-1 do
    if aList.Objects[liIndex] <> nil then
      AddDatabase(TComponent(aList.Objects[liIndex]));

end; {procedure, GetDatabaseNames}

{------------------------------------------------------------------------------}
{ TdaApolloSession.AddDatabase }

procedure TdaApolloSession.AddDatabase(aDatabase: TComponent);
begin

  if daGetApolloConnectionList.IndexOf(aDatabase) < 0 then
    FApolloConnectionList.Add(aDatabase);

end; {procedure, AddDatabase}

{------------------------------------------------------------------------------}
{ TdaApolloSession.ValidDatabaseTypes }
function TdaApolloSession.ValidDatabaseTypes: TppDatabaseTypes;
begin
   Result := [dtOther];
end; {function, ValidDatabaseTypes }

{------------------------------------------------------------------------------}
{ TdaApolloSession.DefaultSQLType }

function TdaApolloSession.DefaultSQLType(aDatabaseType: TppDatabaseType): TppSQLType;
begin
  Result := sqSQL2;
end; {function, DefaultSQLType}

function TdaApolloSession.GetSearchCriteriaDateFormat(aDatabaseType:
TppDatabaseType; const aDatabaseName: String): String;
begin

  {return a format usable by FormatDate}
  case aDatabaseType of

    dtMSAccess:
      Result := 'YYYY-MM-DD';

    dtMSSQLServer, dtSybaseASA, dtSybaseASE, dtOracle:
      Result := 'YYYY/MM/DD';

    dtAdvantage:
      Result := 'YYYY-MM-DD';

    dtOther:
      Result := 'MM/DD/YYYY';

    else
      Result := 'MM/DD/YYYY';

  end;
end;

{******************************************************************************
 *
 ** Apollo   D A T A S E T
 *
{******************************************************************************}

{------------------------------------------------------------------------------}
{ TdaApolloDataSet.Create }

constructor TdaApolloDataSet.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  FDataSet := nil;

end; {constructor, Create}

{------------------------------------------------------------------------------}
{ TdaApolloDataSet.Destroy }

destructor TdaApolloDataSet.Destroy;
begin
  FDataSet.Free;

  inherited Destroy;
end; {destructor, Destroy}

{------------------------------------------------------------------------------}
{ TdaApolloDataSet.ClassDescription }

class function TdaApolloDataSet.ClassDescription: String;
begin
  Result := 'ApolloDataSet';
end; {class function, ClassDescription}

{------------------------------------------------------------------------------}
{ TdaApolloDataSet.GetActive }

function TdaApolloDataSet.GetActive: Boolean;
begin
  Result := GetDataSet.Active;
end; {function, GetActive}

{------------------------------------------------------------------------------}
{ TdaApolloDataSet.SetActive }

procedure TdaApolloDataSet.SetActive(Value: Boolean);
begin
  GetDataSet.Active := Value;

end; {procedure, SetActive}

{------------------------------------------------------------------------------}
{ TdaApolloDataSet.GetDataSet }

function TdaApolloDataSet.GetDataSet: TDataSet;
begin

  {create Apollo dataset, if needed}
  if (FDataSet = nil) then
    FDataSet := TApolloTable.Create(Self);

  Result := FDataSet;

end; {procedure, GetDataSet}

{------------------------------------------------------------------------------}
{ TdaApolloDataSet.SetDatabaseName }

procedure TdaApolloDataSet.SetDatabaseName(const aDatabaseName: String);
begin
  inherited SetDatabaseName(aDatabaseName);

  {dataset cannot be active to set database property}
  if GetDataSet.Active then
    FDataSet.Active := False;

  {get Apollo database for name}
  FDataSet.DatabaseName := aDatabaseName;

end; {procedure, SetDatabaseName}

{------------------------------------------------------------------------------}
{ TdaApolloDataSet.SetDataName }

procedure TdaApolloDataSet.SetDataName(const aDataName: String);
const
  lcDoubleQuote = #34;
var
  lsTableName: String;
begin

  inherited SetDataName(aDataName);

  {dataset cannot be active to set table name property}
  if GetDataset.Active then
    FDataSet.Active := False;

  {strip off any double quotes which may be added by the DataDictionary Builder}
  lsTableName := StringReplace(aDataName, lcDoubleQuote, '', [rfReplaceAll]);

  FDataSet.TableName := lsTableName;

  {construct an SQL statment that returns an empty result set,
   this is used to get the field information }
{  FDataSet.SQL.Text := 'SELECT * FROM ' + aDataName +
                     ' WHERE ''c'' <> ''c'' ';}

end; {procedure, SetDataName}

{------------------------------------------------------------------------------}
{ TdaApolloDataSet.BuildFieldList }

procedure TdaApolloDataSet.BuildFieldList;
var
  liIndex: Integer;
  lApolloField: TField;
  lField: TppField;
begin

  inherited BuildFieldList;

  {set dataset to active}
  if not(GetDataSet.Active) then
    try
      FDataSet.Active := True;
    except
      on E:Exception do
        Exit;
    end;

  {create TppField objects for each field in the table}
  for liIndex := 0 to FDataSet.FieldCount - 1 do
    begin
      lApolloField := FDataSet.Fields[liIndex];

      lField := TppField.Create(nil);

      lField.TableName  := DataName;
      lField.FieldName := lApolloField.FieldName;
      lField.DataType  := ppConvertFieldType(lApolloField.DataType);

      AddField(lField);
    end;

end; {function, BuildFieldList}

{------------------------------------------------------------------------------}
{ TdaApolloDataSet.GetFieldNamesForSQL }

procedure TdaApolloDataSet.GetFieldNamesForSQL(aList: TStrings; aSQL: TStrings);
var
  lQuery: TApolloQuery;
  lConnection: TApolloConnection;
begin
  aList.Clear;

  {create a temporary Apollo query}
  lQuery := TApolloQuery.Create(Self);
  lConnection := TApolloConnection.Create(Self);

  {set the database and SQL properties}
  lQuery.AccessMethod:=amLocal;
  lQuery.ApolloConnection:=lConnection;
  lQuery.DatabaseName := DatabaseName;
  lQuery.SQL := aSQL;

  {get the field names}
  lQuery.GetFieldNames(aList);

  lConnection.Free;
  lQuery.Free;

end; {procedure, GetFieldNamesForSQL}

{------------------------------------------------------------------------------}
{ TdaApolloDataSet.GetFieldsForSQL }

procedure TdaApolloDataSet.GetFieldsForSQL(aList: TList; aSQL: TStrings);
var
  lQuery: TApolloQuery;
  lConnection: TApolloConnection;
  lApolloField: TField;
  lField: TppField;
  liIndex: Integer;
  lTableItem: TApolloTableItem;
  FileFile,FileAlias,SQLText:String;
begin

  aList.Clear;

  {create a temporary Apollo query}
  lQuery := TApolloQuery.Create(Self);
  lConnection := TApolloConnection.Create(Self);

  try

    for liIndex := 0 to aSQL.Count - 1 do
    begin
      If Pos('FROM',aSQL[liIndex])>0 then begin
        SQLText:=aSQL[liIndex];
        Delete(SQLText,Pos('FROM',SQLText),5);
        While Length(SQLText)>0 do begin
          FileFile:=Copy(SQLText,1,Pos(' ',SQLText)-1);
          Delete(SQLText,1,Pos(' ',SQLText));
          if Pos(' ',SQLText)>0 then begin
            FileAlias:=Copy(SQLText,1,Pos(' ',SQLText)-1);
            Delete(SQLText,1,Pos(' ',SQLText));
            If Pos('JOIN',SQLText)>0 then
              Delete(SQLText,1,Pos('JOIN',SQLText)+4)
            else
              SQLText:='';
          end else begin
            FileAlias:=SQLText;
            SQLText:='';
          end;
          lTableItem := lQuery.TableNames.Add;
          lTableItem.Alias:=FileAlias;
          lTableItem.TableName:=FileFile+'.dbf';
        end;
      end;
    end;

    {assign database and SQL properties}
    lQuery.AccessMethod:=amLocal;
    lQuery.ApolloConnection:=lConnection;
    lQuery.DatabaseName := DatabaseName;
    lQuery.SQL := aSQL;

    {set query to active}
    lQuery.Active := True;

    {create a TppField object for each field in the query}
    for liIndex := 0 to lQuery.FieldCount - 1 do
      begin
        lApolloField := lQuery.Fields[liIndex];

        lField := TppField.Create(nil);

        lField.FieldName    := lApolloField.FieldName;
        lField.DataType     := ppConvertFieldType(lApolloField.DataType);

        aList.Add(lField);
      end;

  finally
    lQuery.Free;
    lConnection.Free;

  end;

end; {procedure, GetFieldsForSQL}



{******************************************************************************
 *
 ** Apollo  Q U E R Y   D A T A V I E W
 *
{******************************************************************************}

{------------------------------------------------------------------------------}
{ TdaApolloQueryDataView.Create }
constructor TdaApolloQueryDataView.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  {notes: 1. must use ChildQuery, ChildDataSource, ChildPipeline etc.
          2. use Self as owner for Query, DataSource etc.
          3. do NOT assign a Name }

  FQuery := TdaChildApolloQuery.Create(Self);

  FDataSource := TppChildDataSource.Create(Self);
  FDataSource.DataSet := FQuery;

end; {constructor, Create}

{------------------------------------------------------------------------------}
{ TdaApolloQueryDataView.Destroy }
destructor TdaApolloQueryDataView.Destroy;
begin
  FDataSource.Free;
  FQuery.Free;

  inherited Destroy;

end; {destructor, Destroy}

{------------------------------------------------------------------------------}
{ TdaApolloQueryDataView.PreviewFormClass }
class function TdaApolloQueryDataView.PreviewFormClass: TFormClass;
begin
  Result := TFormClass(GetClass('TdaPreviewDataDialog'));
end; {class function, PreviewFormClass}

{------------------------------------------------------------------------------}
{ TdaApolloQueryDataView.SessionClass }
class function TdaApolloQueryDataView.SessionClass: TClass;
begin
  Result := TdaApolloSession;
end; {class function, SessionClass}

{------------------------------------------------------------------------------}
{ TdaApolloQueryDataView.ConnectPipelinesToData }
Procedure TdaApolloQueryDataView.ConnectPipelinesToData;
begin
  If DataPipeLineCount=0 then Exit;

  TppDBPipeLine(DataPipelines[0]).DataSource := FDataSource;
end; {Procedure, ConnectPIpelinesToData}

{------------------------------------------------------------------------------}
{ TdaApolloQueryDataView.Init }
procedure TdaApolloQueryDataView.Init;
var
  lDataPipeline: TppChildDBPipeline;
begin
  inherited Init;

  if DataPipelineCount > 0 then Exit;

  {note: DataView's owner must own the DataPipeline }
  lDataPipeline := TppChildDBPipeline(ppComponentCreate(Self, TppChildDBPipeline));
  lDataPipeline.DataSource := FDataSource;

  lDataPipeline.AutoCreateFields := False;

  {add DataPipeline to the dataview }
  lDataPipeline.DataView := Self;
end; {procedure, Init}

{------------------------------------------------------------------------------}
{ TdaApolloQueryDataView.SQLChanged }
procedure TdaApolloQueryDataView.SQLChanged;
Var i:Integer;
    lTableItem: TApolloTableItem;
begin
  If FQuery.Active then
    FQuery.Close;

  FQuery.DatabaseName := SQL.DatabaseName;
  FQuery.SQL := SQL.MagicSQLText;

  FQuery.TableNames.Clear;
  For i:=0 to SQL.SelectTableCount-1 do begin
    lTableItem := FQuery.TableNames.Add;
    lTableItem.Alias:=sql.SelectTables[i].Alias;
    lTableItem.TableName:=sql.SelectTables[i].RawTableName+'.dbf';
  end;

end; {procedure, WizardCompleted}

{******************************************************************************
 *
 ** P R O C E D U R E S   A N D   F U N C T I O N S
 *
{******************************************************************************}

{------------------------------------------------------------------------------}
{ daGetApolloConnectionList }

function daGetApolloConnectionList: TppComponentList;
begin
  if (FApolloConnectionList = nil) then
    FApolloConnectionList := TppComponentList.Create(nil);

  Result := FApolloConnectionList;

end; {function, daGetADSConnectionList}

{------------------------------------------------------------------------------}
{ daGetApolloConnectionNames }

procedure daGetApolloConnectionNames(aList: TStrings);
begin
  {can add code here to read a list from an .ini file, etc.}

end; {function, daGetApolloConnectionNames}


{------------------------------------------------------------------------------}
{ daGetDefaultApolloConnection }

function daGetDefaultApolloConnection: TApolloDatabase;
begin

  {create the default Apollo database, if needed}
  if (FApolloDatabase = nil) then
    begin
      {create default Apollo database}
      FApolloDatabase := TApolloDatabase.Create(nil);
      FApolloDatabase.Name := cDefaultDatabase;
    end;

  Result := FApolloDatabase;

end; {function, daGetDefaultApolloDatabase}

{------------------------------------------------------------------------------}
{ daGetApolloConnectionForName }

function daGetApolloConnectionForName(aDatabaseName: String): TApolloDatabase;
var
  liIndex: Integer;

begin
  Result := nil;

  liIndex := 0;

 {check for a database object with this name}
  while (Result = nil) and (liIndex < daGetApolloConnectionList.Count) do
    begin
      if (AnsiCompareStr(FApolloConnectionList[liIndex].Name, aDatabaseName) = 0) then
        Result :=  TApolloDatabase(FApolloConnectionList[liIndex]);
      Inc(liIndex);
    end;

  if (Result <> nil) then Exit;

  {use the default database object}
  Result := daGetDefaultApolloConnection;

  {set DatabaseName property, if needed}
  if (Result.Name <> aDatabaseName) then
    begin
      Result.Name := aDatabaseName;
    end;

end; {function, daGetApolloConnectionForName}

initialization

  {register the Apollo descendant classes}
  RegisterClasses([TdaChildApolloQuery, TdaChildApolloTable]);

  {register DADE descendant session, dataset, dataview}
  daRegisterSession(TdaApolloSession);
  daRegisterDataSet(TdaApolloDataSet);
  daRegisterDataView(TdaApolloQueryDataView);

  {initialize internal reference variables}
  FApolloDatabase := nil;

finalization

  {free the default database object}
  FApolloDatabase.Free;

  {unregister the Apollo descendant classes}
  UnRegisterClasses([TdaChildApolloQuery, TdaChildApolloTable]);

  {unregister DADE descendant the session, dataset, dataview}
  daUnRegisterSession(TdaApolloSession);
  daUnRegisterDataSet(TdaApolloDataSet);
  daUnRegisterDataView(TdaApolloQueryDataView);

end.

