(*=========================================================================
 Fast Text Search VCL Component for Delphi
  Copyright 1999-2002 Vista Software
  Last Updated: 10-22-2002
==========================================================================*)
                                               
unit FTS32;

interface

uses
  SysUtils, Windows, Classes, Forms, Dialogs;

const
  { FTS File Open Modes }
  FTS_SHARE  = $0;    { SHARE }
  FTS_EXCL   = $1;    { EXCLUSIVE }
  FTS_RDONLY = $2;    { READ-ONLY }

  { FTS Error Codes }
  FTS_CREATEFAIL    = -1;
  FTS_MEMERR        = -2;
  FTS_NULLPTR       = -3;
  FTS_BADSEEK       = -4;
  FTS_BADREAD       = -5;
  FTS_BADWRITE      = -6;
  FTS_RECBOUND      = -7;
  FTS_ISDELETED     = -8;
  FTS_NOTDELETED    = -9;
  FTS_OPENERR       = -10;
  FTS_INTERR        = -11;
  FTS_NORECS        = -13;
  FTS_BADPARMS      = -16;
  FTS_NOMOREHANDLES = -17;
  FTS_BADHANDLE     = -18;
  FTS_BADIHANDLE    = -19;
  FTS_LOCKFAILED    = -20;
  FTS_NOMORELOCKS   = -21;
  FTS_CANNOTUNLOCK  = -22;
  FTS_BADCOMMIT     = -23;

type
  { Verification types for Verify method }
  TVerifyType = ( VBEG,    { verify at begining of string only }
                  VEND,    { verify at end of string only }
                  VAND,    { verify all tokens in target }
                  VONE );  { don't tokenize target }

  TFts = class(TComponent)
  private
    { Private declarations }
    FHandle : Integer;
  protected
    { Protected declarations }
  public
    { Public declarations }
    function Add( pszIdxStr : PChar ): Longint;
    function Close: Integer;
    function CreateIndex( sIdxName : String; iBuffSize, iIdxSize : SmallInt;
               bCase : Boolean; iFilter : SmallInt ): Integer;
    function Delete( lRecno : Longint ): Integer;
    function IsDelete( lRecno : Longint ): Integer;
    function NextRec : Longint;
    function NumRecs :  Longint;
    function Open( sFileName : String; iBuffSize, iMode : SmallInt ): Integer;
    function Replace( pszIdxStr : PChar; lRecno : Longint ): Integer;
    function Setup( pszIdxStr : PChar ): Longint;
    function UnDelete( lRecno : Longint ): short;
    function Verify( pszSrcStr, pszTargetStr : PChar; iType : TVerifyType ):
                     Integer;
    function Version: String;
    property Handle : Integer read FHandle;
  published
    { Published declarations }
  end;

  procedure Register;
  
implementation

var
  DLLHandle: THandle;
  ErrorMode: Integer;
  DLLLoaded: Boolean;

{ FTS32.DLL Function Prototypes }
function FtsAdd( FHandle : Integer; pszIdxStr : PChar ): Longint; stdcall;
                 external 'FTS32.DLL';
function FtsClose( FHandle : Integer ): Integer; stdcall; external 'FTS32.DLL';
function FtsCreate( pszIdxName : PChar; iBuffSize, iIdxSize : SmallInt;
                    bCase : Boolean; iFilter : SmallInt ): Integer; stdcall;
                    external 'FTS32.DLL';
function FtsDelete( FHandle : Integer; lRecno : Longint ): Integer; stdcall;
                    external 'FTS32.DLL';
function FtsIsDelete( FHandle : Integer; lRecno : Longint ): Integer; stdcall;
                      external 'FTS32.DLL';
function FtsNextRec( FHandle : Integer ): Longint; stdcall;
                     external 'FTS32.DLL';
function FtsNumRecs( FHandle : Integer ): Longint; stdcall;
                     external 'FTS32.DLL';
function FtsOpen( pszFileName : PChar; iBuffSize, iMode : SmallInt ): Integer;
                  stdcall; external 'FTS32.DLL';
function FtsReplace( FHandle : Integer; pszIdxStr : PChar; lRecno : Longint ):
                     Integer; stdcall; external 'FTS32.DLL';
function FtsSet( FHandle : Integer; pszIdxStr : PChar ): Longint; stdcall;
                 external 'FTS32.DLL';
function FtsUnDelete( FHandle : Integer; lRecno : Longint ): short; stdcall;
                      external 'FTS32.DLL';
function FtsVerify( FHandle : Integer; pszSrcStr, pszTargetStr : PChar;
                    iType : SmallInt ): Integer; stdcall; external 'FTS32.DLL';
function FtsVersion: PChar; stdcall; external 'FTS32.DLL';

{Tfts.Add}
function Tfts.Add( pszIdxStr : PChar ): Longint;
begin
  Result := FtsAdd( FHandle, pszIdxStr );
end;

{Tfts.Close}
function Tfts.Close: Integer;
begin
  Result := FtsClose( FHandle );
end;

{Tfts.CreateIndex}
function Tfts.CreateIndex( sIdxName : string; iBuffSize, iIdxSize : SmallInt;
                      bCase : Boolean; iFilter : SmallInt ): Integer;
var
  pszIdxName : Array[0..255] of Char;
begin
  StrPCopy( pszIdxName, sIdxName );
	FHandle := FtsCreate( pszIdxName, iBuffSize, iIdxSize, bCase, iFilter );
  Result := FHandle;
end;

{Tfts.Delete}
function Tfts.Delete( lRecno : Longint ): Integer;
begin
  Result := FtsDelete( FHandle, lRecno );
end;

{Tfts.IsDelete}
function Tfts.IsDelete( lRecno : Longint ): Integer;
begin
  Result := FtsIsDelete( FHandle, lRecno );
end;

{Tfts.NextRec}
function Tfts.NextRec: Longint;
begin
  Result := FtsNextRec( FHandle );
end;

{Tfts.NumRecs}
function Tfts.NumRecs: Longint;
begin
  Result := FtsNumRecs( FHandle );
end;

{Tfts.Open}
function Tfts.Open( sFileName : string; iBuffSize, iMode : SmallInt ): Integer;
var
  pszFileName : Array[0..255] of Char;
begin
  StrPCopy( pszFileName, sFileName );
  FHandle := FtsOpen( pszFileName, iBuffSize, iMode );
  Result := FHandle;
end;

{Tfts.Replace}
function Tfts.Replace( pszIdxStr : PChar; lRecno : Longint ): Integer;
begin
  Result := FtsReplace( FHandle, pszIdxStr, lRecno );
end;

{Tfts.Setup}
function Tfts.Setup( pszIdxStr : PChar ): Longint;
begin
  Result := FtsSet( FHandle, pszIdxStr );
end;

{Tfts.UnDelete}
function Tfts.UnDelete( lRecno : Longint ): short;
begin
  Result := FtsUnDelete( FHandle, lRecno );
end;

{Tfts.Verify}
function Tfts.Verify( pszSrcStr, pszTargetStr : PChar; iType : TVerifyType ):
                      Integer;
begin
  Result := FtsVerify( FHandle, pszSrcStr, pszTargetStr, Ord(iType)+1 );
end;

{Tfts.Version}
function TFts.Version: String;
begin
  Result := StrPas( FtsVersion );
end;

{ LoadDLL}
procedure LoadDLL;
begin
  if DLLLoaded then
    Exit;
  ErrorMode := SetErrorMode( $8000 );
  DLLHandle := LoadLibrary('FTS32.DLL');
  if DLLHandle >= 32 then
    DLLLoaded := True
	else
    DLLLoaded := False; { Error: FTS32.DLL could not be loaded !! }
  SetErrorMode(ErrorMode)
end {LoadDLL};


procedure Register;
begin
  RegisterComponents( 'Apollo',[TFts] );
end;

initialization

	// Attempt to load FTS32.DLL, just to test if it is available
  LoadDLL;

  // If FTS32.DLL not loaded, report error & bail
  if (not DLLLoaded) then
  begin
    MessageDlg( 'Fast Text Search (FTS): Unable to load required FTS32.DLL',
                mtError, [mbOk], 0);
    Application.Terminate;
  end;
  FreeLibrary(DLLHandle);   // Unload FTS32.DLL

finalization
  FreeLibrary(DLLHandle);

end.



