(*****************************************************************************)
(* MODULE: TwwApolloTable - TApolloTable for InfoPower 4.x                   *)
(* RIGHTS: Copyright 1999-2002 Vista Software                               *)
(* DESCR.: This module contains classes for InfoPower compatibility          *)
(*****************************************************************************)

Unit ApolloTB;

Interface

uses
  SysUtils, Windows, Messages, Classes, Graphics, Controls, Forms, dialogs,
  wwfilter, wwStr, wwSystem, wwTable, wwtypes, ApoDSet;

Type
  TwwApolloTable =
  Class( TApolloTable )    
    Private
      FControlType    : TStrings;
      FPictureMasks   : TStrings;
      FUsePictureMask : boolean;
      FOnInvalidValue : TwwInvalidValueEvent;
      Function GetControlType : TStrings;
      Procedure SetControlType( sel : TStrings );
      Function GetPictureMasks : TStrings;
      Procedure SetPictureMasks( sel : TStrings );
    Protected
      Procedure DoBeforePost; Override; { For picture support }
    Public
      Constructor Create( AOwner : TComponent ); Override;
      Destructor Destroy; Override;
    Published
      Property ControlType : TStrings
        Read  GetControlType
        Write setControltype;
      Property PictureMasks: TStrings
        Read GetPictureMasks
        Write SetPictureMasks;
      Property ValidateWithMask : boolean
        Read FUsePictureMask
        Write FUsePictureMask;
      Property OnInvalidValue: TwwInvalidValueEvent
        Read FOnInvalidValue
        Write FOnInvalidValue;
  end;

Procedure Register;

implementation

uses
  wwcommon, dbconsts;

Constructor TwwApolloTable.Create( AOwner : TComponent );
begin
  inherited Create( AOwner );
  FControlType    := TStringList.create;
  FPictureMasks   := TStringList.create;
  FUsePictureMask := True;
end;

Destructor TwwApolloTable.Destroy;
begin
  FControlType.Free;
  FPictureMasks.Free;
  FPictureMasks:= NIL;
  Inherited Destroy;
end;

Function TwwApolloTable.GetControltype : TStrings;
begin
  Result := FControlType;
end;

Procedure TwwApolloTable.SetControlType( sel : TStrings );
begin
  FControlType.Assign( sel );
end;

Function TwwApolloTable.GetPictureMasks : TStrings;
begin
  Result:= FPictureMasks
end;

Procedure TwwApolloTable.SetPictureMasks( sel : TStrings );
begin
  FPictureMasks.Assign( sel );
end;

Procedure TwwApolloTable.DoBeforePost;
begin
  Inherited DoBeforePost;
  if FUsePictureMask then
    wwValidatePictureFields( self, FOnInvalidValue );
end;

Procedure Register;
begin
  RegisterComponents('InfoPower', [TwwApolloTable] );
end;

end.