(*============================================================================
  SDE60.PAS
  SDE60.DLL interface code for TApolloDataSet
  
  Copyright 1999-2002 Vista Software
  
  Synchronized with SDE60.DLL version 6.1.0.0

  Last Updated: 10-22-2002	[mo] 
============================================================================*)

unit SDE60;

///////////////// INTERFACE PART ////////////////////////

interface
{M.O: 03-30-01, Modified}

// {$define INTL_LANG}	// Uncomment to force machine collation
// {$define SPAN_LANG}	// Uncomment to add group collation (Spanish is preset)
// {$define DUDN_LANG}	// Uncomment to add Duden collation
                                         
uses  Windows, Forms, Dialogs;

///////// ENGINE CLASS //////////
type
  TSDEEngine = class
  private
    m_bIsLoaded : boolean;
    m_iDllHandle: Integer;
  public
    constructor Create;
    destructor Destroy; override;
    function LoadDLLs: boolean stdcall;
    procedure FreeDLLs;
    procedure SetSystemCollation;
    procedure SetMachineCollation;
    procedure AddDudenCollation;
    procedure AddEtecCollation;
    function GetSystemLocale: PChar stdcall;
    function ViewSystemCharOrder():  Pchar; stdcall;
end;

var

  SDEEngine: TSDEEngine;
  DLLLoaded: Boolean = False; { is DLL (dynamically) loaded already? }

  sx_AddDudenCollation: procedure stdcall;					// [mo] 9-27-01 new collation functions
  sx_AddEtecCollation: procedure stdcall;						// [mo] 9-27-01 new collation functions
  sx_Alias: function(uiWorkArea: Word): PChar stdcall;
  sx_Append: procedure stdcall;
  sx_AppendBlank: procedure stdcall;
  sx_AppendBlankEx: function: SmallInt stdcall;
  sx_AppendEx: function: SmallInt stdcall;
  sx_AppendFrom: function(cpFileName: PChar; iSourceType: SmallInt; cpScopeExpr: PChar): LongBool stdcall;
  sx_BaseDate: function: PChar stdcall;
  sx_BaseName: function: PChar stdcall;
  sx_BlobToFile: function( cpFieldName: PChar; cpFileName: PChar): LongBool stdcall;
  sx_Bof: function: LongBool stdcall;
  sx_Close: procedure stdcall;
  sx_CloseAll: procedure stdcall;
  sx_CloseIndexes: procedure stdcall;
  sx_Commit: procedure stdcall;
  sx_CommitLevel: function (nNewLevel: Integer): Integer stdcall;
  sx_CopyFile: function(cpToFileName: PChar): LongBool stdcall;
  sx_CopyFileText: function(cpTextFileName: PChar; iFileType: SmallInt): LongBool stdcall;
  sx_CopyStructure: function(cpFileName: PChar; cpAlias: PChar): LongBool stdcall;
  sx_CopyStructureExtended: function(cpFileName: PChar): LongBool stdcall;
  sx_Count: function: Integer stdcall;
  sx_CreateExec: function: LongBool stdcall;
  sx_CreateField: procedure(cpName: PChar; cpType: PChar; iLength: SmallInt; iDecimals: SmallInt) stdcall;
  sx_CreateFrom: function(cpFileName: PChar; cpAlias: PChar; cpStruFile: PChar; iRDEType: SmallInt): LongBool stdcall;
  sx_CreateNew: function(cpFileName: PChar; cpAlias: PChar; iRdeType: SmallInt; iNumFields: SmallInt): SmallInt stdcall;
  sx_CreateNewEx: function(cpFileName: PChar; cpAlias: PChar; iRdeType: SmallInt; iNumFields: SmallInt; uiMode: Cardinal): SmallInt stdcall;
  sx_DbfDecrypt: function: LongBool stdcall;
  sx_DbfEncrypt: function: LongBool stdcall;
  sx_DBFilter: function: PChar stdcall;
  sx_DBRlockList: procedure( ulpArray: PChar ); stdcall;
  sx_Decrypt: function(cpBuffer: PChar; cpPassword: PChar; iLen: Integer): PChar stdcall;
  sx_Delete: procedure stdcall;
  sx_Deleted: function: LongBool stdcall;
  sx_Descend: function(cpKeyString: PChar): PChar stdcall;
  sx_Empty: function(cpFieldName: PChar): LongBool stdcall;
  sx_Encrypt: function(cpBuffer: PChar; cpPassword: PChar; iLen: Integer): PChar stdcall;
  sx_Eof: function: LongBool stdcall;
  sx_ErrorLevel: function(iErrorLevel: SmallInt): SmallInt stdcall;  {LAS: 06-20-00, Changed values from Integer to SmallInt}
  sx_EvalLogical: function(cpExpression: PChar): LongBool stdcall;
  sx_EvalNumeric: function(cpExpression: PChar): Double stdcall;
  sx_EvalString: function(cpExpression: PChar): PChar stdcall;
  sx_EvalTest: function(cpExpression: PChar): SmallInt stdcall;
  sx_FieldCount: function: Word stdcall;
  sx_FieldDecimals: function(cpFieldName: PChar): Word stdcall;
  sx_FieldName: function(uiFieldNum: Word): PChar stdcall;
  sx_FieldNum: function(cpFieldName: PChar): Word stdcall;
  sx_FieldOffset: function(cpFieldName: PChar): Word stdcall;
  sx_FieldType: function(cpFieldName: PChar): PChar stdcall;
  sx_FieldWidth: function(cpFieldName: PChar): Word stdcall;
  sx_FilterAlias: function(cpAliasName: PChar; cpFieldName: PChar): LongBool stdcall;
  sx_FilterDlg: function(hwnd: HWND; cpExpr: PChar; cpCaption: PChar; bHasIndexList: WordBool): PChar stdcall;
  sx_FinalizeSession: procedure stdcall;
  sx_Flock: function: LongBool stdcall;
  sx_Found: function: LongBool stdcall;
  sx_GetBlob: function(cpFieldName: PChar; vpVar: Pointer): Integer stdcall;
  sx_GetBlobLength: function(cpFieldName: PChar): Integer stdcall;
  sx_GetByte: function(cpFieldName: PChar): PChar stdcall;
  sx_GetCommitLevel: function (uiWorkArea: SmallInt): Integer stdcall;
  sx_GetDateFormat: function: Integer;							// [mo] 9-27-01 new collation functions
  sx_GetDateJulian: function(cpFieldName: PChar): Integer stdcall;
  sx_GetDateString: function(cpFieldName: PChar): PChar stdcall;
  sx_GetDouble: function(cpFieldName: PChar): Double stdcall;
  sx_GetInteger: function(cpFieldName: PChar): SmallInt stdcall;
  sx_GetLogical: function(cpFieldName: PChar): WordBool stdcall;
  sx_GetLong: function(cpFieldName: PChar): Integer stdcall;
  sx_GetMemo: function( cpFieldName: PChar; uiLineWidth: Word): PChar stdcall;
  sx_GetQueryBit: function(lRecNo: Integer): LongBool stdcall;
  sx_GetRecord: procedure(cpRecord: PChar) stdcall;
  sx_GetScope: function(iWhichScope: SmallInt): PChar stdcall;
  sx_GetString: function(cpFieldName: PChar): PChar stdcall;
  sx_GetSystemCharOrder: function: Pchar stdcall;		// [mo] 9-27-01 new collation functions
  sx_GetSystemLocale: function: Pchar stdcall;			// [mo] 9-27-01 new collation functions
  sx_GetTrimString: function(cpFieldName: PChar): PChar stdcall;
  sx_GetUDFPath: procedure(cpUDFPath: PChar; lMaxBufLen: Integer) stdcall;
  sx_GetXML: function(): PChar; stdcall;
  //sx_GetVariant: function(cpFieldName: String): pChar stdcall;
  sx_Go: procedure(lRecNum: Integer) stdcall;
  sx_GoBottom: procedure stdcall;
  sx_GoTop: procedure stdcall;
  sx_Index: function(cpFileName: PChar; cpExpr: PChar; iOption: SmallInt; bDescend: LongBool; cpCondition: PChar): SmallInt stdcall;
  sx_IndexClose: procedure stdcall;
  sx_IndexCondition: function: PChar stdcall;
  sx_IndexFlip: function: LongBool stdcall;
  sx_IndexKey: function: PChar stdcall;
  sx_IndexKeyField: function: PChar stdcall;
  sx_IndexName: function(iIndex: SmallInt): PChar stdcall;
  sx_IndexOpen: function(cpFileName: PChar): SmallInt stdcall;
  sx_IndexOrd: function: SmallInt stdcall;
  sx_IndexTag: function(cpFileName: PChar; cpTagName: PChar; cpExpr: PChar; iOption: SmallInt; bDescend: LongBool; cpCondition: PChar): SmallInt stdcall; 
  sx_IndexType: function: SmallInt stdcall;
  sx_IsEncrypted: function(iFileOrRec: SmallInt): LongBool stdcall;
  sx_IsNull: function(cpFieldName: PChar): LongBool stdcall;
  sx_KeyAdd: function(cpTagname: PChar): LongBool stdcall;
  sx_KeyData: function: PChar stdcall;
  sx_KeyDrop: function(cpTagname: PChar): LongBool stdcall;
  sx_KeyLength: function: SmallInt stdcall;		//\\ [ac]
  sx_Locate: function(cpExpression: PChar; iDirection: WordBool; bContinue: LongBool): Integer stdcall;
  sx_LockCount: function: Word stdcall;
  sx_Locked: function(lRecNum: Integer): LongBool stdcall;
  sx_MemAlloc: function( lNumberOfBytes: Integer ): PChar stdcall;
  sx_MemDealloc: procedure(vpPtr: Pointer) stdcall;
  sx_MemRealloc: function(vpPtr: Pointer; lSize: Integer): PChar stdcall;
  sx_OpenMode: function: SmallInt stdcall;
  sx_OrderPosGet: function: Double stdcall;
  sx_OrderPosSet: procedure( dPosition: Double ) stdcall;
  sx_OrderRecNo: function: Integer stdcall;
  sx_Pack: procedure stdcall;
  sx_PutBlob: function(cpFieldName: PChar; vpVar: Pointer; lSize: Integer): Integer stdcall;
  sx_PutRecord: procedure(cpRecord: PChar) stdcall;
  sx_PutString: procedure(cpFieldName:PChar; pVal: PChar); stdCall;
  sx_PutMemo: procedure(cpFieldName:PChar; pMemoText: PChar); stdCall;
  sx_PutDateString: procedure(cpFieldName :PChar; pVal  :PChar); stdcall;
  sx_PutInteger: procedure(cpFieldName:PChar; iVal: SmallInt); stdCall;
  sx_PutLong: procedure(cpFieldName:PChar; iVal: LongInt); stdCall;
  sx_PutLogical: procedure(cpFieldName:PChar; bVal:WordBool); stdCall;
  sx_PutDouble: procedure(cpFieldName:PChar; dVal:Double); stdCall;
  //sx_PutVariant: procedure(cpFieldName: String; lpVariant: Pointer);
  sx_Query: function(cpExpression: PChar): Integer stdcall;
  sx_QueryRecCount: function: Integer stdcall;
  sx_QuerySetExact: function(bSetQueryExact: LongBool): LongBool stdcall;  {LAS: 06-23-00, Added}
  sx_QueryTest: function(cpExpression: PChar): SmallInt stdcall;
  sx_Recall: procedure stdcall;
  sx_RecCount: function: Integer stdcall;
  sx_RecNo: function: Integer stdcall;
  sx_RecSize: function: Integer stdcall;
  //sx_RecToString: function (cpRecStruc: PChar; iLength: SmallInt): Integer stdcall;		//\\ ac
  sx_Reindex: procedure stdcall;
  sx_Replace: procedure(cpFieldname: PChar; iDataType: SmallInt; vpData: Pointer) stdcall;
  sx_Rlock: function(lRecNum: Integer): LongBool stdcall;
  sx_RYOFilterActivate: function(iFilterHandle: SmallInt; iBoolOperation: SmallInt): LongBool stdcall;
  sx_RYOFilterCopy: function: SmallInt stdcall;
  sx_RYOFilterCreate: function: SmallInt stdcall;
  sx_RYOFilterDestroy: function(iFilterHandle: SmallInt): LongBool stdcall;
  sx_RYOFilterGetBit: function(iFilterHandle: SmallInt; lRecNo: Integer): LongBool stdcall;
  sx_RYOFilterRestore: function(cpFileName: PChar): LongBool stdcall;
  sx_RYOFilterSave: function(iFilterHandle: SmallInt; cpFileName: PChar): LongBool stdcall;
  sx_RYOFilterSetBit: function(iFilterHandle: SmallInt; lRecNo: Integer; iOnOrOff: SmallInt): LongBool stdcall;
  sx_RYOKeyAdd: function(cpTagname: PChar; cpKey: PChar): LongBool stdcall;
  sx_RYOKeyDrop: function(cpTagname: PChar): LongBool stdcall;
  sx_Seek: function(cpKeyValue: PChar): LongBool stdcall;
  sx_SeekBin: function(cpKeyValue: PChar; uiLength: Word): LongBool stdcall;
  sx_Select: function(uiBaseArea: Word): Word stdcall;
  sx_SetCentury: procedure(uiOnOff: WordBool) stdcall;
  sx_SetCollationRule: function(lRuleType: Integer; cpSrcSymbSet: PChar; cpDstSymbSet: PChar; bResetPrev: LongBool; bOem: LongBool; lReserved: Integer): LongBool stdcall;
  sx_SetDateFormat: procedure(uiDateType: Word) stdcall;
  sx_SetDeleted: procedure(uiDeleted: WordBool) stdcall;
  sx_SetEpoch: function(uiBaseYear: Word): Word stdcall;
  sx_SetErrorFunc: function( pFunc, pData : Integer): Integer; stdcall;
  sx_SetExact: procedure(uiOnOff: WordBool) stdcall;
  sx_SetFilter: procedure(cpExpression: PChar) stdcall;
  sx_SetGaugeFunc: function( pGaugeFunc: Integer ): Integer; stdcall;
  sx_SetGaugeHook: procedure(hwndGauge: HWND) stdcall;
  sx_SetHandles: function(iNumHandles: SmallInt): SmallInt; stdcall;
  sx_SetLockTimeout: procedure(iSeconds: SmallInt) stdcall;
  sx_SetMachineCollation: procedure stdcall;				// [mo] 9-27-01 new collation functions
  sx_SetMemoBlockSize: function(uiBlockSize: Word): Word; stdcall;
  sx_SetOrder: function(iIndex: SmallInt): SmallInt stdcall;
  sx_SetPassword: procedure(cpEncodeKey: PChar) stdcall;
  sx_SetQueryBit: procedure(lRecNo: Integer; bValue: LongBool) stdcall;
  sx_SetRelation: procedure(uiChildArea: Word; cpKeyExpr: PChar) stdcall;
  sx_SetScope: function(cpLowVal: PChar; cpHighVal: PChar): LongBool stdcall;
  sx_SetSoftSeek: procedure(uiOnOff: WordBool) stdcall;
  sx_SetStringType: procedure(uiStringType: Word) stdcall;
  sx_SetSystemCollation: procedure stdcall;					// [mo] 9-27-01 new collation functions
  sx_SetTranslate: procedure(uiOnOff: WordBool) stdcall;
  sx_SetTurboRead: procedure(uiOnOff: WordBool) stdcall;
  sx_SetUDFPath: procedure(cpUDFPath: PChar) stdcall;
  sx_SetUserDelimiter: function(cDelimiter: Char; bMakeTrim: LongBool; bReflectMemo: LongBool): Char stdcall;
  sx_Skip: procedure(lNumRecs: Integer) stdcall;
  sx_SysProp: function(uiSysItem: Word; vpData: Pointer): Integer stdcall;
  sx_TagArea: function(cpTagName: PChar): SmallInt stdcall;
  sx_TagDelete: function(cpTagname: PChar): LongBool stdcall;
  sx_TagName: function(iTagArea: SmallInt): PChar stdcall;
  sx_Unlock: procedure(lRecNum: Integer) stdcall;
  sx_Use: function(cpFileName: PChar; cpAlias: PChar; iOpenMode: SmallInt; iRdeType: SmallInt): SmallInt stdcall;
  sx_UseEx: function(cpFileName: PChar; cpAlias: PChar; iOpenMode: SmallInt; iRdeType: SmallInt; uiMode: Cardinal): SmallInt stdcall;
  sx_Version: function: PChar stdcall;
  sx_WorkArea: function(cpAlias: PChar ): Word stdcall;
  sx_Zap: procedure stdcall;

  procedure LoadDLL;
  procedure UnLoadDLL;  

///////////////// IMPLEMENTAION PART ////////////////////////

implementation

Uses ApCommon;        //MED: 08-30-01, added

var
  DLLHandle: THandle;
  ErrorMode: Integer;

// MED: 08-17-01 added for VB DC
procedure UnLoadDLL;
begin
  if (DLLLoaded) then
    SDEEngine.FreeDlls;
end;


procedure LoadDLL;
begin
  if (not DLLLoaded) then
  begin
    // assign to global values for backaward compatibility
    DLLLoaded := SDEEngine.LoadDlls;
    if DLLLoaded then
        DLLHandle := SDEEngine.m_iDLLHandle;
  end;
end;

constructor TSDEEngine.Create;
begin
  m_bIsLoaded := false;
  m_iDLLHandle := 0;
end;

destructor TSDEEngine.Destroy;
begin
  if m_bIsLoaded then FreeDLLs;
end;

function TSDEEngine.LoadDLLs: boolean;
begin
  if m_bIsLoaded then
  begin
    Result := true;
    Exit;
  end;

  // assign as global for backward compatibility
  try
    ErrorMode := SetErrorMode( $8000 );
    m_iDllHandle := LoadLibrary('SDE60.DLL'); // assign to local values

    if m_iDLLHandle >= 32 then
    begin
      @sx_AddDudenCollation := GetProcAddress(m_iDLLHandle,'sx_AddDudenCollation');		
      Assert(@sx_AddDudenCollation <> nil);
      @sx_AddEtecCollation := GetProcAddress(m_iDLLHandle,'sx_AddEtecCollation');	
      Assert(@sx_AddEtecCollation <> nil);
      @sx_Alias := GetProcAddress(m_iDLLHandle,'sx_Alias');
      Assert(@sx_Alias <> nil);
      @sx_Append := GetProcAddress(m_iDLLHandle,'sx_Append');
      Assert(@sx_Append <> nil);
      @sx_AppendEx := GetProcAddress(m_iDLLHandle,'sx_AppendEx');
      Assert(@sx_AppendEx <> nil);
      @sx_AppendBlank := GetProcAddress(m_iDLLHandle,'sx_AppendBlank');
      Assert(@sx_AppendBlank <> nil);
      @sx_AppendBlankEx := GetProcAddress(m_iDLLHandle,'sx_AppendBlankEx');
      Assert(@sx_AppendBlankEx <> nil);
      @sx_AppendFrom := GetProcAddress(m_iDLLHandle,'sx_AppendFrom');
      Assert(@sx_AppendFrom <> nil);
      @sx_BaseDate := GetProcAddress(m_iDLLHandle,'sx_BaseDate');
      Assert(@sx_BaseDate <> nil);
      @sx_BaseName := GetProcAddress(m_iDLLHandle,'sx_BaseName');
      Assert(@sx_BaseName <> nil);
      @sx_BlobToFile := GetProcAddress(m_iDLLHandle,'sx_BlobToFile');
      Assert(@sx_BlobToFile <> nil);
      @sx_Bof := GetProcAddress(m_iDLLHandle,'sx_Bof');
      Assert(@sx_Bof <> nil);
      @sx_Close := GetProcAddress(m_iDLLHandle,'sx_Close');
      Assert(@sx_Close <> nil);
      @sx_CloseAll := GetProcAddress(m_iDLLHandle,'sx_CloseAll');
      Assert(@sx_CloseAll <> nil);
      @sx_CloseIndexes := GetProcAddress(m_iDLLHandle,'sx_CloseIndexes');
      Assert(@sx_CloseIndexes <> nil);
      @sx_Commit := GetProcAddress(m_iDLLHandle,'sx_Commit');
      Assert(@sx_Commit <> nil);
      @sx_CommitLevel := GetProcAddress(m_iDLLHandle,'sx_CommitLevel');
      Assert(@sx_CommitLevel <> nil);
      @sx_CopyFile := GetProcAddress(m_iDLLHandle,'sx_CopyFile');
      Assert(@sx_CopyFile <> nil);
      @sx_CopyFileText := GetProcAddress(m_iDLLHandle,'sx_CopyFileText');
      Assert(@sx_CopyFileText <> nil);
      @sx_CopyStructure := GetProcAddress(m_iDLLHandle,'sx_CopyStructure');
      Assert(@sx_CopyStructure <> nil);
      @sx_CopyStructureExtended := GetProcAddress(m_iDLLHandle,'sx_CopyStructureExtended');
      Assert(@sx_CopyStructureExtended <> nil);
      @sx_Count := GetProcAddress(m_iDLLHandle,'sx_Count');
      Assert(@sx_Count <> nil);
      @sx_CreateExec := GetProcAddress(m_iDLLHandle,'sx_CreateExec');
      Assert(@sx_CreateExec <> nil);
      @sx_CreateField := GetProcAddress(m_iDLLHandle,'sx_CreateField');
      Assert(@sx_CreateField <> nil);
      @sx_CreateFrom := GetProcAddress(m_iDLLHandle,'sx_CreateFrom');
      Assert(@sx_CreateFrom <> nil);
      @sx_CreateNew := GetProcAddress(m_iDLLHandle,'sx_CreateNew');
      Assert(@sx_CreateNew <> nil);
      @sx_CreateNewEx := GetProcAddress(m_iDLLHandle,'sx_CreateNewEx');
      Assert(@sx_CreateNewEx <> nil);
      @sx_DbfDecrypt := GetProcAddress(m_iDLLHandle,'sx_DbfDecrypt');
      Assert(@sx_DbfDecrypt <> nil);
      @sx_DbfEncrypt := GetProcAddress(m_iDLLHandle,'sx_DbfEncrypt');
      Assert(@sx_DbfEncrypt <> nil);
      @sx_DBFilter := GetProcAddress(m_iDLLHandle,'sx_DBFilter');
      Assert(@sx_DBFilter <> nil);
      @sx_DBRlockList := GetProcAddress(m_iDLLHandle,'sx_DBRlockList');
      Assert(@sx_DBRlockList <> nil);
      @sx_Decrypt := GetProcAddress(m_iDLLHandle,'sx_Decrypt');
      Assert(@sx_Decrypt <> nil);
      @sx_Delete := GetProcAddress(m_iDLLHandle,'sx_Delete');
      Assert(@sx_Delete <> nil);
      @sx_Deleted := GetProcAddress(m_iDLLHandle,'sx_Deleted');
      Assert(@sx_Deleted <> nil);
      @sx_Descend := GetProcAddress(m_iDLLHandle,'sx_Descend');
      Assert(@sx_Descend <> nil);
      @sx_Empty := GetProcAddress(m_iDLLHandle,'sx_Empty');
      Assert(@sx_Empty <> nil);
      @sx_Eof := GetProcAddress(m_iDLLHandle,'sx_Eof');
      Assert(@sx_Eof <> nil);
      @sx_Encrypt := GetProcAddress(m_iDLLHandle,'sx_Encrypt');
      Assert(@sx_Encrypt <> nil);
      @sx_ErrorLevel := GetProcAddress(m_iDLLHandle,'sx_ErrorLevel');
      Assert(@sx_ErrorLevel <> nil);
      @sx_EvalLogical := GetProcAddress(m_iDLLHandle,'sx_EvalLogical');
      Assert(@sx_EvalLogical <> nil);
      @sx_EvalNumeric := GetProcAddress(m_iDLLHandle,'sx_EvalNumeric');
      Assert(@sx_EvalNumeric <> nil);
      @sx_EvalString := GetProcAddress(m_iDLLHandle,'sx_EvalString');
      Assert(@sx_EvalString <> nil);
      @sx_EvalTest := GetProcAddress(m_iDLLHandle,'sx_EvalTest');
      Assert(@sx_EvalTest <> nil);
      @sx_FieldCount := GetProcAddress(m_iDLLHandle,'sx_FieldCount');
      Assert(@sx_FieldCount <> nil);
      @sx_FieldDecimals := GetProcAddress(m_iDLLHandle,'sx_FieldDecimals');
      Assert(@sx_FieldDecimals <> nil);
      @sx_FieldName := GetProcAddress(m_iDLLHandle,'sx_FieldName');
      Assert(@sx_FieldName <> nil);
      @sx_FieldNum := GetProcAddress(m_iDLLHandle,'sx_FieldNum');
      Assert(@sx_FieldNum <> nil);
      @sx_FieldOffset := GetProcAddress(m_iDLLHandle,'sx_FieldOffset');
      Assert(@sx_FieldOffset <> nil);
      @sx_FieldType := GetProcAddress(m_iDLLHandle,'sx_FieldType');
      Assert(@sx_FieldType <> nil);
      @sx_FieldWidth := GetProcAddress(m_iDLLHandle,'sx_FieldWidth');
      Assert(@sx_FieldWidth <> nil);
      @sx_FilterAlias := GetProcAddress(m_iDLLHandle,'sx_FilterAlias');
      Assert(@sx_FilterAlias <> nil);
      @sx_FilterDlg := GetProcAddress(m_iDLLHandle,'sx_FilterDlg');
      Assert(@sx_FilterDlg <> nil);
      @sx_FinalizeSession := GetProcAddress(m_iDLLHandle,'sx_FinalizeSession');
      Assert(@sx_FinalizeSession <> nil);
      @sx_Flock := GetProcAddress(m_iDLLHandle,'sx_Flock');
      Assert(@sx_Flock <> nil);
      @sx_Found := GetProcAddress(m_iDLLHandle,'sx_Found');
      Assert(@sx_Found <> nil);
      @sx_GetBlob := GetProcAddress(m_iDLLHandle,'sx_GetBlob');
      Assert(@sx_GetBlob <> nil);
      @sx_GetBlobLength := GetProcAddress(m_iDLLHandle,'sx_GetBlobLength');
      Assert(@sx_GetBlobLength <> nil);
      @sx_GetByte := GetProcAddress(m_iDLLHandle,'sx_GetByte');
      Assert(@sx_GetByte <> nil);
      sx_GetCommitLevel := GetProcAddress(m_iDLLHandle,'sx_GetCommitLevel');
      Assert(@sx_GetCommitLevel <> nil);
      @sx_GetDateJulian := GetProcAddress(m_iDLLHandle,'sx_GetDateJulian');
      Assert(@sx_GetDateJulian <> nil);
      @sx_GetDateString := GetProcAddress(m_iDLLHandle,'sx_GetDateString');
      Assert(@sx_GetDateString <> nil);
      @sx_GetDouble := GetProcAddress(m_iDLLHandle,'sx_GetDouble');
      Assert(@sx_GetDouble <> nil);
      @sx_GetInteger := GetProcAddress(m_iDLLHandle,'sx_GetInteger');
      Assert(@sx_GetInteger <> nil);
      @sx_GetLogical := GetProcAddress(m_iDLLHandle,'sx_GetLogical');
      Assert(@sx_GetLogical <> nil);
      @sx_GetLong := GetProcAddress(m_iDLLHandle,'sx_GetLong');
      Assert(@sx_GetLong <> nil);
      @sx_GetMemo := GetProcAddress(m_iDLLHandle,'sx_GetMemo');
      Assert(@sx_GetMemo <> nil);
      @sx_GetQueryBit := GetProcAddress(m_iDLLHandle,'sx_GetQueryBit');
      Assert(@sx_GetQueryBit <> nil);
      @sx_GetRecord := GetProcAddress(m_iDLLHandle,'sx_GetRecord');
      Assert(@sx_GetRecord <> nil);
      @sx_GetScope := GetProcAddress(m_iDLLHandle,'sx_GetScope');
      Assert(@sx_GetScope <> nil);
      @sx_GetString := GetProcAddress(m_iDLLHandle,'sx_GetString');
      Assert(@sx_GetString <> nil);
      @sx_GetTrimString := GetProcAddress(m_iDLLHandle,'sx_GetTrimString');
      Assert(@sx_GetTrimString <> nil);
      @sx_GetXML := GetProcAddress(m_iDLLHandle,'sx_GetXML');
      Assert(@sx_GetXML <> nil);
      @sx_Go := GetProcAddress(m_iDLLHandle,'sx_Go');
      Assert(@sx_Go <> nil);
      @sx_GoBottom := GetProcAddress(m_iDLLHandle,'sx_GoBottom');
      Assert(@sx_GoBottom <> nil);
      @sx_GoTop := GetProcAddress(m_iDLLHandle,'sx_GoTop');
      Assert(@sx_GoTop <> nil);
      @sx_Index := GetProcAddress(m_iDLLHandle,'sx_Index');
      Assert(@sx_Index <> nil);
      @sx_IndexClose := GetProcAddress(m_iDLLHandle,'sx_IndexClose');
      Assert(@sx_IndexClose <> nil);
      @sx_IndexCondition := GetProcAddress(m_iDLLHandle,'sx_IndexCondition');
      Assert(@sx_IndexCondition <> nil);
      @sx_IndexFlip := GetProcAddress(m_iDLLHandle,'sx_IndexFlip');
      Assert(@sx_IndexFlip <> nil);
      @sx_IndexKey := GetProcAddress(m_iDLLHandle,'sx_IndexKey');
      Assert(@sx_IndexKey <> nil);
      @sx_IndexKeyField := GetProcAddress(m_iDLLHandle,'sx_IndexKeyField');
      Assert(@sx_IndexKeyField <> nil);
      @sx_IndexName := GetProcAddress(m_iDLLHandle,'sx_IndexName');
      Assert(@sx_IndexName <> nil);
      @sx_IndexOpen := GetProcAddress(m_iDLLHandle,'sx_IndexOpen');
      Assert(@sx_IndexOpen <> nil);
      @sx_IndexOrd := GetProcAddress(m_iDLLHandle,'sx_IndexOrd');
      Assert(@sx_IndexOrd <> nil);
      @sx_IndexTag := GetProcAddress(m_iDLLHandle,'sx_IndexTag');
      Assert(@sx_IndexTag <> nil);
      @sx_IndexType := GetProcAddress(m_iDLLHandle,'sx_IndexType');
      Assert(@sx_IndexType <> nil);
      @sx_IsEncrypted := GetProcAddress(m_iDLLHandle,'sx_IsEncrypted');
      @sx_IsNull := GetProcAddress(m_iDLLHandle,'sx_IsNull');
      Assert(@sx_IsNull <> nil);
      Assert(@sx_IsEncrypted <> nil);
      @sx_KeyAdd := GetProcAddress(m_iDLLHandle,'sx_KeyAdd');
      Assert(@sx_KeyAdd <> nil);
      @sx_KeyData := GetProcAddress(m_iDLLHandle,'sx_KeyData');
      Assert(@sx_KeyData <> nil);
      @sx_KeyDrop := GetProcAddress(m_iDLLHandle,'sx_KeyDrop');
      Assert(@sx_KeyDrop <> nil);
      @sx_KeyLength := GetProcAddress(m_iDLLHandle,'sx_KeyLength');
      Assert(@sx_KeyLength <> nil);
      @sx_Locate := GetProcAddress(m_iDLLHandle,'sx_Locate');
      Assert(@sx_Locate <> nil);
      @sx_LockCount := GetProcAddress(m_iDLLHandle,'sx_LockCount');
      Assert(@sx_LockCount <> nil);
      @sx_Locked := GetProcAddress(m_iDLLHandle,'sx_Locked');
      Assert(@sx_Locked <> nil);
      @sx_MemAlloc := GetProcAddress(m_iDLLHandle,'sx_MemAlloc');
      Assert(@sx_MemAlloc <> nil);
      @sx_MemDealloc := GetProcAddress(m_iDLLHandle,'sx_MemDealloc');
      Assert(@sx_MemDealloc <> nil);
      @sx_MemRealloc := GetProcAddress(m_iDLLHandle,'sx_MemRealloc');
      Assert(@sx_MemRealloc <> nil);
      @sx_OpenMode := GetProcAddress(m_iDLLHandle,'sx_OpenMode');
      Assert(@sx_OpenMode <> nil);
      @sx_OrderPosGet := GetProcAddress(m_iDLLHandle,'sx_OrderPosGet');
      Assert(@sx_OrderPosGet <> nil);
      @sx_OrderPosSet := GetProcAddress(m_iDLLHandle,'sx_OrderPosSet');
      Assert(@sx_OrderPosSet <> nil);
      @sx_OrderRecNo := GetProcAddress(m_iDLLHandle,'sx_OrderRecNo');
      Assert(@sx_OrderRecNo <> nil);
      @sx_Pack := GetProcAddress(m_iDLLHandle,'sx_Pack');
      Assert(@sx_Pack <> nil);
      @sx_PutBlob := GetProcAddress(m_iDLLHandle,'sx_PutBlob');
      Assert(@sx_PutBlob <> nil);
      @sx_PutRecord := GetProcAddress(m_iDLLHandle,'sx_PutRecord');
      Assert(@sx_PutRecord <> nil);
      @sx_PutString := GetProcAddress(m_iDLLHandle,'sx_PutString');
      Assert(@sx_PutString <> nil);
      @sx_PutMemo := GetProcAddress(m_iDLLHandle,'sx_PutMemo');
      Assert(@sx_PutMemo <> nil);
      @sx_PutDateString := GetProcAddress(m_iDLLHandle,'sx_PutDateString');
      Assert(@sx_PutDateString <> nil);
      @sx_PutInteger := GetProcAddress(m_iDLLHandle,'sx_PutInteger');
      Assert(@sx_PutInteger <> nil);
      @sx_PutLong := GetProcAddress(m_iDLLHandle,'sx_PutLong');
      Assert(@sx_PutLong <> nil);
      @sx_PutLogical := GetProcAddress(m_iDLLHandle,'sx_PutLogical');
      Assert(@sx_PutLogical <> nil);
      @sx_PutDouble := GetProcAddress(m_iDLLHandle,'sx_PutDouble');
      Assert(@sx_PutDouble <> nil);
      @sx_Query := GetProcAddress(m_iDLLHandle,'sx_Query');
      Assert(@sx_Query <> nil);
      @sx_QueryRecCount := GetProcAddress(m_iDLLHandle,'sx_QueryRecCount');
      Assert(@sx_QueryRecCount <> nil);
      @sx_QuerySetExact := GetProcAddress(m_iDLLHandle, 'sx_QuerySetExact');   {LAS: 06-23-00, Added}
      Assert(@sx_QuerySetExact <> nil);                                     {LAS: 06-23-00, Added}
      @sx_QueryTest := GetProcAddress(m_iDLLHandle,'sx_QueryTest');
      Assert(@sx_QueryTest <> nil);
      @sx_Recall := GetProcAddress(m_iDLLHandle,'sx_Recall');
      Assert(@sx_Recall <> nil);
      @sx_RecCount := GetProcAddress(m_iDLLHandle,'sx_RecCount');
      Assert(@sx_RecCount <> nil);
      @sx_RecNo := GetProcAddress(m_iDLLHandle,'sx_RecNo');
      Assert(@sx_RecNo <> nil);
      @sx_RecSize := GetProcAddress(m_iDLLHandle,'sx_RecSize');
      Assert(@sx_RecSize <> nil);
      @sx_Reindex := GetProcAddress(m_iDLLHandle,'sx_Reindex');
      Assert(@sx_Reindex <> nil);
      @sx_Replace := GetProcAddress(m_iDLLHandle,'sx_Replace');
      Assert(@sx_Replace <> nil);
      @sx_Rlock := GetProcAddress(m_iDLLHandle,'sx_Rlock');
      Assert(@sx_Rlock <> nil);
      @sx_RYOFilterActivate := GetProcAddress(m_iDLLHandle,'sx_RYOFilterActivate');
      Assert(@sx_RYOFilterActivate <> nil);
      @sx_RYOFilterCopy := GetProcAddress(m_iDLLHandle,'sx_RYOFilterCopy');
      Assert(@sx_RYOFilterCopy <> nil);
      @sx_RYOFilterCreate := GetProcAddress(m_iDLLHandle,'sx_RYOFilterCreate');
      Assert(@sx_RYOFilterCreate <> nil);
      @sx_RYOFilterDestroy := GetProcAddress(m_iDLLHandle,'sx_RYOFilterDestroy');
      Assert(@sx_RYOFilterDestroy <> nil);
      @sx_RYOFilterGetBit := GetProcAddress(m_iDLLHandle,'sx_RYOFilterGetBit');
      Assert(@sx_RYOFilterGetBit <> nil);
      @sx_RYOFilterRestore := GetProcAddress(m_iDLLHandle,'sx_RYOFilterRestore');
      Assert(@sx_RYOFilterRestore <> nil);
      @sx_RYOFilterSave := GetProcAddress(m_iDLLHandle,'sx_RYOFilterSave');
      Assert(@sx_RYOFilterSave <> nil);
      @sx_RYOFilterSetBit := GetProcAddress(m_iDLLHandle,'sx_RYOFilterSetBit');
      Assert(@sx_RYOFilterSetBit <> nil);
      @sx_RYOKeyAdd := GetProcAddress(m_iDLLHandle,'sx_RYOKeyAdd');
      Assert(@sx_RYOKeyAdd <> nil);
      @sx_RYOKeyDrop := GetProcAddress(m_iDLLHandle,'sx_RYOKeyDrop');
      Assert(@sx_RYOKeyDrop <> nil);
      @sx_Seek := GetProcAddress(m_iDLLHandle,'sx_Seek');
      Assert(@sx_Seek <> nil);
      @sx_SeekBin := GetProcAddress(m_iDLLHandle,'sx_SeekBin');
      Assert(@sx_SeekBin <> nil);
      @sx_Select := GetProcAddress(m_iDLLHandle,'sx_Select');
      Assert(@sx_Select <> nil);
      @sx_SetCentury := GetProcAddress(m_iDLLHandle,'sx_SetCentury');
      Assert(@sx_SetCentury <> nil);
      @sx_SetCollationRule := GetProcAddress(m_iDLLHandle,'sx_SetCollationRule');
      Assert(@sx_SetCollationRule <> nil);
      @sx_SetDateFormat := GetProcAddress(m_iDLLHandle,'sx_SetDateFormat');
      Assert(@sx_SetDateFormat <> nil);
      @sx_SetDeleted := GetProcAddress(m_iDLLHandle,'sx_SetDeleted');
      Assert(@sx_SetDeleted <> nil);
      @sx_SetEpoch := GetProcAddress(m_iDLLHandle,'sx_SetEpoch');
      Assert(@sx_SetEpoch <> nil);
      @sx_SetErrorFunc := GetProcAddress(m_iDLLHandle,'sx_SetErrorFunc');
      Assert(@sx_SetErrorFunc <> nil);
      @sx_SetExact := GetProcAddress(m_iDLLHandle,'sx_SetExact');
      Assert(@sx_SetExact <> nil);
      @sx_SetFilter := GetProcAddress(m_iDLLHandle,'sx_SetFilter');
      Assert(@sx_SetFilter <> nil);
      @sx_SetGaugeFunc := GetProcAddress(m_iDLLHandle,'sx_SetGaugeFunc');
      Assert(@sx_SetGaugeFunc <> nil);
      @sx_SetGaugeHook := GetProcAddress(m_iDLLHandle,'sx_SetGaugeHook');
      Assert(@sx_SetGaugeHook <> nil);
      @sx_SetHandles := GetProcAddress(m_iDLLHandle,'sx_SetHandles');
      Assert(@sx_SetHandles <> nil);
      @sx_SetLockTimeout := GetProcAddress(m_iDLLHandle,'sx_SetLockTimeout');
      Assert(@sx_SetLockTimeout <> nil);
      @sx_SetMachineCollation := GetProcAddress(m_iDLLHandle,'sx_SetMachineCollation');
      Assert(@sx_SetMachineCollation <> nil);
      @sx_SetMemoBlockSize := GetProcAddress(m_iDLLHandle,'sx_SetMemoBlockSize');
      Assert(@sx_SetMemoBlockSize <> nil);
      @sx_SetOrder := GetProcAddress(m_iDLLHandle,'sx_SetOrder');
      Assert(@sx_SetOrder <> nil);
      @sx_SetPassword := GetProcAddress(m_iDLLHandle,'sx_SetPassword');
      Assert(@sx_SetPassword <> nil);
      @sx_SetQueryBit := GetProcAddress(m_iDLLHandle,'sx_SetQueryBit');
      Assert(@sx_SetQueryBit <> nil);
      @sx_SetRelation := GetProcAddress(m_iDLLHandle,'sx_SetRelation');
      Assert(@sx_SetRelation <> nil);
      @sx_SetScope := GetProcAddress(m_iDLLHandle,'sx_SetScope');
      Assert(@sx_SetScope <> nil);
      @sx_SetSoftSeek := GetProcAddress(m_iDLLHandle,'sx_SetSoftSeek');
      Assert(@sx_SetSoftSeek <> nil);
      @sx_SetStringType := GetProcAddress(m_iDLLHandle,'sx_SetStringType');
      Assert(@sx_SetStringType <> nil);
      @sx_SetTranslate := GetProcAddress(m_iDLLHandle,'sx_SetTranslate');
      Assert(@sx_SetTranslate <> nil);
      @sx_SetTurboRead := GetProcAddress(m_iDLLHandle,'sx_SetTurboRead');
      Assert(@sx_SetTurboRead <> nil);
      @sx_Skip := GetProcAddress(m_iDLLHandle,'sx_Skip');
      Assert(@sx_Skip <> nil);
      @sx_SetUDFPath  := GetProcAddress(m_iDLLHandle,'sx_SetUDFPath');
      Assert(@sx_SetUDFPath <> nil);
      @sx_SetUserDelimiter := GetProcAddress(m_iDLLHandle,'sx_SetUserDelimiter');
      Assert(@sx_SetUserDelimiter <> nil);
      @sx_GetUDFPath  := GetProcAddress(m_iDLLHandle,'sx_GetUDFPath');
      Assert(@sx_GetUDFPath <> nil);
      @sx_SysProp := GetProcAddress(m_iDLLHandle,'sx_SysProp');
      Assert(@sx_SysProp <> nil);
      @sx_TagArea := GetProcAddress(m_iDLLHandle,'sx_TagArea');
      Assert(@sx_TagArea <> nil);
      @sx_TagName := GetProcAddress(m_iDLLHandle,'sx_TagName');
      Assert(@sx_TagName <> nil);
      @sx_TagDelete := GetProcAddress(m_iDLLHandle,'sx_TagDelete');
      Assert(@sx_TagDelete <> nil);
      @sx_Unlock := GetProcAddress(m_iDLLHandle,'sx_Unlock');
      Assert(@sx_Unlock <> nil);
      @sx_Use := GetProcAddress(m_iDLLHandle,'sx_Use');
      Assert(@sx_Use <> nil);
      @sx_UseEx := GetProcAddress(m_iDLLHandle,'sx_UseEx');
      Assert(@sx_UseEx <> nil);
      @sx_Version := GetProcAddress(m_iDLLHandle,'sx_Version');
      Assert(@sx_Version <> nil);
      @sx_WorkArea := GetProcAddress(m_iDLLHandle,'sx_WorkArea');
      Assert(@sx_WorkArea <> nil);
      @sx_Zap := GetProcAddress(m_iDLLHandle,'sx_Zap');
      Assert(@sx_Zap <> nil);
      @sx_GetSystemLocale := GetProcAddress(m_iDLLHandle,'sx_GetSystemLocale');
      Assert(@sx_GetSystemLocale <> nil);
      @sx_GetSystemCharOrder := GetProcAddress(m_iDLLHandle,'sx_GetSystemCharOrder');
      Assert(@sx_GetSystemCharOrder <> nil);
      @sx_SetSystemCollation := GetProcAddress(m_iDLLHandle,'sx_SetSystemCollation');
      Assert(@sx_SetSystemCollation <> nil);
      @sx_GetDateFormat := GetProcAddress(m_iDLLHandle,'sx_GetDateFormat');
      Assert(@sx_GetDateFormat <> nil);

      m_bIsLoaded := True;

      // Set initial engine string defaults for Delphi/C++Builder
      sx_SetStringType( 1 );

{$ifdef INTL_LANG}     {MO: 03-22-01, Added}
      SetMachineCollation; {MO: 03-22-01, Added}
{$endif}               {MO: 03-22-01, Added}

{$ifdef SPAN_LANG}     {MO: 03-22-01, Added}
      AddEtecCollation;    {MO: 03-22-01, Added}
{$endif}               {MO: 03-22-01, Added}

{$ifdef DUDN_LANG}     {MO: 03-22-01, Added}
      AddDudenCollation;    {MO: 03-22-01, Added}
{$endif}               {MO: 03-22-01, Added}

    end
      else m_bIsLoaded := False; { Error: SDE60.DLL could not be loaded !! }

  except
    m_iDLLHandle := 0;
    m_bIsLoaded := False;
  end;

  SetErrorMode(ErrorMode); // restore error mode

  // If SDE0.DLL not loaded, report error & bail
  // ASP checks SQLDLLoaded and Response.Write
  if (not m_bIsLoaded) AND (not IsApolloASPRunning) then
  begin
    MessageDlg('Unable to load required Apollo DLL: SDE60.DLL', mtError, [mbOk], 0);
    Application.Terminate;
  end;

  Result := m_bIsLoaded;
end;

procedure TSDEEngine.FreeDLLs;
begin
  try
    if m_bIsLoaded then FreeLibrary(m_iDLLHandle);
  except
  end;
  m_iDLLHandle := 0;
  m_bIsLoaded := False;
end;

function TSDEEngine.GetSystemLocale: PChar;
begin
  Result:= sx_GetSystemLocale();
end;

procedure TSDEEngine.SetSystemCollation;
begin
  sx_SetSystemCollation(); // set sytem's alphabet for main collation (according to system language)
end;

procedure TSDEEngine.SetMachineCollation;
begin
  sx_SetMachineCollation();// set collation to direct compare char binary values
end;

procedure TSDEEngine.AddDudenCollation;
begin
  sx_AddDudenCollation();  // force to reset additional ETEC rule;
                           // respects current alphabetical or machine collation
end;

procedure TSDEEngine.AddEtecCollation;
begin
  sx_AddEtecCollation();   // force to reset additional DUDEN rule;
                           // respects current alphabetical or machine collation
end;

function TSDEEngine.ViewSystemCharOrder():  Pchar;
begin
  Result:= Pchar(sx_GetSystemCharOrder());
end;

initialization
  SDEEngine := TSDEEngine.Create;

finalization
  SDEEngine.free;
end.
