unit slang;

{-------------------------------------------------------------------------------------------
  Created by Ren van Stekelenburg (rvstekelenburg@icon.nl)
  October 21, 1999

  This unit bypasses the sLanguage setting, so you can use Apollo without
  manually change the environment settings in the control panel. You do not
  have to reboot the system or restart the program. 

  The Initialization part of this unit sets the Language to ENU.
  This unit must be the first in the USES clause of the .dpr file.
  Just before the application.initialize is called in the dpr file you can
  add the RestoreLang function to restore the original setting

  Example :

		program Project1;

    uses
      slang in 'slang.pas',    // <<---- Add this line !!
      Forms,
      main in 'main.pas';

      begin
				RestoreLang;           // <<---- Add this line !!
       	Application.Initialize;
       	Application.CreateForm(TForm1, Form1);
       	Application.Run;
      end.

  This unit also checks if you are running WinNT and uses the Registry
  instead of the WIN.INI file. This is not required in WinNT 4.0, but you
  never know what NT 5.0 will do (just to make sure ..)
-------------------------------------------------------------------------------------------}

interface

  function RestoreLang:boolean;
  function SetLang(cNewLang:string):string;

implementation

  uses inifiles,registry,windows,sysutils,dialogs;

  var OrgLang:String;  // Global Var for storing the original language


{----------------------------------------------------------------------}
function SetLang(cNewLang:string):string;
{ This function sets a new language and returns the original language.
  The return value is a empty string, if an error has occured. }
var
  Version:TOSVERSIONINFO;
  cOrg:String;
  ini:TIniFile;
  reg:TRegistry;
  lErr:boolean;
begin
  cOrg := '';
  lErr := false;
  ZeroMemory(@Version,sizeof(TOSVERSIONINFO));
  Version.dwOSVersionInfoSize := sizeof(TOSVERSIONINFO);

  if GetVersionEx(Version) then
  begin

    case Version.dwPlatformId of

			// Windows 95
			VER_PLATFORM_WIN32_WINDOWS:
			begin
				try
					ini := TIniFile.Create('WIN.INI');
					cOrg := ini.ReadString('Intl','sLanguage','');
					if CompareText(cOrg,cNewLang)<>0 then
          begin
						ini.WriteString('Intl','sLanguage',cNewLang);
					end;
					if length(cOrg)=0 then
          	cOrg := cNewLang;
					ini.free;
				except
					lErr := true;
				end;
			end;

			// Windows NT
			VER_PLATFORM_WIN32_NT:
			begin
				try
					reg := TRegistry.Create;
					if reg.OpenKey('Control Panel\International',FALSE) then
          begin
						if reg.ValueExists('sLanguage') then
  	        begin
							cOrg := reg.ReadString('sLanguage');
							if CompareText(cOrg,cNewLang)<>0 then
        		  begin
								reg.WriteString('sLanguage',cNewLang);
							end;
	          end
  	        else
    	      begin
							reg.WriteString('sLanguage',cNewLang);
							cOrg := cNewLang;
						end;
						reg.CloseKey;
					end
    	    else
      	  begin
						lErr := true;
					end;
					reg.Free;
				except
					lErr := true;
				end;
  		end;

   	// Otherwise
		else
    	lErr := true;
    end;
  end
  else
  begin
		lErr := true;
  end;

  if lErr then
  begin
		cOrg := '';  // informing that something has gone wrong..
		assert(true,'Error setting temporary language!');
  end;
  Result := cOrg;
end;


{----------------------------------------------------------------------}
function RestoreLang:boolean;
{ RestoreLang, sets the original language back }
begin
  Result := false;
  if length(OrgLang)>0 then  // if Orglang is empty, something has gone wrong
  begin
		if Length(SetLang(OrgLang))>0 then
    begin
			Result := true;
    end;
  end;
end;


{----------------------------------------------------------------------}
initialization
  OrgLang := SetLang('ENU');

end.

