(*============================================================================
  ApoDSet.PAS
  Main TApolloTable code (Plus TApolloDataSet)

  Copyright 1999-2002 Vista Software
  Last Updated: 10-22-2002

  NOTE: Tab stops were set at 2 spaces during development of this unit. If you
        are using a tab stop setting other than 2 spaces, the code indents
        may not line up accurately in your editor.

-10-21-00, Add {$T-} to compiler's directive to turn off "Typed @ Operator"
-06-28-01, FOptimisticCommit default value is set into False [mo]
- APT, APX, APM
============================================================================*)

unit ApoDSet;

interface

//{$I CS_BUILD.PAS}  // Modify this file to toggle between C/S build and std

{$define APOLLO_CS}  // Remove dot for Client/Server build

//MED : 09-10-01, added

{$ifdef VER150}  {MED: 09-19-02, Added for D7 support}
  {$define DELPHI5_OR_LATER}
{$endif}

{$ifdef VER140}
  {$define DELPHI5_OR_LATER}
{$endif}

{$ifdef VER130}
  {$define DELPHI5_OR_LATER}
{$endif}

uses
  DB, Classes, SysUtils, DBCommon, Windows,
  SDE60, ApoEnv, ApCommon, ApConn, Dialogs

  {$ifdef VER150}  {MED: 09-19-02, Added for D7 support}
  , Variants
  {$endif}

  {$ifdef VER140}  {LAS: 05-29-00, Added for D6 support}
  , Variants
  {$endif}

  {$ifdef EVAL}
  , Registry
  {$endif}
  ;

const

{$ifdef VER120}      // D4
	{$define NOT_D3_OR_CB3}
{$endif}

{$ifdef VER125}      // CB4
	{$define NOT_D3_OR_CB3}
{$endif}

{$ifdef DELPHI5_OR_LATER}      // D5 or D6
  {$define NOT_D3_OR_CB3}
{$endif}


  
  // ------------------- Exception Handler Error Messages -------------------//
  ERR_NOT_IN_EDIT_OR_INSERT_MODE =
  	'Data set must be in Edit or Insert mode to update';
  ERR_MUST_CLOSE_TABLE =
  	'Cannot perform this operation on an open table';
  ERR_MUST_OPEN_TABLE =
  	'Cannot perform this operation on a closed table';
  ERR_MUST_OPEN_EXCLUSIVE =
  	'Table must be opened in Exclusive mode to perform this operation';
  ERR_TABLE_IS_READONLY =
  	'Cannot modify a read-only data set';
  ERR_LOCATE_BEFORE_NEXT =
  	'Must call ApolloLocate before ApolloLocateNext';
  ERR_CREATENEW_BEFORE_CREATEFIELD =
  	'Must call CreateNew before CreateField';
  ERR_CREATENEW_BEFORE_CREATEEXEC =
  	'Must call CreateNew before CreateExec';
  ERR_INVALID_DATABASENAME =
  	'Invalid DatabaseName value';
  ERR_INVALID_INDEXNAME =
  	'Invalid IndexName value';
  ERR_INVALID_INDEX =
  	'Invalid Index File or Tag';
  ERR_INVALID_QUERY_EXPRESSION =
  	'Invalid Query Expression';
  ERR_UNABLE_TO_OPEN_TABLE =
  	'Unable to open table.' + #13 + 'File: ';
  ERR_RECORD_LOCK_FAILURE =
  	'Record Lock Failure. May be locked by another user or task.';
  ERR_MISSING_TABLENAME_PROPERTY =
  	'Missing TableName property';
  ERR_TABLE_DOES_NOT_EXIST =
  	'Table does not exist.' + #13 +
    'File or directory does not exist.' + #13 +
    'File: ';
  ERR_TABLE_IS_EMPTY =
  	'Cannot perform this operation on an empty table';
  ERR_SCOPE_SET_FAILED =
  	'Scope Set failed';
  ERR_CIRCULAR_DATASOURCE_REFERENCE =
  	'Circular Datasource reference not allowed';
	ERR_VALUE_OUT_OF_BOUNDS =
  	'Translate error. Value out of bounds';
  ERR_BLOB_WRITE_ERROR =
  	'Blob Write Error';
	ERR_MISSING_MEMO_FILE =
  	'Associated MEMO file does not exist.' + #13 + 'File: ';
	ERR_INDEX_NAME_REQUIRED =
  	'Missing Parameter to Index Method: Index File Name';
	ERR_INDEX_TAG_NAME_REQUIRED =
  	'Missing Parameter to IndexTag Method: Index Tag Name';
	ERR_INDEX_EXPRESSION_REQUIRED =
  	'Missing Parameter to Index or IndexTag Method: Index Expression';
	ERR_INDEX_FILE_NOT_FOUND =
    'Index file does not exist.' + #13 +
    'File: ';
  ERR_NO_INDEX_ACTIVE =
  	'This operation requires an active index order';
	ERR_MUST_CLOSE_TABLE_TO_CHANGE_DATA_ACCESS =
  	'Cannot change data access method while active';
	ERR_ONLY_FOX_SUPPORTED =
		'Only TableType of ttSXFOX supported in daServer mode';
	ERR_NO_APOLLOCONNECTION_SPECIFIED =
  	'No ApolloConnection specified';
	ERR_NOT_CONNECTED =
  	'Not connected';
	ERR_FIELD_NOT_FOUND =
  	'Field not found: ';
	ERR_INVALID_DATE_VALUE =
  	'Invalid Date value';
	ERR_OUT_OF_MEMORY_WHILE_EXPANDING_ARRAY =
  	'Out of memory while expanding dynamic array';
	ERR_INVALID_INDEX_EXPRESSION =
  	'Invalid index expression';
	ERR_MUST_ENTER_LOGIN_INFO =
  	'Must include User Name and Password to open table ' +
    'through the Apollo Database Server';
	ERR_INVALID_FIELD_NAME =
  	'Invalid field name. Field names must begin with a letter';
	ERR_INVALID_FIELD_TYPE =                              {LAS: 05-23-00, Added}
  	'Invalid field type';                               {LAS: 05-23-00, Added}
	ERR_INVALID_FIELD_LEN =                               {LAS: 07-16-00, Added}
  	'Invalid field length.';                            {LAS: 07-16-00, Added}
  ERR_UNABLE_TO_CREATE_TABLE =                          {LAS: 05-23-00, Added}
  	'Unable to create table.' + #13 + 'File: ';         {LAS: 05-23-00, Added}
	ERR_MISSING_FIELD_DEFINITIONS =                       {LAS: 05-24-00, Added}
  	'Missing required field definitions';               {LAS: 05-24-00, Added}
  ERR_MISSING_DATABASENAME_PROPERTY =                   {LAS: 05-26-00, Added}
  	'Missing DatabaseName property';                    {LAS: 05-26-00, Added}
	JULIAN_ADJUSTMENT  = 1721425;
  ERR_CANT_CHANGE_COMMIT_LEVEL =
    'Cannot change the commit level while the Table is opened';

	iaMonthTot : Array[0..13] of Integer =
  	(0,0,31,59,90,120,151,181,212,243,273,304,334,365);

  FLD_NULL    = #0;
  FLD_NOTNULL = #1;

	// Open Modes
  OPEN_READWRITE = 0;
  OPEN_READONLY  = 1;
  OPEN_EXCLUSIVE = 2;

  // RDE types    TApolloTableType = (ttSXNONE, ttSXNTX, ttSXFOX, ttSXNSX, ttSXNSX_DBT);
  SDENTX      = 1;
  SDEFOX      = 2;
  SDENSX      = 3;
  SDENSX_DBT  = 4;        {MED: 03-23-01, added}

  // Text file types for AppendFrom
  COMMA_DELIM = 21;
  SDF_FILE    = 22;
  TAB_DELIM   = 23;
		SPACE_DELIM = 24;
		
  // OEM Source types for AppendFrom
  OEMNTX = 31;
  OEMFOX = 32;
  OEMNSX = 33;

 { MED: 03-27-01, commented- no need.  TBD
  // Outdated RDE Names
  SIXNTX = 1;
  SIXFOX = 2;
  SIXNSX = 3;
  DBFNTX = 1;
  DBFIDX = 2;
  DBFNSX = 3;
}
  // Index Types
  INDEX_STANDARD           = 1;
  INDEX_STANDARD_UNIQUE    = 2;
  INDEX_CONDITIONAL        = 3;
  INDEX_CONDITIONAL_UNIQUE = 4;

  // Date types
  AMERICAN    = 0;
  ANSI        = 1;
  BRITISH     = 2;
  FRENCH      = 3;
  GERMAN      = 4;
  ITALIAN     = 5;
  SPANISH     = 6;
  WIN_DEFAULT = 99;

  // Epoch Default
  EPOCH_DEFAULT = 1980;

  // Data type identifiers for sx_Replace
  R_INTEGER  = 1;
  R_LONG     = 2;
  R_DOUBLE   = 8;
  R_JULIAN   = 32;
  R_LOGICAL  = 128;
  R_CHAR     = 1024;
  R_CHAREX   = 1025;
  R_DATESTR  = 1056;
  R_MEMO     = 3072;
  R_BITMAP   = 4096;
  R_BLOBFILE = 8192;

  // sx_QueryTest Results
  OPTIMIZE_NONE = 0;
  OPTIMIZE_PART = 1;
  OPTIMIZE_FULL = 2;

  // sx_EvalTest Results
  EVAL_CHARACTER  = 1;
  EVAL_NUMERIC    = 2;
  EVAL_LOGICAL    = 3;
  EVAL_DATESTRING = 4;  

  // sx_Index(tag) iOptions
  IDX_NONE   = 0;
  IDX_UNIQUE = 1;
  IDX_EMPTY  = 2;

  // sx_ErrorLevel uiErrorLevels
  {MED: 10-28-00, implement this for 7.0 in ApCommon}
  //TApolloErrorLevel = (ERRLEVEL_NONE, ERRLEVEL_FATAL, ERRLEVEL_STANDARD); {MED: 10-28-00, added}

  ERRLEVEL_NONE     = 0;
  ERRLEVEL_FATAL    = 1;
  ERRLEVEL_STANDARD = 2;

  // RYO BOOL Operations for RYOFilterActivate
  RYOFILTER_NEW    = 1;
  RYOFILTER_AND    = 2;
  RYOFILTER_OR     = 3;
  RYOFILTER_XOR    = 4;
  RYOFILTER_ANDNOT = 5;
  RYOFILTER_ORNOT  = 6;
  RYOFILTER_XORNOT = 7;


  // sx_SysProp Constants
  // Global Task Information
  SDE_SP_GETSOFTSEEK      	= 1000; // Get the softseek flag
  SDE_SP_SETSOFTSEEK      	= 1001; // Set the softseek flag
  SDE_SP_GETEXACT         	= 1002; // Get the extact flag
  SDE_SP_SETEXACT         	= 1003; // Set the extact flag
	 SDE_SP_GETTRANSLATE 	  	= 1004; // Get the translate flag
  SDE_SP_GETDELETED       	= 1006; // Get the deleted flag
  SDE_SP_PUTOBUFFER	      	= 1007; // Write the optimistic buffer on commit  {LAS: 06-09-00, Added}
  SDE_SP_GETOBUFFER       	= 1008; // Get the optimistic buffer flag
  SDE_SP_SETOBUFFER       	= 1009; // Set the optimistic buffer flag
  SDE_SP_GETSTRINGTYPE    	= 1010; // Get the stringtype flag
  SDE_SP_SETSTRINGTYPE    	= 1011; // Set the stringtype flag
  SDE_SP_GETDISABLEAUTO   	= 1012; // Get the disable auto open flag
  SDE_SP_SETDISABLEAUTO   	= 1013; // Set the disable auto open flag

  SDE_SP_SETOEMCOLLATE    	= 1101; // Set the collation sequence for OEM tables.
  SDE_SP_GETOEMCOLLATE    	= 1111; // Get the collation sequence for OEM tables.
  SDE_SP_SETCHRCOLLATE    	= 1102; // Set the collation sequence for Windows tables.
  SDE_SP_GETCHRCOLLATE    	= 1122; // Get the collation sequence for Windows tables.
	SDE_SP_SETLGTRCOLLATE   	= 1103; // Set the ligatures collation dimmension
	SDE_SP_GETLGTRCOLLATE   	= 1133; // Get the ligatures collation dimmension
  SDE_SP_GETDUDENCOLLATE  	= 1104; // Get the German DUDEN collate flag
  SDE_SP_SETDUDENCOLLATE  	= 1105; // Set the German DUDEN collate flag
  SDE_SP_GETLIMITCASECONV 	= 1106; // Get limit case conv to A-Z, a-z if True
  SDE_SP_SETLIMITCASECONV 	= 1107; // Set limit case conv to A-Z, a-z if True
 	SDE_SP_SETSPECIALCOLLATE	=	1108; // Set the international collation like DUDEN collate flag
	SDE_SP_GETSPECIALCOLLATE	=	1109; // Set the international collation like DUDEN collate flag
	SDE_SP_GETLANGUAGECOLLATE	=	1110; // Get language, according to collation done
  

  // Behavior settings which bridge the differences between 1.40 and 2.00
  SDE_SP_GETADDQUERY       = 1300; // Get the AddQueryFlag
  SDE_SP_SETADDQUERY       = 1301; // Set the AddQueryFlag
  SDE_SP_GETUSECONDITIONAL = 1302; // Get the bUseConditional flag
  SDE_SP_SETUSECONDITIONAL = 1303; // Get the bUseConditional flag
  SDE_SP_SETWRITEBLOBHDR   = 1305; // Set the bWriteBlobHdr
	SDE_SP_GETQUERYRELAXFLAG = 1306; // Get flag that dictates rules of query
	SDE_SP_SETQUERYRELAXFLAG = 1307; // Set flag that dictates rules of query

  // WorkArea information
  SDE_SP_GETDRIVER 				= 2000;	// Get the active driver
	SDE_SP_SETSTRDEFLEN			= 2001;	// Set the default lenght for STR, if 2nd parameter is absent and field lenght zero
	SDE_SP_SETSTRDEFDEC			= 2002;	// Set the default decimals for STR, if 3d parameter is absent and field lenght zero
	SDE_SP_SETDEFAPPEND			= 2003;	// Set default behavior for ordering ordering for non-unique key like FOX/Clipper
  SDE_SP_SETDBTNSX        = 2004;	// MED: 03-23-01, added NSX_DBT support // differs from sde60.h

  // Index information for current workarea
  SDE_SP_GETINDEXCOUNT = 3000;     // Get the number of indexes
  SDE_SP_GETDESCENDING = 3002;     // Get the descending flag
  SDE_SP_GETEMPTY      = 3004;     // Get the empty index flag

  {MED:03-21-01, added}

  FRecOrdinalNeedsRefresh = -2;
type

	// Cached Update Support {MED: 05-01-00, Added}
  TUpdateKind         = (ukModify, ukInsert, ukDelete);
  TUpdateStatus       = (usUnmodified, usModified, usInserted, usDeleted);
  TUpdateAction       = (uaFail, uaAbort, uaSkip, uaRetry, uaApplied);
  TUpdateRecordTypes  = set of (rtModified, rtInserted, rtDeleted, rtUnmodified);
  TUpdateRecordEvent  = procedure(DataSet: TDataSet; UpdateKind: TUpdateKind;
                        var UpdateAction: TUpdateAction) of object;
  TUpdateErrorEvent   = procedure(DataSet: TDataSet; E: EDatabaseError;
                        UpdateKind: TUpdateKind; var UpdateAction: TUpdateAction) of object;


	// Fetch Support {MED: 05-01-00, Added}
  PFetchList = ^FetchList;
  FetchList = record
    GetResult  : TGetResult;
    RecNum     : Integer;             // Physical Server Record Number  {LAS: 05-31-00, Changed from RecNo}
    RecOrdinal : Integer;             // OrderPosGet (ordinal number)
    Data       : PChar;               // current data
    Next       : Boolean;
  end;


	// Cached Update Support {MED: 05-01-00, Added}
  PRecList = ^RecList;
  RecList = record
    upStatus   : TUpdateStatus;
    upAction   : TUpdateAction;
    SDEDeleted : WordBool;             // TRUE when sx_deleted = TRUE
    BackIndex  : Integer;              // <--index to backup cached
    RecNo      : Integer;
    RecNum     : Integer;              // Physical Server Record Number  {LAS: 05-31-00, Changed from RecNo}
    Data       : String;               // current data
  end;

  TServerRecNo = array of Integer;

  EApolloError = class(Exception);

{$ifdef NOT_D3_OR_CB3}
  TLockList = Array of Integer;
{$endif}

  // This is embedded into the record buffer
  TApBlobData = packed record
	  Size: Integer;
    Data: PChar;
    SDEPointer: Array[0..9] of Char;
    Modified: ByteBool;
    Loaded: ByteBool;
  end;
  PBlobData = ^TApBlobData;

  PRecInfo = ^TRecInfo;
  TRecInfo = packed record
    Bookmark: Integer;
    BookmarkFlag: TBookmarkFlag;
  end;

  TDateFld = record
    Year  : Array[0..3] of Char;
    Month : Array[0..1] of Char;
    Day   : Array[0..1] of Char;
  end;

  TFieldInfo = class
    SDEOffset: Integer;
    NativeOffset: Integer;
    Width: Integer;
    Decimals: Integer;
  end;

  TDbfStruct = record
   sName     : String;
   sType     : String;
   iLength   : Integer;
   iDecimals : Integer;
  end;

  TApolloTable = class;
  TApBlobStream = class(TMemoryStream)
  private
    FMode: TBlobStreamMode;
    FField: TBlobField;
    FDataSet: TApolloTable;
  public
    constructor Create(Field: TBlobField; Mode: TBlobStreamMode);
    destructor Destroy; override;
    function Write( const Buffer; Count: Integer ): Integer; override;
  end;

  TApolloSearchObject = Class(TObject)        {MED: 04-20-00, made it thread-safe!}
  private
    SearchComponentDataLast,
    SearchComponentDataCompLast,
    SearchComponentScreenLast,
    SearchComponentScreenCompLast,
    SearchComponentScreenFrameLast:Integer;
    function SearchComponent(CompClass:TClass): TComponent;
  public
    function SearchNextComponent(CompClass:TClass): TComponent;
    function SearchFirstComponent(CompClass:TClass): TComponent;
  end;

  {============================================================================
		TApolloTable
  ============================================================================}
  TCommandProtocol = (cpVcl, cpBypassVCL);

  TEOFAction = (eaEOF, eaMoveLast);
	TSequencedCount = (scAllCount, scFilterOnlyCount, scRawCount); // MED:03-05-2002 addedCount-nasty bug!

  TApolloRangeSet = (rsNone, rsStart, rsEnd);  {LAS: 05-23-00, Added}

  TApolloTable = class( TDataSet )
  private

    FIndirectlyAssigned : Boolean;     // MED: 03-07-2002, added to support OLEDB WORKAREA=1

  protected

    FCommitLevel      : TApolloCommitLevel;

    FStructList       : TStrings;         // Table Structure {MED:05-20-00, added to make it lessen the calls on the server}

		FLastIndexNameVar : String;
    FInTrans          : Bool;             // When Transaction Processing is ON
    FServerRecNo      : TServerRecNo;     // dynamic array for server physical record #

    FFetchCount           : Integer;          // # of record to fetch at a time
    FFetchLastRec         : Integer;          // our pointer to last record read
    FFetchSavedDirection  : TGetMode;
    FFetchReadCount       : Integer;          // # of records read on last call-med
    AFetch                : PFetchList;
    FFetchRecs            : TList;

    FRecords            : TList;    // our internal cached records
    FBackRecords        : TList;    // Our Original Records
    ARecord             : PRecList;
    FRecNo              : Integer;  // cached update record # OR physical recno when cached update is FALSE
    fFRecOrdinal         : Integer;  // MED: 05-21-00, added to replace OrderPosGet

    FCachedUpdates    : Boolean;
    FUpdatesPending   : Boolean;
    FUpdateRecordTypes: TUpdateRecordTypes; // determines what type of records are visible
    FOnUpdateRecord   : TUpdateRecordEvent;
    FOnUpdateError    : TUpdateErrorEvent;
  	FCountNeedsRefresh: WordBool;
		FCentury: Boolean;   {LAS: 04-05-00, Added}
    FExact  : Boolean;     {MED: 05-20-00, Added}
    FSoftSeek  : Boolean;     {MED: 05-20-00, Added}
    FDateFormat  : TApolloDateFrmt;     {MED: 05-20-00, Added}
    FTranslate  : WordBool;            // MED: 04-07-01    
    FMemoBlockSize: Integer; {MED: 05-20-00, Added}

    FErrorMsg   : String;       {MED: 11-21-00, added to support ApolloASP}

	  FErrorFunc: Integer;
    FRecSize    : Integer;
  	FEpoch      : Integer;
  	FErrorLevel : Integer;
    FDeleted    : WordBool;
		FEOFAction  : TEOFAction;
		FRowCount   : Integer;
    FRecCount   : Integer;   {LAS: 05-30-00, Added}
    FSequencedCount: TSequencedCount;

    FDatabaseName: String;
    FDatabase: TApolloDatabase;
    FTableType: TApolloTableType;
    FTurboRead: WordBool;
    FApplyEnvSettings : WordBool;       // MED: 05-21-00, added

    FRangeStart: Variant;        {LAS: 05-23-00, Added}
    FRangeEnd: Variant;          {LAS: 05-23-00, Added}
    FRangeSet: TApolloRangeSet;  {LAS: 05-23-00, Added}
		FAddingIndex: Boolean;       {LAS: 05-23-00, Added}

    FPassword: String;
    iWA: SmallInt;
		FLastScopeVar: String;
    FScopeBegin: String;
    FScopeEnd: String;
    FScoped: Boolean;
		FSpeedMode: WordBool;
    FDataPath: String;
    FSequenced: Boolean;
    FExtraIndexes: TStringList;
    FIndexDefs: TIndexDefs;
    FMasterLink: TMasterDataLink;
    FRecBufSize: Integer;
    FRecInfoOfs: Integer;
    FBlobCacheOfs: Integer;
    FCalcFieldsOfs: Integer;
    //    FAliasName,               {MED: 05-20-00, made it local- no need for global}
    FTableName: String;

    FExclusive: WordBool;
    FReadOnly: WordBool;

  	LFieldOffsets: TList;
    FCursorOpen: WordBool;
    FFirstCrack, FLastCrack: WordBool;
    LLocateExpression: String;
    LLocateDirection: WordBool;
    FIndexName: String;
    FAppending: WordBool;
		FDummyNum: Integer;   // To get RecCount info to show up in Object Inspector
    FDummyString: String; // To get Version info to show up in Object Inspector
    FOldRec: Integer;
    FWriteBlobHdr: WordBool;
    LSDERecordBuffer: PChar;
    LOpeningTable: Boolean;
		FFilterStr: String;
    FFiltered: Boolean;
    LFilterRecordBuffer: PChar;
    FIsTableEncrypted: WordBool;
    FSDEFieldCount: Integer;
    FLocateIndexOrd: Integer;
    FIndexOrd       : Integer;
    FIndexCondition : String;
    FOptimisticCommit: Boolean;
    FAccessMethod: TAccessMethod;
    FIsEncrypted: WordBool;

{$ifdef NOT_D3_OR_CB3}
    FLockList : TLockList;
		FUpdateLockList : WordBool;
{$endif}


    FApolloConnection : TApolloConnection;

		function GetFRecOrdinal: integer;
		property FRecOrdinal : integer read GetFRecOrdinal write fFRecOrdinal;

    procedure SetTranslate( Value: WordBool );

    procedure SetApolloConnection(Val:TApolloConnection);
		procedure SetAccessMethod( Value: TAccessMethod );

    procedure RaiseApollo(msg:String);{MED:11-22-00, added}
    ///////////////////////////////////////////////////////////////////
    // private CachedUpdates methods

    procedure ClearFetchedItems;
    function FetchRecord(
                          FirstCrack, LastCrack:WordBool;
                          FEOFAction:Integer;
                          GetMode: TGetMode;
                          var iRecNo : Integer): TGetResult;


    ///////////////////////////////////////////////////////////////////

    procedure OpenTable;
		procedure SetTableType( Value: TApolloTableType );
		procedure SetCommitLevel( Value: TApolloCommitLevel );
    procedure CheckMasterRange;
    procedure MasterChanged(Sender: TObject);
    procedure MasterDisabled(Sender: TObject);
    function  GetMasterSource: TDataSource;
    procedure SetMasterSource(Value: TDataSource);
    function  GetMasterFields: String;
    procedure SetMasterFields(const Value: String);
		function  GetMasterVal: String;
    procedure SetIndexName(Value: String);
    procedure GetDataPath;
    procedure DoOnNewRecord; override;
    procedure InternalOpen; override;
    procedure InternalClose; override;
    procedure InternalInitFieldDefs; override;
    procedure CreateFields; Reintroduce; {LAS: 05-18-00, Changed back to virtual from reintroduce}
    function GetBlobData(Field: TField): pBlobData;
{$ifdef NOT_D3_OR_CB3}
    function  GetLockList(Index: Integer): Integer;
{$endif}
		function  GetApolloBOF : WordBool;
    function  GetApolloEOF : WordBool;
    function  GetDatabase: TApolloDatabase;
    procedure SetDatabaseName(Value: String);
    procedure SetTableName(Value: String);
    function  GetExists: Boolean;
		function  GetIndexExt: String;
		function  GetMemoExt: String;
		function  GetIndexFieldCount: Integer;
    function  GetSXRecno: Integer;
    function  GetSXRecCount: Integer;
		function  GetUniqueAliasName( sTableName: String ): String;
    function  ApolloTableType( Value: TApolloTableType): Integer;
    function  GetDeleted: WordBool;
    procedure SetExclusive( Value: WordBool );
    procedure SetExtraIndexes( Value : TStringList );
		procedure SetReadOnly( Value: WordBool );
		procedure SetSequenced( Value: Boolean );
    function  GetVersion: String;
    procedure SetScopeBegin( Value: String );
    procedure SetScoped( Value: Boolean );
    procedure SetScopeEnd( Value: String );
    procedure SetSpeedMode( Value: WordBool );
    procedure WriteNativeToSDE(Buffer: PChar);
    procedure ReadSDEToNative(Buffer: PChar);
    procedure SetWriteBlobHdr( Value : WordBool );
    procedure SetSxQuery( Value: String );
		procedure UpdateFieldAndIndexDefs;
		procedure InternalRefresh; override;
    procedure ClearBuffers; override;
    procedure ClearCalcFields(Buffer: PChar); override;
    function  AllocRecordBuffer: PChar; override;
    procedure FreeRecordBuffer(var Buffer: PChar); override;
    procedure GetBookmarkData(Buffer: PChar; Data: Pointer); override;
    function  GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; override;
    function  GetLockCount: Word;
    function  GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean):
                                                        TGetResult; override;
    function  GetRecordSize: Word; override;
    procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); override;
    procedure InternalDelete; override;
    procedure InternalFirst; override;

    procedure InternalGotoBookmark(Bookmark: Pointer); override;

    procedure InternalHandleException; override;
    procedure InternalInitRecord(Buffer: PChar); override;
    procedure InternalLast; override;
    procedure InternalPost; override;
    procedure InternalSetToRecord(Buffer: PChar); override;
    function  IsCursorOpen: Boolean; override;
    procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override;
    procedure SetBookmarkData(Buffer: PChar; Data: Pointer); override;
    procedure SetFieldData(Field: TField; Buffer: Pointer); override;
    function  GetCanModify: Boolean; override;
    procedure RefreshInternalCalcFields(Buffer: PChar); override;
		function  PathFromAlias: String;
    procedure Loaded; override;
    procedure DoBeforeEdit; override;
  	procedure DoBeforeInsert; override;
    procedure DoAfterInsert; override;
    procedure DoAfterCancel; override;
    procedure DoAfterScroll; override;
		procedure SetFilterText(const Value: string); override;  {LAS: 04-06-00, Added}
    procedure ExtraIndexesChanged( Sender: TObject ); virtual;

   { New Protected Methods }
		procedure ActivateIndexes; virtual;
		procedure BeginScroll; virtual;
		procedure EndScroll; virtual;

    { Additional overrides (optional) }
    function  GetRecordCount: Integer; override;
    function  GetRecNo: Integer; override;
    procedure SetRecNo(Value: Integer); override;
    procedure UpdateIndexDefs; override;
    function  GetBOF: Boolean; virtual;    {LAS: 04-17-00, Added}
    function  GetEOF: Boolean; virtual;
    property  NativeBufferSize: Integer read FRecBufSize;
    function  SelectForSxCall: Boolean;
		function GetApVersion: String;
    function SetToApolloEnv: TApolloEnv;

    {-------- Allows local/server tables ------------------}
    function  cs_Alias(uiWorkArea: Word): String;
    procedure cs_Append;
    procedure cs_AppendBlank;

    function  cs_AppendFrom( cpFileName: PChar; iSourceType: SmallInt;
                             cpScopeExpr: PChar): LongBool;
    function  cs_BaseDate: PChar;
    function  cs_BaseName: PChar;
    function  cs_BlobToFile( cpFieldName: PChar;
                             cpFileName: PChar): LongBool;
    function  cs_Bof: WordBool;
    procedure cs_Close;
    procedure cs_CloseAll;
    procedure cs_CloseIndexes;
    procedure cs_Commit;
    function  cs_CopyFile( cpToFileName: PChar): LongBool;  {LAS: 06-23-00, Changed return type from WordBool}
    function  cs_CopyFileText( cpTextFileName: PChar;
                               iFileType: SmallInt): LongBool;  {LAS: 06-23-00, Changed return type from WordBool}
    function  cs_CopyStructure( cpFileName: PChar;
                                cpAlias: PChar): LongBool;  {LAS: 06-23-00, Changed return type from WordBool}
    function  cs_CopyStructureExtended(cpFileName: PChar): LongBool;  {LAS: 06-23-00, Changed return type from WordBool}
    function  cs_Count: Integer;
    function  cs_CreateExec: LongBool;  {LAS: 06-23-00, Changed return type from WordBool}
    procedure cs_CreateField(cpName: PChar; cpType: PChar;
                            iLength: SmallInt; iDecimals: SmallInt);
    function  cs_CreateFrom(cpFileName: PChar; cpAlias: PChar;
                            cpStruFile: PChar; iRDEType: SmallInt): LongBool;  {LAS: 06-23-00, Changed return type from WordBool}
    function  cs_CreateNew(cpFileName: PChar; cpAlias: PChar;
                           iRdeType: SmallInt; iNumFields: SmallInt): SmallInt;
    function  cs_DbfDecrypt: WordBool;
    function  cs_DbfEncrypt: WordBool;
    function  cs_DBFilter: PChar;
{$ifdef NOT_D3_OR_CB3}
		procedure cs_DBRLockList( ulpArray: PChar );
{$endif}
    function  cs_Decrypt(cpBuffer: PChar; cpPassword: PChar;
                         iLen: Integer): PChar;

    procedure cs_Delete;

    function  cs_Deleted: WordBool;
		function  cs_DeleteFile( sFileName : String ): WordBool;
    function  cs_Descend(cpKeyString: PChar): PChar;
    function  cs_Empty(cpFieldName: PChar): WordBool;
    function  cs_Eof: WordBool;
    function  cs_Encrypt(cpBuffer: PChar; cpPassword: PChar;
                         iLen: Integer): PChar;
    function  cs_EvalLogical(cpExpression: PChar): WordBool;
    function  cs_EvalNumeric(cpExpression: String): Double;
    function  cs_EvalString(cpExpression: PChar): PChar;
    function  cs_EvalTest(cpExpression: PChar): SmallInt;
    function  cs_FieldCount: Word;
    function  cs_FieldDecimals(cpFieldName: PChar): Word;
    function  cs_FieldName(uiFieldNum: Word): PChar;
    function  cs_FieldNum(cpFieldName: PChar): Word;
    function  cs_FieldOffset(cpFieldName: PChar): Word;
    function  cs_FieldType(cpFieldName: PChar): PChar;
    function  cs_FieldWidth(cpFieldName: PChar): Word;
		function  cs_FileExists( sFileName : String ): WordBool;
    function  cs_FilterAlias(cpAliasName: PChar; cpFieldName: PChar): WordBool;
    function  cs_FilterDlg(hwnd: HWND; cpExpr: PChar; cpCaption: PChar;
                            bHasIndexList: WordBool): PChar;
    function  cs_Flock: WordBool;
    function  cs_Found: WordBool;
    function  cs_GetBlob(cpFieldName: PChar; vpVar: Pointer): Integer;
    function  cs_GetBlobLength(cpFieldName: PChar): Integer;
    function  cs_GetByte(cpFieldName: PChar): PChar;
    function  cs_GetDateJulian(cpFieldName: PChar): Integer;
    function  cs_GetDateString(cpFieldName: PChar): PChar;
    function  cs_GetDouble(cpFieldName: PChar): Double;
    function  cs_GetInteger(cpFieldName: PChar): SmallInt;
    function  cs_GetLogical(cpFieldName: PChar): WordBool;
    function  cs_GetLong(cpFieldName: PChar): Integer;
    function  cs_GetMemo( cpFieldName: PChar; uiLineWidth: Word): PChar;
    function  cs_GetQueryBit(lRecNo: Integer): WordBool;
    procedure cs_GetRecord(cpRecord: PChar);
    function  cs_GetScope(iWhichScope: SmallInt): PChar;
    function  cs_GetString(cpFieldName: PChar): PChar;
    function  cs_GetTrimString(cpFieldName: PChar): PChar;

		procedure cs_Go(lRecNum: Integer);  // CRVG Changed from Function to Proc
    procedure cs_GoBottom;
    procedure cs_GoTop;

    function  cs_Index(cpFileName: PChar; cpExpr: PChar; iOption: SmallInt;
                       bDescend: Boolean; cpCondition: PChar): SmallInt;
    procedure cs_IndexClose;
    function  cs_IndexCondition: PChar;
    function  cs_IndexFlip: WordBool;
    function  cs_IndexKey: PChar;
    function  cs_IndexKeyField: PChar;
    function  cs_IndexName(iIndex: SmallInt): PChar;
    function  cs_IndexOpen(cpFileName: PChar): SmallInt;
    function  cs_IndexOrd: SmallInt;
    function  cs_IndexTag(cpFileName: PChar; cpTagName: PChar; cpExpr: PChar;
                          iOption: SmallInt; bDescend: Boolean;
                          cpCondition: PChar): SmallInt;
    function  cs_IndexType: SmallInt;
    function  cs_IsEncrypted(iFileOrRec: SmallInt): WordBool;
    function  cs_KeyAdd(cpTagname: PChar): WordBool;
    function  cs_KeyData: PChar;
    function  cs_KeyDrop(cpTagname: PChar): WordBool;
    function  cs_Locate(cpExpression: PChar; iDirection: WordBool;
                        bContinue: WordBool): Integer;
    function  cs_LockCount: Word; 
    function  cs_Locked(lRecNum: Integer): WordBool;
    procedure cs_MemDealloc(vpPtr: Pointer);
    function  cs_OpenMode:SmallInt;
    function  cs_OrderPosGet: Double;
    procedure cs_OrderPosSet( dPosition: Double );
    function  cs_OrderRecNo: Integer;
    procedure cs_Pack;
    function  cs_PutBlob(cpFieldName: PChar; vpVar: Pointer;
                         iSize: Integer): Integer;
    procedure cs_PutRecord(cpRecord: PChar);
    function  cs_Query(cpExpression: PChar): Integer;
    function  cs_QueryRecCount: Integer;
    function  cs_QueryTest(cpExpression: PChar): SmallInt;
    procedure cs_Recall;
    function  cs_RecCount: Integer;
    function  cs_RecNo: Integer;
    function  cs_RecSize: Integer;
    procedure cs_Reindex;
    procedure cs_Replace(cpFieldname: PChar; iDataType: SmallInt; vpData: Pointer);
    function  cs_Rlock(lRecNum: Integer): WordBool;
    function  cs_RYOFilterActivate(iFilterHandle: SmallInt;
                                   iBoolOperation: SmallInt): WordBool;
    function  cs_RYOFilterCopy: SmallInt;
    function  cs_RYOFilterCreate: SmallInt;
    function  cs_RYOFilterDestroy(iFilterHandle: SmallInt): WordBool;
    function  cs_RYOFilterGetBit(iFilterHandle: SmallInt;
                                 lRecNo: Integer): WordBool;
    function  cs_RYOFilterRestore(cpFileName: PChar): WordBool;
    function  cs_RYOFilterSave(iFilterHandle: SmallInt;
                               cpFileName: PChar): WordBool;
    function  cs_RYOFilterSetBit(iFilterHandle: SmallInt; lRecNo: Integer;
                                 iOnOrOff: SmallInt): WordBool;
    function  cs_RYOKeyAdd(cpTagname: PChar; cpKey: PChar): WordBool;
    function  cs_RYOKeyDrop(cpTagname: PChar): WordBool;
    function  cs_Seek(cpKeyValue: PChar): WordBool;
    function  cs_SeekBin(cpKeyValue: PChar; uiLength: Word): WordBool;
    function  cs_Select(uiBaseArea: Word): Word;
    procedure cs_SetFilter(cpExpression: PChar);
    procedure cs_SetGaugeFunc( pFunc: Integer );
    procedure cs_SetGaugeHook(hwndGauge: HWND);
    function  cs_SetOrder(iIndex: SmallInt): SmallInt;
    procedure cs_SetPassword(cpEncodeKey: PChar);
    procedure cs_SetQueryBit(lRecNo: Integer; bValue: WordBool);
    procedure cs_SetRelation(uiChildArea: Word; cpKeyExpr: PChar);
    function  cs_SetScope(cpLowVal: PChar; cpHighVal: PChar): WordBool;

    procedure cs_SetStringType(uiStringType: Word);
    procedure cs_SetTranslate(uiOnOff: WordBool);
    procedure cs_SetTurboRead(uiOnOff: WordBool);
    procedure cs_Skip(lNumRecs: Integer);
		function  cs_SysProp( uiSysItem: Word; vpData: Pointer ): Integer;
    function  cs_TagArea(cpTagName: PChar): SmallInt;
    function  cs_TagName(iTagArea: SmallInt): PChar;
    procedure cs_Unlock(lRecNum: Integer);
    function  cs_Use(cpFileName: PChar; cpAlias: PChar; iOpenMode: SmallInt;
                     iRdeType: SmallInt): SmallInt;
    function  cs_UseEx(cpFileName: PChar; cpAlias: PChar; iOpenMode: SmallInt;
                     iRdeType: SmallInt; uiMode : Cardinal): SmallInt;
    function  cs_Version: PChar;
    function  cs_WorkArea(cpAlias: PChar ): Word;
    procedure cs_Zap;

    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure SetFetchCount( Val: Integer );
    function SeekVal( tmpFields: TList; const KeyValues: Variant;
             Options: TLocateOptions; iOldOrder: Integer ): Boolean;

  public

    // MED: 03-31-2001 moved from public to here
    FFieldName     : array of String;  {MED: 10-22-00, added the following}
    FFieldWidth    : array of Integer;
    FFieldDecimals : array of Integer;
    FFieldOffset   : array of Integer;
    FFieldType     : array of String;
    FFieldCount    : integer;

    constructor Create( Owner: TComponent ); override;
    destructor Destroy; override;

    {MED: 03-26-01, Moved here from protected}
		function  GetTableType:TApolloTableType;

    {MED: 07-30-01}
    function AssignWorkArea(WA:Integer):Boolean; // Access:TAccessMethod; Conn:TApolloConnection): Boolean;

    {MED: 03-08-01, Moved here from private}
    procedure Cancel; override;                    {MED: 10-24-00, added}
    {MED: 11-21-00, moved to public for OLE DB Provider}
    property ErrorMsg : String read FErrorMsg write FErrorMsg;

    {MED: 11-15-00, moved to public for OLE DB Provider}
    ///////////////////////////////////////
    ////////////////////////////////////////


    property InTransaction  : Bool read FInTrans write FInTrans;
    property Records : TList read FRecords;
    //property UpdatesPending: Boolean read FUpdatesPending;
    //property UpdateRecordTypes: TUpdateRecordTypes read FUpdateRecordTypes;
    //function UpdateStatus : TUpdateStatus; reintroduce;  {LAS: 05-02-00, Added reintroduce}

    function  CreateNew( FileName: String; RDEType: SmallInt;
                         NumFields: SmallInt): WordBool;
		function  IndexOpen( sFileName: String ): SmallInt;
    function  CreateBlobStream( Field: TField;
                                Mode: TBlobStreamMode ): TStream; override;
    function  CompareBookmarks( Bookmark1,
                                Bookmark2: TBookmark ): Integer; override;

    function  IsSequenced: Boolean; override;
    function  GetFieldData( Field: TField; Buffer: Pointer ): Boolean; override;
		procedure GetIndexNames(List: TStrings);
    property  WorkArea: SmallInt read iWA ;

    // Delphi versions of the direct SDE calls
		function  Alias: String;
		property  ApolloBOF : WordBool read GetApolloBOF;
		procedure ApolloDelete;
    property  ApolloEOF : WordBool read GetApolloEOF;
		function  ApolloFieldCount: Word;
		procedure ApolloGetRecord( cpRecord: PChar );
    function  ApolloLocate( Expression: String; Backwards: WordBool;
                         Continue: WordBool): Integer;
    function  ApolloLocateNext: Integer;
		procedure Append;
		procedure AppendBlank;
		function  AppendFrom( sFileName: String; iSourceType: SmallInt;
                         sScopeExpr: String ): WordBool;
		function  BaseDate: String;
		function  BaseName: String;
		function  BlobToFile( sFieldName, sFileName: String ): WordBool;
		property  Bof: Boolean read GetBOF;    {LAS: 04-17-00, Added}
		function  BookmarkValid(Bookmark: TBookmark): Boolean; override;
		procedure CancelRange;
		procedure CloseIndexes;
    procedure Commit;

		function  CopyFile( sToFileName: String ): LongBool;  {LAS: 06-23-00, Changed return type from WordBool}
		function  CopyFileExt( sToFileName: String ): LongBool;  {LAS: 06-23-00, Changed return type from WordBool}

		function  CopyFileText( sTextFileName: String; iFileType: SmallInt): LongBool;  {LAS: 06-23-00, Changed return type from WordBool}
		function  CopyStructure( sFileName: String ): LongBool;   {LAS: 06-23-00, Changed return type from WordBool}
		function  CopyStructureExtended( sFileName: String ): LongBool;  {LAS: 06-23-00, Changed return type from WordBool}
    function  Count: Integer;
    function  CreateExec: LongBool;  {LAS: 06-23-00, Changed return type from WordBool}
		function  CreateFrom( sNewFileName, sStruExtFileName: String;
                         iRdeType : SmallInt ): LongBool;  {LAS: 06-23-00, Changed return type from WordBool}
    procedure CreateField( FieldName: String; FieldType: String;
                           FieldLen: SmallInt; FieldDec: SmallInt);
    property  Database: TApolloDatabase read GetDatabase;
		property  DataPath: String read FDataPath;
		function  DbfDecrypt: WordBool;
		function  DbfEncrypt: WordBool;
    function  DBFilter: String;
		function  Decrypt( cpBuffer : PChar; sPassWord : String;
                       iLen: Integer ): PChar;
    procedure Delete;
    property  Deleted: WordBool read GetDeleted;
    procedure DeleteTable;
		function  Descend( Value: String ): String;
		function  Empty( Value: String ): WordBool;
    procedure EmptyTable;
		function  Encrypt( cpBuffer : PChar; sPassWord : String;
                       iLen: Integer ): PChar;
		property  Eof: Boolean read GetEOF;
		function  EvalLogical( Value: String ): WordBool;
		function  EvalNumeric( Value: String ): Double;
		function  EvalString( Value: String ): String;
		function  EvalTest( Value: String ): SmallInt;
    property  Exists: Boolean read GetExists;
		function  FieldDecimals( Value: String ): Word;
		function  FieldName( Value: Integer ): String;
		function  FieldNum( Value: String ): Word;
		function  FieldOffset( Value: String ): Word;
		function  FieldType( Value: String ): String;
		function  FieldWidth( Value: String ): Word;
		function  FilterAlias( sAliasName, sFieldName: String ): WordBool;
		function  FilterDlg( winHandle: HWnd; sExpr, sCaption: String;
                                   bHasIndexList: WordBool ): String;
		function  FindKey( const KeyValues: Array of Const ): Boolean;
		function  FLock: WordBool;
		function  Found: WordBool;
		function  GetFullTableName: String;   {MED: 03-07-01, moved here}
		function  GetBlob( sFieldName: String; vpVar: Pointer ): Integer;
		function  GetBlobLength( sFieldName: String ): Integer;
		function  GetByte( sFieldName: String ): Char;
		function  GetDate( sFieldName: String ): TDateTime;
		function  GetDateJulian( sFieldName: String ): Integer;
		function  GetDateString( sFieldName: String ): String;
		function  GetDouble( sFieldName: String ): Double;
		function  GetInteger( sFieldName: String ): Integer;
		function  GetLogical( sFieldName: String ): WordBool;
		function  GetLong( sFieldName: String ): Integer;
		function  GetMemo( sFieldName: String; iLineWidth: Integer): PChar;
		function  GetQueryBit( lRecNum : Integer ) : WordBool;
		function  GetScope( iWhichScope: Integer ): String;
		function  GetString( sFieldName: String ): String;
		function  GetStringEx( sFieldName: String ): PChar;
		function  GetTrimString( sFieldName: String ): String;
    procedure Go( Value : Integer );
    procedure GoTop;
    procedure GoBottom;
    property  Handle: SmallInt read iWA;
		function  Index( sFileName, sExpression: String; Options: SmallInt;
                     Descending: Boolean; sForClause: String ) : SmallInt;
		procedure IndexClose;
		function  IndexCondition: String;
		property  IndexExt: String read GetIndexExt;
		function  IndexKey: String;
		function  IndexKeyField: String;
		property  IndexFieldCount: Integer read GetIndexFieldCount;
		function  IndexFileName( iIndex: Integer ): String;  // = Apollo.IndexName
    function  IndexFlip: WordBool;
		function  IndexOrd: Integer;
    function  IndexTag( sFileName, sTagName, sExpression: String;
                        Options: SmallInt; Descending: Boolean;
                        sForClause: String): SmallInt;
		function  IndexType: Integer;
		function  IsEncrypted( iFileOrRec: Integer ): WordBool;
		function  KeyAdd( sTagName: String ): WordBool;
		function  KeyData: String;
		function  KeyDrop( sTagName: String ): WordBool;
		property  LockCount: Word read GetLockCount;
		function  Locked( lRecNum: Integer ): WordBool;
{$ifdef NOT_D3_OR_CB3}
    property  LockList[Index: Integer]: Integer read GetLockList;
{$endif}
		procedure MemDealloc( cpBuff: PChar );
    function MoveBy(Distance: Integer): Integer; virtual;  {LAS: 04-17-00, Added}
		procedure Next; virtual;  {LAS: 04-17-00, Added}
		function  OpenMode: SmallInt;
    property  OptimisticCommit: Boolean read FOptimisticCommit
                                        write FOptimisticCommit default False; {M.O.: 06-28-01, default changed from True}
		function  OrderPosGet: Double;
    procedure OrderPosSet( Value: Double );
		function  OrderRecNo: Integer;
		procedure Pack;
		procedure Prior; virtual;  {LAS: 04-17-00, Added}
		function  PutBlob( sFieldName: String; vpVar: Pointer;
                       lSize: Integer ): Integer;
		procedure PutRecord( cpRecord: PChar );
		function  Query( const Value: String ): Integer;
		function  QueryRecCount: Integer;
		function  QueryTest( Value: String ): Integer;
		procedure Recall;
    property  RecordCount: Integer read GetSxRecCount;
		function  RecSize: Integer;
		procedure RefreshCount;
    procedure Reindex;
		procedure Replace( sFieldName: String; iDataType: SmallInt; vpData: Pointer);
    function  RLock( Value: Integer ): WordBool;
		function  RYOFilterActivate( iFilterHandle, iBoolOperation: SmallInt ): WordBool;
		function  RYOFilterCopy : Integer;
		function  RYOFilterCreate : Integer;
    function  RYOFilterDestroy( iFilterHandle: SmallInt ): WordBool;
		function  RYOFilterGetBit( iFilterHandle: SmallInt; lRecNo: Integer ): WordBool;
		function  RYOFilterSave( iFilterHandle: SmallInt; sFileName: String ): WordBool;
		function  RYOFilterRestore( sFileName: String ): WordBool;
		function  RYOFilterSetBit( iFilterHandle: SmallInt; lRecNo : Integer;
                               bSetBitOn : WordBool ): WordBool;
		function  RYOKeyAdd( sTagName, sKeyValue: String ): WordBool;
		function  RYOKeyDrop( sTagName: String ): WordBool;
    function  Seek( Value: String ): WordBool;
		function  SeekBin( Value: String; iLength: Integer ): WordBool;
    procedure SetFilter( Value: Variant );
    procedure SetFiltered( Value: Boolean ); override;
    procedure SetGaugeFunc( pFunc: Integer );
    procedure SetGaugeHook( Handle: THandle );
		function  SetOrder( iIndex: Integer ): Integer;
		procedure SetPassword( sEncodeKey: String );
		procedure SetQueryBit( lRecNum : Integer; bValue : WordBool );
		procedure SetRange( const aStartVal, aEndVal: Array of Const );
		procedure SetRelation( iChildArea: Integer; sKeyExpr: String );
		function  SetScope( sLowVal, sHiVal: String ): WordBool;
		procedure SetTurboRead( Const Value: WordBool );
    procedure Skip( Value : Integer );
		function  SysProp( uiSysItem : Word; vpData : Pointer ): Integer;
		function  TagArea( sTagName: String ): Integer;
		function  TagName( iTagArea: Integer ): String;
		procedure Unlock( lRecNum: Integer );
		procedure Zap;
		property  WriteBlobHeader: WordBool read FWriteBlobHdr
                                        write SetWriteBlobHdr default True;
	  function  Lookup( const KeyFields: String; const KeyValues: Variant;
                      const ResultFields: String ): Variant; override;
		property  LocateIndexOrd: Integer read FLocateIndexOrd write FLocateIndexOrd;
    function  Locate(const KeyFields: string; const KeyValues: Variant;
                                   Options: TLocateOptions): Boolean; override;
    procedure Resync(Mode: TResyncMode); override;
		procedure GetBinary( sField: String; var Buffer: PChar );
		procedure PutBinary( sField: String; Buffer: PChar );

    ///////////////////////////////////////////////////
    // MED: 09-09-01, added
    ///////////////////////////////////////////////////
    procedure SetSystemCollation;
    procedure SetMachineCollation;
    procedure AddDudenCollation;
    procedure AddEtecCollation;

    // Additional TTable-emulation methods {LAS: 05-23-00, Added}
    procedure CreateTable;
		procedure AddIndex(const Name, Fields: string; Options: TIndexOptions);
    procedure CloseIndexFile(const IdxName: string);
		procedure OpenIndexFile(const IdxName: string);
    procedure FindNearest(const KeyValues: array of const);
		procedure DeleteIndex(const IdxName: string);
		procedure EditKey;
		procedure SetKey;
		procedure GotoCurrent( Table: TApolloTable );
		function  GotoKey: Boolean;
		procedure GotoNearest;
		procedure EditRangeStart;
		procedure EditRangeEnd;
		procedure SetRangeStart;
		procedure SetRangeEnd;
		procedure ApplyRange;

  published

    property ApolloConnection: TApolloConnection read FApolloConnection
                                                write SetApolloConnection;
    property AccessMethod: TAccessMethod read FAccessMethod
                                          write SetAccessMethod default amLocal;

    property Active;
    property AutoCalcFields;
    property DatabaseName: String read FDatabaseName write SetDatabaseName;
    //property CachedUpdates: Boolean read FCachedUpdates write SetCachedUpdates;  {MED: 05-01-00, Moved to Published}
    property EOFAction: TEOFAction read FEOFAction write FEOFAction
                                                   default eaMoveLast;
    property Exclusive: WordBool read FExclusive write SetExclusive
                                                 default False;
		property ExtraIndexes: TStringList read FExtraIndexes write SetExtraIndexes;
    property FetchCount : Integer read FFetchCount write SetFetchCount default 50;  {LAS: 05-26-00, Raised default from 1 to 50}
		//    property Filter : String read FFilterStr write SetSxQuery;  {CRVG : 2001-05-01, Changed as below}
		//    property Filtered : Boolean read FFiltered write SetFiltered default False;  {CRVG :2001-05-01, changed as below}

    property Filter;   { CRVG: 2001-05-01, Changed}
    property Filtered; { CRVG: 2001-05-01, Changed}


{$ifdef NOT_D3_OR_CB3}
    property FieldDefs;
    property IndexDefs: TIndexDefs read FIndexDefs write FIndexDefs;  {LAS: 07-28-00, Added write}
{$endif}
    property IndexName: String read FIndexName write SetIndexName;

    property CommitLevel : TApolloCommitLevel Read FCommitLevel Write SetCommitLevel default clNormal;

		property Password : String read FPassword write SetPassword;
		property ReadOnly: WordBool read FReadOnly write SetReadOnly
                                               default False;
    property RecCount: Integer read GetSxRecCount write FDummyNum
                                               default 0;
    property RecNo: Integer read GetSxRecNo write Go default 0;
		property ScopeBegin: String read FScopeBegin write SetScopeBegin;
    property Scoped: Boolean read FScoped write SetScoped default False;
		property ScopeEnd: String read FScopeEnd write SetScopeEnd;
    property Sequenced: Boolean read IsSequenced write SetSequenced
                                                 default True;
    property SequencedCount: TSequencedCount read FSequencedCount
                                           write FSequencedCount default scRawCount;

    property SpeedMode: WordBool read FSpeedMode write SetSpeedMode default False;
    property TableName: String read FTableName write SetTableName;
    property TurboRead: WordBool read FTurboRead write SetTurboRead
                                             default False;                     

    property OEMTranslate : WordBool read FTranslate write SetTranslate;

    property MasterFields: String read GetMasterFields write SetMasterFields;
    property MasterSource: TDataSource read GetMasterSource write SetMasterSource;
    property TableType: TApolloTableType read FTableType write SetTableType default ttSXFOX;
    property Version: String read GetVersion write FDummyString stored False;   {LAS: 04-11-00, Added stored}

		// Event-Handlers
    property AfterCancel;
    property AfterClose;
    property AfterDelete;
    property AfterEdit;
    property AfterInsert;
    property AfterOpen;
    property AfterPost;
{$ifdef DELPHI5_OR_LATER }
    property AfterRefresh;
{$endif}
    property AfterScroll;
    property BeforeCancel;
    property BeforeClose;
    property BeforeDelete;
    property BeforeEdit;
    property BeforeInsert;
    property BeforeOpen;
    property BeforePost;
{$ifdef DELPHI5_OR_LATER }
    property BeforeRefresh;
{$endif}
    property BeforeScroll;
    property OnCalcFields;
    property OnDeleteError;
    property OnEditError;
    property OnFilterRecord;
    property OnNewRecord;
    property OnPostError;

		// For cached updates
    property OnUpdateRecord : TUpdateRecordEvent read FOnUpdateRecord write FOnUpdateRecord;
    property OnUpdateError  : TUpdateErrorEvent read FOnUpdateError write FOnUpdateError;
  end;

	// Backwards compatibility
	TApolloDataSet = class( TApolloTable );

	function YearToJulian( iYr : Integer ): Integer;
	function MonthDay( iYear, iDays: Integer; var ipMonthPtr: Integer;
                                            var ipDayPtr: Integer ): Boolean;
	function JulianToDT( dJulian: Double ): TDateTime;
	function DTtoJulian( DT: TDateTime ): Double;
 	function JulianDayYear( iYear, iMonth, iDay: Integer ): Integer;
	function GetApDsVer: String;
	function GetFieldsFromExpression( sIndexKey: String ): String;


{$ifdef EVAL}
	{$I EvalDefs.PAS}    {LAS: 05-05-00, Added EVAL support}
{$endif}

implementation

uses Forms;

{$ifdef EVAL}
	{$I EvalTimeOut.PAS}    {LAS: 05-05-00, Added EVAL support}
{$endif}

{ ===================================================================

  MED: 07-30-01, added
  assigns a sx_use workarea to the table.  This allows opening a table via sx_Use
  and then operating on it using TApolloTable
 ===================================================================}
function TApolloTable.AssignWorkArea(WA:Integer):Boolean;// Access:TAccessMethod; Conn:TApolloConnection):Boolean;
var
  sTemp : String;
  bOpenNew : Boolean;
begin
  Result := FALSE;

  if WA < 1 then
    exit;

  FIndirectlyAssigned := TRUE;

  bOpenNew := TRUE;

  if Active then        // table is open already.  We need to set the scope and order state
  begin
    if iWA <> WA then   // this is a different table
      Active := FALSE
    else                  // same table.
      bOpenNew := FALSE;
  end;

  cs_Select(WA);

  if bOpenNew then
  begin
    iWA             := WA;
    FTableName      := ExtractFileName(cs_BaseName);
  end;

  FIndexOrd   := cs_IndexOrd;
  FFilterStr  := cs_DbFilter;     // extracts the filter/Query condition
  FScopeBegin := cs_GetScope(0);
	FScopeEnd   := cs_GetScope(1);
  FScoped     := ( FScopeBegin <> '');// ((sLowVal + sHiVal) > '') ;


  if FTableType =  ttSXNTX then
    FIndexName := String( cs_IndexName(FIndexOrd ))
  else
    FIndexName := String( cs_TagName( FIndexOrd ));

  if FAccessMethod = amLocal then
  begin
    FSDEFieldCount  := sx_FieldCount;     // Get raw SDE field count for table
    FFieldCount     := FSDEFieldCount;    // MED: 10-21-00, added-will replace FSDEFieldCount eventually
    FRecSize        := sx_RecSize;
    // Is table encrypted?
    FIsEncrypted := IsEncrypted(0);

    if (FIndexName <> '') then
    begin
      sTemp := FIndexName;
      SetIndexName( '' );       // Toggle index name off...
      SetIndexName( sTemp );    // ...and back on, to re-initialize
    end;

    ActivateIndexes;
    IndexDefs.Updated := False;
    UpdateIndexDefs;
    FRowCount       := sx_RecCount;
    FIndexCondition := sx_IndexCondition;
  end
  else
  begin
    FSDEFieldCount  := cs_FieldCount;     // Get raw SDE field count for table
    FFieldCount     := FSDEFieldCount;    // MED: 10-21-00, added-will replace FSDEFieldCount eventually
    FRecSize        := cs_RecSize;
    // Is table encrypted?
    FIsEncrypted := IsEncrypted(0);

    // The server sets active index to FIndexName already for structural index
    // the following is only for the inferior ttSXNTX drivers! -- MED
    if (FIndexName <> '') AND (FTableType = ttSXNTX) then
    begin
      sTemp := FIndexName;
      SetIndexName( '' );       // Toggle index name off...
      SetIndexName( sTemp );    // ...and back on, to re-initialize
    end;
    ActivateIndexes;
    IndexDefs.Updated := TRUE;
    FRowCount       := cs_RecCount;
    FIndexCondition := cs_IndexCondition;
  end;

	FRecCount := FRowCount; {LAS: 06-14-00, Moved down}
  FCountNeedsRefresh := True;

  if bOpenNew then
    Active := TRUE;

  Result := Active;
end;

function TApolloSearchObject.SearchComponent(CompClass:TClass): TComponent;
var
  Temp1: TComponent;
	i, j : Integer;
  dm : TDataModule;
  frm : TForm;
  bFound : Boolean;

{$ifdef DELPHI5_OR_LATER }
  Temp2: TComponent;
	k : Integer;
{$endif}
begin
	Result := nil;

  // Check all DataModules first -- no need to check frames here.
  // Datamodule does not allow frames
  bFound := FALSE;

  //\\ Delphi 6 bug. D6 does not report the correct # of DataModules
  //\\showmessage('DataModuleCount ' + inttostr(Screen.DataModuleCount) );

  if Screen.DataModuleCount > 0 then
  begin
    for i := SearchComponentDataLast to Screen.DataModuleCount-1 do
    begin
      dm := Screen.DataModules[i] as TDataModule;
      for j := SearchComponentDataCompLast to dm.ComponentCount - 1 do
      begin
        Temp1 := dm.Components[j];
        if Temp1.InheritsFrom(CompClass) then
        begin
          SearchComponentDataCompLast := j + 1;
          Result := Temp1;
          bFound := TRUE;
          break;
        end;
      end;
      if bFound then
        break;
      SearchComponentDataCompLast := 0;
    end;
    SearchComponentDataLast := i;
  end;

  if not bFound  then    // we didn't find in datamodule. try the forms
  begin
    // Check all other forms
	  for i := SearchComponentScreenLast to Screen.FormCount-1 do
  	begin
	    frm := Screen.Forms[i] as TForm;
	    for j := SearchComponentScreenCompLast to frm.ComponentCount - 1 do
      begin
	      Temp1 := frm.Components[j];
        if Temp1.InheritsFrom(CompClass) then
        begin
          SearchComponentScreenCompLast := j + 1;
          Result := Temp1;
          bFound := TRUE;
          break;
        end
{$ifdef DELPHI5_OR_LATER }
        else
        if Temp1.InheritsFrom(TFrame) then     // TFrame might have some embedded
        begin
          for k := SearchComponentScreenFrameLast to Temp1.ComponentCount-1 do
          begin
            Temp2 := Temp1.Components[k];
            if Temp2.InheritsFrom(CompClass) then
            begin
              SearchComponentScreenFrameLast := k + 1;
              Result := Temp2;
              bFound := TRUE;
              break;
            end;
          end;
          if bFound then
            break;
          SearchComponentScreenFrameLast := 0;
        end
{$endif};
	  	end;
      if bFound then
        break;
      SearchComponentScreenCompLast := 0 ;
    end;
    SearchComponentScreenLast := i; {MED:08-17-00, moved it up before end; as per Michael Benowitz suggestion}
  end;
end;

function TApolloSearchObject.SearchNextComponent(CompClass:TClass): TComponent;
begin
  Result := SearchComponent(CompClass);
end;

function TApolloSearchObject.SearchFirstComponent(CompClass:TClass): TComponent;
begin
  SearchComponentDataLast       := 0;
  SearchComponentDataCompLast   := 0;
  SearchComponentScreenLast     := 0;
  SearchComponentScreenCompLast := 0;
  SearchComponentScreenFrameLast:= 0;
  Result := SearchComponent(CompClass);
end;



{MED:11-22-00, added}
procedure TApolloTable.RaiseApollo(msg:String);
begin
  FErrorMsg := msg;
  raise EApolloError.Create( msg );
end;

{==============================================================================
  GetApDSVer - Returns Version Info of JUST the ApolloDataSet
 ==============================================================================}
function GetApDSVer: String;
const
{$ifdef EVAL}         {LAS: 04-25-00, Changed for EVAL}
	EVAL = '[EVAL] ';   {LAS: 04-25-00, Changed for EVAL}
{$else}
	EVAL = '';
{$endif}
begin
	{LAS: 04-25-00, Changed for EVAL}
	Result := 'Apollo VCL ' + EVAL + APDS_VER + ', ' +
                        DateToStr( EncodeDate( VER_YEAR, VER_MONTH, VER_DAY ));
end;

{==============================================================================
	YearToJulian
	Calculates the number of days to the year

   This calculation takes into account the fact that:
   1)  Years divisible by 400 are always leap years.
   2)  Years divisible by 100 but not 400 are not leap years.
   3)  Otherwise, years divisible by four are leap years.

   Since we do not want to consider the current year, we will subtract
   the year by 1 before doing the calculation.
 ==============================================================================}
function YearToJulian( iYr : Integer ): Integer;
begin
   Dec( iYr );
   Result := (iYr*365) + Trunc(iYr/4) - Trunc(iYr/100) + Trunc(iYr/400);
end;

{==============================================================================
	MonthDay
  Given the year and the day of the year, sets the month and day of month.
 ==============================================================================}
function MonthDay( iYear, iDays: Integer; var ipMonthPtr: Integer;
                                          var ipDayPtr: Integer ): Boolean;
var
	isLeap, i, j: Integer;
begin
	if (iYear mod 4 = 0) and (iYear mod 100 <> 0) or (iYear mod 400 = 0) then
	  isLeap := 1
  else
  	isLeap := 0;

  if (iDays <= 59) then
		isLeap := 0;

	for i := 2 to 13 do
	begin
		if (iDays <= iaMonthTot[i] + isLeap) then
		begin
    	j := (i-1);
    	ipMonthPtr := j;
			if (j <= 2) then
				isLeap := 0;
      ipDayPtr := iDays - iaMonthTot[j] - isLeap;
      Result := True;
      exit;
		end;
  end;
	ipDayPtr   := 0;
	ipMonthPtr := 0;
	Result := False;
end;

{==============================================================================
	JulianToDT
	Converts a long Julian date to TDateTime
 ==============================================================================}
function JulianToDT( dJulian: Double ): TDateTime;
var
  iYear,
  iMonth,
  iDay,
  iNumDays,
  iMaxDays,
	lTotDays: Integer;

  // Left Pad Values with Zeros
  function PadZ( iVal, iLen: Integer ): String;
  begin
    Result := IntToStr( iVal );
    while Length( Result ) < iLen do
      Result := '0' + Result;
  end;

begin
  // null date in an index is stored as 1.0E100
  // if you don't want 1.0E100 then check the value
  // you are converting for 0 first and act
  // appropriately instead of calling this function
  // ----------------------------------------------
  if (dJulian = 1.0E100) or (dJulian = 0) then
	begin
  	Result := 0; // '        ';'
    exit;
  end;

  lTotDays := Trunc( dJulian ) - JULIAN_ADJUSTMENT;
  iYear    := Trunc( lTotDays / 365.2425 ) + 1;
  iNumDays := lTotDays - YearToJulian( iYear );
  if (iNumDays <= 0) then
  begin
		Dec( iYear );
		iNumDays := (lTotDays - YearToJulian(iYear));
	end;

  if (iYear mod 4 = 0) and (iYear mod 100 <> 0) or (iYear mod 400 = 0) then
		iMaxDays := 366
  else
		iMaxDays := 365;

  if (iNumDays > iMaxDays) then
	begin
		Inc( iYear );
		Dec( iNumDays, iMaxDays );
	end;

	if ( MonthDay(iYear, iNumDays, iMonth, iDay) = False ) then
		raise EApolloError.Create(ERR_INVALID_DATE_VALUE );

	Result := EncodeDate( iYear, iMonth, iDay );
end;


{==============================================================================
	JulianDayYear
   Returns an (Integer) day of the year starting from 1.
   Ex.    Jan 1, returns  1

   Returns -1  if it is an illegal date.
 ==============================================================================}
function JulianDayYear( iYear, iMonth, iDay: Integer ): Integer;
var
	iIsLeap,
	iMonthDays: Integer;
begin
	if (iYear mod 4 = 0) and (iYear mod 100 <> 0) or (iYear mod 400 = 0) then
		iIsLeap := 1
  else
  	iIsLeap := 0;

  iMonthDays := iaMonthTot[iMonth + 1] - iaMonthTot[iMonth];
	if (iMonth = 2) then
		Inc( iMonthDays, iIsLeap );

  if (iYear < 0) or (iMonth < 1) or (iMonth > 12) or (iDay  < 1) or
                                                     (iDay > iMonthDays) then
	begin
		Result := -1;  // Illegal Date
		exit;
  end;

  if (iMonth <= 2) then
		iIsLeap := 0;

	Result := (iaMonthTot[iMonth] + iDay + iIsLeap);
end;

{==============================================================================
	DTtoJulian
   Converts from TDateTime to Julian

   Can be used for checking the validity of a date.

   Returns
		>0 -  Valid date
    -1 -  Illegal Date
    -2 -  NULL Date (dbf_date is all blank; converts to 1.0E100)
 ==============================================================================}
function DTToJulian( DT: TDateTime ): Double;
var
	iYear,
	iMonth,
	iDay: Word;
	iDayYear : Integer;
begin
  DecodeDate( DT, iYear, iMonth, iDay );

  if (iYear = 0) then
  begin
		Result := -2;
    exit;
  end;

  if (iYear < 1000) then
	begin
   	Result := -1;
    exit;
  end;

	iDayYear := JulianDayYear( iYear, iMonth, iDay );
	if (iDayYear < 1) then
		Result := -1
	else
		Result := YearToJulian(iYear) + iDayYear + JULIAN_ADJUSTMENT;
end;

{==============================================================================
  GetApVersion - Returns Version Info for SDE .DLLs and ApolloDataSet
 ==============================================================================}
function TApolloTable.GetApVersion: String;
begin
	if (FAccessMethod = amServer) then
  begin
  	if ((FApolloConnection <> nil) and FApolloConnection.Connected) then
			Result := cs_Version     // Server connection found and open
    else
    	Result := 'SDE ?.??.?? 32-bit';  // Otherwise, no .DLL version info
  end
  else
		Result := sx_Version;  // Return Local SDE DLL version info

	Result := Result + ', ' + GetApDSVer;
end;


{==============================================================================
  SetApolloConnection for Client/Server
 ==============================================================================}
procedure TApolloTable.SetApolloConnection( Val: TApolloConnection );
begin
	if FApolloConnection <> Val then
	begin
		FApolloConnection := Val;
    if csDesigning in ComponentState then
      if FApolloConnection = nil then
        FAccessMethod := amLocal
      else
        FAccessMethod := amServer;
	end;
end;


{==============================================================================
  TApolloTable
 ==============================================================================}
{==============================================================================
  Create
 ==============================================================================}
constructor TApolloTable.Create(Owner: TComponent);
begin
  inherited;

	// Set Up Defaults
  FTranslate      := FALSE;
  FCommitLevel    := clNormal;  

  iWA := 0;              // No active work area
	FAppending := False;   // In Append / AppendBlank flag
  FDataPath := '';       // Physical path of table (DatabaseName location)
  FCursorOpen := False;  // Table cursor open flag
  FSequenced := True;    // Proportional scroll bar in DBGrid control
  FTableType := ttSXFOX; // Default table type to FoxPro

	FLastIndexNameVar := '';     // For saving last IndexName value
	FLastScopeVar := '';         // For saving last Scope value
	FScopeBegin := '';           // Low scope value
  FScopeEnd := '';             // High scope value
  FScoped := False;            // Is SetScope active
  FWriteBlobHdr := True;       // SetWriteBlobHeader property default
  FFilterStr := '';            // Temporary holder for active filter string
  FLocateIndexOrd := 0;        // LocateIndexOrd property default
  FIndexOrd       := 0;        // IndexOrd default
  FIndexCondition := '';       // Index condition default
  FOptimisticCommit := True;   // OptimisticCommit property default
  FRowCount := -1;             // Initial value used by GetRecNo calculation
	FRecCount := -1;             // Initial total RecCount value  {LAS: 05-30-00, Added}
	// FRecOrdinal := -1; CRVG 2001-11-23          // Initial Ordinal Record position  {LAS: 06-01-00, Added}
  fFRecOrdinal := -1;           // Initial Ordinal Record position  {LAS: 06-01-00, Added}
	FCountNeedsRefresh := True;  // Indicates if FCount needs to be refreshed
  FSequencedCount := scRawCount;    // Default value for SequencedCount property

  FEOFAction := eaMoveLast;    // Default value for EOFAction property
  FAccessMethod := amLocal;    // Default value for AccessMethod property
	FEpoch := EPOCH_DEFAULT;     // Default value for Epoch setting
  FErrorLevel := 1;            // Default Error Level
  FErrorFunc := -1;            // Error Func Flag
  FCachedUpdates := False;     // Default value for Cached Updates
  FCentury := True;            // Default Century setting  {LAS: 04-05-00, Added}
  FFetchCount := 50;           // Default FetchCount value   {LAS: 05-26-00, Raised from 1 to 50}
  FRangeSet := rsNone;         // Range Set Mode  {LAS: 05-23-00, Added}

  FErrorMsg    := '';

{$ifdef NOT_D3_OR_CB3}
	FUpdateLockList := True;      // Indicates if RLock List needs refreshing
{$endif}

  // Create required objects
  FExtraIndexes := TStringList.Create;    // ExtraIndexes property
  FExtraIndexes.OnChange := ExtraIndexesChanged;

  LFieldOffsets := TList.Create;          // Field Offsets within rec buf
  FIndexDefs := TIndexDefs.Create(Self);  // IndexDefs

  FApolloConnection := nil;

  FRecords      := TList.Create;
  FBackRecords  := TList.Create;
  FFetchRecs    := TList.Create;
  FFetchLastRec := 0;
  FFetchReadCount:= -1;
//  FFetchSavedDirection := 0;
  FStructList   := TStringList.Create;

  FMasterLink := TMasterDataLink.Create(Self);
  FMasterLink.OnMasterChange := MasterChanged;
  FMasterLink.OnMasterDisable := MasterDisabled;

end;


{==============================================================================
 Loaded
==============================================================================}
procedure TApolloTable.Loaded;
begin
  inherited;

	if FAccessMethod = amLocal then
  	LoadDLL;

	if Assigned( FApolloConnection ) then
   	FApolloConnection.DataBaseName := FDatabaseName;
end;

{==============================================================================
  Destroy
 ==============================================================================}
destructor TApolloTable.Destroy;
begin
  inherited;

	if (iWA > 0) or Active then     {LAS: 07-25-00, Added -- just in case}
  begin
  	Close;                        {LAS: 07-25-00, Added -- just in case}
  end;

  while LFieldOffsets.Count > 0 do
  begin
    TObject(LFieldOffsets[0]).Free;
    LFieldOffsets.Delete(0);
  end;
	FExtraIndexes.Free;
  FIndexDefs.Free;
  LFieldOffsets.Free;
  FMasterLink.Free;

  //ClearCachedItems;

  FRecords.Free;
  FRecords := nil;    

  FBackRecords.Free;
  FBackRecords := nil;

  ClearFetchedItems;

  FFetchRecs.Free;
  FFetchRecs := nil;

  FStructList.Free;
  FStructList := nil;
end;

{==============================================================================
  DoBeforeEdit
 ==============================================================================}
procedure TApolloTable.DoBeforeEdit;
begin
	inherited;
	if (iWA > 0) then
  begin
		cs_Select( iWA );

    if FSpeedMode then
    	FRecNo := cs_RecNo
    else
			FRecNo := PRecInfo(ActiveBuffer + FRecInfoOfs).Bookmark;

		if not cs_RLock( FRecNo ) and (FErrorFunc = -1) then
			RaiseApollo( ERR_RECORD_LOCK_FAILURE );

  	// Check for invalid workarea
		if FAccessMethod = amLocal then  {LAS: 07-14-00, Added}
	   	if cs_Alias(iWA) = '' then     {LAS: 07-14-00, Added}
    		exit;                        {LAS: 07-14-00, Added}

		// Only do these steps when Not editing a newly-inserted record
		if not (FRecNo = -1) then
    begin
      // The following three lines force a refresh of the record buffer in case
      // any other applications or processes have edited this record since it
      // was read into the buffer.  Previously used a call to Resync, but this
      // would cause a DBGrid cell to clear out when you double-clicked into it
      // and then pressed the End key, which should have just moved you to the
      // end of the existing data.
      cs_Go( FRecNo );
      cs_GetRecord( LSDERecordBuffer );
      ReadSDEToNative( ActiveBuffer );
		end;
	end;
end;

{==============================================================================
  DoBeforeInsert
 ==============================================================================}
procedure TApolloTable.DoBeforeInsert;
begin
	inherited;
	if (iWA > 0) then
  begin
		cs_Select( iWA );
		FOldRec := PRecInfo(ActiveBuffer + FRecInfoOfs).Bookmark;
	end;
end;

{==============================================================================
  DoAfterInsert
 ==============================================================================}
procedure TApolloTable.DoAfterInsert;
begin
	if State = dsInsert then   {LAS: 05-08-00, Added}
  begin                      {LAS: 05-08-00, Added}
    cs_Select( iWA );
    with PRecInfo(ActiveBuffer + FRecInfoOfs)^ do
    begin
      BookmarkFlag := bfEOF;
      Bookmark := -1;
    end;
    inherited;
  end;                       {LAS: 05-08-00, Added}
end;

{==============================================================================
  DoOnNewRecord
 ==============================================================================}
procedure TApolloTable.DoOnNewRecord;
var
	Fields: TList;
  i: Integer;
  iCount: Integer;
begin
	if FMasterLink.Active and (FMasterLink.Fields.Count > 0) then
  begin
  	Fields := TList.Create;
    GetFieldList( Fields, GetFieldsFromExpression( IndexKey ));
    iCount := FMasterLink.Fields.Count;
		for i := 0 to (iCount-1) do
			FieldByName( TField(Fields[i]).FieldName ).Value :=
                     FMasterLink.DataSet.FieldByName(
                       TField(FMasterLink.Fields[i]).FieldName ).Value;
		Fields.Free;
  end;
	inherited;
end;

{==============================================================================
  DoAfterCancel
 ==============================================================================}
procedure TApolloTable.DoAfterCancel;
begin
	cs_Select( iWA );
  {
	if cs_Locked( cs_RecNo ) then       // MED: 10-24-00, commented- see Cancel below
		cs_Unlock( 0 );
   }
  // Properly reposition cursor after a cancelled Insert
	if (FOldRec > 0) then
  begin
		with PRecInfo(ActiveBuffer + FRecInfoOfs)^ do
    begin
      BookmarkFlag := bfCurrent;
      Bookmark := FOldRec;
    end;

		FOldRec := 0;
    Resync([]);
	end;
	inherited;
end;

{==============================================================================
  Cancel
  MED: 10-24-00, added
 ==============================================================================}
procedure TApolloTable.Cancel;
begin
	inherited;
	cs_Select( iWA );
	cs_Unlock( 0 );  // unlocks all
end;

{==============================================================================
  DoAfterScroll
 ==============================================================================}
procedure TApolloTable.DoAfterScroll;
begin
	if (iWA > 0) then
    UpdateCursorPos;
	inherited;
end;

{==============================================================================
  SetToApolloEnv
 ==============================================================================}
function TApolloTable.SetToApolloEnv: TApolloEnv;
var
  Temp: TComponent;
  SearchObj : TApolloSearchObject;  {LAS: 04-20-00, Added}
begin
  Result := nil;
  FApplyEnvSettings  := FALSE;

	{LAS: 04-20-00, Added--------------------------------}
  SearchObj := TApolloSearchObject.Create;
  try
    Temp := SearchObj.SearchFirstComponent(TApolloEnv);
  finally
    SearchObj.Free;
  end;
	{LAS: 04-20-00, Added--------------------------------}

  if Temp <> nil then
  begin
    if (Temp is TApolloEnv) then
    begin
      Result             := TApolloEnv(Temp);
      if ((FAccessMethod = amServer ) and (atServer in Result.ApplyTo) )
              OR
         ((FAccessMethod = amLocal )  and (atLocal in Result.ApplyTo) ) then
      begin

        FEpoch             := Result.Epoch;
        FErrorLevel        := Result.GetErrorLevel;
        FCountNeedsRefresh := Result.CountNeedsRefresh;
        FErrorFunc         := Result.ErrorFunc;
        FCentury           := Result.Century;   {LAS: 04-05-00, Added}
        FExact             := Result.Exact;
        FSoftSeek          := Result.Softseek;
        FDateFormat        := Result.DateFormat;
        FMemoBlockSize     := Result.MemoBlockSize;
        FDeleted           := Result.Deleted;
        FApplyEnvSettings  := TRUE;
      end;
      // Moved into InternalOpen
    end;
  end;
end;

{==============================================================================
  ApolloBOF
 ==============================================================================}
function TApolloTable.GetApolloBOF: WordBool;
begin
  Result := False;
  if (iWA > 0) then
  begin
    if FSpeedMode then
    begin
      cs_Select( iWA );
      Result := cs_Bof;
    end
    else
      Result := BOF;
  end;
end;

{==============================================================================
  ApolloEOF
 ==============================================================================}
function TApolloTable.GetApolloEOF: WordBool;
begin
  Result := False;
  if (iWA > 0) then
  begin
    if FSpeedMode then
    begin
      cs_Select( iWA );
      Result := cs_EOF;
    end
    else
      Result := EOF;
  end;
end;

{LAS: 04-17-00, Added}
{==============================================================================
  BOF
 ==============================================================================}
function TApolloTable.GetBOF: Boolean;
begin
	if FSpeedMode then
  begin
    Result := False;
    if (iWA > 0) then
    begin
      cs_Select( iWA );
      Result := cs_BOF;
    end;
    if Result = False then
      Result := inherited BOF;
  end
  else
		Result := inherited BOF;
end;

{==============================================================================
  EOF
 ==============================================================================}
function TApolloTable.GetEOF: Boolean;
begin
	Result := False;
  if (iWA > 0) then
  begin
		cs_Select( iWA );
		Result := cs_EOF;
	end;
	if Result = False then
  	Result := inherited EOF;
end;

{==============================================================================
  ApolloGetRecord - Renamed from Apollo.GetRecord to avoid naming conflict with
                    protected GetRecord method.
 ==============================================================================}
procedure TApolloTable.ApolloGetRecord( cpRecord: PChar );
begin
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
    UpdateCursorPos;    {LAS: 06-27-00, Added}
		cs_GetRecord( cpRecord );
	end;
end;

{==============================================================================
  Alias
 ==============================================================================}
function TApolloTable.Alias : String;
begin
	Result := '';
	if (iWA > 0) then
		Result := cs_Alias( iWA );
end;

{==============================================================================
  Append
 ==============================================================================}
procedure TApolloTable.Append;
begin
	if (FSpeedMode = False) then
		inherited
	else
	begin
		// Can't APPEND a ReadOnly table
		if ReadOnly and (FErrorFunc = -1) then
	  	RaiseApollo( ERR_TABLE_IS_READONLY );

		if (iWA > 0) then
	  begin
			cs_Select( iWA );
			cs_AppendBlank;
	    FAppending := True;
		end;
	end;
	FCountNeedsRefresh := True;  {LAS: 05-30-00, Added}
end;

{==============================================================================
  AppendBlank
 ==============================================================================}
procedure TApolloTable.AppendBlank;
begin
	if (FSpeedMode = False) then
		Append
	else
  begin
		// Can't APPENDBLANK a ReadOnly table
		if ReadOnly and (FErrorFunc = -1) then
  		RaiseApollo( ERR_TABLE_IS_READONLY );

		if (iWA > 0) then
  	begin
			cs_Select( iWA );
			cs_AppendBlank;
  	  FAppending := True;
		end;
	end;
  FCountNeedsRefresh := True;  {LAS: 05-30-00, Added}
end;

{==============================================================================
  AppendFrom
 ==============================================================================}
function TApolloTable.AppendFrom( sFileName: String;
                         iSourceType: SmallInt; sScopeExpr: String ): WordBool;
begin
	Result := False;

	// Can't modify a ReadOnly table
	if ReadOnly  and (FErrorFunc = -1) then
  	RaiseApollo( ERR_TABLE_IS_READONLY );

	if (iWA > 0) then
  begin
		cs_Select( iWA );
    BeginScroll;
		Result := cs_AppendFrom( PChar( sFileName ), iSourceType,
  	                         PChar( sScopeExpr ));

    //		cs_Commit;  // MED: 03-25-02, see cs_AppendFrom. // REQUIRED!! Without this, last record is not flushed
		FCountNeedsRefresh := True;
		Resync([]);
    EndScroll;
	end;
end;

{==============================================================================
  ApolloDelete
 ==============================================================================}
procedure TApolloTable.ApolloDelete;
begin
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
		cs_Delete;
    Resync([]);
	end;
end;

{==============================================================================
  ApolloLocate - Locates the FIRST matching record...
 ==============================================================================}
function TApolloTable.ApolloLocate( Expression: String;
                             Backwards: WordBool; Continue: WordBool): Integer;
begin

	Result := 0;
	if (iWA > 0) then
  begin
    BeginScroll;
	  cs_Select( iWA );
  	Result := cs_Locate( PChar(Expression), Backwards, Continue );
	  if (Result > 0) then
  	begin
	    LLocateExpression := Expression;
  	  LLocateDirection := Backwards;

      if not FSpeedMode then
      begin
        with PRecInfo(ActiveBuffer + FRecInfoOfs)^ do
        begin
          BookmarkFlag := bfCurrent;
          Bookmark := Result;
        end;
      end;

    	Resync([]);
	    EndScroll;
	  end
  	else
    	LLocateExpression := '';
	end;
end;

{==============================================================================
  ApolloLocateNext - Locates the NEXT matching record after ApolloLocate call
 ==============================================================================}
function TApolloTable.ApolloLocateNext: Integer;
begin
  Result := ApolloLocate(PChar(LLocateExpression), LLocateDirection, True);
end;

{==============================================================================
  BaseDate
 ==============================================================================}
function TApolloTable.BaseDate: String;
begin
	Result := '';
	if (iWA > 0) then
  begin
		cs_Select( iWA );
		Result := String( cs_BaseDate );
	end;
end;

{==============================================================================
  BaseName
 ==============================================================================}
function TApolloTable.BaseName: String;
begin
	Result := '';
	if (iWA > 0) then
  begin
		cs_Select( iWA );
		Result := String( cs_BaseName );
	end;
end;

{==============================================================================
  BlobToFile
 ==============================================================================}
function TApolloTable.BlobToFile( sFieldName, sFileName: String ): WordBool;
begin
	Result := False;
	if (iWA > 0) then
  begin
		cs_Select( iWA );
		Result := cs_BlobToFile( PChar( sFieldName ), PChar( sFileName ));
	end;
end;

{==============================================================================
  CancelRange
 ==============================================================================}
procedure TApolloTable.CancelRange;
begin
	SetScope( '', '' );
end;

{==============================================================================
  CloseIndexes
 ==============================================================================}
procedure TApolloTable.CloseIndexes;
begin
  if (iWA > 0) then
  begin
    cs_Select( iWA );
    cs_CloseIndexes;
    IndexName := '';
    IndexDefs.Updated := False;
    UpdateIndexDefs;
		FCountNeedsRefresh := True;
    Resync([]);
  end;
end;

{==============================================================================
  Commit
 ==============================================================================}
procedure TApolloTable.Commit;
begin
  if (FSpeedMode = False) then
  begin
    if State in [dsEdit, dsInsert] then
      Post;
  end
  else
  begin
    // Can't modify a ReadOnly table
    if ReadOnly and (FErrorFunc = -1) then
      RaiseApollo( ERR_TABLE_IS_READONLY );

    if (iWA > 0) then
    begin
      cs_Select( iWA );
      cs_Commit;
      FAppending := False;
    end;
  end;
end;

{==============================================================================
  CopyFile
 ==============================================================================}
function TApolloTable.CopyFile(sToFileName: String ): LongBool;  {LAS: 06-23-00, Change return type from WordBool}
begin
	Result := False;
	if (iWA > 0) then
	begin
		cs_Select( iWA );
		Result := cs_CopyFile( PChar( sToFileName ));
	end;
end;

{==============================================================================
  CopyFileExt
 ==============================================================================}
function TApolloTable.CopyFileExt( sToFileName: String ): LongBool;  {LAS: 06-23-00, Change return type from WordBool}
begin
	Result := False;
	if (iWA > 0) then
	begin
		cs_Select( iWA );
		Result := cs_CopyFile( PChar( sToFileName ));
	end;
end;


{==============================================================================
  CopyFileText
 ==============================================================================}
function TApolloTable.CopyFileText( sTextFileName: String;
                                      iFileType: SmallInt ): LongBool;  {LAS: 06-23-00, Changed return type from WordBool}
begin
	Result := False;
	if (iWA > 0) then
  begin
		cs_Select( iWA );
		Result := cs_CopyFileText( PChar( sTextFileName ), iFileType );
	end;
end;

{==============================================================================
  CopyStructure
 ==============================================================================}
function TApolloTable.CopyStructure( sFileName: String ): LongBool;  {LAS: 06-23-00, Changed return type from WordBool}
var
	bWasClosed: Boolean;
begin
	// Save current Active state; Open if table is currently closed
	bWasClosed := (not Active);
	if bWasClosed then
  	Open;

	Result := False;
	if (iWA > 0) then
  begin
		cs_Select( iWA );
		Result := cs_CopyStructure( PChar( sFileName ),
                                PChar( GetUniqueAliasName( sFileName )));
	end;

  // If table was originally closed, close it again
  if bWasClosed then
  	Close;
end;

{==============================================================================
  CopyStructureExtended
 ==============================================================================}
function TApolloTable.CopyStructureExtended( sFileName: String ): LongBool;  {LAS: 06-23-00, Changed return type from WordBool}
var
	dbFieldItem : TDbfStruct;
	i : Integer;
	iUserDBF : Integer;
	iDBF : Integer;
	iFCount : Integer;
	cpName: Array[0..10] of Char;
  cpType: Array[0..1] of Char;
	bWasClosed: Boolean;
begin
	// Save current Active state; Open if table is currently closed
	bWasClosed := (not Active);
	if bWasClosed then
  	Open;

	Result := False;
	if (iWA > 0) then
  begin
		cs_Select( iWA );

//---------------------------------------------------------------------------//
// NOTE: The following single & simple line of code has been replaced with all
//       of the other code below solely for the purpose of hiding SDE's internal
//       long field name support, which is not fully supportable under Apollo.
//
//	Result := cs_CopyStructureExtended( PChar( sFileName ));
//
//---------------------------------------------------------------------------//

		iFCount := cs_FieldCount;
		iUserDBF := iWA;
		iDBF := cs_CreateNew( PChar( sFileName ),
                         PChar( GetUniqueAliasName( sFileName )),
                         ApolloTableType( FTableType ), 4);
		if (iDBF > 0) then
		begin
    	StrPCopy( cpName, 'FIELD_NAME' );
      StrPCopy( cpType, 'C' );
      cs_CreateField( cpName, cpType, 10, 0 );
    	StrPCopy( cpName, 'FIELD_TYPE' );
      cs_CreateField( cpName, cpType,  1, 0 );
    	StrPCopy( cpName, 'FIELD_LEN' );
      StrPCopy( cpType, 'N' );
      cs_CreateField( cpName, cpType,  5, 0 );
    	StrPCopy( cpName, 'FIELD_DEC' );
      cs_CreateField( cpName, cpType,  3, 0 );
		end;

		if (cs_CreateExec) then
		begin
			for i := 0 to iFCount - 1 do
			begin
				cs_Select( iUserDBF );
				dbFieldItem.sName := StrPas( cs_FieldName( i+1 ));
				dbFieldItem.sType := StrPas( cs_FieldType( cs_FieldName( i+1 )));
				dbFieldItem.iLength := cs_FieldWidth( cs_FieldName( i+1 ));
				dbFieldItem.iDecimals := cs_FieldDecimals(cs_FieldName( i+1 ));
				cs_Select(iDBF);
				cs_AppendBlank;
        cs_Replace( PChar('FIELD_NAME'), R_CHAR, PChar( dbFieldItem.sName ));
				cs_Replace( PChar('FIELD_TYPE'), R_CHAR, PChar( dbFieldItem.sType ));
        cs_Replace( PChar('FIELD_LEN'), R_INTEGER, @dbFieldItem.iLength );
				cs_Replace( PChar('FIELD_DEC'), R_INTEGER, @dbFieldItem.iDecimals );
				cs_Commit;
			end;
			cs_Select( iDBF );
			cs_Close;
			cs_Select( iUserDBF );
			Result := True;
		end
		else
			Result := False;
  end;

	// If table was originally closed, close it again
  if bWasClosed then
  	Close;
end;

{==============================================================================
  Count
 ==============================================================================}
function TApolloTable.Count: Integer;
begin
	Result := 0;
	if (iWA > 0) then
  begin
		cs_Select( iWA );
		Result := cs_Count;
	end;
end;

{==============================================================================
  CreateExec
 ==============================================================================}
function TApolloTable.CreateExec: LongBool;   {LAS: 06-23-00, Changed return type from WordBool}
begin
  if (not (iWA < 0)) and (FErrorFunc = -1) then
		RaiseApollo( ERR_CREATENEW_BEFORE_CREATEEXEC );
  cs_Select(-iWA);
  iWA := 0;
  Result := cs_CreateExec;
  cs_Close;
	FFetchReadCount := -1;      {LAS: 07-11-00, Added}
end;

{==============================================================================
  CreateField
 ==============================================================================}
procedure TApolloTable.CreateField( FieldName: String;
                    FieldType: String; FieldLen: SmallInt; FieldDec: SmallInt);
var
  CharFieldName : Array[0..63] of Char;
  CharFieldType : Array[0..1] of Char;
begin
  if (not (iWA < 0)) and (FErrorFunc = -1) then
		RaiseApollo( ERR_CREATENEW_BEFORE_CREATEFIELD );

	// Check that FieldName is not empty   {LAS: 07-16-00, Added}
	if (FieldName = '') then
  	RaiseApollo( ERR_INVALID_FIELD_NAME );

	// Check that FieldType is not empty   {LAS: 07-16-00, Added}
	if (FieldType = '') then
  	RaiseApollo( ERR_INVALID_FIELD_TYPE );

	// Check that FieldLen is not greater than zero   {LAS: 07-16-00, Added}
	if (FieldLen <= 0) then
  	RaiseApollo( ERR_INVALID_FIELD_LEN );

  cs_Select(-iWA);
  StrPCopy( CharFieldType, FieldType );
  StrPCopy( CharFieldName, FieldName );

	// Check that first character is a letter
	if not (CharFieldName[0] in ['A'..'Z','a'..'z']) then
  	RaiseApollo( ERR_INVALID_FIELD_NAME );

  cs_CreateField( @CharFieldName, @CharFieldType, FieldLen, FieldDec );
end;

{==============================================================================
  CreateFrom
 ==============================================================================}
function TApolloTable.CreateFrom( sNewFileName, sStruExtFileName: String;
                                    iRdeType: SmallInt ): LongBool;   {LAS: 06-23-00, Changed return type from WordBool}
var
  cpAlias: Array[0..10] of Char;
  i : Integer;
begin
  Result := FALSE;
	// Does not require an active workarea to do it's business, so just call it!
	if cs_FileExists( sStruExtFileName ) then
  begin

    // This code (for some reason) prevents a 3150 Error after AppendFrom call
    if (Active) then
    begin
			InternalClose;
			OpenTable;
    end;

  	StrPCopy( cpAlias, GetUniqueAliasName( sNewFileName ));
		Result := cs_CreateFrom( PChar( sNewFileName ), cpAlias,
    	                       PChar( sStruExtFileName ), iRdeType );
		{ Make sure newly created table is closed }
		i := cs_WorkArea( cpAlias );
  	if i > 0 then
	  begin
			cs_Select( i );
			cs_Close;
		end;
	end
  else
  	RaiseApollo( ERR_TABLE_DOES_NOT_EXIST + sStruExtFileName );
end;

{==============================================================================
  CreateNew
 ==============================================================================}
function TApolloTable.CreateNew( FileName: String; RDEType: SmallInt;
                                   NumFields: SmallInt): WordBool;
var
  sPath: String;
  cpAliasName,
  cpFileName: Array[0..127] of Char;
  FAliasName : String;
begin
  if AccessMethod = amLocal then                
    FAliasName := GetUniqueAliasName( FileName );

  // Get the path to the table...
  if (ExtractFilePath(FileName) = '') then
  begin
  	if Assigned( FDatabase ) then
	    sPath := Database.DataPath
		else
    	sPath := DatabaseName;

		// If no path was specified, use current directory
		if (sPath = '') then
    	GetDir( 0, sPath );
    if (Length(sPath) > 0) and (sPath[ Length(sPath) ] <> '\') then
    	sPath := sPath + '\';
  end
  else
  	sPath := ExtractFilePath( FileName );

  StrPCopy( cpFileName, sPath + ExtractFileName( FileName ));
  StrPCopy( cpAliasName, FAliasName );

  // Set it as negative so we know if we're in "CreateNew" mode or not.
  iWA := -cs_CreateNew( @cpFileName, @cpAliasName, RDEType, NumFields );

 	Result := (iWA < 0);
end;


{==============================================================================
  DbfDecrypt
 ==============================================================================}
function TApolloTable.DbfDecrypt: WordBool;
begin
	Result := False;

 	// Can't modify a ReadOnly table
	if ReadOnly and (FErrorFunc = -1) then
  	RaiseApollo( ERR_TABLE_IS_READONLY );

	if (iWA > 0) then
  begin
    cs_Select( iWA );
    Result := cs_DbfDecrypt;
    Resync([]);

		// Is table encrypted?
    FIsEncrypted := IsEncrypted(0);
	end;
end;

{==============================================================================
  DbfEncrypt
 ==============================================================================}
function TApolloTable.DbfEncrypt: WordBool;
begin
	Result := False;

	// Can't modify a ReadOnly table
	if ReadOnly and (FErrorFunc = -1) then
  	RaiseApollo( ERR_TABLE_IS_READONLY );

	if (iWA > 0) then
  begin
    cs_Select( iWA );
    Result := cs_DbfEncrypt;
    Resync([]);

		// Is table encrypted?
    FIsEncrypted := IsEncrypted(0);
	end;
end;

{==============================================================================
  DBFilter
 ==============================================================================}
function TApolloTable.DBFilter: String;
begin
	Result := '';
	if (iWA > 0) then
  begin
		cs_Select( iWA );
		Result := String( cs_DBFilter );
	end;
end;

{$ifdef NOT_D3_OR_CB3}
{==============================================================================
  GetLockList
 ==============================================================================}
function TApolloTable.GetLockList(Index: Integer): Integer;
begin
	Result := 0;
	if (iWA > 0) then
  begin
		cs_Select( iWA );

    // Only take the time to get the lock list if it may have changed
    if FUpdateLockList then
    begin
			SetLength(FLockList, cs_LockCount);
			if High( FLockList ) >= 0 then
      begin
				cs_DBRLockList( PChar(FLockList));  // Retrieve Lock List
    	  FUpdateLockList := False;
      end;
    end;

    // Return zero if no records locked or invalid index number
		if (High( FLockList ) >= 0) and (High( FLockList ) >= Index) then
			Result := FLockList[Index];
	end;
end;
{$endif}                          

{==============================================================================
  Decrypt
 ==============================================================================}
function TApolloTable.Decrypt( cpBuffer: PChar; sPassWord: String;
                                 iLen : Integer ): PChar;
begin
	Result := cs_Decrypt( cpBuffer, PChar( sPassword ), iLen );
end;


{==============================================================================
  Delete
 ==============================================================================}
procedure TApolloTable.Delete;
begin
	CheckBrowseMode;   
  DoBeforeDelete;
	DoBeforeScroll;
  InternalDelete;
  DoAfterDelete;
	DoAfterScroll;
end;

{==============================================================================
  Deleted
 ==============================================================================}
function TApolloTable.GetDeleted: WordBool;
begin
	Result := False;
	if SelectForSxCall then
		Result := cs_Deleted;
end;

{==============================================================================
  DeleteTable
 ==============================================================================}
procedure TApolloTable.DeleteTable;
var
	sFileName : String;
begin
 	Close;

  sFileName := GetFullTableName;
  IndexName := '';

	if cs_DeleteFile( sFileName ) then
  begin
    sFileName := ChangeFileExt( sFileName, GetMemoExt );
    if cs_FileExists( sFileName ) then
      cs_DeleteFile( sFileName );

    // Delete any structural .CDX/.NSX
		if (FTableType <> ttSXNTX) then
    begin
      sFileName := ChangeFileExt( sFileName, GetIndexExt );
      if cs_FileExists( sFileName ) then
        cs_DeleteFile( PChar( sFileName ));
		end;
  end;
end;

{==============================================================================
  Descend
 ==============================================================================}
function TApolloTable.Descend( Value: String ): String;
begin
	Result := String( cs_Descend( PChar( Value )));
end;

{==============================================================================
  Empty
 ==============================================================================}
function TApolloTable.Empty( Value: String ): WordBool;
begin
	Result := False;
	if SelectForSxCall then
		Result := cs_Empty( PChar( Value ));
end;

{==============================================================================
  EmptyTable - Emulates TTable method
 ==============================================================================}
procedure TApolloTable.EmptyTable;
var
	bExclusive : Boolean;
  bIsOpen : Boolean;
begin
	// Save current Active and Exclusive settings
	bExclusive := Exclusive;
	bIsOpen := Active;

  // Set Exclusive to True, if not already
	if not bExclusive then
  begin
  	if Active then
		  Close;
  	Exclusive := True;
	end;

  // Open table, if not already open
  if not Active then
  	Open;

  Zap;   // Empty table

  // Reset Exlusive and Active status
  if (not bIsOpen) or (not bExclusive) then
  begin
	  Close;
  	Exclusive := bExclusive;
		Active := bIsOpen;
	end;
end;

{==============================================================================
  Encrypt
==============================================================================}
function TApolloTable.Encrypt( cpBuffer: PChar; sPassWord: String;
                                 iLen: Integer ): PChar;
begin
	Result := cs_Encrypt( cpBuffer, PChar( sPassword ), iLen );
end;

{==============================================================================
  EvalLogical
 ==============================================================================}
function TApolloTable.EvalLogical( Value: String ): WordBool;
begin
	Result := False;
	if SelectForSxCall then
    Result := cs_EvalLogical( PChar( Value ));
end;

{==============================================================================
  EvalNumeric
 ==============================================================================}
function TApolloTable.EvalNumeric( Value: String ): Double;
begin
	Result := 0;
	if SelectForSxCall then
		Result := cs_EvalNumeric( Value );  
end;

{==============================================================================
  EvalString
 ==============================================================================}
function TApolloTable.EvalString( Value: String ): String;
begin
	Result := '';
	if SelectForSxCall then
    Result := String( cs_EvalString( PChar( Value )));
end;

{==============================================================================
  EvalTest
 ==============================================================================}
function TApolloTable.EvalTest( Value: String ): SmallInt;
begin
	Result := 0;
	if (iWA > 0) then
  begin
		cs_Select( iWA );
    Result := cs_EvalTest( PChar( Value ));
	end;
end;

{==============================================================================
  Exclusive (SET)
 ==============================================================================}
procedure TApolloTable.SetExclusive( Value : WordBool );
begin
	if (Value <> FExclusive) then
  begin
		// Table must be closed first
		if (iWA > 0) then
	  	RaiseApollo( ERR_MUST_CLOSE_TABLE );

		FExclusive := Value
  end;
end;

{==============================================================================
  GetFullTableName
 ==============================================================================}
function TApolloTable.GetFullTableName: String;
begin
	if FAccessMethod = amLocal then
		GetDataPath;
	Result := FDataPath + ExtractFileName( TableName );

	if ExtractFileExt( Result ) = '' then
		Result := Result + '.DBF';
end;

{==============================================================================
  Exists  (GET)
 ==============================================================================}
function TApolloTable.GetExists: Boolean;
begin
  Result := cs_FileExists( GetFullTableName );  
end;

{==============================================================================
  ApolloFieldCount
 ==============================================================================}
function TApolloTable.ApolloFieldCount: Word;
begin
	Result := 0;
	if (iWA > 0) then
  begin
		cs_Select( iWA );
    Result := cs_FieldCount;
	end;
end;

{==============================================================================
  FieldDecimals
 ==============================================================================}
function TApolloTable.FieldDecimals( Value: String ): Word;
begin
	Result := 0;
	if (iWA > 0) then
  begin
		cs_Select( iWA );
    Result := cs_FieldDecimals( PChar( Value ));
	end;
end;

{==============================================================================
  FieldName
 ==============================================================================}
function TApolloTable.FieldName( Value: Integer ): String;
begin
	Result := '';
	if (iWA > 0) then
  begin
		cs_Select( iWA );
    Result := String( cs_FieldName( Value ));
	end;
end;

{==============================================================================
  FieldNum
 ==============================================================================}
function TApolloTable.FieldNum( Value: String ): Word;
begin
	Result := 0;
	if (iWA > 0) then
  begin
		cs_Select( iWA );
    Result := cs_FieldNum( PChar( Value ));
	end;
end;

{==============================================================================
  FieldOffset
 ==============================================================================}
function TApolloTable.FieldOffset( Value: String ): Word;
begin
	Result := 0;
	if (iWA > 0) then
  begin
		cs_Select( iWA );
    Result := cs_FieldOffset( PChar( Value ));
	end;
end;

{==============================================================================
  FieldType
 ==============================================================================}
function TApolloTable.FieldType( Value: String ): String;
begin
	Result := '';
	if (iWA > 0) then
  begin
		cs_Select( iWA );
    Result := String( cs_FieldType( PChar( Value )));
	end;
end;

{==============================================================================
  FieldWidth
 ==============================================================================}
function TApolloTable.FieldWidth( Value: String ): Word;
begin
	Result := 0;
	if (iWA > 0) then
  begin
		cs_Select( iWA );
    Result := cs_FieldWidth( PChar( Value ));
	end;
end;

{==============================================================================
  FilterAlias
 ==============================================================================}
function TApolloTable.FilterAlias( sAliasName,
                                           sFieldName: String ): WordBool;
begin
	Result := False;
	if (iWA > 0) then
  begin
		cs_Select( iWA );
    Result := cs_FilterAlias( PChar( sAliasName ), PChar( sFieldName ));
	end;
end;

{==============================================================================
  FilterDlg
 ==============================================================================}
function TApolloTable.FilterDlg( winHandle: HWnd;
                     sExpr, sCaption: String; bHasIndexList: WordBool ): String;
begin
	Result := '';
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
		Result := String( cs_FilterDlg( winHandle, PChar( sExpr ),
                      PChar( sCaption ), bHasIndexList ));
		Resync([]);
	end;
end;

{==============================================================================
  FindKey
 ==============================================================================}
function TApolloTable.FindKey( const KeyValues: Array of Const ): Boolean;

	// Convert TVarRec value into String to pass to Seek
	function ApVarToString( xVar : TVarRec ): String;  {LAS: 05-30-00, renamed from VarToString}
	begin
		case xVar.VType of
			vtBoolean:
				if xVar.VBoolean then
					Result := '.T.'
				else
					Result := '.F.';
			vtChar:
				Result := xVar.VChar;
			vtCurrency:
				Result := FloatToStr( xVar.VCurrency^ );
			vtExtended:
				Result := FloatToStr( xVar.VExtended^ );
			vtInteger:
				Result := IntToStr( xVar.VInteger );
			vtPChar:
				Result := StrPas( xVar.VPChar );
			vtString:
				Result := xVar.VString^;
  		vtPointer,
			vtAnsiString,
			vtWideString:
				Result := StrPas( PChar( xVar.VPointer ));
			vtVariant:
				Result := VarToStr( xVar.VVariant^ );
    else
			Result := '';
		end;
	end;

begin
  Result := False;
  if ((iWA > 0) and (FIndexName > '')) then
  begin
    cs_Select( iWA );
    Result := cs_Seek( PChar( ApVarToString( KeyValues[0] )));  {LAS: 05-30-00, renamed from VarToString}
    if Result then
    begin
      if not FSpeedMode then
      begin
        with PRecInfo(ActiveBuffer + FRecInfoOfs)^ do
        begin
          if Result then
            BookmarkFlag := bfCurrent
          else
            BookmarkFlag := bfEOF;
          Bookmark := cs_RecNo;
        end;
      end;
	    BeginScroll;
      Resync([]);
	    EndScroll;
    end
    else
    	cs_Go( FRecNo );   // Preserve original record position, if not found
  end
	else if (iWA = 0) then
		RaiseApollo( ERR_MUST_OPEN_TABLE )
	else if (FIndexName = '') then
		RaiseApollo( ERR_NO_INDEX_ACTIVE );
end;

{==============================================================================
  FLock
 ==============================================================================}
function TApolloTable.FLock: WordBool;
begin
	Result := False;
	if (iWA > 0) then
  begin
		cs_Select( iWA );
		Result := cs_FLock;
	end;
end;

{==============================================================================
  Found
 ==============================================================================}
function TApolloTable.Found: WordBool;
begin
	Result := False;
	if (iWA > 0) then
  begin
		cs_Select( iWA );
		Result := cs_Found;
	end;
end;

{==============================================================================
  GetBlob
 ==============================================================================}
function TApolloTable.GetBlob( sFieldName: String;
                                       vpVar: Pointer ): Integer;
begin
	Result := 0;
	if SelectForSxCall then
		Result := cs_GetBlob( PChar( sFieldName ), Pointer( vpVar ));
end;

{==============================================================================
  GetBlobLength
 ==============================================================================}
function TApolloTable.GetBlobLength( sFieldName: String ): Integer;
begin
	Result := 0;
	if SelectForSxCall then
		Result := cs_GetBlobLength( PChar( sFieldName ));
end;

{==============================================================================
  GetByte
 ==============================================================================}
function TApolloTable.GetByte( sFieldName: String ): Char;
var
	sVal : String;
begin
  Result := #0;
  if (iWA > 0) then
  begin
    if FSpeedMode then
    begin
      cs_Select( iWA );
			Result := cs_GetByte( PChar( sFieldName ))[0];
    end
    else
      with FieldByName( sFieldName ) do
      begin
        sVal := AsString;
        if sVal <> ''  then
					Result := sVal[1];
      end;
  end;
end;

{==============================================================================
  GetDateJulian
 ==============================================================================}
function TApolloTable.GetDateJulian( sFieldName: String ): Integer;
begin
	Result := 0;
	if SelectForSxCall then
 		Result := cs_GetDateJulian( PChar( sFieldName ));
end;

{==============================================================================
  GetDate
 ==============================================================================}
function TApolloTable.GetDate( sFieldName: String ): TDateTime;
var
	sDate: String;
begin
	Result := 0;
	if (iWA > 0) then
  begin
  	if FSpeedMode then
    begin
			cs_Select( iWA );
      sDate := cs_GetDateString( PChar( sFieldName ));
      if sDate[1] = ' ' then
      	Result := 0
      else
				Result := StrToDate( cs_GetDateString( PChar( sFieldName )));
		end
    else
			Result := FieldByName( sFieldName ).AsDateTime;
	end;
end;

{==============================================================================
  GetDateString
 ==============================================================================}
function TApolloTable.GetDateString( sFieldName: String ): String;
begin
	Result := '';
	if (iWA > 0) then
  begin
  	if FSpeedMode then
    begin
			cs_Select( iWA );
			Result := String( cs_GetDateString( PChar( sFieldName )));
		end
    else
			Result := DateToStr( FieldByName( sFieldName ).AsDateTime );
	end;
end;

{==============================================================================
  GetDouble
 ==============================================================================}
function TApolloTable.GetDouble( sFieldName: String ): Double;
begin
	Result := 0;
	if (iWA > 0) then
  begin
    if FSpeedMode then
    begin
			cs_Select( iWA );
			Result := cs_GetDouble( PChar( sFieldName ));
		end
    else
			Result := FieldByName( sFieldName ).AsFloat;
	end;
end;

{==============================================================================
  GetIndexFieldCount - Returns number of fields in active index expression
 ==============================================================================}
function TApolloTable.GetIndexFieldCount: Integer;
var
	sExpr: String;
  iPos: Integer;
begin
	Result := 0;
	if IndexName > '' then
  begin
    sExpr := GetFieldsFromExpression( IndexKey );
    iPos := Pos( ';', sExpr );
		while (iPos > 0) do
    begin
	  	Inc( Result );
      sExpr := Copy( sExpr, iPos+1, Length( sExpr )-iPos);
	    iPos := Pos( ';', sExpr );
    end;
  end;
end;

{==============================================================================
  GetIndexExt - Returns default (NOT current/active) index extension for driver
 ==============================================================================}
function TApolloTable.GetIndexExt: String;
begin
	case FTableType of
		ttSXFOX:    Result := '.CDX';
		ttSXNTX:    Result := '.NTX';
		ttSXNSX:    Result := '.NSX';
		ttSXNSX_DBT:Result := '.NSX';
	else
		Result := '.CDX';
	end;
end;

{==============================================================================
  GetMemoExt - Returns default memo fileextension for driver
 ==============================================================================}
function TApolloTable.GetMemoExt: String;
begin
	case FTableType of
		ttSXFOX:      Result := '.FPT';
		ttSXNTX:      Result := '.DBT';
		ttSXNSX:      Result := '.SMT';
		ttSXNSX_DBT:  Result := '.DBT';
	else
		Result := '.FPT';
	end;
end;

{==============================================================================
  GetInteger
 ==============================================================================}
function TApolloTable.GetInteger( sFieldName: String ): Integer;
begin
	Result := 0;
	if (iWA > 0) then
  begin
    if FSpeedMode then
    begin
      cs_Select( iWA );
			Result := cs_GetInteger( PChar( sFieldName ));
    end
    else
			Result := FieldByName( sFieldName ).AsInteger;
	end;
end;

{==============================================================================
  GetLogical
 ==============================================================================}
function TApolloTable.GetLogical( sFieldName: String ): WordBool;
begin
	Result := False;
	if (iWA > 0) then
  begin
    if FSpeedMode then
    begin
			cs_Select( iWA );
			Result := cs_GetLogical( PChar( sFieldName ));
		end
    else
			Result := FieldByName( sFieldName ).AsBoolean;
	end;
end;

{==============================================================================
  GetLong
 ==============================================================================}
function TApolloTable.GetLong( sFieldName: String ): Integer;
begin
	Result := 0;
	if (iWA > 0) then
  begin
		if FSpeedMode then
    begin
			cs_Select( iWA );
			Result := cs_GetLong( PChar( sFieldName ));
		end
    else
			Result := FieldByName( sFieldName ).AsInteger;
	end;
end;

{==============================================================================
  GetMemo
 ==============================================================================}
function TApolloTable.GetMemo( sFieldName: String;
                                 iLineWidth: Integer ): PChar;
begin
	Result := nil;
	if SelectForSxCall then
		Result := cs_GetMemo( PChar( sFieldName ), iLineWidth );
end;

{==============================================================================
  GetQueryBit
 ==============================================================================}
function TApolloTable.GetQueryBit( lRecNum: Integer ): WordBool;
begin
	Result := False;
	if (iWA > 0) then
  begin
		cs_Select( iWA );
		Result := cs_GetQueryBit( lRecNum );
	end;
end;

{==============================================================================
  GetScope
 ==============================================================================}
function TApolloTable.GetScope( iWhichScope: Integer ): String;
begin
	Result := '';
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
		Result := String( cs_GetScope( iWhichScope ));
	end;
end;

{==============================================================================
  GetString
 ==============================================================================}
function TApolloTable.GetString( sFieldName: String ): String;
begin
  Result := '';
  if (iWA > 0) then
  begin
    if FSpeedMode then
    begin
      cs_Select( iWA );
      Result := String( cs_GetString( PChar( sFieldName )));
    end
    else
      with FieldByName( sFieldName ) do
      begin
        Result := AsString;
        // Pad result to field length, as GetString normally does
        if DataType = ftString then
          Result := Result + StringOfChar( #32, Size - Length( Result ));
      end;
  end;
end;

{==============================================================================
  GetStringEx - Only here for backwards comatibility. Not needed under 32-bit.
 ==============================================================================}
function TApolloTable.GetStringEx( sFieldName: String ): PChar;
begin
	Result := PChar( GetString( sFieldName ));
end;

{==============================================================================
  GetTrimString
 ==============================================================================}
function TApolloTable.GetTrimString( sFieldName: String ): String;
begin
	Result := '';
  if (iWA > 0) then
  begin
//    if FSpeedMode then           {LAS: 05-12-00, Removed}
//    begin                        {LAS: 05-12-00, Removed}
      cs_Select( iWA );            {LAS: 05-12-00, Removed}
			Result := String( cs_GetTrimString( PChar( sFieldName )));
//    end                                                 {LAS: 05-12-00, Removed}
//    else                                                {LAS: 05-12-00, Removed}
//			Result := FieldByName( sFieldName ).AsString;     {LAS: 05-12-00, Removed}
  end;
end;

{==============================================================================
	Notification
==============================================================================}
procedure TApolloTable.Notification( AComponent: TComponent;
                                             Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);

  if (Operation = opRemove) and (AComponent = FApolloConnection) then
  begin
  	Close;
    FApolloConnection := nil;
  end;
end;

{==============================================================================
  GetBinary - Reads a value containing any ASCII characters from a field
 ==============================================================================}
procedure TApolloTable.GetBinary( sField: String; var Buffer: PChar );
var
  i, iFldOffset, iFldLen: Integer;
begin
  if iWA = 0 then
  	Exit;

	cs_Select( iWA );
  cs_Go( RecNo );
	cs_GetRecord(LSDERecordBuffer);
	iFldOffset := cs_FieldOffset( PChar( sField ));
  iFldLen    := cs_FieldWidth( PChar( sField ));

  for i := 0 to iFldLen-1 do
		Buffer[i] := LSDERecordBuffer[(iFldOffset-1)+i];
end;

{==============================================================================
  PutBinary - Writes a value containing any ASCII characters to a field
 ==============================================================================}
procedure TApolloTable.PutBinary( sField: String; Buffer: PChar );
var
  i, iFldOffset, iFldLen: Integer;
begin
	// Can't modify a ReadOnly table
	if ReadOnly and (FErrorFunc = -1) then
  	RaiseApollo( ERR_TABLE_IS_READONLY );

	// Unless SpeedMode is True, we must be in Edit or Insert mode
	if (FSpeedMode = False) and not (State in [dsEdit, dsInsert]) then
		RaiseApollo( ERR_NOT_IN_EDIT_OR_INSERT_MODE );

  if iWA = 0 then
  	Exit;

	cs_Select( iWA );
	iFldOffset := cs_FieldOffset( PChar( sField ));
  iFldLen    := cs_FieldWidth( PChar( sField ));

  for i := 0 to iFldLen-1 do
    LSDERecordBuffer[(iFldOffset-1)+i] := Buffer[i];

  cs_PutRecord(LSDERecordBuffer);  // Read it into the internal buffer.

	if not FSpeedMode then
	  Resync([]);
end;


{==============================================================================
  WriteNativeToSDE
  This handles the conversion from the native Delphi buffer to the native
  SDE formatted buffer - and then writes it to the dataset.
 ==============================================================================}
procedure TApolloTable.WriteNativeToSDE(Buffer: PChar);
var
  i, j: Integer;
  FieldInfo: TFieldInfo;
  CurField: TField;
  SDEData: PChar;
  NativeData: PChar;
  NullPointer: PChar;
	dVal: TTimeStamp;
  SmallIntValue: SmallInt;
  IntValue : Integer;
  BooleanValue: WordBool;
  FloatValue: Double;
  sVal: String;
  StrLength,
  iYear: Integer;
  sYear: String;
  BlobData: PBlobData;
  iFldCount: Integer;
begin
  // Parse the SDE buffer into the Delphi formatted buffer
  iFldCount := FieldCount;

  if iWA = 0 then
  	Exit;
	cs_Select( iWA );

	// If all fields not selected in Fields Editor, MUST refresh record buffer
  if (iFldCount <> FSDEFieldCount) then
		cs_GetRecord(LSDERecordBuffer);

  for i := 0 to iFldCount-1 do
  begin
    CurField := Fields[i];
    if CurField.FieldNo > 0 then
    begin
      FieldInfo := TFieldInfo(LFieldOffsets[CurField.FieldNo-1]);
      SDEData := LSDERecordBuffer + FieldInfo.SDEOffset;
      NullPointer := Buffer + FieldInfo.NativeOffset;
      NativeData := NullPointer + 1;

      // Translate the SDE buffer to the Native Delphi Buffer.
      case CurField.DataType of
        ftBoolean:
        begin
          if NullPointer[0] = FLD_NULL then
						SDEData[0] := ' '
          else
					begin
	    	    Move(NativeData^, BooleanValue, SizeOf(BooleanValue));
  		      if (BooleanValue) then
    	      	SDEData[0] := 'T'
      	    else
        	  	SDEData[0] := 'F';
          end;
        end;
        ftDate:
        begin
          if NullPointer[0] = FLD_NULL then
          begin
      	    sVal := '        ';
          end
          else
          begin
            Move(NativeData^, dVal.Date, SizeOf(dVal.Date));
            dVal.Time := 0;

						sVal := FormatDateTime( 'yyyymmdd', TimeStampToDateTime( dVal ));
            if Copy( sVal, 1, 2 ) = '00' then
            begin
	            sYear := DateToStr( TimeStampToDateTime( dVal ));
							if Pos( 'Y', Uppercase(ShortDateFormat)) = 1 then
								sYear := Copy( sYear, 1, Pos( DateSeparator, sYear)-1 )
        	    else
  	       	  begin
								for j := Length( sYear ) downto 1 do
									if sYear[j] = DateSeparator then
    		          begin
										sYear := Copy( sYear, j+1, Length(sYear)-j );
										break;
									end;
							end;
							iYear := StrToInt( sYear );
							if iYear < 100 then
							begin
          	   	Inc( iYear, (FEpoch div 100) * 100 );
								if (iYear < FEpoch) then
									Inc( iYear, 100 )
								else if (iYear > (FEpoch + 100)) then
									Dec( iYear, 100 );
							end;
							sVal := Trim( IntToStr(iYear)) + Copy( FormatDateTime( 'yyyymmdd',
  	      	                                TimeStampToDateTime( dVal )), 5,4 );
						end;
					end;
          Move(PChar(sVal)^, SdeData^, 8);
        end;
        ftSmallInt:
        begin
          if NullPointer[0] = FLD_NULL then
          	sVal := ''
          else
          begin
  	        Move(NativeData^, SmallIntValue, SizeOf(SmallIntValue));
    	      sVal := IntToStr( SmallIntValue );
		      	if (Length(sVal) > FieldInfo.Width) then
							RaiseApollo( ERR_VALUE_OUT_OF_BOUNDS );
          end;
    	    while Length( sVal ) < FieldInfo.Width do
	  				sVal := ' ' + sVal;
  	      Move(PChar(sVal)^, SDEData^, FieldInfo.Width);
        end;

        ftInteger:
        begin
          if NullPointer[0] = FLD_NULL then
          	sVal := ''
          else
          begin
  	        Move(NativeData^, IntValue, SizeOf(IntValue));
    	      sVal := IntToStr( IntValue );
		      	if (Length(sVal) > FieldInfo.Width) then
							RaiseApollo( ERR_VALUE_OUT_OF_BOUNDS );
          end;
    	    while Length( sVal ) < FieldInfo.Width do
	  				sVal := ' ' + sVal;
  	      Move(PChar(sVal)^, SDEData^, FieldInfo.Width);
        end;
        ftCurrency,
        ftFloat:
        begin
          if NullPointer[0] = FLD_NULL then
          	sVal := ''
          else
          begin
        	  Move(NativeData^, FloatValue, SizeOf(FloatValue));
          	sVal := Format( '%*.*f', [FieldInfo.Width,
                                      FieldInfo.Decimals, FloatValue]);

            // Support Regional settings
            for j := 1 to Length( sVal ) do
            	if sVal[j] = DecimalSeparator then
              	sVal[j] := '.'
							else if sVal[j] = ThousandSeparator then
              	sVal[j] := ' ';
          end;

        	while Length( sVal ) < FieldInfo.Width do
  					sVal := ' ' + sVal;
  	      Move(PChar(sVal)^, SDEData^, FieldInfo.Width);
        end;
        ftString:
        begin
          if NullPointer[0] = FLD_NULL then
 	        	StrLength := 0
   	      else
     	    	StrLength := Length( PChar( NativeData ));

     		  Move(NativeData^, SDEData^, StrLength);
 	      	for j := StrLength to FieldInfo.Width-1 do
           	SDEData[j] := #32;
        end;
        ftMemo,
        ftBlob:
        begin
          BlobData := PBlobData( Buffer + FBlobCacheOfs +
                                (CurField.Offset * SizeOf( TApBlobData )));
          Move(BlobData.SDEPointer, SDEData^, 10);
        end;
      end;
    end;
  end;
  
  if (State = dsInsert) then
  begin
		if fIsTableEncrypted then
      LSDERecordBuffer[0] := 'E'
    else
      LSDERecordBuffer[0] := ' ';
  end
  else if cs_Deleted then
  begin
		if fIsTableEncrypted then
      LSDERecordBuffer[0] := 'D'
    else
    	LSDERecordBuffer[0] := '*';
  end;

  cs_PutRecord(LSDERecordBuffer);  // Write it to the table
end;

{==============================================================================
  ReadSDEToNative
  This reads the SDE Buffer and then converts it to the native format.
 ==============================================================================}
procedure TApolloTable.ReadSDEToNative(Buffer: PChar);
var
  i, j: Integer;
  FieldInfo: TFieldInfo;
  CurField: TField;

  SDEData: PChar;
  NativeData: PChar;
  NullPointer: PChar;
  DateFld: TDateFld;
  SmallIntValue: SmallInt;
  IntValue : Integer;
  BooleanValue: WordBool;
  FloatValue: Double;
  sVal: String;
  LDate: Integer;
  BlobData: PBlobData;
  iFldCount: Integer;
begin
  // Parse the SDE buffer into the Delphi formatted buffer
  iFldCount := FieldCount;
  for i := 0 to iFldCount-1 do
  begin
    CurField := Fields[i];
    if CurField.FieldNo > 0 then
    begin
      FieldInfo := TFieldInfo(LFieldOffsets[CurField.FieldNo-1]);
      SDEData := LSDERecordBuffer + FieldInfo.SDEOffset;
      NullPointer := Buffer + FieldInfo.NativeOffset;
      NativeData := NullPointer + 1;
      if (FIsEncrypted and (FPassword = '')) then
      begin
				NullPointer[0] := FLD_NOTNULL;
				Move(SDEData^, NativeData^, CurField.Size);
				NativeData[CurField.Size] := #0;
				for j := CurField.Size-1 downto 0 do
	  	    case CurField.DataType of
            ftString:
	          	NativeData[j] := '*'
						else
	          	NativeData[j] := #0;
          end;
			end
      else
      begin
	      // Translate the SDE buffer to the Native Delphi Buffer.
  	    case CurField.DataType of
    	    ftBoolean:
            begin
              // If the field contains a 'T' or 'Y', it's True -- else False
              sVal := Copy(SDEData, 1, FieldInfo.Width);
              if Trim(sVal) = '' then
                NullPointer[0] := FLD_NULL
              else
              begin
                NullPointer[0] := FLD_NOTNULL;
                BooleanValue := (SDEData[0] in ['Y','T']);
                Move(BooleanValue, NativeData^, SizeOf( BooleanValue ));
              end;
            end;
        	ftDate:
            begin
              DateFld := TDateFld(Pointer(SDEData)^);
              if (DateFld.Year[1] = ' ') or (DateFld.Year[1] = #0) or (DateFld.Month[1] = #0 ) or (DateFld.Day[1] = #0)then // CRVG
              begin
                NullPointer[0] := FLD_NULL;
                LDate := 0;
              end
              else
              begin
                NullPointer[0] := FLD_NOTNULL;
                try
                	if (DateFld.Month = '00') or (DateFld.Day = '00') or
                     (DateFld.Month[1] = #0 ) or (DateFld.Day[1] = #0)   then
                  	LDate := 0
                  else
	                  LDate := DateTimeToTimeStamp( Trunc( EncodeDate(
 		                                 StrToIntDef( DateFld.Year , 0 ),
   		                               StrToIntDef( DateFld.Month, 0 ),
     		                             StrToIntDef( DateFld.Day  , 0 )))).Date;
                except
                  LDate := 0;
                end;
              end;
              Move(LDate, NativeData^, SizeOf(LDate));
            end;
    	    ftSmallInt:
            begin
              sVal := Copy(SDEData, 1, FieldInfo.Width);
              if Trim(sVal) = '' then
                NullPointer[0] := FLD_NULL
              else
              begin
                NullPointer[0] := FLD_NOTNULL;
                try
                  SmallIntValue := StrToInt(sVal) // CRVG : 2002-06-13 Add Trim to avoid extra errors in run-time
                except
                  SmallIntValue := 0;
                end;
                Move( SmallIntValue, NativeData^, SizeOf(SmallIntValue));
              end;
            end;
        	ftInteger:
            begin
              sVal := Copy(SDEData, 1, FieldInfo.Width);
              if Trim(sVal) = '' then
                NullPointer[0] := FLD_NULL
              else
              begin
                NullPointer[0] := FLD_NOTNULL;
                try
                  IntValue := StrToInt(sVal)    // CRVG : 2002-06-13 Add Trim to avoid extra errors in run-time
                except
                  IntValue := 0;
                end;
                Move( IntValue, NativeData^, SizeOf(IntValue));
              end;
            end;
  	      ftFloat:
            begin
              sVal := Copy(SDEData, 1, FieldInfo.Width );
              if Trim(sVal) = '' then
                NullPointer[0] := FLD_NULL
              else
              begin
                NullPointer[0] := FLD_NOTNULL;
                try
                  // Support Regional settings
                  for j := 1 to Length( sVal ) do
                    case sVal[j] of
                      '.' : sVal[j] := DecimalSeparator;
                      ',' : sVal[j] := ' ';
                    end;
                  FloatValue := StrToFloat(sVal)
                except
                  FloatValue := 0.0;
                end;
                Move( FloatValue, NativeData^, SizeOf(FloatValue));
              end;
            end;
	        ftString:
            begin
              sVal := Copy(SDEData, 1, FieldInfo.Width );
              if Trim(sVal) = '' then
                NullPointer[0] := FLD_NULL
              else
                NullPointer[0] := FLD_NOTNULL;
                
              Move(SDEData^, NativeData^, FieldInfo.Width {CurField.Size});  {LAS: 07-26-00, Changed}
              NativeData[ FieldInfo.Width] := #0;
              for j := FieldInfo.Width-1 downto 0 do
              begin
                if NativeData[j] = #32 then
                  NativeData[j] := #0
                else
                  break;
              end;
            end;
    	    ftMemo,
      	  ftBlob:
            begin
              BlobData := PBlobData( Buffer + FBlobCacheOfs +
                                   ( CurField.Offset * SizeOf( TApBlobData )));

              // This next bit is precautionary. It's likely this data has
              // already been cleared by InitRecordBuffer.
              // FreeMem( BlobData.Data );   {LAS: 04-04-00, Removed.  Caused AV w/ Lookup fields}
              BlobData.Data := nil;
              BlobData.Size := 0;
              BlobData.Modified := False;
              BlobData.Loaded := False;

              // Only fill buffer if not at EOF (unless Inserting)
              if (not ((State = dsInsert) and Eof)) then
                // Move the sde data pointer into our buffer
                Move( SDEData^, BlobData.SDEPointer, 10 );

              // Is this field NULL?
              if BlobData.SDEPointer <> '          ' then
                NullPointer[0] := FLD_NOTNULL
              else
                NullPointer[0] := FLD_NULL;
            end;
  	    end;
      end;
    end;
  end;
end;

{=BEGIN=== AREA FOR ALL THE NAVIGATION OPERATIONS AND RECORD MOVEMENTS ========}

{==============================================================================
  BeginScroll
 ==============================================================================}
procedure TApolloTable.BeginScroll;
begin
  if FSpeedMode = False then
  begin
    CheckBrowseMode;
		if not ControlsDisabled then  {LAS: 06-01-00, Added}
  	  DoBeforeScroll;
  end;
end;

{==============================================================================
  EndScroll
 ==============================================================================}
procedure TApolloTable.EndScroll;
begin
  if (FSpeedMode = False)then // {MED:08-24-00, commented} and (not ControlsDisabled) then  {LAS: 06-01-00, Added ControlsDisabled check}
 	  DoAfterScroll;
end;

{==============================================================================
  SelectForSxCall
 ==============================================================================}
function TApolloTable.SelectForSxCall: Boolean;
begin
	Result := False;
	if (iWA > 0) then
	begin
		Result := True;
		cs_Select( iWA );
		if (FSpeedMode = False) and (State <> dsCalcFields) then
    	UpdateCursorPos;
	end;
end;

{==============================================================================
  Go
 ==============================================================================}
procedure TApolloTable.Go( Value: Integer );
begin
  if (iWA > 0) then
  begin
  	if FSpeedMode then
    begin
    	cs_Select( iWA );
    	cs_Go( Value );
    end
    else
    begin
	    BeginScroll;
      with PRecInfo(ActiveBuffer + FRecInfoOfs)^ do
      begin
        BookmarkFlag := bfCurrent;
        Bookmark := Value;
      end;
		  FFetchReadCount := -1;      {LAS: 06-26-00, Added}
	    Resync([]);
  	  EndScroll;
    end;
  end;
end;

{==============================================================================
  GoTop
 ==============================================================================}
procedure TApolloTable.GoTop;
begin
  if (iWA > 0) then
  begin
	  FFetchReadCount := -1;       {LAS: 06-29-00, Added}
    if (FSpeedMode = False) then
      First
    else
    begin
      cs_Select( iWA );
      cs_GoTop;
    end;
  end;
end;

{==============================================================================
  InternalFirst
  Go to the crack before the first record.  After this is called, a call to
  GetRecord with a gmNext should position on the first record of the dataset.
  This does NOT position on the first record - it positions before the first
  record.  That's why the FFirstCrack is set to True.
 ==============================================================================}
procedure TApolloTable.InternalFirst;
begin
  if (iWA > 0) then
  begin
			// Workaround for QuickReports preview bug
      if FAccessMethod = amLocal then  {LAS: 07-07-00, Added}
	    	if cs_Alias(iWA) = '' then     {LAS: 07-07-00, Added}
  	    	OpenTable;                   {LAS: 07-07-00, Added}

      cs_Select( iWA );
      cs_GoTop;
      //cs_Skip(-1);   {LAS: 06-29-00, Removed}
  end;

  FFetchReadCount := -1;
  FFirstCrack     := TRUE;
  FLastCrack      := False;  {LAS: 07-28-00, Added}
	FRecNo          := 0;   {LAS: 06-09-00, Added}

end;

{==============================================================================
  InternalRefresh
 ==============================================================================}
procedure TApolloTable.InternalRefresh;
var                           {LAS: 07-25-00, Added}
	sVal: String;               {LAS: 07-25-00, Added}
  iRec: Integer;
begin
	sVal := DBFilter;           {LAS: 07-25-00, Added}
	if not (sVal = '') then     {LAS: 07-25-00, Added}
  begin
    iRec := cs_RecNo;         {LAS: 07-28-00, Added} {MED: 08-24-00, moved before if}
  	if cs_Querytest( PChar( sVal )) > 0 then   {LAS: 07-28-00, Added}
    begin
  		Query( PChar( sVal ));    {LAS: 07-25-00, Added}
	    cs_Go( iRec );            {LAS: 07-28-00, Added}
    end;
  end
  else                        {LAS: 07-25-00, Added}
		Resync([]);
end;

{==============================================================================
  GoBottom
 ==============================================================================}
procedure TApolloTable.GoBottom;
begin
  if (iWA > 0) then
  begin
	  FFetchReadCount := -1;    {LAS: 06-29-00, Added}
    if (FSpeedMode = False) then
      Last
    else
    begin
      cs_Select( iWA );
      cs_GoBottom;
    end;
  end;
end;

{==============================================================================
  InternalLast
  Go to the crack after the last record.  After this is called, a call to
  GetRecord with a gmPrior should position on the last record of the dataset.
  This does NOT position on the last record - it positions after the last
  record.  That's why the FLastCrack is set to True.
 ==============================================================================}
procedure TApolloTable.InternalLast;
begin
  if (iWA > 0) then
  begin
    cs_Select( iWA );
    cs_GoBottom;
//      cs_Skip(1);    {LAS: 06-29-00, Removed}
  end;
  FFetchReadCount := -1;
  FLastCrack      := True;
  FFirstCrack     := False;  {LAS: 07-28-00, Added}
end;

{LAS: 04-17-00, Added}
{==============================================================================
  MoveBy
 ==============================================================================}
function TApolloTable.MoveBy(Distance: Integer): Integer;
begin
//	if Abs(Distance) > 1 then   {LAS: 07-16-00, Added}
	  FFetchReadCount := -1;    {LAS: 06-29-00, Added}
	if FSpeedMode then
  begin
  	Skip( Distance );
    Result := Distance;  // Always returns Distance when SpeedMode True
  end
  else
		Result := inherited MoveBy( Distance );
end;

{LAS: 04-17-00, Added}
{==============================================================================
  Next
 ==============================================================================}
procedure TApolloTable.Next;
begin
	if FSpeedMode then
  	Skip( 1 )
  else
		inherited;
end;

{LAS: 04-17-00, Added}
{==============================================================================
  Prior
 ==============================================================================}
procedure TApolloTable.Prior;
begin
	if FSpeedMode then
  	Skip( -1 )
  else
		inherited;
end;

{==============================================================================
  Skip
 ==============================================================================}
procedure TApolloTable.Skip( Value: Integer) ;
begin
  if (iWA > 0) then
  begin
  	// For multiple skips, we may need to refetch the block of records
  	if Abs( Value ) > 1 then    {LAS: 07-05-00, Added}
		  FFetchReadCount := -1;    {LAS: 06-29-00, Added}

  	// If moving forward, we're no longer on FirstCrack
  	if Value > 0 then          {LAS: 07-05-00, Added}
			FFirstCrack := False     {LAS: 07-05-00, Added}
  	// If moving backwards, we're no longer on LastCrack
  	else if Value < 0 then     {LAS: 07-05-00, Added}
			FLastCrack := False;    {LAS: 07-05-00, Added}

    if (FSpeedMode = False) then
			MoveBy( Value )
    else
    begin
		  cs_Select( iWA );
      cs_Skip( Value );
    end;
  end;
end;

{==============================================================================
  Seek
 ==============================================================================}
function TApolloTable.Seek( Value: String ): WordBool;
begin
  Result := False;
  if ((iWA > 0) and (FIndexName > '')) then
  begin
    BeginScroll;
	  FFetchReadCount := -1;      {LAS: 06-20-00, Added}

    cs_Select( iWA );
    Result := cs_Seek( PChar( Value ));

    FRecNo := cs_RecNo;  {LAS: 07-24-00, Added}

//		if not FSpeedMode then    {LAS: 07-24-00, Removed}
    begin
      with PRecInfo(ActiveBuffer + FRecInfoOfs)^ do
      begin
				if Result then
	        BookmarkFlag := bfCurrent
        else
        begin
        	cs_GoBottom;           {LAS: 07-24-00, Added}
        	FLastCrack := True;    {LAS: 07-24-00, Added}
	        BookmarkFlag := bfEOF;
        end;

        Bookmark := FRecNo;  // cs_RecNo;  {LAS: 07-24-00, Changed}
      end;
    end;

   	Resync([]);
    EndScroll;
  end
	else if (iWA = 0) then
		RaiseApollo( ERR_MUST_OPEN_TABLE )
	else if (FIndexName = '') then
  begin
		RaiseApollo( ERR_NO_INDEX_ACTIVE );
  end;
end;

{==============================================================================
  SeekBin
 ==============================================================================}
function TApolloTable.SeekBin( Value: String; iLength: Integer ): WordBool;
begin
  Result := False;
  if ((iWA > 0) and (FIndexName > '')) then
  begin
    BeginScroll;
	  FFetchReadCount := -1;      {LAS: 06-20-00, Added}
    cs_Select( iWA );
    Result := cs_SeekBin( PChar( Value ), iLength );
    if Result = False then
      cs_GoBottom;

		if not FSpeedMode then
    begin
      with PRecInfo(ActiveBuffer + FRecInfoOfs)^ do
      begin
				if Result then
	        BookmarkFlag := bfCurrent
        else
	        BookmarkFlag := bfEOF;
        Bookmark := cs_RecNo;
      end;
    end;

    Resync([]);
    EndScroll;
  end
	else if (iWA = 0) then
		RaiseApollo( ERR_MUST_OPEN_TABLE )
	else if (FIndexName = '') then
		RaiseApollo( ERR_NO_INDEX_ACTIVE );
end;

procedure TApolloTable.ClearFetchedItems;
var
  i :Integer;
begin
  for i := 0 to (FFetchRecs.Count - 1) do
  begin
    AFetch := FFetchRecs.Items[i];
    if Assigned(AFetch^.Data) then        // MED: 04-20-01, added
      FreeMem(AFetch^.Data);

    Dispose(AFetch);
  end;
  FFetchRecs.Clear;
  FFetchLastRec   := 0 ;
  FFetchReadCount := -1;
end;

{==============================================================================
  SetFetchCount
 ==============================================================================}
procedure TApolloTable.SetFetchCount(Val:Integer);
begin
  // Table must be closed first
  if (iWA > 0) then                                     {LAS: 07-30-00, Added}
    raise EApolloError.Create( ERR_MUST_CLOSE_TABLE );  {LAS: 07-30-00, Added}

	if FFetchCount <> Val then
	begin
    if Val < 1 then
    	Val := 1;

    ClearFetchedItems;

		FFetchCount := Val;
    while FFetchRecs.Count < Val do
    begin
      New(AFetch);
      AFetch^.RecNum     := 0;
      AFetch^.RecOrdinal := 0;
      AFetch^.Data       := nil;
      FFetchRecs.Add(AFetch);
    end;
	end;
end;

{MED: 05-21-00, Added to speed up fetching in server mode}
function TApolloTable.FetchRecord(FirstCrack, LastCrack:WordBool;
                          FEOFAction:Integer;
                          GetMode: TGetMode;
                          var iRecNo : Integer): TGetResult;
  procedure FetchAdjustSize;
  begin
    while FFetchRecs.Count < FFetchCount do
    begin
      New(AFetch);
      AFetch^.RecNum     := 0;   {LAS: 05-31-00, Renamed from RecNo}
      AFetch^.RecOrdinal := 0;
      AFetch^.Data       := AllocMem(FRecSize);
      FFetchRecs.Add(AFetch);
    end;
  end;
begin

  FetchAdjustSize;

  if (FFetchSavedDirection = GetMode) and (FFetchLastRec < (FFetchReadCount-1) ) then
  begin
    inc(FFetchLastRec);
    AFetch := FFetchRecs.Items[FFetchLastRec];
  end
  else
  begin
    FFetchLastRec := 0;
    FApolloConnection.ap_MultiTraverseTable( FFetchReadCount,
                  FirstCrack,LastCrack, FEOFAction,GetMode, FFetchSavedDirection,
                  LSDERecordBuffer, FRecSize, FFetchCount, FRecNo, FFetchRecs);

    AFetch := FFetchRecs.Items[0];
  end;
  CopyMemory(pchar(LSDERecordBuffer), pchar(AFetch^.Data), FRecSize);
  Result              := AFetch^.GetResult;
  iRecNo              := AFetch^.RecNum;    {LAS: 05-31-00, Renamed from RecNo}
  FRecOrdinal         := AFetch^.RecOrdinal;
  FRecNo              := iRecNo;
  FFetchSavedDirection:= GetMode;
end;

{==============================================================================
  GetRecord
  This is the routine that does all of the navigation of the dataset.  TDataSet
  calls this with a buffer and then this navigates and fills in the buffer.
 ==============================================================================}
function TApolloTable.GetRecord( Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean ): TGetResult;
var
  iRecNo, iSavedRecNo: Integer;
  CurState: TDataSetState;
  bAccept : Boolean;
begin
  begin
    cs_Select( iWA );
    bAccept := True;

    repeat

      if (FAccessMethod = amLocal) then
      begin
        if FFirstCrack or FLastCrack then
          iSavedRecNo := 0
        else
          iSavedRecNo := cs_RecNo;

        Result := grOK;
        case GetMode of
          gmNext:
            begin
              if FFirstCrack then
              begin
                cs_GoTop;
              end
              else
              begin
                cs_Skip(1);
              end;
              if cs_EOF or (cs_RecNo = iSavedRecNo) then
              begin
                cs_GoBottom;
                Result := grEOF;
              end
            end;
          gmPrior:
            begin
              if FLastCrack then
                cs_GoBottom
              else
                cs_Skip(-1);
              if cs_BOF or (cs_RecNo = iSavedRecNo) then
              begin
                cs_GoTop;
                Result := grBOF;
              end;
            end;
          gmCurrent:
            begin
              if (cs_BOF and cs_EOF) then
              begin
                Result := grEOF       // We are sure there are no records
              end
              else if cs_EOF and (FEOFAction = eaMoveLast) then
              begin
                // To see if cs_BOF failed
                cs_GoBottom;
                if cs_EOF then
                  Result := grEOF;
              end;
            end;
        end;
        iRecNo      := cs_RecNo;
        FRecNo      := iRecNo;
				FRecOrdinal := FRecOrdinalNeedsRefresh; // CRVG 2001-10-22
      end
      else
      begin
        Result := FetchRecord(FFirstCrack, FLastCrack, Ord(FEOFAction),
                              GetMode, iRecNo);
      end;

  	  // Clear the buffer
    	InternalInitRecord(Buffer);

  	  if (Result = grOK) then
    	begin
        if (FAccessMethod = amLocal) then
          cs_GetRecord(LSDERecordBuffer);

      	ReadSDEToNative(Buffer);
  	    with PRecInfo(Buffer + FRecInfoOfs)^ do
 	  	  begin
   	  	  BookmarkFlag := bfCurrent;
        	Bookmark     := iRecNo;
	 	    end;
  			// If we have any calculated fields, process the OnCalcFields event
	  		if (CalcFieldsSize > 0) then
	     	begin
		      CurState := State;
 			    SetTempState( dsCalcFields );
     			try
	       		CalculateFields( Buffer );
		      finally
 			      RestoreState( CurState );
   			  end;
  			end;

	  		if ((Assigned(OnFilterRecord)) and (Filtered) and (Result = grOK)) then
      	begin
			  	// Call the filtered event!
				  CurState := State;
  				SetTempState( dsFilter );
	  			bAccept := True;
		  		try
			  		// protect from user errors with the try except...
					try
						LFilterRecordBuffer := Buffer;
						OnFilterRecord( Self, bAccept );
				    cs_Select( iWA );   {LAS: 07-30-00, Added in case event changed workarea}
					except
					end;
  		    finally
					  RestoreState( CurState );
  				end;
				  if (bAccept = False) then   // Failed filter event
	        begin
					  if GetMode = gmCurrent then
						  Result := grEOF;
  					if FFirstCrack then
	  					Result := grBOF;
        	  if FLastCrack then
          		Result := grEOF;
				  	if (Result = grOK) then
					  	Continue;               // Go back to the top of the repeat loop
          end;
        end;
      end
      else if ((Result = grError) and DoCheck) then
        RaiseApollo( ERR_TABLE_IS_EMPTY );

  	Until (bAccept or (Result = grBOF) or (Result = grEOF));
  end;
  FFirstCrack := False;
  FLastCrack  := False;
end;

{==============================================================================
  Resync
 ==============================================================================}
procedure TApolloTable.Resync(Mode: TResyncMode);
begin
  if (iWA > 0) and (FSpeedMode = False) and (not (State = dsInactive)) then   {LAS: 05-23-00, Added dsINactive Check}
  begin
  	cs_Select( iWA );
  	cs_Go( RecNo );
    inherited Resync(Mode);
	end;
end;

{==============================================================================
  SetWriteBlobHdr
 ==============================================================================}
procedure TApolloTable.SetWriteBlobHdr( Value: WordBool );
begin
	if (FWriteBlobHdr <> Value) then
  begin
		FWriteBlobHdr := Value;
    SysProp( SDE_SP_SETWRITEBLOBHDR, Pointer( SmallInt(Value)) );
	end;
end;

{==============================================================================
  SetAccessMethod
 ==============================================================================}
procedure TApolloTable.SetAccessMethod( Value: TAccessMethod );
begin
	if (Value <> FAccessMethod) then
  begin
    if Active then
    	RaiseApollo( ERR_MUST_CLOSE_TABLE_TO_CHANGE_DATA_ACCESS );

		FAccessMethod := Value;

    if FAccessMethod = amLocal then
    begin
    	LoadDLL;
    end;
    ClearFetchedItems;
    //ClearCachedItems;
  end;
end;


{==============================================================================
  SetTableType
 ==============================================================================}
procedure TApolloTable.SetTableType( Value: TApolloTableType );
begin
	if (Value <> FTableType) then
  begin
		FTableType := Value;
		UpdateFieldAndIndexDefs;
  end;
end;

procedure TApolloTable.SetCommitLevel( Value: TApolloCommitLevel );
begin
	if (Value <> FCommitlevel) then
  begin
    if Active then
      begin
			  RaiseApollo( ERR_CANT_CHANGE_COMMIT_LEVEL );
      end;
		FCommitlevel := Value;
  end;
end;


{==============================================================================
  SysProp
 ==============================================================================}
function TApolloTable.SysProp( uiSysItem: Word; vpData: Pointer ): Integer;
begin
  Result := cs_SysProp(uiSysItem, vpData);   {LAS: 06-13-00, Added}
end;


{=END===== AREA FOR ALL THE NAVIGATION OPERATIONS AND RECORD MOVEMENTS ========}

{==============================================================================
  Index
 ==============================================================================}
function TApolloTable.Index( sFileName, sExpression: String;
                               Options: SmallInt; Descending: Boolean;
                               sForClause: String ) : SmallInt;
var
  PFileName: PChar;
  PForClause: PChar;
begin
  Result := 0;
  PFileName := nil;
	if (iWA > 0) then
  begin
	  cs_Select( iWA );

	  if ((sFileName = '') or (sFileName = #0)) then
  		RaiseApollo( ERR_INDEX_NAME_REQUIRED )
		else
  		PFileName := PChar( sFileName );

	  if ((sExpression = '') or (sExpression = #0)) then
  		RaiseApollo( ERR_INDEX_EXPRESSION_REQUIRED );

	  if (cs_EvalTest(pchar(sExpression)) = 0 ) then
  		RaiseApollo( ERR_INVALID_INDEX_EXPRESSION +' ['+sExpression+']' );

	  if ((sForClause = '') or (sForClause = #0)) then
  		PForClause := nil
	  else
			PForClause := PChar( sForClause );

	  Result := cs_Index( PFileName, PChar(sExpression),
                        Options, Descending, PForClause);

		if not FAddingIndex then     {LAS: 05-23-00, Added}
    begin                        {LAS: 05-23-00, Added}
      FIndexName := sFileName;
      IndexDefs.Updated := False;
      UpdateIndexDefs;
      FCountNeedsRefresh := True;
      if (sForClause > '') then
        GoTop
      else
        Resync([]);
    end;                         {LAS: 05-23-00, Added}
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  IndexClose
 ==============================================================================}
procedure TApolloTable.IndexClose;
begin
	if (iWA > 0) then
  begin
    cs_Select( iWA );
    cs_IndexClose;
    FIndexName := '';
    IndexDefs.Updated := False;
    UpdateIndexDefs;
		FCountNeedsRefresh := True;
    Resync([]);
	end;
end;

{==============================================================================
  IndexCondition
 ==============================================================================}
function TApolloTable.IndexCondition: String;
begin
	Result := '';

	if (iWA > 0) then
  begin
		if (FIndexName = '') then
			RaiseApollo( ERR_NO_INDEX_ACTIVE );

	  cs_Select( iWA );
		Result := String( cs_IndexCondition );
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  IndexKey
 ==============================================================================}
function TApolloTable.IndexKey: String;
begin
	Result := '';
	if (iWA > 0) then
  begin
		if (FIndexName = '') then
			RaiseApollo( ERR_NO_INDEX_ACTIVE );

	  cs_Select( iWA );
		Result := String( cs_IndexKey );
	end
{  else
		RaiseApollo( ERR_MUST_OPEN_TABLE )};  {LAS: 06-19-00, Removed.  Caused problems with Master/Child}
end;

{==============================================================================
  IndexKeyField
 ==============================================================================}
function TApolloTable.IndexKeyField: String;
begin
	Result := '';
	if (iWA > 0) then
  begin
		if (FIndexName = '') then
			RaiseApollo( ERR_NO_INDEX_ACTIVE );

	  cs_Select( iWA );
		Result := String( cs_IndexKeyField );
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  IndexFileName
 ==============================================================================}
function TApolloTable.IndexFileName( iIndex: Integer ): String;
begin
	Result := '';
	if (iWA > 0) then
  begin
//		if (FIndexName = '') then                            {LAS: 04-26-00, Removed}
//			RaiseApollo( ERR_NO_INDEX_ACTIVE );  {LAS: 04-26-00, Removed}

	  cs_Select( iWA );
		Result := String( cs_IndexName( iIndex ));
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  IndexFlip
 ==============================================================================}
function TApolloTable.IndexFlip: WordBool;
begin
  Result := FALSE;
	if (iWA > 0) then
  begin
		if (FIndexName = '') then
			RaiseApollo( ERR_NO_INDEX_ACTIVE );

    cs_Select( iWA );
    Result := cs_IndexFlip;
    Resync([]);
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  IndexOpen
 ==============================================================================}
function TApolloTable.IndexOpen( sFileName: String ): SmallInt;
var
	cpFile : Array[0..255] of Char;
begin
  Result := 0;
  if FAccessMethod = amLocal then
  begin
    // Make sure index file exists
    if (not cs_FileExists( sFileName )) and (FErrorFunc = -1) then
		  RaiseApollo( ERR_INDEX_FILE_NOT_FOUND + sFileName );

  	if (iWA > 0) then
    begin
	    cs_Select( iWA );
      StrPCopy( cpFile, sFileName );
	  	Result := cs_IndexOpen( cpFile );

      // Update IndexDefs
	  	IndexDefs.Updated := False;
		  UpdateIndexDefs;

      // Update IndexName property
	  	if (FTableType = ttSXNTX) or  (Pos( '.IDX', UpperCase( sFileName )) > 0) then
      begin
	      cs_SetOrder( Result );
				FCountNeedsRefresh := True;
        FIndexName := sFileName;
	      CheckMasterRange;
  	    Resync([]);
  		end;
	  end
	  else
			RaiseApollo( ERR_MUST_OPEN_TABLE );
  end
  else
  begin
	  if (iWA > 0) then
    begin
	    cs_Select( iWA );
      StrPCopy( cpFile, sFileName );
	  	Result := cs_IndexOpen( cpFile );

      // Update IndexDefs
			FCountNeedsRefresh := True;
  		IndexDefs.Updated := False;
	  	UpdateIndexDefs;
  	end
    else
			RaiseApollo( ERR_MUST_OPEN_TABLE );
  end;
end;

{==============================================================================
  IndexOrd
 ==============================================================================}
function TApolloTable.IndexOrd: Integer;
begin
  Result := 0;
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
		Result := cs_IndexOrd;
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  IndexTag
 ==============================================================================}
function TApolloTable.IndexTag(sFileName, sTagName, sExpression: String;
                                       Options: SmallInt; Descending: Boolean;
                                       sForClause: String ) : SmallInt;
var
  PFileName: PChar;
  PForClause: PChar;
begin
  Result := 0;
	if (iWA > 0) then
  begin
	  cs_Select( iWA );

	  if ((sFileName = '') or (sFileName = #0)) then
  		PFileName := nil
		else
  		PFileName := PChar( sFileName );

	  if ((sTagName = '') or (sTagName = #0)) and (FErrorFunc = -1) then
  		RaiseApollo( ERR_INDEX_TAG_NAME_REQUIRED );

	  if ((sExpression = '') or (sExpression = #0)) and (FErrorFunc = -1) then
  		RaiseApollo( ERR_INDEX_EXPRESSION_REQUIRED );

	  if (cs_EvalTest(pchar(sExpression)) = 0 ) then
  		RaiseApollo( ERR_INVALID_INDEX_EXPRESSION +
            ' ['+sExpression+']' );

	  if ((sForClause = '') or (sForClause = #0)) then
  		PForClause := nil
	  else
			PForClause := PChar( sForClause );
                                      {LAS: 06-13-00, Added Trim around sTagname}
	  Result := cs_IndexTag( PFileName, PChar( Trim( sTagName)), PChar(sExpression),
                           Options, Descending, PForClause);

		if not FAddingIndex then     {LAS: 05-23-00, Added}
    begin                        {LAS: 05-23-00, Added}
      FIndexName := sTagName;
      IndexDefs.Updated := False;
      UpdateIndexDefs;
      FCountNeedsRefresh := True;
      if (sForClause > '') then
        GoTop
      else
        Resync([]);
		end;                       {LAS: 05-23-00, Added}
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  IndexType
 ==============================================================================}
function TApolloTable.IndexType: Integer;
begin
  Result := 0;
	if (iWA > 0) then
  begin
		if (FIndexName = '') then
			RaiseApollo( ERR_NO_INDEX_ACTIVE );

	  cs_Select( iWA );
		Result := cs_IndexType;
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  IsEncrypted
 ==============================================================================}
function TApolloTable.IsEncrypted( iFileOrRec: Integer ): WordBool;
begin
  Result := FALSE;
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
		Result := cs_IsEncrypted( iFileOrRec );
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  KeyAdd
 ==============================================================================}
function TApolloTable.KeyAdd( sTagName: String ): WordBool;
begin
  Result := FALSE;
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
		Result := cs_KeyAdd( PChar( sTagName ));
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  KeyData
 ==============================================================================}
function TApolloTable.KeyData: String;
begin
	Result := '';
	if SelectForSxCall then
		Result := String( cs_KeyData );
end;

{==============================================================================
  KeyDrop
 ==============================================================================}
function TApolloTable.KeyDrop( sTagName: String ): WordBool;
begin
	Result := False;
	if SelectForSxCall then  
		Result := cs_KeyDrop( PChar( sTagName ));
end;

{==============================================================================
  LockCount  
 ==============================================================================}
function TApolloTable.GetLockCount: Word;
begin
	Result := 0;
	if SelectForSxCall then
		Result := cs_LockCount;
end;

{==============================================================================
  Locked
 ==============================================================================}
function TApolloTable.Locked( lRecNum: Integer ): WordBool;
begin
	Result := False;
	if SelectForSxCall then
		Result := cs_Locked( lRecNum );
end;

{==============================================================================
  MemDealloc
 ==============================================================================}
procedure TApolloTable.MemDealloc( cpBuff: PChar );
begin
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
		cs_MemDealloc( cpBuff );
	end;
end;

{==============================================================================
  OpenMode
 ==============================================================================}
function TApolloTable.OpenMode: SmallInt;
begin
	Result := 0;
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
		Result := cs_OpenMode;
 	end;
end;

{==============================================================================
  OrderPosGet
 ==============================================================================}
function TApolloTable.OrderPosGet: Double;
begin
	Result := 0;
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
		Result := cs_OrderPosGet;
	end;
end;

{==============================================================================
  OrderPosSet
 ==============================================================================}
procedure TApolloTable.OrderPosSet( Value: Double );
begin
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
		cs_OrderPosSet( Value );

		with PRecInfo(ActiveBuffer + FRecInfoOfs)^ do
    begin
      BookmarkFlag := bfCurrent;
      Bookmark := cs_RecNo;
    end;
		Resync([rmExact]);
	end;
end;

{==============================================================================
  OrderRecNo
 ==============================================================================}
function TApolloTable.OrderRecNo: Integer;
begin
	Result := 0;
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
		UpdateCursorPos;
		Result := cs_OrderRecNo;
	end;
end;

{==============================================================================
  Pack
 ==============================================================================}
procedure TApolloTable.Pack;
begin
	// Can't PACK a ReadOnly table
	if ReadOnly and (FErrorFunc = -1) then
  	RaiseApollo( ERR_TABLE_IS_READONLY );

	// Can't PACK if table not opened EXCLUSIVE
	if not Exclusive and (FErrorFunc = -1) then
  	RaiseApollo( ERR_MUST_OPEN_EXCLUSIVE );

	if (iWA > 0) then
  begin
    cs_Select( iWA );
    cs_Pack;
    FCountNeedsRefresh := True;  {LAS: 05-30-00, Added}
    Resync([]);
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  PutBlob
 ==============================================================================}
function TApolloTable.PutBlob( sFieldName: String; vpVar: Pointer;
                                 lSize: Integer ): Integer;
var
	OldState : TDataSetState;
  iFldNum : Integer;
  BlobData: PBlobData;
begin
  Result := 0;

  if ReadOnly and (FErrorFunc = -1) then
    RaiseApollo( ERR_TABLE_IS_READONLY );

  if (FSpeedMode) then
  begin
    if (iWA > 0) then
    begin
      OldState := State;
      if (OldState in [dsEdit, dsInsert]) then
	      Post;
      cs_Select( iWA );
      Result := cs_PutBlob( PChar(sFieldName), vpVar, lSize );
      if (OldState in [dsEdit, dsInsert]) then
      	Edit;
    end;
  end
  else
  begin
    if not (State in [dsEdit, dsInsert]) then
    	RaiseApollo( ERR_NOT_IN_EDIT_OR_INSERT_MODE );

	  iFldNum := FieldNum( sFieldName );
    BlobData := PBlobData( ActiveBuffer + FBlobCacheOfs +
                         ( Fields[iFldNum-1].Offset * SizeOf( TApBlobData )));
    FreeMem( BlobData.Data );

    BlobData.Size := lSize;     // Set the size
    BlobData.Modified := True;  // It's been modified

    // OK, copy the data
    BlobData.Data := AllocMem( lSize+1 );
    Move( vpVar^, BlobData.Data^, lSize );
    BlobData.Data[lSize] := #0;
    Result := lSize;
  end;
end;

{==============================================================================
  PutRecord
 ==============================================================================}
procedure TApolloTable.PutRecord( cpRecord: PChar );
begin
	// Can't modify a ReadOnly table
	if ReadOnly and (FErrorFunc = -1) then
  	RaiseApollo( ERR_TABLE_IS_READONLY );

	// Unless SpeedMode is True, we must be in Edit or Insert mode
	if (FSpeedMode = False) and not (State in [dsEdit, dsInsert]) then
		RaiseApollo( ERR_NOT_IN_EDIT_OR_INSERT_MODE );

	if (iWA > 0) then
  begin
	  cs_Select( iWA );
		cs_PutRecord( cpRecord );
	  Resync([]);
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  Query
 ==============================================================================}
function TApolloTable.Query( const Value: String ): Integer;
var
	sVal: String;
begin
	Result := 0;
	if (iWA > 0) then
  begin
	  cs_Select( iWA );

  	if (Value = '') or (Value = #0) then
    	sVal := ''
		else
    	sVal := Value;

		// If SpeedMode is TRUE, do it and exit early
    if FSpeedMode then
    begin
			Result := cs_Query( PChar( sVal ));
      exit;  // Early exit
    end;

    if (CalcFieldsSize > 0) then
    begin
    	// Try to check if a calculated field is actually used within the query
      // expression and, if so, raise an exception and bail. (FUTURE FEATURE)
    end;

    if (sVal > '') and (cs_EvalTest( PChar( sVal )) <> 3) then
			RaiseApollo( ERR_INVALID_QUERY_EXPRESSION );

    BeginScroll;

		Result    := cs_Query( PChar( sVal ));

    //FRecCount := cs_RecCount;          {MED: 10-21-00, added}
    //FRowCount := cs_QueryRecCount;     {MED: 10-21-00, added}
		//if not FSpeedMode then    {MED: 03-05-2002, Commented- FSpeedMode exists early-see above}
    begin
      with PRecInfo(ActiveBuffer + FRecInfoOfs)^ do
      begin
				if cs_Eof  then
	        BookmarkFlag := bfEOF
        else
	        BookmarkFlag := bfCurrent;
        Bookmark := Result; //cs_RecNo;     {MED: 03-05-2002, changed it to Result same as cs_RecNo here}
      end;
    end;
    FRecNo   := Result;                      {MED: 03-05-2002, added since cs_Recno was changed}
    FCountNeedsRefresh := True;

		Resync([]);
    EndScroll;

	end;
end;

{==============================================================================
  SetSxQuery
 ==============================================================================}
procedure TApolloTable.SetSxQuery( Value : String );
begin
	FFilterStr := Value;
  if (iWA > 0) then
  begin
    if FFiltered then
      Query( Value )
    else
      Query( '' );
  end;
end;

{==============================================================================
  SetFiltered
 ==============================================================================}
procedure TApolloTable.SetFiltered( Value : Boolean );
begin
	inherited SetFiltered( Value );
	FFiltered := Value;
	SetSxQuery( FFilterStr );
end;

{LAS: 04-06-00, Added}
{==============================================================================
  SetFilterText
 ==============================================================================}
procedure TApolloTable.SetFilterText(const Value: string);
begin
  FFilterStr := Value;
  SetSxQuery( FFilterStr );         {LAS: 04-24-00, Added}
  inherited SetFilterText(Value);   {CRVG: 2001-05-01}
end;

{==============================================================================
  QueryRecCount
 ==============================================================================}
function TApolloTable.QueryRecCount: Integer;
begin
	Result := 0;
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
		Result := cs_QueryRecCount;
	end;
end;

{==============================================================================
  QueryTest
 ==============================================================================}
function TApolloTable.QueryTest( Value: String ): Integer;
begin
  Result := 0;
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
		Result := cs_QueryTest( PChar( Value ));
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  ReadOnly (SET)
 ==============================================================================}
procedure TApolloTable.SetReadOnly( Value : WordBool );
begin
	if (Value <> FReadOnly) then
  begin
		if (iWA > 0) then
	  	RaiseApollo( ERR_MUST_CLOSE_TABLE );
    FReadOnly := Value;
  end;
end;

{==============================================================================
  Recall
 ==============================================================================}
procedure TApolloTable.Recall;
begin
	// Can't modify a ReadOnly table
	if ReadOnly and (FErrorFunc = -1) then
  	RaiseApollo( ERR_TABLE_IS_READONLY );

	if IsEmpty and (FErrorFunc = -1) then
  	RaiseApollo( ERR_TABLE_IS_EMPTY );

	if (iWA > 0) then
  begin
	  cs_Select( iWA );
    cs_Recall;
    Resync([]);
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  RecCount -- (Called GetSxRecCount to avoid naming conflict with
               protected GetRecCount)
 ==============================================================================}
function TApolloTable.GetSxRecCount : Integer;
begin
	Result := 0;
	if (iWA > 0) then
  begin
	  cs_Select( iWA );  {LAS: 07-18-00, Moved above IF statement below}

    {MED: 09-06-00, commented as per Stansilav suggestion.}
	 // if FCountNeedsRefresh or (FRowCount = -1) then  {LAS: 07-13-00, Added}
  	  RefreshCount;                                 {LAS: 07-13-00, Added}

		Result := FRecCount;  {LAS: 05-30-00, Changed from cs_RecCount}
	end;
end;

{==============================================================================
  RecNo -- (Called GetSxRecno to avoid naming conflict with protected GetRecNo)
 ==============================================================================}
function TApolloTable.GetSxRecNo: Integer;
var
  DataBuffer: PChar;
begin
  Result := 0;
  if (iWA > 0) then
  begin
  if FSpeedMode then
  begin
    cs_Select( iWA );
    Result := cs_RecNo;
  end
  else
    begin
      if State = dsCalcFields then
        DataBuffer := CalcBuffer
      else
        DataBuffer := ActiveBuffer;
      Result := PRecInfo(DataBuffer + FRecInfoOfs).Bookmark;
    end;
  end;
end;

{==============================================================================
  RecSize
 ==============================================================================}
function TApolloTable.RecSize: Integer;
begin
  Result := 0;
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
		Result := cs_RecSize;
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  RefreshCount
 ==============================================================================}
procedure TApolloTable.RefreshCount;
begin
	if iWA > 0 then              {LAS: 08-13-00, Added}
  begin                        {LAS: 08-13-00, Added}
    cs_Select(iWA);            {LAS: 08-13-00, Added}
    FRecCount := cs_RecCount;  {LAS: 05-30-00, Added}
    case Ord( FSequencedCount ) of
      Ord( scAllCount ):
        FRowCount := Count;        // Most accurate, slowest
      Ord( scFilterOnlyCount ):
      begin
        FRowCount := QueryRecCount; // Fast for queries/scopes, otherwise slow
      end;
      Ord( scRawCount ):
        FRowCount := FRecCount;    // Fastest, but may not be accurate  {LAS: Changed from RecCount to FRecCount}
    end;

    FCountNeedsRefresh := False;
  end;                        {LAS: 08-13-00, Added}
end;

{==============================================================================
  Reindex
 ==============================================================================}
procedure TApolloTable.Reindex;
begin
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
		cs_Reindex;
		Resync([]);
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  Replace
 ==============================================================================}
procedure TApolloTable.Replace( sFieldName: String; iDataType: SmallInt;
                                  vpData: Pointer );
var
	lVal : Integer;
  tVal : TTimeStamp;
  cpDate : Array[0..10] of Char;

begin
	// Can't modify a ReadOnly table
	if ReadOnly and (FErrorFunc = -1) then
 		RaiseApollo( ERR_TABLE_IS_READONLY );

	// Unless SpeedMode is True, we must be in Edit or Insert mode
//	if (FSpeedMode = False) and not (State in [dsEdit, dsInsert]) then   {LAS: 07-24-00, Removed}
//		RaiseApollo( ERR_NOT_IN_EDIT_OR_INSERT_MODE );       {LAS: 07-24-00, Removed}

	if (FSpeedMode = False) then
  begin
		case iDataType of
		 	R_CHAR, R_CHAREX:
	     	begin  {LAS: 06-25-00, Modified to accurately handle 1-byte values}
        {MED: 10-13-00, Shawn reported a problem with the first line which seems to be true
            Change it to one line using the String typecast
        		FieldByName( sFieldName ).AsString := String( pchar(vpData) );
        }
         	if StrLen( @vpData )= 1 then                            //MED: 10-13-00, changed from Length to StrLen and took out the StrPas
         		FieldByName( sFieldName ).AsString := StrPas( @vpData )
         	else
						FieldByName( sFieldName ).AsString := StrPas( vpData );
        end;
      R_INTEGER:
        FieldByName( sFieldName ).AsInteger := Integer( vpData^ );
      R_LONG:
        FieldByName( sFieldName ).AsInteger := Integer( vpData^ );
      R_DOUBLE:
        FieldByName( sFieldName ).AsFloat := Double( vpData^ );
      R_LOGICAL:
        FieldByName( sFieldName ).AsBoolean := Boolean( vpData^ );
			R_JULIAN:
      	begin
					lVal := Integer( vpData^ );
					FieldByName( sFieldName ).AsDateTime := JulianToDT( lVal );
				end;
    	R_DATESTR:
      	begin
        	Move( vpData^, cpDate, SizeOf( cpDate ));
          tVal := DateTimeToTimeStamp( StrToDateTime( cpDate ));
					SetFieldData( FindField( sFieldName ), @tVal.Date );
				end;
    	R_MEMO:
      	begin
					FieldByName( sFieldName ).AsString := PChar( vpData );
				end;
    	R_BLOBFILE:
      	begin
					(FieldByName( sFieldName ) as TBlobField).LoadFromFile(
                                                              String( vpData ));
				end;
	    else
				SetFieldData( FindField( sFieldName ), vpData )
		end;
	end
	else
	begin
		if (iWA > 0) then
  	begin
	  	cs_Select( iWA );
			cs_Replace( PChar( sFieldName ), iDataType, vpData );
		end
	  else
			RaiseApollo( ERR_MUST_OPEN_TABLE );
	end;
end;

{==============================================================================
  RLock
 ==============================================================================}
function TApolloTable.RLock( Value: Integer ): WordBool;
begin
  Result := FALSE;
	if (iWA > 0) then
  begin
		cs_Select( iWA );
    Result := cs_RLock( Value );
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  RYOFilterActivate
 ==============================================================================}
function TApolloTable.RYOFilterActivate(
                          iFilterHandle, iBoolOperation: SmallInt ): WordBool;
begin
  Result := FALSE;
	if (iWA > 0) then
  begin
    cs_Select( iWA );
    Result := cs_RYOFilterActivate( iFilterHandle, iBoolOperation );
		FCountNeedsRefresh := True;
    Resync([]);
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  RYOFilterCopy
 ==============================================================================}
function TApolloTable.RYOFilterCopy : Integer;
begin
  Result := 0;
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
	  Result := cs_RYOFilterCopy;
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  RYOFilterCreate
 ==============================================================================}
function TApolloTable.RYOFilterCreate : Integer;
begin
  Result := 0;
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
	  Result := cs_RYOFilterCreate;
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  RYOFilterDestroy
 ==============================================================================}
function TApolloTable.RYOFilterDestroy( iFilterHandle: SmallInt ): WordBool;
var
  iOrd : Integer;
begin
  Result := FALSE;
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
	  Result := cs_RYOFilterDestroy( iFilterHandle );
    if iFilterHandle = 0 then
    begin
      iOrd := FIndexOrd;
      SetOrder(0);
      SetOrder(iOrd);
    end;
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  RYOFilterGetBit
 ==============================================================================}
function TApolloTable.RYOFilterGetBit( iFilterHandle: SmallInt;
                                                   lRecNo: Integer ): WordBool;
begin
  Result := FALSE;
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
	  Result := cs_RYOFilterGetBit( iFilterHandle, lRecNo );
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  RYOFilterSave
 ==============================================================================}
function TApolloTable.RYOFilterSave( iFilterHandle: SmallInt;
                                       sFileName: String): WordBool;
begin
  Result := FALSE;
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
	  Result := cs_RYOFilterSave( iFilterHandle, PChar( sFileName ));
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  RYOFilterRestore
 ==============================================================================}
function TApolloTable.RYOFilterRestore( sFileName : String ): WordBool;
begin
  Result := FALSE;
	if (iWA > 0) then
  begin
    cs_Select( iWA );
    Result := cs_RYOFilterRestore( PChar( sFileName ));

    cs_GoTop;
		if not FSpeedMode then
    begin
      with PRecInfo(ActiveBuffer + FRecInfoOfs)^ do
      begin
				if cs_Eof  then
	        BookmarkFlag := bfEOF
        else
	        BookmarkFlag := bfCurrent;
        Bookmark := cs_RecNo;
      end;
    end;

		FCountNeedsRefresh := True;
    Resync([]);
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  RYOFilterSetBit
 ==============================================================================}
function TApolloTable.RYOFilterSetBit( iFilterHandle: SmallInt;
                           lRecNo: Integer; bSetBitOn: WordBool ): WordBool;
begin
  Result := FALSE;
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
	  Result := cs_RYOFilterSetBit( iFilterHandle, lRecNo, SmallInt( bSetBitOn ));
		FCountNeedsRefresh := True;
    Resync([]);
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  RYOKeyAdd
 ==============================================================================}
function TApolloTable.RYOKeyAdd( sTagName, sKeyValue: String ): WordBool;
begin
  Result := FALSE;
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
	  Result := cs_RYOKeyAdd( PChar( sTagName ), PChar( sKeyValue ));
		FCountNeedsRefresh := True;
    Resync([]);
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  RYOKeyDrop
 ==============================================================================}
function TApolloTable.RYOKeyDrop( sTagName: String ): WordBool;
begin
  Result := FALSE;
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
	  Result := cs_RYOKeyDrop( PChar( sTagName ));
		FCountNeedsRefresh := True;
    //Resync([]);     {MED: 08-24-00, commented.  }
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  ExtraIndexesChanged
 ==============================================================================}
procedure TApolloTable.ExtraIndexesChanged( Sender: TObject );
begin
  IndexDefs.Updated := False;
  UpdateIndexDefs;
end;

{==============================================================================
  SetExtraIndexes
 ==============================================================================}
procedure TApolloTable.SetExtraIndexes( Value: TStringList );
begin
  if (Value <> FExtraIndexes) then
  begin
  	IndexName := '';
		FExtraIndexes.Assign( Value );
		// If the table is already opened, open the indexes
		if (iWA > 0) then
    begin
      ActivateIndexes;
			IndexDefs.Updated := False;
			UpdateIndexDefs;
    end;
	end;
end;

{==============================================================================
  SetFilter
 ==============================================================================}
procedure TApolloTable.SetFilter( Value: Variant );
var
	sVal: String;
begin
	if (iWA > 0) then
  begin
  	if (Value = '') or (Value = #0) then
    	sVal := ''
		else
    	sVal := Value;

	  cs_Select( iWA );

		// If SpeedMode is TRUE, do it and exit early
    if FSpeedMode then
    begin
			cs_SetFilter( PChar( sVal ));
      exit;  // Early exit
    end;

    if (CalcFieldsSize > 0) then
    begin
    	// Try to check if a calculated field is actually used within the query
      // expression and, if so, raise an exception and bail. (FUTURE FEATURE)
    end;

    if (sVal > '') and (cs_EvalTest( PChar( sVal )) <> 3) then
			RaiseApollo( ERR_INVALID_QUERY_EXPRESSION );

    BeginScroll;
		cs_SetFilter( PChar( sVal ));
    cs_GoTop;

		if not FSpeedMode then
    begin
      with PRecInfo(ActiveBuffer + FRecInfoOfs)^ do
      begin
				if cs_Eof  then
	        BookmarkFlag := bfEOF
        else
	        BookmarkFlag := bfCurrent;
        Bookmark := cs_RecNo;
      end;
    end;

    FCountNeedsRefresh := True;
		Resync([]);
    EndScroll;
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  SetGaugeHook
 ==============================================================================}
procedure TApolloTable.SetGaugeFunc( pFunc: Integer );
begin
	// Global, so no need to make sure that a table is already open
  cs_SetGaugeFunc( pFunc );
end;

{==============================================================================
  SetGaugeHook
 ==============================================================================}
procedure TApolloTable.SetGaugeHook( Handle: THandle );
begin
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
	  cs_SetGaugeHook( Handle );
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{==============================================================================
  SetOrder
 ==============================================================================}
function TApolloTable.SetOrder( iIndex: Integer ): Integer;
begin
  Result := 0;
  if (iWA > 0) then
  begin
    CheckBrowseMode;
    cs_Select( iWA );

    Result := cs_SetOrder( iIndex );
    if FTableType =  ttSXNTX then
      FIndexName := String( cs_IndexName( iIndex ))
    else
      FIndexName := String( cs_TagName( iIndex ));

     FCountNeedsRefresh := True;

     Go( RecNo );      // 08-08-01 - CRVG
  end
  else
    RaiseApollo( ERR_MUST_OPEN_TABLE );
end;


{==============================================================================
  SetPassword
 ==============================================================================}
procedure TApolloTable.SetPassword( sEncodeKey: String );
begin
	FPassword := sEncodeKey;
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
		cs_SetPassword( PChar( FPassword ));
    Resync([]);
	end;
end;

{==============================================================================
  SetQueryBit
 ==============================================================================}
procedure TApolloTable.SetQueryBit( lRecNum: Integer; bValue: WordBool );
begin
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
		cs_SetQueryBit( lRecNum, bValue );
	end
  else                                                 
		RaiseApollo( ERR_MUST_OPEN_TABLE );  
end;

{==============================================================================
  SetRange
 ==============================================================================}
procedure TApolloTable.SetRange( const aStartVal, aEndVal: Array of Const );

	// Convert TVarRec value into String to pass to SetScope
	function VarToString( xVar : TVarRec ): String;
	begin
		case xVar.VType of
			vtBoolean:
				if xVar.VBoolean then
					Result := '.T.'
				else
					Result := '.F.';
			vtChar:
				Result := xVar.VChar;
			vtCurrency:
				Result := FloatToStr( xVar.VCurrency^ );
			vtExtended:
				Result := FloatToStr( xVar.VExtended^ );
			vtInteger:
				Result := IntToStr( xVar.VInteger );
			vtPChar:
				Result := StrPas( xVar.VPChar );
			vtString:
				Result := xVar.VString^;
  		vtPointer,
			vtAnsiString,
			vtWideString:
				Result := StrPas( PChar( xVar.VPointer ));
			vtVariant:
				Result := VarToStr( xVar.VVariant^ );
    else
			Result := '';
		end;
	end;

begin
	SetScope( VarToString( aStartVal[0] ), VarToString( aEndVal[0] ));
end;

{==============================================================================
  SetRelation
 ==============================================================================}
procedure TApolloTable.SetRelation( iChildArea: Integer; sKeyExpr: String );
begin
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
		cs_SetRelation( iChildArea, PChar( sKeyExpr ));
	end
  else                                                 
		RaiseApollo( ERR_MUST_OPEN_TABLE );  
end;

{==============================================================================
  SetScope
 ==============================================================================}
function TApolloTable.SetScope( sLowVal, sHiVal: String ): WordBool;
var
	lCurRec : Integer;
  sTempVal:  String;  {LAS: 06-21-00, Added}
begin
	// Make sure an index order is active

  Result := FALSE;
	if (iWA > 0) then
  begin
		CheckBrowseMode;
    cs_Select( iWA );
		lCurRec := RecNo;

    //===========================================================
    // M.O.: Commented out
    {LAS: 06-21-00, Added}  {LAS: 07-28-00, Added #0 check below}
    // Make sure sLowVal is not higher than sHiVal; If so, exchange them
    //if (sHiVal <> #0) and (sHiVal > '') and (sLowVal > sHiVal) then    {LAS: 07-03-00, Added sHiVal empty check}
    //begin
    //  sTempVal := sLowVal;
    //  sLowVal := sHiVal;
    //  sHiVal := sTempVal;
    //end;
    //===========================================================

    Result    := cs_SetScope( PChar( sLowVal ), PChar( sHiVal ));
    //FRecCount := cs_RecCount;     {MED: 10-21-00, added}
    //FRowCount := cs_RecCount;     {MED: 10-21-00, added}

    // If at least one record meets Scope condition, preserve record pointer
		if not cs_Eof then
    begin
	   	cs_Go( lCurRec );
      // Force record position reset in case current record is now out of scope
			cs_Skip(1);
			if cs_Eof then
      	cs_GoBottom
      else
	    	cs_Skip(-1);
    end;

		if not FSpeedMode then
    begin
      with PRecInfo(ActiveBuffer + FRecInfoOfs)^ do
      begin
				if cs_Eof  then
	        BookmarkFlag := bfEOF
        else
	        BookmarkFlag := bfCurrent;
        Bookmark := cs_RecNo;
      end;
    end;

    FCountNeedsRefresh := True;
    Resync([]);

    // M.O.: modified in order to get scoped status correctly
    sTempVal := cs_GetScope(0);
    FScoped := ( sTempVal <> '');// ((sLowVal + sHiVal) > '') ;
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );

end;

{==============================================================================
  SetSpeedMode
 ==============================================================================}
procedure TApolloTable.SetSpeedMode( Value: WordBool );
begin
  FSpeedMode := Value;
  if iWA > 0 then              {LAS: 04-20-00, Added}
  begin                        {LAS: 04-20-00, Added}
    if (Value = True) then
      UpdateCursorPos
    else
    begin                                                         {LAS: 04-17-00, Added}
      PRecInfo(ActiveBuffer + FRecInfoOfs).Bookmark := cs_RecNo;  {LAS: 04-17-00, Added}
      Resync([]);
    end;                                                          {LAS: 04-17-00, Added}
  end;                         {LAS: 04-20-00, Added}
end;

{==============================================================================
  SetTranslate
 ==============================================================================}
procedure TApolloTable.SetTranslate( Value: WordBool );
begin
  if (Value <> FTranslate) then
  begin
    FTranslate := Value;
  	if (iWA > 0) then
    begin
	    cs_Select( iWA );
  		cs_SetTranslate( FTranslate );
	  	Resync([]);
  	end;
  end;
end;

{==============================================================================
	GetFRecOrdinal  // CRVG 2001-10-22
 ==============================================================================}
function TApolloTable.GetFRecOrdinal: integer;
begin
	if ffRecOrdinal = FRecOrdinalNeedsRefresh then
		ffRecOrdinal := Trunc(sx_OrderPosGet * FRowCount);
	Result := ffRecOrdinal;
end;

{==============================================================================
  SetTurboRead
 ==============================================================================}
procedure TApolloTable.SetTurboRead( Const Value: WordBool );
begin
	FTurboRead := Value;
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
		cs_SetTurboRead( Value );
	end;
end;

{==============================================================================
  TagArea
 ==============================================================================}
function TApolloTable.TagArea( sTagName: String ): Integer;
begin
	Result := 0;
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
		Result := cs_TagArea( PChar( sTagName ));
	end;
end;

{==============================================================================
  TagName
 ==============================================================================}
function TApolloTable.TagName( iTagArea: Integer ): String;
begin
	Result := '';
	if (iWA > 0) then
  begin
	  cs_Select( iWA );
		Result := cs_TagName( iTagArea );
	end;
end;

{==============================================================================
  Unlock
 ==============================================================================}
procedure TApolloTable.Unlock( lRecNum: Integer );
begin
	if (iWA > 0) then
  begin
  	cs_Select( iWA );
		cs_Unlock( lRecNum );
	end;
end;

{==============================================================================
  Version
 ==============================================================================}
function TApolloTable.GetVersion: String;
begin
	Result := GetApVersion;
end;

{==============================================================================
  Zap
 ==============================================================================}
procedure TApolloTable.Zap;
begin
	// Can't ZAP a ReadOnly table
	if ReadOnly and (FErrorFunc = -1) then
	 	RaiseApollo( ERR_TABLE_IS_READONLY );

	// Can't ZAP if table not opened EXCLUSIVE
	if not Exclusive and (FErrorFunc = -1) then
  	RaiseApollo( ERR_MUST_OPEN_EXCLUSIVE );

	if (iWA > 0) then
  begin
    cs_Select( iWA );
    cs_Zap;
    FCountNeedsRefresh := True;  {LAS: 05-30-00, Added}
    Resync([]);
	end;
end;

(*============================================================================*)
(*                          END OF MAIN SDE FUNCTIONALITY                     *)
(*============================================================================*)


{==============================================================================
  ActivateIndexes (used with ExtraIndexes property)
 ==============================================================================}
procedure TApolloTable.ActivateIndexes;
var
	iCnt : Integer;
	iOrd : Integer;
	iOpenVal : Integer;
	sIFName : String;
begin

	// No ExtraIndexes?  If none, early exit
	if FExtraIndexes.Count = 0 then
  	Exit;

	if (iWA > 0) then
  begin
		cs_Select( iWA );
		iOrd := IndexOrd;
    cs_CloseIndexes;

    // Reopen structural index, if any
		sIFName := ChangeFileExt( TableName, GetIndexExt );  {LAS: 05-23-00, Added}
		if not (FTableType = ttSXNTX) then                   {LAS: 05-30-00, Added}
			if cs_FileExists( sIFName ) then                   {LAS: 05-23-00, Added}
				cs_IndexOpen( PChar( sIFName ));                 {LAS: 05-23-00, Added}

		cs_SetOrder( 0 );

		for  iCnt := 1 to FExtraIndexes.Count do
		begin
			sIFName := FExtraIndexes.Strings[iCnt-1];

      if Trim( sIFName ) = '' then    {LAS: 04-28-00, Added}
      	continue;                     {LAS: 04-28-00, Added}

			if (ExtractFilePath( sIFName ) = '') then
				sIFName := FDataPath + sIFName;

			if (ExtractFileExt(sIFName) = '') then
			begin
				case Ord( FTableType ) of
					Ord( ttSXNTX ):
						sIFName := ChangeFileExt( sIFName, '.NTX' );

					Ord( ttSXFOX ):
						sIFName := ChangeFileExt( sIFName, '.IDX' );

					Ord( ttSXNSX ):
						sIFName := ChangeFileExt( sIFName, '.NSX' );

					Ord( ttSXNSX_DBT ):
						sIFName := ChangeFileExt( sIFName, '.NSX' );

				end;
			end;

      if not cs_FileExists( sIFName ) and (FErrorFunc = -1) then
				RaiseApollo( ERR_INVALID_INDEXNAME + ' : ' +
  	                               FExtraIndexes.Strings[iCnt-1] );

      iOpenVal := cs_IndexOpen( PChar( sIFName ));

			// If no order was previously active; We'll set order to the first one
      // opened here once we're done
			if (iOrd = 0) then
      begin
				iOrd := iOpenVal;
	      FIndexName := sIFName;
      end;
		end;
		cs_SetOrder( iOrd );
	end;

	IndexDefs.Updated := False;
	UpdateIndexDefs;
	FCountNeedsRefresh := True;

	if FCursorOpen then
		Resync([])
	else
		UpdateCursorPos;
end;

{==============================================================================
  GetDatabase
 ==============================================================================}
function TApolloTable.GetDatabase: TApolloDatabase;
var
  bFound : Boolean;
  sPath : String;
  SearchRec : TSearchRec;
  Temp: TComponent;
  SearchObj : TApolloSearchObject;  {LAS: 04-20-00, Added}
begin
	Result := FDatabase;
  bFound := False;

  if Assigned(FDataBase) or (FAccessMethod = amServer) then   {LAS: 06-15-00, Added amServer check}
    exit;

  SearchObj := TApolloSearchObject.Create;
  try
    Temp := SearchObj.SearchFirstComponent(TApolloDataBase);
    while Temp <> nil do
    begin
      if (Temp is TApolloDataBase) then
      begin
        if (UpperCase(TApolloDatabase(temp).DataPathAlias) = UpperCase(DatabaseName)) then
        begin
          FDatabase := TApolloDataBase(temp);
          bFound    := TRUE;
          break;
        end;
      end;
      Temp := SearchObj.SearchNextComponent(TApolloDataBase);
    end;
  finally
    SearchObj.Free;
  end;

  if not bFound then
  begin
    // we didn't find it
		sPath := DatabaseName;

		// Strip backslash from end of path
		if (Length(sPath) > 0) and (sPath[ Length(sPath) ] = '\') then
		  sPath := Copy( sPath, 1, Length( sPath )-1 );

		// if (sPath = '') then sPath := '\';
    // sPath := ExtractFileDir( sPath );
    // ShowMessage('sPath >'+sPath+'<');

    if SysUtils.FindFirst( sPath, faDirectory, SearchRec ) <> 0 then
    begin
      // Only raise exception if DatabaseName is not empty but also not valid
      if (Length(DatabaseName) > 0) {and (FErrorFunc = -1)} then   {LAS: 05-05-00, Removed ErrorFunc check}
        RaiseApollo( ERR_INVALID_DATABASENAME +' : ' + DatabaseName );
		end;
    SysUtils.FindClose( SearchRec );
  end;
	Result := FDatabase;
end;

{==============================================================================
  UpdateFieldAndIndexDefs
 ==============================================================================}
procedure TApolloTable.UpdateFieldAndIndexDefs;
begin
	FieldDefs.Clear;
{$ifdef NOT_D3_OR_CB3}
	if (csDesigning in ComponentState) then
		Fields.Clear;
{$endif}

	// Update FieldDefs & IndexDefs
	if (not Active) then
  begin
{$ifndef APOLLO_CS}
		if (not (csReading in ComponentState)) then
    begin
			if Exists then
      begin
				OpenTable;
  			InternalClose;
    	end;
    end;
{$endif}
	end
  else
  begin
		InternalInitFieldDefs;
    IndexDefs.Updated := False;
    UpdateIndexDefs;
  end;
end;

{==============================================================================
  SetTableName
 ==============================================================================}
procedure TApolloTable.SetTableName( Value: String );
begin
  if (Value <> FTableName) then
  begin
		if (iWA = 0) and (not Active) then   {LAS: 06-15-00, Added Active check, to be sure}
		begin
      //FPassword := '';  09-12-00, commented
	    FTableName := Value;
			UpdateFieldAndIndexDefs;   {LAS: 04-20-00, Moved up}
    end
    else
	  	RaiseApollo( ERR_MUST_CLOSE_TABLE );
  end;
//	UpdateFieldAndIndexDefs;   {LAS: 04-20-00, Moved up}
end;

{==============================================================================
  SetDatabaseName
 ==============================================================================}
procedure TApolloTable.SetDatabaseName( Value: String );
//var                         {LAS: 05-09-00, Removed}
//	sTblName : String;        {LAS: 05-09-00, Removed}
begin
  if (Value <> FDatabaseName) then
  begin
    FDatabase := nil;
    FDatabaseName := Value;

    // If local physical path, change AccessMethod to amLocal {LAS: 06-03-00, Added}
    if (Pos(':', FDatabaseName) > 0) or (Pos('\', FDatabaseName) > 0) then
    	FAccessMethod := amLocal;

		if Assigned( FApolloConnection ) then
    	FApolloConnection.DataBaseName := Value;

	  if ((Active) and (FTableName > '')) then
  	begin
			FDataPath := PathFromAlias;

      // Add backslash to Path if not there already
	    if (Length(FDataPath) > 0) and (FDataPath[Length(FDataPath)] <> '\') then
      	FDataPath := FDataPath + '\';

			// Append *.DBF to end of file, if no extension there
//			sTblName := FDataPath + TableName;                {LAS: 05-09-00, Removed}
//			if (ExtractFileExt( sTblName ) = '') then         {LAS: 05-09-00, Removed}
//				sTblName := sTblName + '.DBF';                  {LAS: 05-09-00, Removed}

      if not Exists then
    		Active := False;

			UpdateFieldAndIndexDefs;
		end;
	end;

  // Toggle DataSource.Enabled to force data to display at design-time
  if Assigned( DataSource ) {and (DataSource.Enabled)} then   {LAS: 05-09-00, Moved Enabled check down}
  begin
		if  DataSource.Enabled then  {LAS: 05-09-00, Added}
    begin                        {LAS: 05-09-00, Added}
	    DataSource.Enabled := False;
  	  DataSource.Enabled := True;
    end;                         {LAS: 05-09-00, Added}
  end;
end;

{==============================================================================
  SetScopeBegin
 ==============================================================================}
procedure TApolloTable.SetScopeBegin(Value: String);
begin
	FScopeBegin := Value;
	if FScoped then
		SetScope( FScopeBegin, FScopeEnd );  // Refresh Scope
end;

{==============================================================================
  SetScoped
 ==============================================================================}
procedure TApolloTable.SetScoped(Value: Boolean);
begin
	if FScoped <> Value then
  begin
		if Value then
			FScoped := SetScope( FScopeBegin, FScopeEnd )  // Set Scope
	  else
    begin
  		SetScope( '', '' );  // Clear Scope
      FScoped := False;
    end;
  end;
end;

{==============================================================================
  SetScopeEnd
 ==============================================================================}
procedure TApolloTable.SetScopeEnd(Value: String);
begin
	FScopeEnd := Value;
	if FScoped then
		SetScope( FScopeBegin, FScopeEnd );  // Refresh Scope
end;

{==============================================================================
  SetIndexName
 ==============================================================================}
procedure TApolloTable.SetIndexName( Value: String );
var
  NewArea: SmallInt;
  sTemp : String;

  {LAS: 08-12-00, Added}
  // Returns just the base file name, without extension; UpperCases result
  function GetBaseName( sVal: String ): String;
  begin
  	sVal := ExtractFileName( sVal );
  	Result := UpperCase(Copy( sVal, 1, Pos('.', sVal + '.' )-1 ));
  end;

begin
  if FAccessMethod = amLocal then
  begin
    if (FIndexName <> Value) then
    begin
	    // If base .NTX file names are same, bail now!
    	if (FTableType = ttSXNTX) then
      	if GetBaseName( FIndexName ) = GetBaseName( Value ) then   {LAS: 08-12-00, Changed ExtractFileName to GetBaseName}
        	Exit;

      FIndexName := Value;
	    if (iWA > 0) then
  	  begin
   	  	cs_Select( iWA );
 	  		if Active then
		  		CheckBrowseMode;
			  if Value = '' then
    	   	NewArea := 0
	  		else
		  	begin
					// Find area number in ExtraIndexes list
	        if (FTableType = ttSXNTX) or
             (ExtractFileExt( UpperCase( Value )) = '.IDX') then  {LAS: 08-12-00, Added}
	        begin
  	      	// Check in ExtraIndexes list first
    	    	NewArea := (FExtraIndexes.IndexOf( Value ) + 1);
      	    if (NewArea = 0) then
        	  begin
  	        	sTemp := Value;
    	      	if (ExtractFilePath( sTemp ) = '') then
		  				 	sTemp := FDataPath + sTemp;
        	  	if (ExtractFileExt( sTemp ) = '') then
              	if FTableType = ttSXNTX then           {LAS: 08-12-00, Added}
					  			sTemp := sTemp + '.NTX'
                else                                   {LAS: 08-12-00, Added}
	                sTemp := sTemp + '.IDX';             {LAS: 08-12-00, Added}
					  	if cs_FileExists( sTemp ) then
  	          begin
							  NewArea := cs_IndexOpen( PChar( sTemp ));
        	      IndexDefs.Updated := False;
          	    UpdateIndexDefs;
		  				end;
			  		end;
				  end
  				else
	      		NewArea := cs_TagArea( PChar( Value ));

		      if (Active and (NewArea = 0)) and (FErrorFunc = -1) then
			    	RaiseApollo( ERR_INVALID_INDEXNAME + ' : ' + Value );
  			end;

	     	cs_SetOrder( NewArea );
	    	FCountNeedsRefresh := True;
  	    UpdateCursorPos;
    	  if Active then
  	 	  begin
    	 	  CheckMasterRange;
      	 	Resync([]);
	      end;
      end;
	  end;
  end
  else
  begin
    // If base .NTX file names are same, bail now!
    if (FTableType = ttSXNTX) then                               {LAS: 08-12-00, Added}
      if GetBaseName( FIndexName ) = GetBaseName( Value ) then   {LAS: 08-12-00, Added}
        Exit;                                                    {LAS: 08-12-00, Added}

    if (FIndexName <> Value) then
    begin
      FIndexName := Value;
	    if (iWA > 0) then
    	begin
 	    	cs_Select( iWA );
 		  	if Active then
			  	CheckBrowseMode;
  			if Value = '' then
    	   	NewArea := 0
		  	else
			  begin
  	    	// Find area number in ExtraIndexes list
  				if (FTableType = ttSXNTX) then
	        begin
  	      	// Check in ExtraIndexes list first
    	    	NewArea := (FExtraIndexes.IndexOf( Value ) + 1);
      	    if (NewArea = 0) then
        	  begin
  	        	sTemp := Value;
    	      	if (ExtractFilePath( sTemp ) = '') then
		  					sTemp := FDataPath + sTemp;
        	  	if (ExtractFileExt( sTemp ) = '') then
              	if FTableType = ttSXNTX then           {LAS: 08-12-00, Added}
					  			sTemp := sTemp + '.NTX'
                else                                   {LAS: 08-12-00, Added}
	                sTemp := sTemp + '.IDX';             {LAS: 08-12-00, Added}
					  	if cs_FileExists( sTemp ) then
  	          begin
							  NewArea := cs_IndexOpen( PChar( sTemp ));
        	      IndexDefs.Updated := False;
          	    UpdateIndexDefs;
		  				end;
			  		end;
				  end
  				else
	      		NewArea := cs_TagArea( PChar( Value ));

		      if Active and (NewArea = 0) then
			    	RaiseApollo( ERR_INVALID_INDEXNAME + ' : ' + Value );
  			end;
	    	cs_SetOrder( NewArea );
	    	FCountNeedsRefresh := True;
  	    UpdateCursorPos;
    	  if Active then
  	 	  begin
      	 	Resync([]);
	      end;
      end;
	  end;
  end;
end;

{==============================================================================
 GetUniqueAliasName
 This is a private method that creates a unique alias name from the table name.

 Because the same file could be open more than once, we'll need to open the
 second (or more) file with a different alias.  This makes sure we have a
 unique alias.
 ==============================================================================}
function TApolloTable.GetUniqueAliasName( sTableName: String ): String;
var
  i : Integer;
begin
	Result := sTableName;

	if not (sTableName = EmptyStr) then
  begin
	  // Strip off any path
		Result := UpperCase( ExtractFileName( Result ));

	  // Strip of any file extention
  	if Pos('.', Result) > 0 then
  		Result := Copy( Result, 1, Pos('.', Result)-1);

		// Truncate file name to eight characters, if longer
  	if Length( Result ) > 8 then
  		Result := Copy( Result, 1, 8 );

		// Replace certain ASCII values with an underscore
		for i := 1 to Length( Result ) do
  		if not (Result[i] in ['A'..'Z']) then                            
    		Result[i] := '_';

		// Make sure Database Engine .DLLs are loaded for local tables
		if FAccessMethod = amLocal then   {LAS: 05-03-00, Added}
    	LoadDLL;                        {LAS: 05-03-00, Added}

		// Append numeric value to end of alias name, if name already in use
    i := 1;
    while (cs_WorkArea(PChar( Result )) > 0) do
    begin
      Result := Copy( Result, 1, 8 - Length( IntToStr(i) )) + IntToStr(i);
      Inc( i );
    end;
  end;
end;

{==============================================================================
  ApolloTableType
  Converts the Enumertaed TTableType to the SDE internal representation
 ==============================================================================}
function TApolloTable.ApolloTableType( Value: TApolloTableType ): Integer;
begin
  // Map to the internal types
  case Value of
    ttSXNTX: Result     := SDENTX;
    ttSXFOX: Result     := SDEFOX;
    ttSXNSX: Result     := SDENSX;
    ttSXNSX_DBT: Result := SDENSX_DBT;
  else
    Result := SDEFOX;
  end;
end;

{==============================================================================
  PathFromAlias - If TApolloTable Alias is used, gets physical path
 ==============================================================================}
function TApolloTable.PathFromAlias: String;
begin
	Result := DatabaseName;  // Default

  if ((Pos( '\', Result ) > 0) or    // Check for characters that would
      (Pos( ':', Result ) > 0) or    //   make it clear that this is a physical
      (Pos( '.', Result ) > 0)) and  //    file path and not an alias.
			not(Result = '') then   // If NOT empty, bail now
  	exit;

	if Assigned( Database ) then
		Result := Database.DataPath;

  // If no path was specified, use .DBF current directory
  if (Result = '') then
    GetDir( 0, Result );

  if (Length(Result) > 0) and (Result[ Length(Result) ] <> '\') then
    Result := Result + '\';
end;

{==============================================================================
  GetDataPath
  Determines physical path of the table.  Sets FDataPath variable.
 ==============================================================================}
procedure TApolloTable.GetDataPath;
begin
  // Get the path to the table...
  if (ExtractFilePath(FTableName) = '') then
	  FDataPath := PathFromAlias
  else
  	FDataPath := ExtractFilePath( FTableName );

  // Add backslash to Path if not there already
  if (Length(FDataPath) > 0) and (FDataPath[Length(FDataPath)] <> '\') then
    FDataPath := FDataPath + '\';
end;

{==============================================================================
  GetTableType  // MED: 03-27-01, changed from CheckTableType
  Prevents attempting to open a table w/ memo under the wrong table type.
  Does not automatically switch table types.
=========================================================
 MED: 03-23-01, This method WILL FAIL checking on NSX tables using DBT memos
 Note : there is nothing in the DBF that could tell us that
 we use DBT exenstion for MEMO file.
==============================================================================}
function TApolloTable.GetTableType:TApolloTableType;
var
	hDBF: Integer;
  Ch: Char;
  sMemoExt, sMemoFile: String;
begin
  Result := ttSXNONE;   // [Med] 09-26-01 changed from ttSXFOX

  if FAccessMethod = amLocal then
  begin
  	GetDataPath;

	  hDBF := FileOpen( FDatabaseName+FTableName, fmOpenRead or fmShareDenyNone );
    if (hDBF > 0) then
    begin
  	  sMemoExt := '';
      smemoFile := '';
      FileRead( hDBF, Ch, SizeOf( Ch ));
      FileClose( hDBF );

      // Clipper memo used? (regular or encrypted)
      if ((Ch = Char($83)) or (Ch = Char($86))) then
      begin
        sMemoFile := ChangeFileExt( FDatabaseName+FTableName, '.DBT' );
        Result := ttSXNTX;
      end

      // FoxPro memo used? (regular or encrypted)
      else if ((Ch = Char($F5)) or (Ch = Char($F6)) or (Ch = Char($30))) then
      begin
        sMemoFile := ChangeFileExt( FDatabaseName+FTableName, '.FPT' );
        Result    := ttSXFOX;
      end

      // HiPer-SIx memo used? (regular or encrypted)
      else if ((Ch = Char($E5)) or (Ch = Char($E6))) then
      begin
        sMemoFile := ChangeFileExt( FDatabaseName+FTableName, '.SMT' );
        Result    := ttSXNSX;
      end;
      //===================================================

      if (sMemoFile > '') and (not FileExists( sMemoFile )) and
                                                     (not (Ch = Char($30)) ) then
      begin
        Result := ttSXNONE;
        RaiseApollo( ERR_MISSING_MEMO_FILE + UpperCase(sMemoFile) );
      end;
    end
    else
      RaiseApollo(ERR_UNABLE_TO_OPEN_TABLE+ FDatabaseName+FTableName);
  end
  else
  begin
    Result := TApolloTableType(FApolloConnection.ap_GetTableType(FDatabaseName, FTableName));
  end;

  FTableType := Result;
end;

{==============================================================================
  OpenTable
  This opens the table and makes sure it was opened correctly.  It also creates
  a valid alias.  This is a private method used by InternalOpen and
  InternalCreateFieldDefs.
 ==============================================================================}
procedure TApolloTable.OpenTable;
var
	sTemp, FAliasName : String;
	iError: Integer;
  InternalTableType: SmallInt;  {LAS: 06-29-00, Added}
  InternalOpenMode: SmallInt;   {LAS: 06-29-00, Added}
  sError          : String;
{$ifdef NOT_D3_OR_CB3}
	Present: TDateTime;
	Year, Month, Day: Word;
{$endif}
begin
  if not FIndirectlyAssigned then
  	iWA := 0;

	if (TableName = '') and (FErrorFunc = -1) then
  begin
		RaiseApollo( Name + ' : ' + ERR_MISSING_TABLENAME_PROPERTY );
    exit;
	end;

	if FAccessMethod = amLocal then
  begin
  	LoadDLL;
    //SwitchTableType;     {MED: 03-26-01, commented.}
    FAliasName := GetUniqueAliasName( FTableName ); {MED: 05-20-00, Moved here }
  end;

  InternalTableType := ApolloTableType( FTableType );

  if Exclusive then        InternalOpenMode := OPEN_EXCLUSIVE
  else if ReadOnly then    InternalOpenMode := OPEN_READONLY
  else                     InternalOpenMode := OPEN_READWRITE;

  SetToApolloEnv;

  // Unable to find file. Table/path does not exist?

  if FAccessMethod = amLocal then
    if (not Exists) and (FErrorFunc = -1) then
      RaiseApollo( ERR_TABLE_DOES_NOT_EXIST + GetFullTableName );

{$ifdef NOT_D3_OR_CB3}
	// Set system TwoDigitCenturyWindow variable
  Present := Now;
  DecodeDate( Present, Year, Month, Day );
	TwoDigitYearCenturyWindow := Abs(Year - FEpoch) mod 100;
{$endif}


   // Open table
  if FAccessMethod = amLocal then
  begin
    if FApplyEnvSettings then
    begin
      sx_SetEpoch(FEpoch);
      sx_SetDeleted(FDeleted);
      sx_SetCentury(FCentury);
      sx_SetMemoBlockSize(FMemoBlockSize);
      sx_SetDateFormat(Ord(FDateFormat));
      sx_SetSoftSeek(FSoftSeek);
      sx_SetExact(FExact);
    end;

    if (FErrorFunc = -1) then
		  sx_ErrorLevel( 0 );

    if iWA < 1 then         // MED: 07-30-01, added to make AssignTable work
      iWA := cs_UseEx( PChar( FDataPath + ExtractFileName(FTableName)),
                            PChar( FAliasName ), InternalOpenMode,
                            InternalTableType, Ord(FCommitLevel) );

    if iWA < 1 then
    begin
      RaiseApollo('Unable to open table: '+ GetFullTableName );
      exit;
    end;

    sx_ErrorLevel( FErrorLevel );

    if (iWA > 0) then
    begin
      sx_SetTurboRead( FTurboRead );
      sx_SetTranslate (FTranslate);

      FSDEFieldCount  := sx_FieldCount;     // Get raw SDE field count for table
      FFieldCount     := FSDEFieldCount;    // MED: 10-21-00, added-will replace FSDEFieldCount eventually
      FRecSize        := sx_RecSize;
      // Is table encrypted?
      FIsEncrypted := IsEncrypted(0);
      if (FPassword > '') then
        sx_SetPassword( PChar( FPassword ));

      sx_SetTurboRead( FTurboRead );
      if (FIndexName <> '') then
      begin
        sTemp := FIndexName;
        SetIndexName( '' );       // Toggle index name off...
        SetIndexName( sTemp );    // ...and back on, to re-initialize
      end;
      ActivateIndexes;
      IndexDefs.Updated := False;
      UpdateIndexDefs;
      FRowCount       := sx_RecCount;
      FIndexCondition := sx_IndexCondition;
    end;
  end
  else
  begin

  	{LAS: 05-26-00, Added FDatabaseName check. Cannot be empty on remote tables}
    if FDatabaseName = '' then
    	sError := ERR_MISSING_DATABASENAME_PROPERTY
    else
      if iWA < 1 then         // MED: 07-30-01, added to make AssignTable work
	      iWA := FApolloConnection.ap_QuickUse(
                     FDataBaseName,
                     ExtractFileName(FTableName),
                     InternalOpenMode,
                     InternalTableType,
                     FRowCount, FSDEFieldCount, FRecSize, FIsEncrypted,
                     FPassword, FTurboRead,
                     FIndexName,
                     FEpoch, Ord(FDateFormat), FMemoBlockSize,
                     FDeleted, FCentury, FSoftSeek, FExact,
                     FTranslate,            // MED: 04-07-01
                     FApplyEnvSettings,
                     FIndexCondition,
                     Ord(FCommitLevel), // CRVG Added support for CommitLevel
                     FIndexOrd,
                     FIndexDefs,
                     iError,
                     sError);


    if iError > 0 then
    begin
      RaiseApollo(sError);
      iWA := 0;  // Just to be sure    {LAS: 05-26-00, Added}
      exit;                            {LAS: 05-26-00, Added}
    end;

    if (iWA > 0) then
    begin
      // The server sets active index to FIndexName already for structural index
      // the following is only for the inferior ttSXNTX drivers! -- MED
      if (FIndexName <> '') AND (FTableType = ttSXNTX) then
      begin
        sTemp := FIndexName;
        SetIndexName( '' );       // Toggle index name off...
        SetIndexName( sTemp );    // ...and back on, to re-initialize
      end;
      ActivateIndexes;
      IndexDefs.Updated := TRUE;
//      FRecCount := FRowCount; {LAS: 06-08-00, Added / 06-14-00, Moved down}
    end;
	end;

	FRecCount := FRowCount; {LAS: 06-14-00, Moved down}

  // Unable to open file
  if (iWA = 0) then
  begin
    if (FErrorFunc = -1) then
    begin //---------------------------( LAS: 04-20-00, Added Login Info Check )
    	if (FAccessMethod = amServer) then
       if (Assigned(FApolloConnection) and ((FApolloConnection.User = '') or
                                        (FApolloConnection.Password = ''))) then
	      RaiseApollo( ERR_MUST_ENTER_LOGIN_INFO );
        //-------------------------( LAS: 04-20-00, Added Login Info Check )

			{LAS: 06-15-00, Changed to enable error message for server table open failure}
			// Otherwise, for both Local or Remote, display generic open error message
      RaiseApollo( ERR_UNABLE_TO_OPEN_TABLE + GetFullTableName );
    end;
    Exit;
  end;

  FCountNeedsRefresh := True;
{$ifdef NOT_D3_OR_CB3}
  FUpdateLockList := True;
{$endif}

end;


{==============================================================================
  InternalOpen
 ==============================================================================}
procedure TApolloTable.InternalOpen;
var
  i, iFldCount: Integer;
  CurField: TField;
  sVal: String;
begin
  if iWA  < 1 then          // MED: 07-30-01, added to make AssignTable work
  begin
    FRowCount       := 0;
    FSDEFieldCount  := 0;
    FFieldCount     := FSDEFieldCount;    // MED: 10-21-00, added-will replace FSDEFieldCount eventually
  end;
  FFetchLastRec   := 0;
  FFetchReadCount := -1;

  ClearFetchedItems;

  if Active then
  	Close;

  if FAccessMethod = amServer then
  begin
    if FApolloConnection = nil then
     RaiseApollo( ERR_NO_APOLLOCONNECTION_SPECIFIED );

    if not FApolloConnection.Connected then
    begin
      iWA := 0;
      Active := False;
      RaiseApollo( ERR_NOT_CONNECTED );
    end;
	end;

  LOpeningTable := True;
  try

    InternalInitFieldDefs;

    if (iWA = 0) then
      exit;

    if DefaultFields then
      CreateFields;

    BindFields(True);    // Bind the fields
    FRecNo         := 0;
    FCalcFieldsOfs := 0;

    // Determine the base record size & set the relationship of field info/size.
    iFldCount := FieldCount;
    for i := 0 to iFldCount-1 do
    begin
      CurField := Fields[i];
      if CurField.FieldNo > 0 then
      begin
        TFieldInfo( LFieldOffsets[CurField.FieldNo-1]).NativeOffset :=
                                                              FCalcFieldsOfs;
        if CurField.DataSize = 0 then
          Inc( FCalcFieldsOfs, 1 )  // Leave enough room for NULL indicator
        else
          Inc( FCalcFieldsOfs, CurField.DataSize+1 );
      end;
    end;

    // Allocate the memory for the SDE buffer...
    LSDERecordBuffer := AllocMem(FRecSize);  {MED: 05-20-00 changed from cs_RecSize}

    FCursorOpen := True;
    FBlobCacheOfs := FCalcFieldsOfs + CalcFieldsSize;
    FRecInfoOfs := FBlobCacheOfs + (BlobFieldCount * SizeOf(TApBlobData));
    FRecBufSize := FRecInfoOfs + SizeOf(TRecInfo);

    // Tell TDataSet how big our Bookmarks are (REQUIRED)
    BookmarkSize := SizeOf( Integer );

    if FFiltered then
      cs_Query( PChar( FFilterStr ));

    // If this is Detail of Master/Detail relationship, Initialize Child range
    if Assigned( MasterSource ) then
    begin
      sVal := GetMasterVal;
      cs_Select( iWA );
      cs_SetScope( PChar(sVal), PChar(sVal));
    end;

    cs_GoTop;

    FFirstCrack     := True;
    FFetchLastRec   := 0;
    FFetchReadCount := -1;

    // Toggle DataSource.Enabled to force data to display at design-time
    if Assigned( DataSource ) and (DataSource.Enabled) then
    begin
      DataSource.Enabled := False;
      DataSource.Enabled := True;
    end;

  finally
    LOpeningTable := False;
  end;
end;


{==============================================================================
  InternalClose
  Make sure you clean up everything that you did in InternalOpen and everything
  you allocated since.
 ==============================================================================}
procedure TApolloTable.InternalClose;
begin
	FRowCount := -1;             {LAS: 08-12-00, Added}
	FRecCount := -1;             {LAS: 08-12-00, Added}
	FCountNeedsRefresh := True;  {LAS: 08-12-00, Added}
  FDataPath := '';
  FCursorOpen := False;
  BindFields( False );        // Unbind the fields
  FreeMem(LSDERecordBuffer);
  LSDERecordBuffer := nil;

  // Destroy the fields if we created them
  if (DefaultFields) then
  	DestroyFields;

  while LFieldOffsets.Count > 0 do
  begin
    TObject(LFieldOffsets[0]).Free;
    LFieldOffsets.Delete(0);
  end;
  LFieldOffsets.Clear;

  if (iWA > 0) then
  begin
  	if cs_Alias( iWA ) > '' then   {LAS: 06-20-00, Added.  Insures no AV when other work area crashes}
    begin                          {LAS: 06-20-00, Added}
	    cs_Select( iWA );
  	  cs_Close;
    end;                           {LAS: 06-20-00, Added}
    iWA := 0;
  end;
end;

{==============================================================================
  UpdateIndexDefs
 ==============================================================================}
procedure TApolloTable.UpdateIndexDefs;
var
  Options: TIndexOptions;
  sIndexKey, sName, sSource, sFields: String;
  i, iCurOrder : Integer;
  iNumIndexes: Word;

	// Extract base file name - no path, no extension
  function BaseFileName( sFile: String ): String;
  begin
  	sFile := UpperCase( ExtractFileName( sFile ));
  	Result := Copy( sFile, 1, Pos( '.', sFile )-1 );
  end;
begin

	if not (iWA > 0) then
  	Exit;
	if not FIndexDefs.Updated then
	begin
  	cs_Select( iWA );

    cs_SetStringType(1);

		iNumIndexes := cs_SysProp( SDE_SP_GETINDEXCOUNT, nil );
		IndexDefs.Clear;
    if iNumIndexes > 0 then
    begin
    	iCurOrder := cs_IndexOrd;
			for i := 0 to iNumIndexes - 1 do
			begin
      	// Get index file name
				if FTableType = ttSXNTX then
        begin
	      	sSource := String( cs_IndexName( i+1 ));
  	      if sSource = '' then
    	    	continue;
        end;

      	cs_SetOrder( i+1 );
        sIndexKey := UpperCase( cs_IndexKey );

        // Get tag name (same as index file name for .NTX)
				if FTableType = ttSXNTX then
					sName := sSource
				else
					sName := String( cs_TagName( i+1 ));

        // Clear options
				Options := [];

{$ifdef NOT_D3_OR_CB3}
				// Non-Maintained index?
        if (FTableType = ttSXNTX) or ((sSource > '') and
           (BaseFileName( sSource ) <> BaseFileName( TableName ))) then
					Include( Options, ixNonMaintained );
{$endif}

        // Unique Index?
				if (cs_IndexType in [2,4]) then
					Include( Options, ixUnique );

        // Case-Insensitive?
				if (Pos('UPPER(', sIndexKey) > 0) then
					Include( Options, ixCaseInsensitive );

				// Descending Index?
        if (cs_SysProp( SDE_SP_GETDESCENDING, nil ) <> 0) then
					Include( Options, ixDescending );

				{LAS: 05-29-00, Added}
				// If more than one field, or if a function was used, = ixExpression
        if (Pos( '+', sIndexKey ) > 0) or (Pos( '(', sIndexKey ) > 0) or
           (Pos( '-', sIndexKey ) > 0) then
					Include( Options, ixExpression );

				// Get Key Fields
        if (ixExpression in Options) then
{$ifdef NOT_D3_OR_CB3}            {LAS: 07-01-00, Added}
          sFields := ''
{$else}                           {LAS: 07-01-00, Added}
					sFields := sIndexKey    {LAS: 07-01-00, Added}
{$endif}                          {LAS: 07-01-00, Added}
        else
        begin
          if Pos( '+', sIndexKey ) = 0 then
            sFields := String( cs_IndexKeyField )
          else
            sFields := GetFieldsFromExpression( sIndexKey );
        end;

				// If more than one field, or if a function was used, = ixExpression
//        if (String( sIndexKey ) <> sFields) then         {LAS: 05-29-00, Moved up}
//					Include( Options, ixExpression );              {LAS: 05-29-00, Moved up}

				if (sName = EmptyStr) or (sIndexKey = EmptyStr) then
					RaiseApollo( ERR_INVALID_INDEX );

				with IndexDefs do
				begin
					Add( sName, sFields, Options );

{$ifdef NOT_D3_OR_CB3}
					// Can't assign Expression or Fields properties under Delphi 3
					if ixExpression in Options then
						Items[i].Expression := String( sIndexKey )
					else                                         {LAS: 05-29-00, Readded}
						Items[i].Fields := sFields;

					if not (ixNonMaintained in Options) then
          	Items[i].Source := ExtractFileName( cs_IndexName(i+1));
{$endif}

					if sSource <> '' then
						Items[Count - 1].Source := sSource;
				end;
			end;
      cs_SetOrder( iCurOrder );
    end;
		IndexDefs.Updated := True;
	end;
end;

{==============================================================================
  InternalInitFieldDefs
  This creates the TFieldDefs based on the table structure.
  We have to open the table to get the field defs.
 ==============================================================================}
procedure TApolloTable.InternalInitFieldDefs;
var
  i, iFldCount: Integer;
  FldInfo: TFieldInfo;
  FldName, FldType: String;

  j: Integer;
  msPtr: Pointer;
begin

  if FIndirectlyAssigned then   // MED: 03-07-2002, added to support OLEDB WORKAREA=1
  	OpenTable
  else if (iWA = 0) then
	  	OpenTable;

  if (iWA = 0) then
  	exit;

	while LFieldOffsets.Count > 0 do
	begin
		TObject(LFieldOffsets[0]).Free;
		LFieldOffsets.Delete(0);
	end;
	LFieldOffsets.Clear; // If the previous loop is needed, this line too
	FieldDefs.Clear;

  if (iWA = 0) then
  	exit;
	cs_Select( iWA );
  iFldCount := FSDEFieldCount; // MED: 05-20-00, changed from cs_FieldCount;
  FFieldCount:= FSDEFieldCount;

  {MED: 10-21-00, added}
  SetLength(FFieldName, FFieldCount+1);  // 1- based array
  SetLength(FFieldType, FFieldCount+1);  // 1- based array
  SetLength(FFieldWidth, FFieldCount+1);  // 1- based array
  SetLength(FFieldDecimals, FFieldCount+1);  // 1- based array
  SetLength(FFieldOffSet, FFieldCount+1);  // 1- based array

  if FAccessMethod = amLocal then
  begin
  	for i := 1 to iFldCount do
	  begin
		  FldInfo := TFieldInfo.Create;
  		FldName := cs_FieldName( i );
	  	FldType := cs_FieldType( PChar( FldName ));
		  FldInfo.SDEOffset := cs_FieldOffset( PChar( FldName ))-1;
  		FldInfo.Width := cs_FieldWidth( PChar( FldName ));
	  	FldInfo.Decimals := cs_FieldDecimals( PChar( FldName ));
      {MED: 10-21-00, added}
      FFieldName[i]    := UpperCase(fldName);
      FFieldType[i]    := FldType;
      FFieldWidth[i]   := FldInfo.Width;
      FFieldDecimals[i]:= FldInfo.Decimals;
      FFieldOffSet[i]  := FldInfo.SDEOffSet;

		  LFieldOffsets.Add( FldInfo );
      case FldType[1] of
        'C': FieldDefs.Add( FldName, ftString, FldInfo.Width, False );
        'I': FieldDefs.Add( FldName, ftInteger, 0, False );
        'N':
        	if (FldInfo.Decimals > 0) or (FldInfo.Width > 4 ) then
            FieldDefs.Add( FldName, ftFloat, 0, False )
          else
          	FieldDefs.Add( FldName, ftSmallInt, 0, False );
        'M': FieldDefs.Add( FldName, ftMemo, 0, False );
        'D': FieldDefs.Add( FldName, ftDate, 0, False );
        'L': FieldDefs.Add( FldName, ftBoolean, 0, False );
        'B',
        'G',
        'P': FieldDefs.Add( FldName, ftBlob, 0, False );
      end;
  	end;

    // If we're just doing an InfoQuery and the table's not already open,
    // then we need to close the table on the way out.
    if (not FCursorOpen) and (not LOpeningTable)
    then begin
      cs_Select( iWA );
      cs_Close;
      iWA := 0;
    end;
  end
  else
  begin
    FStructList.Clear;
    msPtr := FApolloConnection.ap_GetStructure;
    if (msPtr <> nil) then
    begin
      FStructList.LoadFromStream(TMemoryStream(msPtr) );
      j := 0;
      for i := 1 to iFldCount do
      begin
        FldInfo           := TFieldInfo.Create;
        FldName           := Trim( FStructList.strings[j] );
        FldType           := Trim( FStructList.Strings[j+1] );
        FldInfo.Width     := StrToInt( '$' + FStructList.strings[j+2] );
        FldInfo.Decimals  := StrToInt( '$' + FStructList.Strings[j+3] );
        FldInfo.SDEOffset := StrToInt( '$' + FStructList.strings[j+4] )-1;

        {MED: 10-21-00, added}
        FFieldName[i]    := UpperCase(fldName);
   		  FFieldType[i]    := FldType;
	   	  FFieldWidth[i]   := FldInfo.Width;
		    FFieldDecimals[i]:= FldInfo.Decimals;
		    FFieldOffSet[i]  := FldInfo.SDEOffSet;

        Inc( j, 5 );
        LFieldOffsets.Add( FldInfo );
        case FldType[1] of
          'C': FieldDefs.Add( FldName, ftString, FldInfo.Width, False );
          'I': FieldDefs.Add( FldName, ftInteger, 0, False );
          'N':
              if (FldInfo.Decimals > 0) or (FldInfo.Width > 4 ) then
                FieldDefs.Add( FldName, ftFloat, 0, False )
              else
                FieldDefs.Add( FldName, ftSmallInt, 0, False );
          'M': FieldDefs.Add( FldName, ftMemo, 0, False );
          'D': FieldDefs.Add( FldName, ftDate, 0, False );
          'L': FieldDefs.Add( FldName, ftBoolean, 0, False );
          'B',
          'G',
          'P': FieldDefs.Add( FldName, ftBlob, 0, False );
        end;
      end;
    End;
    // If we're just doing an InfoQuery and the table's not already open,
    // then we need to close the table on the way out.
    if (not FCursorOpen) and (not LOpeningTable)  then
    begin
      iWA := 0;
    end;
  end;
end;

{==============================================================================
  IsCursorOpen
 ==============================================================================}
function TApolloTable.IsCursorOpen: Boolean;
begin
  Result := FCursorOpen;
end;

{==============================================================================
  GetIndexNames
 ==============================================================================}
procedure TApolloTable.GetIndexNames(List: TStrings);
var
	i : Integer;
begin
	IndexDefs.Updated := False;
	UpdateIndexDefs;
	List.BeginUpdate;
	try
		List.Clear;
		for i := 0 to FIndexDefs.Count - 1 do
    	with FIndexDefs[i] do
				if Name <> '' then
        	List.Add( Name );
	finally
		List.EndUpdate;
	end;
end;

{==============================================================================
  GetFieldsFromExpression
 ==============================================================================}
function GetFieldsFromExpression( sIndexKey: String ): String;
var
	i: Integer;
	sVal: String;
	iCount: Integer;
	sTmpField: String;
	bOpenPars: Boolean;
	bCompleteField: Boolean;
begin
	Result := '';
	sVal := '';
	if sIndexKey <> '' then
	begin
  	sIndexKey := sIndexKey + '+';
		iCount := Length( sIndexKey );
		sTmpField := '';
		bOpenPars	:= False;
		for i := 1 to iCount do
		begin
			bCompleteField := False;
			case sIndexKey[i] of
        '(':
          begin
            bOpenPars := True;
            sTmpField := '';
          end;
        ',':
					bCompleteField := bOpenPars;
        ')':
          begin
            bOpenPars := False;
            bCompleteField := True;
          end;
        '+',
        '-':
            bCompleteField := True;
      else
        if (not bCompleteField) and (sIndexKey[i] <> ' ') then
          sTmpField := sTmpField + sIndexKey[i];
      end;

      if bCompleteField then
      begin
        if (sTmpField<>'') and (Pos('=', sTmpField)=0) and
           (Pos( '<', sTmpField) = 0) and (Pos( '>', sTmpField ) = 0) and
           (Pos( '$', sTmpField) = 0) and (Pos( '#', sTmpField ) = 0) then
          if not (sTmpField[1] in ['0'..'9']) then
            sVal := sVal + sTmpField + ';';
        sTmpField := '';
      end;
    end;
    if (Length( sVal ) > 0) and (sVal[ Length( sVal )] = ';') then   {LAS: 04-06-00, Added}
    	sVal := Copy( sVal, 1, Length( sVal )-1 );                     {LAS: 04-06-00, Added}
  end;
  Result := sVal;
end;



{==============================================================================
  CreateFields
  This goes back through the fields and sets additional properties about them
  based on things we get from the SDE (decimal size, etc).
 ==============================================================================}
procedure TApolloTable.CreateFields;
var
  i: Integer;
  Field: TField;
  FieldWidth,
  FieldDecimals: Integer;
  DisplayFormat: String;
begin
  if (iWA = 0) then
  	exit;

	cs_Select( iWA );
  if FAccessMethod = amLocal then
  begin
    for i := 0 to FieldDefs.Count - 1 do
    begin
    	with FieldDefs[i] do
      begin
    	  if (DataType <> ftUnknown) then
        begin
	  	    Field := CreateField(Self);
      		case Field.DataType of
          	ftCurrency,
		      	ftFloat:
            	begin
			          FieldWidth := cs_FieldWidth(PChar(Field.FieldName));
      			    FieldDecimals := cs_FieldDecimals(PChar(Field.FieldName));
  			        DisplayFormat := '0';
	  		        if (FieldDecimals > 0) then
                begin
			            DisplayFormat := DisplayFormat + '.';
			            while FieldDecimals > 0 do
			            begin
			              DisplayFormat := DisplayFormat + '0';
			              Dec(FieldDecimals);
			            end;
  			        end;
          			  (Field as TFloatField).DisplayFormat := DisplayFormat;

                // Precision is minimum 2, so add 1
			          (Field as TFloatField).Precision := FieldWidth + 1;
			        end;
          end;
	    	end;
		  end;
  	end;
  end
  else
  begin
    for i := 0 to FieldDefs.Count - 1 do
    begin
      with FieldDefs[i] do
      begin
        if (DataType <> ftUnknown) then
        begin
          Field := CreateField(Self);
          case Field.DataType of
            ftCurrency,
            ftFloat:
              begin
                FieldWidth    := StrToInt( '$' + FStructList.Strings[(i*5) + 2 ]);
                FieldDecimals := StrToInt( '$' + FStructList.Strings[(i*5) + 3 ]);

                DisplayFormat := '0';
                if (FieldDecimals > 0) then
                begin
                  DisplayFormat := DisplayFormat + '.';
                  while FieldDecimals > 0 do
                  begin
                    DisplayFormat := DisplayFormat + '0';
                    Dec(FieldDecimals);
                  end;
                end;
                (Field as TFloatField).DisplayFormat := DisplayFormat;

                // Precision is minimum 2, so add 1
                (Field as TFloatField).Precision := FieldWidth + 1;
              end;
          end;
        end;
      end;
    end;
  end;
end;

{==============================================================================
  InternalHandleException
 ==============================================================================}
procedure TApolloTable.InternalHandleException;
begin
  Application.HandleException( Self );  // This might not be thread safe!
end;

{==============================================================================
  InternalGotoBookmark
 ==============================================================================}
procedure TApolloTable.InternalGotoBookmark(Bookmark: Pointer);
begin
  Try
    Go( Integer( Bookmark^ ));
  Except
    RaiseApollo('Invalid bookmark');    //MED: 07-31-01, added
  End;
end;


{==============================================================================
  InternalSetToRecord
 ==============================================================================}
procedure TApolloTable.InternalSetToRecord( Buffer: PChar );
begin

  cs_Select( iWA );
  cs_Go( PRecInfo(Buffer + FRecInfoOfs).Bookmark );
(*
  if FIndexName = '' then                              {LAS: 06-01-00, Added}
    FRecOrdinal := FRecNo                              {LAS: 06-01-00, Added}
  else                                                 {LAS: 06-01-00, Added}
		FRecOrdinal := Trunc(cs_OrderPosGet * FRowCount);  {LAS: 06-01-00, Added}
*)
end;

{==============================================================================
  BookmarkValid
 ==============================================================================}
function TApolloTable.BookmarkValid(Bookmark: TBookmark): Boolean;
begin
	Result := not(Bookmark = nil);
end;

{==============================================================================
  GetBookmarkFlag
 ==============================================================================}
function TApolloTable.GetBookmarkFlag(Buffer: PChar): TBookmarkFlag;
begin
  Result := PRecInfo(Buffer + FRecInfoOfs).BookmarkFlag;
end;

{==============================================================================
  SetBookmarkFlag
 ==============================================================================}
procedure TApolloTable.SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag);
begin
  PRecInfo(Buffer + FRecInfoOfs).BookmarkFlag := Value;
end;

{==============================================================================
  GetBookmarkData
  This gets the TBookmarkStr data that is used by the multi-select grid.
 ==============================================================================}
procedure TApolloTable.GetBookmarkData(Buffer: PChar; Data: Pointer);
begin
  PInteger(Data)^ := PRecInfo(Buffer + FRecInfoOfs).Bookmark;
end;

{==============================================================================
  SetBookmarkData
 ==============================================================================}
procedure TApolloTable.SetBookmarkData(Buffer: PChar; Data: Pointer);
begin
	if (Data <> nil) then
	  PRecInfo(Buffer + FRecInfoOfs).Bookmark := PInteger(Data)^;
end;

{==============================================================================
  GetRecordSize
  This returns the size of the actual data only - not any additional data on the
  end of the buffer.
 ==============================================================================}
function TApolloTable.GetRecordSize: Word;
begin
  Result := FRecInfoOfs;
end;

{==============================================================================
  AllocRecordBuffer
  Allocate the memory for the record buffer.  Also, any blob Pointers need to
  be allocated here.
 ==============================================================================}
function TApolloTable.AllocRecordBuffer: PChar;
var
  i: Integer;
  BlobData: PBlobData;
begin
  Result := StrAlloc(FRecBufSize);

  for i := 0 to BlobFieldCount-1 do
  begin
    BlobData := PBlobData( Result + FBlobCacheOfs + (i * SizeOf(TApBlobData)));
    BlobData.Size := 0;
    BlobData.Data := nil;
    FillChar( BlobData.SDEPointer, 10, 32 );
    BlobData.Modified := False;
    BlobData.Loaded := False;
  end;
end;

{==============================================================================
  FreeRecordBuffer
  Free the record buffer - Make sure all memory that is allocated by both the
  AllocRecordBuffer function - plus any blob retrieval data - is freed here.
 ==============================================================================}
procedure TApolloTable.FreeRecordBuffer(var Buffer: PChar);
var
  i: Integer;
  BlobData: PBlobData;
begin
  for i := 0 to BlobFieldCount-1 do                                    
  begin
    BlobData := PBlobData( Buffer + FBlobCacheOfs + (i * SizeOf( TApBlobData )));
    FreeMem( BlobData.Data );
    BlobData.Data := nil;
    BlobData.Modified := False;
    BlobData.Loaded := False;
  end;

	  {LAS: 06-13-00, Removed. Caused memory leak.  However, this IF line was
     previously added to fix an AV that could occur if the Buffer was nil!}
//	if (Assigned(Buffer) and (Buffer <> EmptyStr)) then
  StrDispose( Buffer );
end;

{==============================================================================
  InternalInitRecord
  Initialize the record buffer.  This is called whenever a NEW (insert/append)
  record is created so that it shows up with blank values.  When scrolling,
  this is NOT called because TDataSet just recycles - and writes new data over
  the old stuff.
 ==============================================================================}
procedure TApolloTable.InternalInitRecord(Buffer: PChar);
var
	i : Integer;
  BlobData: PBlobData;
begin
	FillChar( Buffer^, FBlobCacheOfs, 0 );
	PRecInfo( Buffer + FRecInfoOfs ).Bookmark := 0;

	if BlobFieldCount > 0 then
  begin
		for i := 0 to BlobFieldCount-1 do
    begin
      BlobData := PBlobData( Buffer + FBlobCacheOfs + (i * SizeOf(TApBlobData)));
      FreeMem( BlobData.Data );
      BlobData.Data := nil;       // Make sure it's NIL!
      FillChar( BlobData.SDEPointer, SizeOf( BlobData.SDEPointer ), 32 );
      BlobData.Size := 0;         // Clear the *size
      BlobData.Modified := False; // Clear the modified flag
      BlobData.Loaded := False;
    end;
	end;
end;

{==============================================================================
  RefreshInternalCalcFields
 ==============================================================================}
procedure TApolloTable.RefreshInternalCalcFields(Buffer: PChar);
begin
  CalculateFields( Buffer );
end;

{==============================================================================
  CreateBlobStream
 ==============================================================================}
function TApolloTable.CreateBlobStream( Field: TField;
                                          Mode: TBlobStreamMode): TStream;
begin
	Result := TApBlobStream.Create(Field as TBlobField, Mode);
end;

{==============================================================================
  ClearBuffers
 ==============================================================================}
procedure TApolloTable.ClearBuffers;
begin
	inherited;
	if iWA > 0 then
		InternalInitRecord( ActiveBuffer );
end;

{==============================================================================
  ClearCalcFields - This is to clean calculated fields buffer
 ==============================================================================}
procedure TApolloTable.ClearCalcFields(Buffer: PChar);
begin
  FillChar( Buffer[FCalcFieldsOfs], CalcFieldsSize, 0 );
end;

{==============================================================================
  GetFieldData
  This is where all the work of converting from native SDE format to Delphi
  formatting is done.
 ==============================================================================}
function TApolloTable.GetFieldData( Field: TField; Buffer: Pointer): Boolean;
var
  OffSet:  Integer;
  DataBuffer: PChar;
  FieldData: PChar;
begin
	Result := False;
  if (not FCursorOpen) then
  	Exit;

  if Field.FieldNo > 0 then
    OffSet := TFieldInfo(LFieldOffsets[Field.FieldNo-1]).NativeOffSet
  else
    OffSet := FCalcFieldsOfs + Field.OffSet;

  if State = dsCalcFields then
    DataBuffer := CalcBuffer
	else if State = dsFilter then
  	DataBuffer := LFilterRecordBuffer
  else
    DataBuffer := ActiveBuffer;

  FieldData := DataBuffer+Offset;
	try
    if FieldData[0] = FLD_NULL then
      Result := False
    else
    begin
      Result := True;
      Inc(FieldData);
      if Assigned(Buffer) then
        Move(FieldData^, Buffer^, Field.DataSize);
    end;
  except
  	Result := False;
  end;
end;

{==============================================================================
  SetFieldData
  This is the routine that writes the data from the field back into the record
  buffer.  Any differences between the SDE and Delphi internal storage types
  are reconciled here.  Supports NULL fields - where the SDE itself does not.
 ==============================================================================}
procedure TApolloTable.SetFieldData( Field: TField; Buffer: Pointer );
var
  OffSet:  Integer;
  DataBuffer: PChar;
  FieldData: PChar;
begin
  if State = dsInactive then
  	exit;

  if Field.FieldNo > 0 then
  begin
		if not (State in [dsEdit, dsInsert, dsSetKey ]) then   {LAS: 05-23-00, Added dsSetKey check}
			RaiseApollo( ERR_NOT_IN_EDIT_OR_INSERT_MODE );

  	Field.Validate(Buffer);
    OffSet := TFieldInfo(LFieldOffsets[Field.FieldNo-1]).NativeOffSet
	end
  else  // Calculated fields have Field.FieldNo values of -1
    OffSet := FCalcFieldsOfs + Field.OffSet;

  if State = dsCalcFields then
    DataBuffer := CalcBuffer
  else
    DataBuffer := ActiveBuffer;

  FieldData := DataBuffer+Offset;

  if (Assigned(Buffer)) then
  begin
    // Not Null                                                    
    FieldData[0] := FLD_NOTNULL;
    Inc( FieldData );
    {MED: 09-08-00, commented again.  Translate error }
    //Buffer   := PChar(TrimRight(PChar(Buffer)));   {Stansilav:09-08-00, to make it compatible with TTable}
    Move(Buffer^, FieldData^, Field.DataSize);
  end
  else
  begin
    // Null data
    FieldData[0] := FLD_NULL;
    Inc( FieldData );
    FillChar(FieldData^, Field.DataSize, 0);
  end;

  if not (State in [dsCalcFields, dsFilter, dsNewValue]) then
    DataEvent(deFieldChange, Integer(Field));

  {LAS: 05-23-00, Added}
	if (State = dsSetKey) then
  begin
		if FRangeSet = rsStart then
    	FRangeStart := FieldByName(Field.FieldName).Value
		else if FRangeSet = rsEnd then
    	FRangeEnd := FieldByName(Field.FieldName).Value;
	end;

end;

{==============================================================================
  InternalPost
  This routine writes the data from the Active Buffer into the dataset.
 ==============================================================================}
procedure TApolloTable.InternalPost;
var
  i, iFldCount: Integer;
  BlobData: PBlobData;
begin
  if (iWA > 0) then
  begin
    cs_Select( iWA );

    if (not FAppending) and (State = dsInsert) then
    begin
	    FCountNeedsRefresh := True;
      cs_AppendBlank;
    end;

    // Write entire record buffer to disk
    WriteNativeToSDE(ActiveBuffer);

    if (BlobFieldCount > 0) then
    begin
    	iFldCount := FieldCount;
      for i := 0 to iFldCount-1 do
      begin
				if Fields[i].IsBlob then
				begin
          BlobData := PBlobData( ActiveBuffer + FBlobCacheOfs +
                               ( Fields[i].Offset * SizeOf( TApBlobData )));
          if (BlobData.Modified) then
          begin
            if ((Fields[i] as TBlobField).BlobType = ftMemo) then
            begin
              cs_Replace( PChar(Fields[i].FieldName), R_MEMO, BlobData.Data );
            end
            else
            begin
              if cs_PutBlob( PChar(Fields[i].FieldName), BlobData.Data,
                                           BlobData.Size) <> BlobData.Size then
              	RaiseApollo( ERR_BLOB_WRITE_ERROR );
            end;
          end;
       	end;
      end;
    end;

//    if not FOptimisticCommit then // Changed to CommitLevel
 	  cs_Commit;
    cs_Unlock( 0 );

    if not FSpeedMode then
    begin
			{LAS: 07-25-00, Removed. This was causing improper record positioning when posting on last record}
//    	if not (State = dsInsert) then   {LAS: 07-14-00, Added}
//      begin                            {LAS: 07-14-00, Added}
//	      cs_Skip(-1);
//  	    if (not cs_Bof) then
//    	    cs_Skip(1);
//      end;                             {LAS: 07-14-00, Added}

      with PRecInfo(ActiveBuffer + FRecInfoOfs)^ do
      begin
        if cs_Eof  then
          BookmarkFlag := bfEOF
        else
          BookmarkFlag := bfCurrent;
        Bookmark := cs_RecNo;
        FRecNo := Bookmark;  {LAS: 07-25-00, Added}
      end;
    end;
		FOldRec := 0;
//    FCountNeedsRefresh := True;   {LAS: 05-30-00, Moved up}
    FAppending := False;
  end;
end;

{==============================================================================
  InternalAddRecord
  This is seldom called - basically only when the AppendRecord or InsertRecord
  of the TDataSet are called.
 ==============================================================================}
procedure TApolloTable.InternalAddRecord(Buffer: Pointer; Append: Boolean);
begin
  // Can't modify a ReadOnly table
  if ReadOnly and (FErrorFunc = -1) then
    RaiseApollo( ERR_TABLE_IS_READONLY );

  if (iWA > 0) then
  begin
    cs_Select( iWA );

    cs_AppendBlank;
    WriteNativeToSDE(Buffer);
    cs_Commit;

    // CRVG 2001-09-27 Added
    with PRecInfo(ActiveBuffer + FRecInfoOfs)^ do
    begin
      if cs_Eof  then
        BookmarkFlag := bfEOF
      else
        BookmarkFlag := bfCurrent;
      Bookmark := cs_RecNo;
      FRecNo := Bookmark;
    end;
  end;
end;


{==============================================================================
  InternalDelete
  Delete the current record (ActiveBuffer) from the dataset.  After Deleting,
  if the deleted record was removed from view (by SetDeleted=False or a
  conditional index FOR !DELETED()), postion the cursor on the NEXT record
  after the deleted one.
 ==============================================================================}
procedure TApolloTable.InternalDelete;
var
  iThisRecNo, iNextRecNo: Integer;
begin
	// Can't modify a ReadOnly table
	if ReadOnly and (FErrorFunc = -1) then
  	RaiseApollo( ERR_TABLE_IS_READONLY );

	// Can't delete a record if there are none  (Go figure!)
	if IsEmpty and (FErrorFunc = -1) then
  	RaiseApollo( ERR_TABLE_IS_EMPTY );

	if (iWA > 0) then
  begin
	  cs_Select( iWA );
		if not FSpeedMode then
    	UpdateCursorPos;

		// Simple Delete if no conditional index is active and SetDeleted is False
	  if ((Pos( 'DELETED()', UpperCase(cs_IndexCondition)) = 0) and
				(cs_SysProp( SDE_SP_GETDELETED, nil ) = 0)) then
	  	cs_Delete               // Delete it.
	  else
    //  Otherwise, determine where to position cursor after record is deleted
		begin
	  	iThisRecno := cs_RecNo; // Save our current record number
		  cs_Skip(1);             // Skip to NEXT one (to move off the deleted one).
		  if (cs_EOF) then        // If it's EOF, we need to know that.
  			iNextRecno := -1
		  else
  			iNextRecno := cs_RecNo;
	  	cs_Go(iThisRecno);        // Move back to record to delete
	  	cs_Delete;                // Delete it.
		  if (iNextRecno < 0) then  // Move to another record
  			cs_GoBottom
		  else
  			cs_Go(iNextRecno);
	  end;

		if not FSpeedMode then
    begin
      with PRecInfo(ActiveBuffer + FRecInfoOfs)^ do
      begin
				if cs_Eof  then
	        BookmarkFlag := bfEOF
        else
	        BookmarkFlag := bfCurrent;
        Bookmark := cs_RecNo;
      end;
    end;
    FCountNeedsRefresh := True;

    Resync([]);
	end;
end;

{==============================================================================
  GetRecordCount
  Returns the count of records in the dataset.  This should be the count of
  viewable records - so if there's a filter in place, it should honor that
  filter.
 ==============================================================================}
function TApolloTable.GetRecordCount: Integer;
begin
  if FCountNeedsRefresh or (FRowCount = -1) then
    RefreshCount;

	Result := FRowCount;
end;

{==============================================================================
  GetRecNo
  Get's the record number (as in the scrollable location in the view) vs the
  Record Count.  As you scroll through a dataset, the first record should be
  #1 - so if an index is in effect, it must report the index location, not the
  actual record number.

  Called only when Sequenced := TRUE
 ==============================================================================}
function TApolloTable.GetRecNo: Integer;
begin
	Result := 0;
	if (iWA > 0) and (FCursorOpen) then
	begin
		cs_Select( iWA );
		UpdateCursorPos;

		if (FIndexOrd = 0) or (FTableType = ttSXNTX) then
    begin
			Result := FRecNo;       {MED: 05-21-00, Changed to FRecNo from cs_RecNo}
    end
		else
    begin
			if FCountNeedsRefresh or (FRowCount = -1) then
      	RefreshCount;

      {MED: 05-21-00, Changed to FRecOrdinal - no more calling the server! }
			Result := FRecOrdinal;  // Trunc(cs_OrderPosGet * FRowCount);
    end;
	end;
end;

{==============================================================================
  SetRecNo
  Move to a record number - this corresponds to GetRecNo.  When the scrollbar
  moves, this function is called with the number.  Makes thumb cursors work.
 ==============================================================================}
procedure TApolloTable.SetRecNo( Value: Integer );
begin
	if (iWA > 0) then
  begin

    CheckBrowseMode;
	  cs_Select( iWA );

  	cs_Go( Value );     // MED: 11-18-00-uncommented again. Blobs disappear in ApolloQuery if commented
                        // this should not affect the scrollbar problem
                        
  {		if cs_IndexOrd = 0 then   // MED: 10-21-00, commented- master/child thumb bar problem-see scrollBarProblem.zip
      RecNo := Value
    	//cs_Go( Value )
		else}
    begin
      if FRowCount > 0 then
				cs_OrderPosSet( Value / FRowCount );
		end;

    with PRecInfo(ActiveBuffer + FRecInfoOfs)^ do
    begin
      BookmarkFlag := bfCurrent;
      Bookmark := cs_RecNo;
    end;
		Resync([{rmExact}]);  {LAS: 04-04-00, Removed rmExact}
	end;
end;

{==============================================================================
  IsSequenced
  Return False from this if you don't want the scrollbar on grids to track
  with record number (ie:  BOF has the thumb at top, EOF has thumb at bottom,
  all others have thumb in middle).
  If you want the GetRecno and GetRecordCount to be queried whenever the
  cursor moves (so scrollbar works), then set this to True.
 ==============================================================================}
function TApolloTable.IsSequenced: Boolean;
begin
	if FTableType = ttSXNTX then
  	Result := False
	else
		Result := FSequenced;
end;

procedure TApolloTable.SetSequenced( Value : Boolean );
begin
  if (Value <> FSequenced) then
  begin
    FSequenced := Value;
    UpdateCursorPos;
    Resync([]);
  end;
end;

{==============================================================================
  CompareBookmarks
  This is used by the grid's multi-select to know if a record should be
  highlighted or not.
 ==============================================================================}

function TApolloTable.CompareBookmarks(
                                     Bookmark1, Bookmark2: TBookmark): Integer;
var
  iRecno : Integer;
  iResult, Recno1, Recno2 : Integer;
  KeyValue1, KeyValue2 : String;
  IndexKey1, IndexKey2 : String;

begin
  // nothing to do if one of the bookmark is NIL
  if (Bookmark1 = nil) or (Bookmark2 = nil) then
        Result := 2
  else
  begin

      if FIndexName='' then
      begin
         Recno1 := Integer( Bookmark1^ );
         Recno2 := Integer( Bookmark2^ );

        if (Recno1 = Recno2) then
          Result := 0
        else if (Recno1 < Recno2) then
          Result := -1
        else
          Result := 1;

      end else begin
         cs_Select(iWA);

         iRecno := cs_Recno;

         cs_Go( Integer( Bookmark1^ ));
         Recno1 := cs_Recno;
         IndexKey1 := cs_IndexKey;

         iResult := cs_EvalTest( PChar(IndexKey1) );
         case iResult of
           EVAL_CHARACTER : KeyValue1  := cs_EvalString(PChar(IndexKey1));
           EVAL_NUMERIC   : KeyValue1  := FloatToStr(cs_EvalNumeric(PChar(IndexKey1)));
           EVAL_LOGICAL   : KeyValue1  := IntToStr(Ord((cs_EvalLogical(PChar(IndexKey1)))));
           EVAL_DATESTRING: KeyValue1  := cs_EvalString(PChar(IndexKey1));
         else
            KeyValue1  := cs_EvalString(PChar(IndexKey1));
         end;

         cs_Go( Integer( Bookmark2^ ));
         Recno2 := cs_Recno;
         IndexKey2 := cs_IndexKey;
         case iResult of
           EVAL_CHARACTER : KeyValue2  := cs_EvalString(PChar(IndexKey2));
           EVAL_NUMERIC   : KeyValue2  := FloatToStr(cs_EvalNumeric(PChar(IndexKey2)));
           EVAL_LOGICAL   : KeyValue2  := IntToStr(Ord((cs_EvalLogical(PChar(IndexKey2)))));
           EVAL_DATESTRING: KeyValue2  := cs_EvalString(PChar(IndexKey2));
         else
            KeyValue2  := cs_EvalString(PChar(IndexKey2));
         end;

         if AnsiCompareStr(KeyValue1, KeyValue2) = 0 then
         begin
            if (Recno1 = Recno2) then
              Result := 0
            else if (Recno1 < Recno2) then
              Result := -1
            else
              Result := 1;
         end
         else
         begin
           if  cs_EvalLogical( PChar('"'+ KeyValue1 + '" < "' + KeyValue2 +'"') )
             xor (cs_SysProp(SDE_SP_GETDESCENDING,Nil)<>0)
           then
             Result := -1
           else
             Result := 1
         end;
        cs_Go (iRecno);         
      end;
  end;
end;

{==============================================================================
  GetCanModify
 ==============================================================================}
function TApolloTable.GetCanModify: Boolean;
begin
  Result := not ReadOnly;
end;

{==============================================================================
  CheckMasterRange
 ==============================================================================}
procedure TApolloTable.CheckMasterRange;
begin
  if FMasterLink.Active and (FMasterLink.Fields.Count > 0) then
    MasterChanged(nil);
end;

{==============================================================================
  GetMasterVal
 ==============================================================================}
function TApolloTable.GetMasterVal: String;
var
  i, iFldCount: Integer;
  sTempVal: String;
  Field : TField;

  function GetChildFieldName( iFldNo : Integer) : String;
  var
    sField, sIndexFields : String;
    i, iPos, iPosLen, iPosFin : Integer;
  begin
    sIndexFields := AnsiUpperCase(GetFieldsFromExpression(cs_IndexKey));
    iPos := Pos(';', sIndexFields);

      if iPos = 0 then
        begin
          Result := sIndexFields;
          Exit;
        end;

    Result := '';

    i := -1;
    while iPos > 0 do
    begin
      iPosLen := Length(sIndexFields);
      iPos    := Pos(';', sIndexFields)-1;
      iPosFin := 0;

        if (iPos > 0) And (iPos < iPosLen) then
          iPosFin := iPos;
        if (iPosLen > 0) And (iPosLen < iPos) then
          iPosFin := iPosLen;
        if iPosFin = 0 then
          iPosFin := Length(sIndexFields);

      sField := Copy(sIndexFields, 1, iPosFin );
      i := i + 1;
      if i = iFldNo then
        begin
          Result := sField;
          Exit;
        end;

      sIndexFields := Copy( sIndexFields, Pos(';', sIndexFields)+1, Length(sIndexFields));
    end;
  end;
begin
  Result := '';

  iFldCount := FMasterLink.Fields.Count;
  for i := 0 to iFldCount-1 do
  begin
    Field := TField(FMasterLink.Fields[i]);
    case Field.DataType of
      ftDateTime,
      ftDate:
        sTempVal := FormatDateTime( 'yyyymmdd', Field.AsDateTime );
      ftBoolean:
        begin
          if Field.AsBoolean then
            sTempVal := 'T'
          else
            sTempVal := 'F';
        end;
      ftSmallInt, ftInteger, ftFloat, ftCurrency:
        with Field do
        begin
{$ifndef NOT_D3_OR_CB3}
					if (Pos( 'STR(' + Field.FieldName, UpperCase(cs_IndexKey)) > 0) then
{$else}
					// if (Pos( 'STR(' + Field.FieldName, UpperCase(TApolloTable(
          //                       FMasterLink.DetailDataSet).IndexKey)) > 0) then
          
          // Changed to extract the name of the field from the child index key instead master's.

					 if (Pos( 'STR(' + GetChildFieldName(i), UpperCase(TApolloTable(
                               FMasterLink.DetailDataSet).IndexKey)) > 0) then
{$endif}
            sTempVal := StringOfChar( #32,
                        TApolloTable(FMasterLink.DataSet).FieldWidth(
                        FieldName ) - Length( AsString )) + AsString
          else
            sTempVal := StringOfChar( '0',
                        TApolloTable(FMasterLink.DataSet).FieldWidth(
                        FieldName ) - Length( AsString )) + AsString;
        end;
      ftString:
        with Field do
          sTempVal := AsString + StringOfChar( #32, Size-Length( AsString ));
    else
        sTempVal := Field.AsString;
    end;    { case }
    Result := Result + sTempVal;
  end;    { for }
end;

{==============================================================================
  MasterChanged
 ==============================================================================}
procedure TApolloTable.MasterChanged(Sender: TObject);
var
  sVal: String;
begin
	if Active then       {LAS: 06-01-00, Added}
	  CheckBrowseMode;
  if (FMasterLink.Fields.Count > 0) then
  	sVal := GetMasterVal;

	if (iWA > 0) then
  begin
    cs_Select( iWA );
    if (sVal <> FLastScopeVar) or (IndexName <> FLastIndexNameVar) then
    begin
			FLastScopeVar := sVal;
      FLastIndexNameVar := IndexName;

      if IndexName = '' then
 		    RaiseApollo( ERR_NO_INDEX_ACTIVE );

	    if not cs_SetScope(PChar(sVal), PChar(sVal)) then
 		    RaiseApollo( ERR_SCOPE_SET_FAILED );
			First;
		end;
	end;
end;

{==============================================================================
  MasterDisabled
 ==============================================================================}
procedure TApolloTable.MasterDisabled(Sender: TObject);
begin
  if (iWA > 0) then
  begin
    cs_Select( iWA );
    cs_SetScope( nil, nil );
    Resync([]);
  end;
end;

{==============================================================================
  GetMasterSource
 ==============================================================================}
function TApolloTable.GetMasterSource: TDataSource;
begin
  Result := FMasterLink.DataSource;
end;

{==============================================================================
  SetMasterSource
 ==============================================================================}
procedure TApolloTable.SetMasterSource(Value: TDataSource);
begin
  if IsLinkedTo(Value) then
    RaiseApollo( ERR_CIRCULAR_DATASOURCE_REFERENCE );
  FMasterLink.DataSource := Value;

  {LAS: 06-01-00, Added}
  if (FMasterLink.FieldNames <> '') and Assigned(FMasterLink.DataSet)
	then
  begin
		FLastScopeVar := '';
		MasterChanged( nil );
	end;
end;

{==============================================================================
  GetMasterFields
 ==============================================================================}
function TApolloTable.GetMasterFields: String;
begin
  Result := FMasterLink.FieldNames;
end;

{==============================================================================
  SetMasterFields
 ==============================================================================}
procedure TApolloTable.SetMasterFields(const Value: String);
begin
  FMasterLink.FieldNames := Value;

  {LAS: 06-01-00, Added}
  if (FMasterLink.FieldNames <> '') and Assigned(FMasterLink.DataSet)
	then
  begin
		FLastScopeVar := '';
		MasterChanged( nil );
  end;
end;



{==============================================================================
  Lookup
 ==============================================================================}

function TApolloTable.Lookup( const KeyFields: String; const KeyValues: Variant;
                              const ResultFields: String): Variant;
var
  sDate, sKey: String;
  CurState: TDataSetState;
  iOldIndexOrd: Integer;
  FType : String;
begin
  Result := Null;

  // Catch for when populating a table that has a lookup field from scratch.
  // Otherwise we'd get a Variant conversion error as Keyvalues is always null.
  if VarIsNull( KeyValues ) then
  begin
   Exit; {CRVG : 2001-05-09}
  end
  else
  begin
    if Not VarIsArray(KeyValues) then   {CRVG : 2001-05-09}
      sKey := String( KeyValues );
  end;

  // if there is no index available and if there are more than two fields,
	// lookup can't deal with it
  // calling locate to resolve it     {CRVG : 2001-05-09}
  if (FIndexOrd = 0) Or VarIsArray(KeyValues) then
  begin
    if Locate( KeyFields, KeyValues, [] ) then
      Result := FieldByName( ResultFields ).AsVariant;
    Exit;
  end;

    cs_Select( iWA );

    FType := cs_FieldType( PChar(KeyFields) ) ;   // MED: 03-26-01, stop calling server so much!

    {LAS: 04-24-00, Added ----------------------------------------------}
    if (FType  = 'N') then
      sKey := StringOfChar(#32, cs_FieldWidth( PChar(KeyFields) ) - Length(sKey)) + sKey

    else if (FType  = 'C') then
      sKey := sKey + StringOfChar( #32, cs_FieldWidth( PChar( KeyFields )) - Length(sKey))

    else if (FType = 'D') then
      sKey := FormatDateTime( 'yyyymmdd', StrToDate( sKey ))

    else if (FType = 'L') then
    begin
      if StrToInt( sKey ) = -1 then
        sKey := 'T'
      else
        sKey := 'F';
    end;

    if FLocateIndexOrd > 0 then
    begin
      iOldIndexOrd := cs_IndexOrd;
      cs_SetOrder( FLocateIndexOrd );
      cs_Seek( PChar( sKey ));
      cs_SetOrder( iOldIndexOrd );
    end
    else
      cs_Seek( PChar( sKey ));

    if cs_Found then
    begin
      CurState := SetTempState( dsCalcFields );
      try
        cs_GetRecord(LSDERecordBuffer);  {TempBuffer?}
        ReadSDEToNative( TempBuffer );
        CalculateFields( TempBuffer );

        FType := cs_FieldType( PChar(ResultFields) ) ;

        if (FType = 'N') then
        begin
          if (cs_FieldDecimals( PChar(ResultFields) ) > 0) or
             (cs_FieldWidth(PChar(ResultFields) ) > 4) then
            Result := cs_GetDouble(PChar(ResultFields))
          else
            Result := cs_GetInteger(PChar(ResultFields));
        end

        else if (FType = 'I') then
        begin
          Result := cs_GetInteger(PChar(ResultFields));
        end

        else if (FType = 'C') then
          Result := String( cs_GetTrimString( PChar( ResultFields )))

        else if (FType = 'D') then
        begin
          sDate := cs_GetDateString( PChar( ResultFields ));
          if sDate[1] = ' ' then
            Result := 0
          else
            Result := StrToDate( cs_GetDateString( PChar( ResultFields )));
        end
        else if (FType  = 'L') then
        begin
         if StrToInt( sKey ) = -1 then
            Result := True
          else Result := False
        end;

      finally
        RestoreState( CurState );
      end;
    end;
end;

// SeekVal - Finds matching Locate value for any type using Seek
function TApolloTable.SeekVal( tmpFields: TList; const KeyValues: Variant;
                    Options: TLocateOptions; iOldOrder: Integer ): Boolean;
var
  sVal: String;
  bOldExact: Boolean;
  sIndexKey : String;
begin

  case TField(tmpFields[0]).DataType of
    ftBoolean:
      if KeyValues then
        sVal := '.T.'
      else
        sVal := '.F.';
    ftSmallInt, ftInteger :
      sVal := IntToStr(KeyValues);
    ftFloat:
      sVal := FloatToStr(KeyValues);
    ftDate:
		sVal := FormatDateTime( 'yyyymmdd', StrToDateTime( KeyValues ));  {LAS: 05-03-00, Changed from 'yyyymmmdd'}
    ftString:
      begin
        sVal := KeyValues;
        // Handle Case-Insensitive Option
        sIndexKey := UpperCase(cs_IndexKey);
        if (Pos('UPPER(', sIndexKey) > 0) or
           ((loCaseInsensitive in Options) and
            (Pos('UPPER(', sIndexKey) > 0)) then
          sVal := AnsiUpperCase( sVal );  {MED:10-05-00, changed from UpperCase - Stansilav}
      end;
  end;

  bOldExact := (cs_SysProp( SDE_SP_GETEXACT, nil) <> 0);
  cs_SysProp( SDE_SP_SETEXACT, Pointer( not(loPartialKey in Options)) );
  Result := cs_Seek( PChar(sVal));

  if not(loPartialKey in Options) then
	  Result := cs_Found;

  cs_SysProp( SDE_SP_SETEXACT, Pointer( bOldExact ));
  cs_SetOrder(iOldOrder);
  FFetchReadCount := -1;    {LAS: 06-29-00, Added}
end;

{==============================================================================
  Locate
  // NOTE: Uses Query/Seek to find matching records. Preserves any existing
  //       (active) Query, except for RYO filters.
==============================================================================}
function TApolloTable.Locate( const KeyFields: String; const KeyValues: Variant;
                              Options: TLocateOptions): Boolean;
var
	i: Integer;
	tmpFields: TList;
  sOldFilt: String;
	sVal: String;
  lRecNo: Integer;
  lCurRec: Integer;
	sFld: String;
  iFldCount: Integer;
  iOldOrder, iIdxCount: Integer;
  bIdxOk: Boolean;
  // sIndexKey: String; // CRVG Moved to SeekVal
  tmpStr : String;
begin
	Result := False;
	if iWA = 0 then
  	Exit;

	BeginScroll;

	tmpFields := TList.Create;
	try
	  // Delphi feature to parse the field names into a list of fields
		GetFieldList(tmpFields, UpperCase(KeyFields));  {LAS: 07-23-00, Added UpperCase}

    cs_Select( iWA );

    // Preserve current record position, in case of failed Locate
    lCurRec := RecNo;

    // Attempt to find matching index
    iOldOrder := cs_IndexOrd;
		iIdxCount := cs_SysProp( SDE_SP_GETINDEXCOUNT, nil );
    bIdxOk := True;

    // If specific index order was requested, use it only
    if FLocateIndexOrd > 0 then
    begin
    	cs_SetOrder( FLocateIndexOrd );
    	Result := SeekVal( tmpFields, KeyValues, Options, iOldOrder );
    end
    else       // Otherwise, make some assumptions and take our chances!
		begin
      if ((tmpFields.Count = 1) and (iIdxCount > 0)) then
      begin
        // Does current index order match already
        if (UpperCase(KeyFields) = cs_IndexKeyField) then   {LAS: 07-23-00, Added UpperCase}
        begin
          if (loCaseInsensitive in Options) then
          begin
            if (Pos('UPPER(', UpperCase(cs_IndexKey)) = 0) then
            begin
              bIdxOk := False;
              for i := 1 to iIdxCount do
              begin
                cs_SetOrder(i);
                if (UpperCase(KeyFields) = cs_IndexKeyField) and   {LAS: 07-23-00, Added UpperCase}
                   (Pos('UPPER(', UpperCase(cs_IndexKey)) > 0 ) then
                begin
                  bIdxOk := True;
                  break;
                end;
              end;
            end;
          end;
        end
        else // Current index was not a match, so look for any other one
        begin
          // If search is case-insensitive we must look for index built UPPER()
          if (loCaseInsensitive in Options) then
          begin
            bIdxOk := False;
            for i := 1 to iIdxCount do
            begin
              cs_SetOrder(i);
              if (UpperCase(KeyFields) = cs_IndexKeyField) and   {LAS: 07-23-00, Added UpperCase}
                 (Pos('UPPER(', UpperCase(cs_IndexKey)) > 0 ) then
              begin
                bIdxOk := True;
                break;
              end;
            end;
          end
          else  // Otherwise, just find any standard matching index
          begin
            bIdxOk := False;
            for i := 1 to iIdxCount do
            begin
              cs_SetOrder(i);
              if ({UpperCase(}KeyFields{)} = cs_IndexKeyField) then  {LAS: 07-23-00, Added UpperCase}
              begin
                bIdxOk := True;
                break;
              end;
            end;
          end;
        end;
      end;

      // If exact match for index was not found, restore original now
      if bIdxOk = False then
        cs_SetOrder( iOldOrder );

      { SPECIAL CASE:  If only one field/value passed in AND if currently-active
        index is built on that same field, use SEEK instead of QUERY. }
      if (UpperCase(KeyFields) = cs_IndexKeyField) then     {LAS: 07-23-00, Added UpperCase}
        begin
          if KeyValues <> Null then
      	    Result := SeekVal( tmpFields, KeyValues, Options, iOldOrder )
          else
            begin
              Result := False;
              Exit;
            end;
        end;


      // If SPECIAL CASE above could not be used, or a Seek was not successful,
      // try using QUERY
      if Result = False then
      begin
        sOldFilt := DBFilter;
        if tmpFields.Count = 1 then
        begin
          case TField(tmpFields[0]).DataType of
            ftBoolean:
              if KeyValues then
                sVal := TField(tmpFields[0]).FieldName + '= .T.'
              else
                sVal := TField(tmpFields[0]).FieldName + '= .F.';
            ftSmallInt, ftInteger:
              sVal := TField(tmpFields[0]).FieldName + '=' + IntToStr(KeyValues);
            ftFloat:
              sVal := TField(tmpFields[0]).FieldName + '=' + FloatToStr(KeyValues);
            ftDate:
            	if Length( KeyValues ) < 8 then
								sVal := 'LEFT( DTOS(' + TField(tmpFields[0]).FieldName + '), ' +
                        IntToStr( Length( IntToStr( KeyValues ))) + ') = "' +
                        IntToStr( KeyValues ) + '"'
							else
	              sVal := 'DTOS(' + TField(tmpFields[0]).FieldName + ') = "' +
                        FormatDateTime( 'yyyymmdd', KeyValues ) + '"';  {LAS: 07-23-00, Added double-quotes "}
            ftString:
              begin
                // Handle Case-Insensitive Option
                if (loCaseInsensitive in Options) then
                  sFld := 'UPPER(' + TField(tmpFields[0]).FieldName + ')'
                else
                  sFld := TField(tmpFields[0]).FieldName;

                // Pad string value if Partial Key option not set
                if (loPartialKey in Options) then
                  sVal := sFld + '="' + KeyValues + '"'
                else
                  sVal := sFld + '="' + KeyValues + StringOfChar( #32,
                         TField(tmpFields[0]).Size - Length( KeyValues )) + '"';
              end;
          end;
        end
        else
        begin
          iFldCount := tmpFields.Count;
          for i := 0 to iFldCount-1 do
          begin
            if sVal > '' then
              sVal := sVal + ' .and. ';
            case TField(tmpFields[i]).DataType of
              ftBoolean:
                if KeyValues[i] then
                  sVal := sVal + TField(tmpFields[i]).FieldName + '= .T.'
                else
                  sVal := sVal + TField(tmpFields[i]).FieldName + '= .F.';
              ftSmallInt, ftInteger:
                sVal := sVal + TField(tmpFields[i]).FieldName + '=' +
                                                         IntToStr(KeyValues[i]);
              ftFloat:
                sVal := sVal + TField(tmpFields[i]).FieldName + '=' +
                                                         FloatToStr(KeyValues[i]);
              ftDate:
                sVal := sVal + 'DTOS(' + TField(tmpFields[i]).FieldName +
                    ') = "' + FormatDateTime( 'yyyymmdd', KeyValues[i] ) + '"';  {LAS: 07-23-00, Added double-quotes "}
              ftString:
                begin
                  // Handle Case-Insensitive Option
                  if (loCaseInsensitive in Options) then
                    sFld := 'UPPER(' + TField(tmpFields[i]).FieldName + ')'
                  else
                    sFld := TField(tmpFields[i]).FieldName;

                  // Pad string value if Partial Key option not set
                  if (loPartialKey in Options) then
                    sVal := sVal + sFld + '="' + KeyValues[i] + '"'
                  else
                     if VarisNull( KeyValues[i] ) then
                       tmpStr := ''
                     else
                       tmpStr := String(KeyValues[i]);

                    sVal := sVal + sFld + '="' + tmpStr +
                            StringOfChar( #32, TField(tmpFields[i]).Size -
                                               Length( tmpStr )) + '"';
                end;
            end;
          end;
        end;

        // Handle Case-Insensitive Option
        if (loCaseInsensitive in Options) then
          sVal := AnsiUpperCase( sVal );  {MED:10-05-00, changed from UpperCase - Stansilav}

				if sOldFilt > '' then
					sVal := '(' + sVal + ') .and. (' + sOldFilt + ')';

        Result := (cs_Query( PChar( sVal )) <> 0);
        lRecNo :=  cs_RecNo;
        cs_Query( PChar( sOldFilt ));
        if Result = True then
          cs_Go( lRecNo );
      end;
  	end;

    // If match not found, restore original record pointer position
    if Result then
    begin
      if not FSpeedMode then
      begin
        with PRecInfo(ActiveBuffer + FRecInfoOfs)^ do
        begin
          BookmarkFlag := bfCurrent;
          Bookmark := cs_RecNo;
        end;
      end;
	    Resync([rmExact, rmCenter]);
      EndScroll;
    end
		else
    	cs_Go( lCurRec );
	finally
		tmpFields.Free;
	end;
end;

{==============================================================================
  GetBlobData
 ==============================================================================}
function TApolloTable.GetBlobData(Field: TField): PBlobData;
begin
	if (State = dsFilter) then
		Result := PBlobData( LFilterRecordBuffer + FBlobCacheOfs +
                       ( Field.Offset * SizeOf( TApBlobData )))
	else
		Result := PBlobData( ActiveBuffer + FBlobCacheOfs +
                       ( Field.Offset * SizeOf( TApBlobData )));
end;

{==============================================================================
  TApBlobStream
  Adds functions to write data out to the dataset when destroyed - and know
  when the blob has changed.
 ==============================================================================}
{==============================================================================
  Create
 ==============================================================================}
constructor TApBlobStream.Create(Field: TBlobField; Mode: TBlobStreamMode);
var
	bFromDisk: WordBool;
	iCurRecno, iOldRecno: Integer;
  cpData: PChar;
  FBlobData: PBlobData;
begin
	FMode := Mode;
	FField := Field;
	FDataSet := TApolloTable(Field.DataSet);
	inherited Create;

	FBlobData := FDataSet.GetBlobData(field);
//	if (Mode <> bmWrite) then // CRVG Removed : 2001-12-07 : and not (Field.IsNull)
	if ((Mode <> bmWrite) and not (Field.IsNull)) or (FDataSet.State = dsCalcFields) then // CRVG 2001-12-21, Changed by the line above
	begin
		bFromDisk := (FDataSet.State = dsOldValue) or
                 (not (FDataSet.State in [dsEdit, dsInsert])) or // CRVG Added : 2001-11-02
                 (not FBlobData.Loaded) or
                 (FDataSet.SpeedMode);  {LAS: 06-22-00, Added FSpeedMode check}
		if bFromDisk then  // Load it from disk rather than from memory.
		begin
			FDataSet.cs_Select( FDataSet.Workarea );
			iOldRecno := FDataSet.RecNo;   {LAS: 06-06-00, Changed from FDataSet.cs_RecNo;}

      if (FDataSet.State = dsFilter) then
        iCurRecno := PRecInfo(FDataSet.LFilterRecordBuffer +
                        FDataSet.FRecInfoOfs).Bookmark
      else
        //   iCurRecno := PRecInfo(FDataSet.ActiveBuffer + FDataset.FRecInfoOfs).Bookmark;
        iCurRecno := FDataSet.Recno;  // CRVG Added : 2001-11-02  Changed to get the recno directly from DataSet.

			FDataSet.cs_Go( iCurRecno );

      FBlobData.Size := 0;                   {LAS: 06-29-00, Added, just in case}
      if FDataSet.TableType <> ttSXNTX then  {LAS: 06-29-00, Added}
				FBlobData.Size := FDataSet.cs_GetBlobLength( PChar( Field.FieldName ));  {LAS: 06-20-00, Added}

			if (FBlobData.Size < $FFFF) and ((Field as TBlobField).BlobType = ftMemo) then
			begin
				cpData := FDataSet.cs_GetMemo( PChar( Field.FieldName ), 0 );
				if (cpData = nil) then
					cpData := '';

				// Get the length
        if FDataSet.TableType = ttSXNTX then  {LAS: 06-29-00, Added}
					FBlobData.Size := StrLen( cpData );

				// Make sure it's not already assigned.
				FreeMem( FBlobData.Data );

				// Alloc some memory - with space for the 0
				FBlobData.Data := AllocMem( FBlobData.Size+1 );

				// Copy the data to the buffer.
				StrLCopy( FBlobData.Data, cpData, FBlobData.Size+1 );

				// Dealloc the SX memory
				if FBlobData.Size > 0 then
        begin
					FDataSet.cs_MemDealloc( cpData );
        end;
			end
			else
      begin
//				FBlobData.Size := FDataSet.cs_GetBlobLength( PChar( Field.FieldName ));  {LAS: 06-20-00, Moved up}
				FreeMem( FBlobData.Data );      // Make sure it's not already assigned
				if fBlobData.Size > 0 then
				begin
	        FBlobData.Data := AllocMem( FBlobData.Size+1 );  // Alloc some memory
					FDataSet.cs_GetBlob( PChar( Field.FieldName ), FBlobData.Data );
				end;
			end;

			FBlobData.Loaded := True;
			FDataSet.cs_Go( iOldRecno );
		end;
	end;

	// If the data was already loaded, then BlobData is going to be pointing
	// at it.  If it's NOT already loaded, it will be too.  Don't need to do
	// any more processing.
	if (Mode <> bmWrite) then
	begin
		Size := FBlobData.Size;
		Seek( 0, soFromBeginning );
		WriteBuffer( FBlobData.Data^, FBlobData.Size );
		Seek( 0, soFromBeginning );
	end;
end;

{==============================================================================
  Write
 ==============================================================================}
function TApBlobStream.Write( const Buffer; Count: Integer): Integer;
begin
  Result := inherited Write( Buffer, Count );
end;

{==============================================================================
  Destroy
 ==============================================================================}
destructor TApBlobStream.Destroy;
var
  BlobData: PBlobData;
begin

  if ((FMode = bmWrite) or (FMode = bmReadWrite)) then
  begin
    // Set the contents of the stream into the field.
    Seek( 0, soFromBeginning );

    BlobData := PBlobData( FDataSet.ActiveBuffer +
                         ( FDataSet as TApolloTable).FBlobCacheOfs +
                         ( FField.Offset * SizeOf( TApBlobData )));

    FreeMem(BlobData.Data);
    BlobData.Size := Size;
    // Allocate an extra byte - which we FORCE into a null terminator.  On a
    // memo, the cs_ function looks for the null terminator.  For a blob, we
    // specify the size - which doesn't include the null terminator we added.
    BlobData.Data := AllocMem( Size + 1 );  // Extra byte for Null Terminator
    ReadBuffer( BlobData.Data^, Size );
    BlobData.Data[Size] := #0;              // Append the null terminator
    BlobData.Modified := True;

    (FDataSet as TApolloTable).DataEvent(deFieldChange, Integer(FField));
  end;
  inherited Destroy;
end;

//---------------------------------------------------------------------------//
// Functions to map to both SDE and Server methods
//---------------------------------------------------------------------------//
function TApolloTable.cs_Alias(uiWorkArea: Word): String;
begin
	if FAccessMethod = amServer then
  begin
  	Result := String( FApolloConnection.ap_Alias(uiWorkArea))
  end
  else
  	Result := String(sx_Alias(uiWorkArea));
end;

procedure TApolloTable.cs_Append;
begin
	if FAccessMethod = amServer then
  	FApolloConnection.ap_Append
  else
		sx_Append;

  FRecNo := cs_RecNo;
  FRowCount := cs_RecCount;
end;

procedure TApolloTable.cs_AppendBlank;
begin
	if FAccessMethod = amServer then
  	FApolloConnection.ap_AppendBlank
  else
		sx_AppendBlank;

  FRecNo := cs_RecNo;
  FRowCount := cs_RecCount;
end;

function TApolloTable.cs_AppendFrom( cpFileName: PChar;
                          iSourceType: SmallInt; cpScopeExpr: PChar): LongBool;
begin
	if FAccessMethod = amServer then		Result := FApolloConnection.ap_AppendFrom( cpFileName, iSourceType, cpScopeExpr )
  else                            	  Result := sx_AppendFrom( cpFileName, iSourceType, cpScopeExpr );

  // 03-13-02 MED, added
  cs_Commit;  //MED:03-25-02, moved here // REQUIRED!! Without this, last record is not flushed
  FRecNo := cs_RecNo;
  FRowCount := cs_RecCount;
end;

function TApolloTable.cs_BaseDate: PChar;
begin
	if FAccessMethod = amServer then
  	Result := FApolloConnection.ap_BaseDate
  else
		Result := sx_BaseDate;
end;

function TApolloTable.cs_BaseName: PChar;
begin
	if FAccessMethod = amServer then
  	Result := FApolloConnection.ap_BaseName
  else
		Result := sx_BaseName;
end;

function TApolloTable.cs_BlobToFile( cpFieldName: PChar;
                                             cpFileName: PChar): LongBool;
begin
	if FAccessMethod = amServer then
	  Result := FApolloConnection.ap_BlobToFile( cpFieldName, cpFileName, FRecNo )  {LAS: 06-26-00, Added FRecNo param}
  else
	  Result := sx_BlobToFile( cpFieldName, cpFileName );
end;

function TApolloTable.cs_Bof: WordBool;
begin
  if FCachedUpdates then
  begin
    Result := (FRecNo <= 0 )
  end
  else
	if FAccessMethod = amServer then
  	Result := FApolloConnection.ap_Bof
  else
		Result := sx_Bof;
end;

procedure TApolloTable.cs_Close;
begin
	if FAccessMethod = amServer then
  	FApolloConnection.ap_Close
  else
		sx_Close;
end;

procedure TApolloTable.cs_CloseAll;
begin
	if FAccessMethod = amServer then
  	FApolloConnection.ap_CloseAll
  else
		sx_CloseAll;
end;

procedure TApolloTable.cs_CloseIndexes;
begin
	if FAccessMethod = amServer then  	FApolloConnection.ap_CloseIndexes
  else                            		sx_CloseIndexes;
end;

procedure TApolloTable.cs_Commit;
begin
	if FAccessMethod = amServer then
  	FApolloConnection.ap_Commit
  else
		sx_Commit;
end;

function TApolloTable.cs_CopyFile( cpToFileName: PChar): LongBool;  {LAS: 06-23-00, Changed return type from WordBool}
begin
	if FAccessMethod = amServer then
  	Result := FApolloConnection.ap_CopyFile(cpToFileName)
  else
		Result := sx_CopyFile(cpToFileName);
end;

function TApolloTable.cs_CopyFileText( cpTextFileName: PChar;
                                                iFileType: SmallInt): LongBool;  {LAS: 06-23-00, Changed return type from WordBool}
begin
	if FAccessMethod = amServer then
	  Result := FApolloConnection.ap_CopyFileText(cpTextFileName, iFileType)
  else
	  Result := sx_CopyFileText(cpTextFileName, iFileType);
end;

function TApolloTable.cs_CopyStructure( cpFileName: PChar;
                                                cpAlias: PChar): LongBool;  {LAS: 06-23-00, Changed return type from WordBool}
begin
	if FAccessMethod = amServer then
	  Result := FApolloConnection.ap_CopyStructure(cpFileName, cpAlias)
  else
	  Result := sx_CopyStructure(cpFileName, cpAlias);
end;

function TApolloTable.cs_CopyStructureExtended(cpFileName: PChar): LongBool;  {LAS: 06-23-00, Changed return type from WordBool}
begin
	if FAccessMethod = amServer then
	  Result := FApolloConnection.ap_CopyStructureExtended(cpFileName)
  else
	  Result := sx_CopyStructureExtended(cpFileName);
end;

function TApolloTable.cs_Count: Integer;
begin
 	if FAccessMethod = amServer then
	  Result := FApolloConnection.ap_Count
  else
	  Result := sx_Count;
end;

function TApolloTable.cs_CreateExec: LongBool;  {LAS: 06-23-00, Changed return type from WordBool}
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_CreateExec
  else                            	  Result := sx_CreateExec;
end;

procedure TApolloTable.cs_CreateField( cpName: PChar; cpType: PChar;
                                       iLength: SmallInt; iDecimals: SmallInt);
begin
	if FAccessMethod = amServer then  FApolloConnection.ap_CreateField( cpName, cpType, iLength, iDecimals )
  else                              sx_CreateField( cpName, cpType, iLength, iDecimals );
end;

function TApolloTable.cs_CreateFrom( cpFileName: PChar; cpAlias: PChar;
                              cpStruFile: PChar; iRDEType: SmallInt): LongBool;  {LAS: 06-23-00, Changed return type from WordBool}
begin
	if FAccessMethod = amServer then		Result := FApolloConnection.ap_CreateFrom( cpFileName, cpAlias, cpStruFile, iRDEType )
  else                            	  Result := sx_CreateFrom(cpFileName, cpAlias, cpStruFile, iRDEType);
end;

function TApolloTable.cs_CreateNew( cpFileName: PChar; cpAlias: PChar;
                            iRdeType: SmallInt; iNumFields: SmallInt): SmallInt;
begin
	if FAccessMethod = amServer then
	  Result := FApolloConnection.ap_CreateNew( pchar(FDataBaseName),
              Pchar(ExtractFileName(cpFileName)), iRDEType, iNumFields )
  else
	  Result := sx_CreateNew( cpFileName, cpAlias, iRDEType, iNumFields );
end;

function TApolloTable.cs_DbfDecrypt: WordBool;
begin
	if FAccessMethod = amServer then
	  Result := FApolloConnection.ap_DbfDecrypt
  else
	  Result := sx_DbfDecrypt;
end;

function TApolloTable.cs_DbfEncrypt: WordBool;
begin
	if FAccessMethod = amServer then
	  Result := FApolloConnection.ap_DbfEncrypt
  else
	  Result := sx_DbfEncrypt;
end;

function TApolloTable.cs_DBFilter: PChar;
begin
	if FAccessMethod = amServer then
	  Result := FApolloConnection.ap_DBFilter
  else
	  Result := sx_DBFilter;
end;

{$ifdef NOT_D3_OR_CB3}

procedure TApolloTable.cs_DBRlockList( ulpArray: PChar );
begin
	if FAccessMethod = amServer then
	  FApolloConnection.ap_DBRLockList( ulpArray )
  else
	  sx_DBRLockList( ulpArray );
end;
{$endif}

function TApolloTable.cs_Decrypt( cpBuffer: PChar; cpPassword: PChar;
                                          iLen: Integer): PChar;
begin
	if FAccessMethod = amServer then
	  Result := FApolloConnection.ap_Decrypt( cpBuffer, cpPassword, iLen )
  else
	  Result := sx_Decrypt( cpBuffer, cpPassword, iLen );
end;

procedure TApolloTable.cs_Delete;
begin
	if FAccessMethod = amServer then
  	FApolloConnection.ap_Delete
  else
	  sx_Delete;
end;

function TApolloTable.cs_Deleted: WordBool;
begin
	if (FAccessMethod = amServer) then
  begin
		FApolloConnection.ap_Go( FRecNo );      {LAS: 07-20-00, Added}
    Result := FApolloConnection.ap_Deleted;
  end
  else
	  Result := sx_Deleted;
end;

function TApolloTable.cs_DeleteFile( sFileName : String ): WordBool;
begin
	if (FAccessMethod = amServer) then    Result := FApolloConnection.DeleteFile(FDataBaseName, sFileName )
	else                              		Result := SysUtils.DeleteFile( sFileName );
end;

function TApolloTable.cs_Descend( cpKeyString: PChar ): PChar;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_Descend( cpKeyString )
  else                            	  Result := sx_Descend( cpKeyString );
end;

function TApolloTable.cs_Empty(cpFieldName: PChar): WordBool;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_Empty(cpFieldName)
  else                            	  Result := sx_Empty(cpFieldName);
end;

function TApolloTable.cs_Eof: WordBool;
begin
	if FAccessMethod = amServer then  Result := FApolloConnection.ap_Eof
  else                              Result := sx_Eof;
end;

function TApolloTable.cs_Encrypt( cpBuffer: PChar; cpPassword: PChar;
                                          iLen: Integer): PChar;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_Encrypt( cpBuffer, cpPassword, iLen )
  else                            	  Result := sx_Encrypt( cpBuffer, cpPassword, iLen );
end;

function TApolloTable.cs_EvalLogical(cpExpression: PChar): WordBool;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_EvalLogical( cpExpression )
  else                            	  Result := sx_EvalLogical( cpExpression );
end;

function TApolloTable.cs_EvalNumeric(cpExpression: String): Double;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_EvalNumeric( PChar( cpExpression ))
  else                            		Result := sx_EvalNumeric( PChar( cpExpression ));
end;

function TApolloTable.cs_EvalString( cpExpression: PChar ): PChar;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_EvalString( cpExpression )
  else                            	  Result := sx_EvalString( cpExpression );
end;

function TApolloTable.cs_EvalTest(cpExpression: PChar): SmallInt;
begin
	if FAccessMethod = amServer then  	Result := FApolloConnection.ap_EvalTest( cpExpression )
  else                            	  Result := sx_EvalTest( cpExpression );
end;

function TApolloTable.cs_FieldCount: Word;
begin
	if FAccessMethod = amServer then  	Result := FApolloConnection.ap_FieldCount
  else                            	  Result := sx_FieldCount;
end;

function TApolloTable.cs_FieldDecimals( cpFieldName: PChar ): Word;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_FieldDecimals( cpFieldName )
  else                            	  Result := sx_FieldDecimals( cpFieldName );
end;

function TApolloTable.cs_FieldName( uiFieldNum: Word ): PChar;
begin
	if FAccessMethod = amServer then  	Result := FApolloConnection.ap_FieldName( uiFieldNum )
  else                            	  Result := sx_FieldName( uiFieldNum );
end;

function TApolloTable.cs_FieldNum( cpFieldName: PChar ): Word;
begin
	if FAccessMethod = amServer then  	Result := FApolloConnection.ap_FieldNum( cpFieldName )
  else                            	  Result := sx_FieldNum( cpFieldName );
end;

function TApolloTable.cs_FieldOffset( cpFieldName: PChar ): Word;
begin
	if FAccessMethod = amServer then  	Result := FApolloConnection.ap_FieldOffset( cpFieldName )
  else                            	  Result := sx_FieldOffset( cpFieldName );
end;

function TApolloTable.cs_FieldType( cpFieldName: PChar ): PChar;
begin
	if FAccessMethod = amServer then  	Result := FApolloConnection.ap_FieldType( cpFieldName )
  else                            		Result := sx_FieldType( cpFieldName );
end;

function TApolloTable.cs_FieldWidth( cpFieldName: PChar ): Word;
begin
	if FAccessMethod = amServer then  	Result := FApolloConnection.ap_FieldWidth( cpFieldName )
  else                            	  Result := sx_FieldWidth( cpFieldName );
end;

function TApolloTable.cs_FileExists( sFileName : String ): WordBool;
begin
	if FAccessMethod = amServer then    Result := FApolloConnection.FileExists( FDataBaseName, sFileName )
	else                            		Result := FileExists( sFileName );
end;

function TApolloTable.cs_FilterAlias( cpAliasName: PChar;
                                              cpFieldName: PChar ): WordBool;
begin
	if FAccessMethod = amServer then
  	Result := FApolloConnection.ap_FilterAlias( cpAliasName, cpFieldName )
  else
	  Result := sx_FilterAlias( cpAliasName, cpFieldName );
end;

function TApolloTable.cs_FilterDlg( hwnd: HWND; cpExpr: PChar;
                            cpCaption: PChar; bHasIndexList: WordBool ): PChar;
begin
	if FAccessMethod = amServer then  	Result := FApolloConnection.ap_FilterDlg( hwnd, cpExpr, cpCaption, bHasIndexList )
  else                            	  Result := sx_FilterDlg( hwnd, cpExpr, cpCaption, bHasIndexList );
end;

function TApolloTable.cs_Flock: WordBool;
begin
	if FAccessMethod = amServer then  	Result := FApolloConnection.ap_Flock
  else                            		Result := sx_Flock;
end;

function TApolloTable.cs_Found: WordBool;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_Found
  else                            		Result := sx_Found;
end;

function TApolloTable.cs_GetBlob( cpFieldName: PChar;  vpVar: Pointer): Integer;
begin
	if FAccessMethod = amServer then  	Result := FApolloConnection.ap_GetBlob( cpFieldName, vpVar )
  else                            		Result := sx_GetBlob( cpFieldName, vpVar );
end;

function TApolloTable.cs_GetBlobLength( cpFieldName: PChar ): Integer;
begin
	if FAccessMethod = amServer then  	Result := FApolloConnection.ap_GetBlobLength( cpFieldName )
  else                            		Result := sx_GetBlobLength( cpFieldName );
end;

function TApolloTable.cs_GetByte( cpFieldName: PChar ): PChar;
begin
	if FAccessMethod = amServer then  	Result := FApolloConnection.ap_GetByte( cpFieldName)
  else                            		Result := sx_GetByte( cpFieldName );
end;

function TApolloTable.cs_GetDateJulian( cpFieldName: PChar ): Integer;
begin
	if FAccessMethod = amServer then  	Result := FApolloConnection.ap_GetDateJulian( cpFieldName )
  else                            		Result := sx_GetDateJulian( cpFieldName );
end;

function TApolloTable.cs_GetDateString( cpFieldName: PChar ): PChar;
begin
	if FAccessMethod = amServer then  	Result := FApolloConnection.ap_GetDateString( cpFieldName )
  else                            		Result := sx_GetDateString( cpFieldName );
end;

function TApolloTable.cs_GetDouble( cpFieldName: PChar ): Double;
begin
	if FAccessMethod = amServer then  	Result := FApolloConnection.ap_GetDouble( cpFieldName )
  else                            	  Result := sx_GetDouble( cpFieldName );
end;

function TApolloTable.cs_GetInteger( cpFieldName: PChar ): SmallInt;
begin
	if FAccessMethod = amServer then  	Result := FApolloConnection.ap_GetInteger( cpFieldName )
  else                             		Result := sx_GetInteger( cpFieldName );
end;

function TApolloTable.cs_GetLogical( cpFieldName: PChar ): WordBool;
begin
	if FAccessMethod = amServer then  	Result := FApolloConnection.ap_GetLogical( cpFieldName )
  else                            	  Result := sx_GetLogical( cpFieldName );
end;

function TApolloTable.cs_GetLong( cpFieldName: PChar ): Integer;
begin
	if FAccessMethod = amServer then  	Result := FApolloConnection.ap_GetLong( cpFieldName )
  else                            		Result := sx_GetLong( cpFieldName );
end;

function TApolloTable.cs_GetMemo( cpFieldName: PChar;
                                          uiLineWidth: Word): PChar;
begin
	if FAccessMethod = amServer then
  begin
    FApolloConnection.ap_Go(FRecNO);
	  Result := FApolloConnection.ap_GetMemo( cpFieldName, uiLineWidth )  {LAS: 07-16-00, Added FRecNo param}
  end
  else
		Result := sx_GetMemo( cpFieldName, uiLineWidth );
end;


function TApolloTable.cs_GetQueryBit(lRecNo: Integer): WordBool;
begin
	if FAccessMethod = amServer then
  	Result := FApolloConnection.ap_GetQueryBit( lRecNo )
  else
		Result := sx_GetQueryBit( lRecNo );
end;

procedure TApolloTable.cs_GetRecord( cpRecord: PChar );
begin
  if FAccessMethod = amServer then  	FApolloConnection.ap_GetRecord( cpRecord, FRecSize )
  else                              	sx_GetRecord( cpRecord );
end;

function TApolloTable.cs_GetScope( iWhichScope: SmallInt ): PChar;
begin
	if FAccessMethod = amServer then  	Result := FApolloConnection.ap_GetScope( iWhichScope )
  else                            		Result := sx_GetScope( iWhichScope );
end;

function TApolloTable.cs_GetString( cpFieldName: PChar ): PChar;
begin
	if FAccessMethod = amServer then  	Result := FApolloConnection.ap_GetString( cpFieldName )
  else                            		Result := sx_GetString( cpFieldName );
end;

function TApolloTable.cs_GetTrimString(cpFieldName: PChar): PChar;
begin
	if FAccessMethod = amServer then  	Result := FApolloConnection.ap_GetTrimString( cpFieldName )  {LAS: 06-09-00, Added FRecNo param} {MED:08-29-00, changed back}
  else                            		Result := sx_GetTrimString( cpFieldName );
end;

{LAS: 06-01-00, Changed to function, and rewritten}
procedure TApolloTable.cs_Go(lRecNum: Integer);
begin
  if FAccessMethod = amServer then
    FRecOrdinal := FApolloConnection.ap_Go(lRecNum)
  else
  begin
    sx_Go(lRecNum);
    if IndexName = '' then
      FRecOrdinal := lRecNum
    else
      FRecOrdinal := FRecOrdinalNeedsRefresh;
  end;
  FRecNo := lRecNum;
end;


procedure TApolloTable.cs_GoBottom;
begin
	if FAccessMethod = amServer then  	FApolloConnection.ap_GoBottom
  else                            		sx_GoBottom;
end;

procedure TApolloTable.cs_GoTop;
begin
	if FAccessMethod = amServer then  	FApolloConnection.ap_GoTop
  else                              	sx_GoTop;

	FRecNo := 1;  {LAS: 06-09-00, Added}
end;

function TApolloTable.cs_Index( cpFileName: PChar; cpExpr: PChar;
           iOption: SmallInt; bDescend: Boolean; cpCondition: PChar): SmallInt;
begin
	if FAccessMethod = amServer then  	Result := FApolloConnection.ap_Index( cpFileName, cpExpr, iOption, bDescend, cpCondition )
  else                            	  Result := sx_Index( cpFileName, cpExpr, iOption, bDescend, cpCondition );

  FIndexOrd := Result;
end;

procedure TApolloTable.cs_IndexClose;
begin
	if FAccessMethod = amServer then  	FApolloConnection.ap_IndexClose
  else                              	sx_IndexClose;

  FIndexOrd := 0;
end;

function TApolloTable.cs_IndexCondition: PChar;
begin
	if FAccessMethod = amServer then  	Result := FApolloConnection.ap_IndexCondition
  else                            		Result := sx_IndexCondition;
  FIndexCondition := Result;
end;

function TApolloTable.cs_IndexFlip: WordBool;
begin
  if FAccessMethod = amServer then	  Result := FApolloConnection.ap_IndexFlip
  else                            	Result := sx_IndexFlip;
end;

function TApolloTable.cs_IndexKey: PChar;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_IndexKey
  else                            		Result := sx_IndexKey;
end;

function TApolloTable.cs_IndexKeyField: PChar;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_IndexKeyField
  else                            	  Result := sx_IndexKeyField;
end;

function TApolloTable.cs_IndexName( iIndex: SmallInt ): PChar;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_IndexName( iIndex )
  else                            	  Result := sx_IndexName( iIndex );
end;

function TApolloTable.cs_IndexOpen(cpFileName: PChar): SmallInt;
begin
	if FAccessMethod = amServer then  	Result := FApolloConnection.ap_IndexOpen( cpFileName )
  else                            	  Result := sx_IndexOpen( cpFileName );
  FIndexOrd := Result;
end;

function TApolloTable.cs_IndexOrd: SmallInt;
begin
  if FAccessMethod = amServer then	  Result := FApolloConnection.ap_IndexOrd
  else                            	  Result := sx_IndexOrd;

  FIndexOrd := Result;
end;

function TApolloTable.cs_IndexTag( cpFileName: PChar; cpTagName: PChar;
                          cpExpr: PChar; iOption: SmallInt; bDescend: Boolean;
                          cpCondition: PChar): SmallInt;
begin
	if FAccessMethod = amServer then
	  Result := FApolloConnection.ap_IndexTag( cpFileName, cpTagName, cpExpr,
                           iOption, bDescend, cpCondition )
  else
	  Result := sx_IndexTag( cpFileName, cpTagName, cpExpr,
                           iOption, bDescend, cpCondition );
  FIndexOrd := Result;
end;

function TApolloTable.cs_IndexType: SmallInt;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_IndexType
  else                            	  Result := sx_IndexType;
end;

function TApolloTable.cs_IsEncrypted(iFileOrRec: SmallInt): WordBool;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_IsEncrypted( iFileOrRec )
  else                            	  Result := sx_IsEncrypted( iFileOrRec );
end;

function TApolloTable.cs_KeyAdd( cpTagname: PChar ): WordBool;
begin
	if FAccessMethod = amServer then  	Result := FApolloConnection.ap_KeyAdd( cpTagname )
  else                            	  Result := sx_KeyAdd( cpTagname );
end;

function TApolloTable.cs_KeyData: PChar;
begin
	if FAccessMethod = amServer then		Result := FApolloConnection.ap_KeyData
  else                            	  Result := sx_KeyData;
end;

function TApolloTable.cs_KeyDrop(cpTagname: PChar): WordBool;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_KeyDrop( cpTagname )
  else                            	  Result := sx_KeyDrop( cpTagname );
end;

function TApolloTable.cs_Locate( cpExpression: PChar;
                          iDirection: WordBool; bContinue: WordBool ): Integer;
begin
	if FAccessMethod = amServer then  	Result := FApolloConnection.ap_Locate( cpExpression, iDirection, bContinue )
  else                            	  Result := sx_Locate( cpExpression, iDirection, bContinue );
end;

function TApolloTable.cs_LockCount: Word;
begin
	if FAccessMethod = amServer then		Result := FApolloConnection.ap_LockCount
  else                            		Result := sx_LockCount;
end;

function TApolloTable.cs_Locked(lRecNum: Integer): WordBool;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_Locked(lRecNum)
  else                            	  Result := sx_Locked(lRecNum);
end;

procedure TApolloTable.cs_MemDealloc( vpPtr: Pointer );
begin
	if FAccessMethod = amServer then	  FApolloConnection.ap_MemDealloc(vpPtr)
  else                            	  sx_MemDealloc(vpPtr);
end;

function TApolloTable.cs_OpenMode: SmallInt;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_OpenMode
  else                            	  Result := sx_OpenMode;
end;

function TApolloTable.cs_OrderPosGet: Double;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_OrderPosGet
  else                            	  Result := sx_OrderPosGet;
end;

procedure TApolloTable.cs_OrderPosSet( dPosition: Double );
begin
	if FAccessMethod = amServer then	  FApolloConnection.ap_OrderPosSet( dPosition )
  else                            	  sx_OrderPosSet( dPosition );
end;

function TApolloTable.cs_OrderRecNo: Integer;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_OrderRecNo
  else                            	  Result := sx_OrderRecNo;
end;

procedure TApolloTable.cs_Pack;
begin
	if FAccessMethod = amServer then	  FApolloConnection.ap_Pack
  else                            	  sx_Pack;
end;

function TApolloTable.cs_PutBlob( cpFieldName: PChar; vpVar: Pointer;
                                          iSize: Integer): Integer;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_PutBlob( cpFieldName, vpVar, iSize )
  else                            	  Result := sx_PutBlob( cpFieldName, vpVar, iSize );
end;

procedure TApolloTable.cs_PutRecord( cpRecord: PChar );
begin
	if FAccessMethod = amServer then	  FApolloConnection.ap_PutRecord( cpRecord, FRecSize)
	else                            	  sx_PutRecord( cpRecord );
end;

// have SDE sx_Query return FCount instead
function TApolloTable.cs_Query(cpExpression: PChar): Integer;
begin
	if FAccessMethod = amServer then
  begin
	  Result := FApolloConnection.ap_QueryRowCount( cpExpression, FRowCount);
    { MED: 03-05-2002 added to speed it up
      if FApolloConnection.ap_QueryTest( cpExpression ) > 0 then
	    FRowCount := FApolloConnection.ap_QueryRecCount
    else
	    FRowCount := FApolloConnection.ap_RecCount;}
  end
  else
  begin
	  Result := sx_Query( cpExpression );
    if sx_QueryTest( cpExpression ) > 0 then
	    FRowCount := sx_QueryRecCount
    else
	    FRowCount := sx_RecCount;
  end;
end;

function TApolloTable.cs_QueryRecCount: Integer;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_QueryRecCount
  else                            	  Result := sx_QueryRecCount;
end;

function TApolloTable.cs_QueryTest(cpExpression: PChar): SmallInt;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_QueryTest( cpExpression )
  else                            	  Result := sx_QueryTest( cpExpression );
end;

procedure TApolloTable.cs_Recall;
begin
	if FAccessMethod = amServer then	  FApolloConnection.ap_Recall
  else                            	  sx_Recall;
end;

function TApolloTable.cs_RecCount: Integer;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_RecCount
  else                            	  Result := sx_RecCount;
end;

function TApolloTable.cs_RecNo: Integer;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_RecNo
  else                            	  Result := sx_RecNo;

  FRecNo := Result;
end;

function TApolloTable.cs_RecSize: Integer;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_RecSize
  else                            	  Result := sx_RecSize;

end;

procedure TApolloTable.cs_Reindex;
begin
	if FAccessMethod = amServer then	  FApolloConnection.ap_Reindex
  else                            	  sx_Reindex;
end;

procedure TApolloTable.cs_Replace( cpFieldname: PChar;
                                         iDataType: SmallInt; vpData: Pointer );
begin
	if FAccessMethod = amServer then	  FApolloConnection.ap_Replace( cpFieldName, iDataType, vpData )
  else                            	  sx_Replace( cpFieldName, iDataType, vpData );
end;

function TApolloTable.cs_Rlock( lRecNum: Integer ): WordBool;
begin
	if FAccessMethod = amServer then  	Result := FApolloConnection.ap_Rlock( lRecNum )
  else                              	Result := sx_Rlock( lRecNum );

{$ifdef NOT_D3_OR_CB3}
	FUpdateLockList := True;
{$endif}
end;

function TApolloTable.cs_RYOFilterActivate( iFilterHandle: SmallInt;
                                           iBoolOperation: SmallInt ): WordBool;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_RYOFilterActivate( iFilterHandle, iBoolOperation )
  else                            	  Result := sx_RYOFilterActivate( iFilterHandle, iBoolOperation );
end;

function TApolloTable.cs_RYOFilterCopy: SmallInt;
begin
	if FAccessMethod = amServer then 	Result := FApolloConnection.ap_RYOFilterCopy
  else                              Result := sx_RYOFilterCopy;
end;

function TApolloTable.cs_RYOFilterCreate: SmallInt;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_RYOFilterCreate
  else                            	  Result := sx_RYOFilterCreate;
end;

function TApolloTable.cs_RYOFilterDestroy( iFilterHandle: SmallInt ): WordBool;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_RYOFilterDestroy( iFilterHandle )
  else                            		Result := sx_RYOFilterDestroy( iFilterHandle );
end;

function TApolloTable.cs_RYOFilterGetBit( iFilterHandle: SmallInt;
                                                  lRecNo: Integer ): WordBool;
begin
	if FAccessMethod = amServer then
	  Result := FApolloConnection.ap_RYOFilterGetBit( iFilterHandle, lRecNo )
  else
	  Result := sx_RYOFilterGetBit( iFilterHandle, lRecNo );
end;

function TApolloTable.cs_RYOFilterRestore( cpFileName: PChar): WordBool;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_RYOFilterRestore( cpFileName )
  else                            	  Result := sx_RYOFilterRestore( cpFileName );
end;

function TApolloTable.cs_RYOFilterSave( iFilterHandle: SmallInt;
                                                cpFileName: PChar ): WordBool;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_RYOFilterSave( iFilterHandle, cpFileName )
  else                            	  Result := sx_RYOFilterSave( iFilterHandle, cpFileName );
end;

function TApolloTable.cs_RYOFilterSetBit( iFilterHandle: SmallInt;
                                lRecNo: Integer; iOnOrOff: SmallInt ): WordBool;
begin
	if FAccessMethod = amServer then
  	Result := FApolloConnection.ap_RYOFilterSetBit( iFilterHandle, lRecNo, iOnorOff )
  else
	  Result := sx_RYOFilterSetBit( iFilterHandle, lRecNo, iOnorOff );
end;

function TApolloTable.cs_RYOKeyAdd( cpTagname: PChar;
                                            cpKey: PChar ): WordBool;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_RYOKeyAdd( cpTagname, cpKey )
  else                            	  Result := sx_RYOKeyAdd( cpTagname, cpKey );
end;

function TApolloTable.cs_RYOKeyDrop(cpTagname: PChar): WordBool;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_RYOKeyDrop( cpTagname )
  else                            	  Result := sx_RYOKeyDrop( cpTagname );
end;

function TApolloTable.cs_Seek( cpKeyValue: PChar ): WordBool;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_Seek( cpKeyValue )
  else                            	  Result := sx_Seek( cpKeyValue );
end;

function TApolloTable.cs_SeekBin( cpKeyValue: PChar;
                                          uiLength: Word): WordBool;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_SeekBin( cpKeyValue, uiLength )
  else                            	  Result := sx_SeekBin( cpKeyValue, uiLength );
end;

// does nothing when CachedUpdates is TRUE
function TApolloTable.cs_Select(uiBaseArea: Word): Word;
begin
	if FAccessMethod = amServer then	    Result := FApolloConnection.ap_Select(uiBaseArea)
  else                            	    Result := sx_Select(uiBaseArea);
end;

procedure TApolloTable.cs_SetFilter(cpExpression: PChar);
begin
	if FAccessMethod = amServer then	  FApolloConnection.ap_SetFilter(cpExpression)
  else                            	  sx_SetFilter(cpExpression);
end;

procedure TApolloTable.cs_SetGaugeFunc( pFunc: Integer );
begin
	if FAccessMethod = amServer then	  // Do nothing.  Not for use on remote tables
  else	  sx_SetGaugeFunc( pFunc );
end;

procedure TApolloTable.cs_SetGaugeHook( hwndGauge: HWND );
begin
	if FAccessMethod = amServer then	  FApolloConnection.ap_SetGaugeHook(hwndGauge)
  else                            	  sx_SetGaugeHook(hwndGauge);
end;

function TApolloTable.cs_SetOrder( iIndex: SmallInt ): SmallInt;
begin
	if FAccessMethod = amServer then
  	Result := FApolloConnection.ap_SetOrder(iIndex)
  else
	  Result := sx_SetOrder(iIndex);

  FIndexOrd := iIndex;
end;

procedure TApolloTable.cs_SetPassword( cpEncodeKey: PChar );
begin
	if FAccessMethod = amServer then	  FApolloConnection.ap_SetPassword(cpEncodeKey)
  else                            	  sx_SetPassword(cpEncodeKey);
end;

procedure TApolloTable.cs_SetQueryBit( lRecNo: Integer; bValue: WordBool );
begin
	if FAccessMethod = amServer then	  FApolloConnection.ap_SetQueryBit(lRecNo, bValue)
  else                            	  sx_SetQueryBit(lRecNo, bValue);
end;

procedure TApolloTable.cs_SetRelation( uiChildArea: Word;cpKeyExpr: PChar );
begin
	if FAccessMethod = amServer then	  FApolloConnection.ap_SetRelation(uiChildArea, cpKeyExpr)
  else                            	  sx_SetRelation(uiChildArea, cpKeyExpr);
end;

function TApolloTable.cs_SetScope( cpLowVal: PChar; cpHighVal: PChar): WordBool;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_SetScope(cpLowVal, cpHighVal)
  else                            	  Result := sx_SetScope(cpLowVal, cpHighVal);
end;

procedure TApolloTable.cs_SetStringType(uiStringType: Word);
begin
	if FAccessMethod = amServer then	  FApolloConnection.ap_SetStringType( uiStringType )
  else                            	  sx_SetStringType( uiStringType );
end;

procedure TApolloTable.cs_SetTranslate(uiOnOff: WordBool);
begin
	if FAccessMethod = amServer then	  FApolloConnection.ap_SetTranslate( uiOnOff )
  else                            	  sx_SetTranslate( uiOnOff );
end;

procedure TApolloTable.cs_SetTurboRead( uiOnOff: WordBool );
begin
	if FAccessMethod = amServer then  	FApolloConnection.ap_SetTurboRead( uiOnOff )
  else                            	  sx_SetTurboRead( uiOnOff );
end;

procedure TApolloTable.cs_Skip( lNumRecs: Integer );
begin
	if FAccessMethod = amServer then	  FApolloConnection.ap_Skip( lNumRecs )
  else                            	  sx_Skip( lNumRecs );
end;

function TApolloTable.cs_SysProp( uiSysItem: Word; vpData: Pointer ): Integer;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_SysProp(uiSysItem, vpData)
  else
	begin        {LAS: 06-13-00, Added}
  	LoadDLL;   {LAS: 06-13-00, Added}
	  Result := sx_SysProp(uiSysItem, vpData);
  end;         {LAS: 06-13-00, Added}
end;

function TApolloTable.cs_TagArea( cpTagName: PChar ): SmallInt;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_TagArea(cpTagName)
  else                            	  Result := sx_TagArea(cpTagName);
end;

function TApolloTable.cs_TagName(iTagArea: SmallInt): PChar;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_TagName(iTagArea)
  else                            	  Result := sx_TagName(iTagArea);
end;

procedure TApolloTable.cs_Unlock(lRecNum: Integer);
begin
	if FAccessMethod = amServer then	  FApolloConnection.ap_UnLock(lRecNum)
  else                            	  sx_Unlock(lRecNum);

{$ifdef NOT_D3_OR_CB3}
	FUpdateLockList := True;
{$endif}
end;

function TApolloTable.cs_Use( cpFileName: PChar; cpAlias: PChar;
                             iOpenMode: SmallInt; iRdeType: SmallInt): SmallInt;
begin
{LAS: 04-25-00, Added EVAL Support}
{$ifdef EVAL}
	if not NewCheckTimeLimit(REGKEY,
                           REGKEY_KILL_NAME,
                           REGKEY_CHECK_NAME,
                           'Apollo 6.1 (ApolloTable)', TRUE) then
  begin
  	Result := 0;
    Exit;
  end;
{$endif}

	if FAccessMethod = amServer then
	  Result := FApolloConnection.ap_Use( cpFileName, cpAlias, iOpenMode, iRdeType    )
  else
	  Result := sx_Use( cpFileName, cpAlias, iOpenMode, iRdeType );

end;

function TApolloTable.cs_UseEx( cpFileName: PChar; cpAlias: PChar;
                             iOpenMode: SmallInt; iRdeType: SmallInt; uiMode : Cardinal): SmallInt;
begin
{LAS: 04-25-00, Added EVAL Support}
{$ifdef EVAL}
	if not NewCheckTimeLimit(REGKEY,
                           REGKEY_KILL_NAME,
                           REGKEY_CHECK_NAME,
                           'Apollo 6.1 (ApolloTable)', TRUE) then
  begin
  	Result := 0;
    Exit;
  end;
{$endif}

	if FAccessMethod = amServer then
	  Result := FApolloConnection.ap_UseEx( cpFileName, cpAlias, iOpenMode, iRdeType, uiMode)
  else
	  Result := sx_UseEx( cpFileName, cpAlias, iOpenMode, iRdeType, uiMode );

end;


function TApolloTable.cs_Version: PChar;
begin
	if FAccessMethod = amServer then
	  Result := FApolloConnection.ap_Version
  else
	  Result := sx_Version;
end;

function TApolloTable.cs_WorkArea(cpAlias: PChar ): Word;
begin
	if FAccessMethod = amServer then	  Result := FApolloConnection.ap_WorkArea(cpAlias)
  else                            	  Result := sx_WorkArea(cpAlias);
end;

procedure TApolloTable.cs_Zap;
begin
	if FAccessMethod = amServer then
	  FApolloConnection.ap_Zap
  else
	  sx_Zap;
end;


///////////////////////////////////////////////////
//MED: 09-09-01, added
///////////////////////////////////////////////////
procedure TApolloTable.SetSystemCollation;
begin
	if FAccessMethod = amServer then	  FApolloConnection.ap_SetSystemCollation
  else                            	  SDEEngine.SetSystemCollation;
end;

procedure TApolloTable.SetMachineCollation;
begin
	if FAccessMethod = amServer then	  FApolloConnection.ap_SetMachineCollation
  else                            	  SDEEngine.SetMachineCollation;
end;

procedure TApolloTable.AddDudenCollation;
begin
	if FAccessMethod = amServer then	  FApolloConnection.ap_AddDudenCollation
  else                            	  SDEEngine.AddDudenCollation;
end;

procedure TApolloTable.AddEtecCollation;
begin
	if FAccessMethod = amServer then	  FApolloConnection.ap_AddEtecCollation
  else                            	  SDEEngine.AddEtecCollation;
end;


{LAS: 05-23-00, Added}
{==============================================================================
  AddIndex
 ==============================================================================}
procedure TApolloTable.AddIndex( const Name, Fields: string;
                                 Options: TIndexOptions);
var
	sExpr, sFld, sFields: String;
  iUnique, iPos: Integer;
  bDesc: Boolean;
begin
	if iWA > 0 then
  begin
    // Build expression
    iPos := Pos( ';', Fields );
    if iPos = 0 then
      sExpr := Fields
    else
    begin
      // If multple fields, must convert non-strings to strings
      sFields := Fields + ';';  // Seed end with trailing semi-colon
      while (iPos > 0) do
      begin
        // Grab next field from list
        sFld := Copy( sFields, 1, iPos-1 );

        // Remove this field from list
        sFields := Copy( sFields, iPos+1, Length( sFields )-iPos);
        iPos := Pos( ';', sFields );  // And get next field delimiter

        case Ord( FieldByName( sFld ).DataType ) of
          // Convert date to string
          Ord(ftDate):
            sFld := 'DTOS(' + sFld + ')';
          // Convert Numbers to string (except currency)
          Ord(ftInteger), Ord(ftFloat), Ord(ftSmallInt):
            sFld := 'STR(' + sFld + ',' +
                           IntToStr(cs_FieldWidth( PChar(sFld ))) + ',0)';
          // Convert Currency to string
          Ord(ftCurrency):
            sFld := 'STR(' + sFld + ',' +
                           IntToStr(cs_FieldWidth( PChar(sFld ))) + ',2)';
          // Convert Boolean to string
          Ord(ftBoolean):
            sFld := 'IIF(' + sFld + ',"1","0")';  // Double-check this!!!
        end;
        sExpr := sExpr + '+' + sFld;
      end;
      // Removed leading plus sign
      if sExpr[1] = '+' then
        sExpr := Copy( sExpr, 2, Length( sExpr )-1 );
    end;

    // Unique Index?
    if ixUnique in Options then
      iUnique := IDX_UNIQUE
    else
      iUnique := IDX_NONE;

    // Descending index?
    bDesc := (ixDescending in Options);

    // Case-insensitive index?
    if ixCaseInsensitive in Options then
      sExpr := 'Upper(' + sExpr + ')';

    if ixNonMaintained in Options then
      Index( Name, sExpr, iUnique, bDesc, '' )
    else
      IndexTag( #0, Name, sExpr, iUnique, bDesc, '' );
  end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{LAS: 05-23-00, Added}
{==============================================================================
  CreateTable
 ==============================================================================}
procedure TApolloTable.CreateTable;
var
	sDir: String;
	i, iCnt: Integer;
	sFieldList: TStringList;
  sFileName, sFieldName, sFieldType: String;
	iFieldSize, iFieldDec: Integer;
	cpName: Array[0..10] of Char;
  cpType: Array[0..1] of Char;
  dType: TFieldType;
begin
	// Must close active table first if open
	if (iWA > 0) then
  	Close;

  sFieldList := TStringList.Create;
  try
		if FAccessMethod = amLocal then
    begin
      GetDatapath;    {MED:08-25-00, added}
      sDir := FDataPath;
    end
    else
      sDir := '';

  	sFileName := ExtractFileName( FTableName );
    if ExtractFileExt( sFileName ) = '' then
	  	sFileName := sFileName + '.DBF';

    if Fields.Count > 0 then
    	iCnt := Fields.Count
    else if FieldDefs.Count > 0 then
    	iCnt := FieldDefs.Count
    else
    begin
			sFieldList.Free;
    	RaiseApollo( ERR_MISSING_FIELD_DEFINITIONS );
      exit;
    end;

		iWA := cs_CreateNew( PChar(sDir + sFileName),
                         PChar( GetUniqueAliasName( sFileName )),
                         SmallInt(FTableType), //+1,    {Add one since type is zero-based}{MED:06-27-01, made it same}
                         iCnt );
  	if (iWA < 0) then
    begin
			sFieldList.Free;
			RaiseApollo( ERR_UNABLE_TO_CREATE_TABLE + sDir + sFileName );
      exit;
    end;

		iFieldSize := 0;
		iFieldDec  := 0;

		for i := 0 to iCnt - 1 do
		begin
	    if Fields.Count > 0 then
      begin
				sFieldName := Fields[i].FieldName;
				dType      := Fields[i].DataType;
			end
      else
      begin
				sFieldName := FieldDefs[i].Name;
				dType      := FieldDefs[i].DataType;
      end;

      case dType of
        // list of possible types accepted in TxQuery parser
        ftString:
          begin
            sFieldType := 'C';
            if Fields.Count > 0 then
              iFieldSize := Fields[i].Size
            else
              iFieldSize := FieldDefs[i].Size;
            iFieldDec  := 0;
          end;
        ftInteger:
					begin
            sFieldType := 'N';
            iFieldSize := 11;
            iFieldDec  := 0;
					end;
				ftSmallInt, ftAutoInc:
          begin
            sFieldType := 'N';
            iFieldSize := 6;
            iFieldDec  := 0;
          end;
        ftBoolean:
          begin
            sFieldType := 'L';
            iFieldSize := 1;   // Non-adjustable size
            iFieldDec  := 0;
          end;
       ftDate, ftTime, ftDateTime:
          begin
            sFieldType := 'D';
            iFieldSize := 8;   // Non-adjustable size
            iFieldDec  := 0;
          end;
       ftCurrency:
					begin
						sFieldType:= 'N';
            iFieldSize := 20;
            iFieldDec  := 2;
          end;
       ftFloat:
          begin
            sFieldType:= 'N';
            iFieldSize := 20;
            iFieldDec  := 4;
          end;
        ftBlob, ftMemo:
          begin
            sFieldType := 'M';
            iFieldSize := 10;  // Non-adjustable size
            iFieldDec  := 0;
          end;
	      else                                                  
  	    begin
        	sFieldList.Free;
					RaiseApollo( ERR_INVALID_FIELD_TYPE + ' [' +
                                     IntToStr( Ord( dType )) + ']' );
        	Exit;
	      end;
      end;
			StrPCopy( cpName, sFieldName );
      StrPCopy( cpType, sFieldType );

			// Check that first character is a letter
			if not (cpName[0] in ['A'..'Z','a'..'z']) then
		  	RaiseApollo( ERR_INVALID_FIELD_NAME );

      cs_CreateField( cpName, cpType, iFieldSize, iFieldDec );
   	end;
    cs_CreateExec;

		if IndexDefs.Count > 0 then
    begin
			if iWA > 0 then
      begin
				FCursorOpen := True;
	      FSDEFieldCount := FieldDefs.Count;
        FFieldCount    := FSDEFieldCount;    // MED: 10-21-00, added-will replace FSDEFieldCount eventually
				InternalInitFieldDefs;
        CreateFields;
      	FAddingIndex := True;
		   	for i := 0 to IndexDefs.Count-1 do
					AddIndex( IndexDefs[i].Name, IndexDefs[i].Fields, IndexDefs[i].Options );

				FAddingIndex := False;
      end;
    end;

  finally
    cs_Close;
    iWA := 0;
    FCursorOpen := False;
		sFieldList.Free;
  end;
end;

{LAS: 05-23-00, Added}
{==============================================================================
  OpenIndexFile
 ==============================================================================}
procedure TApolloTable.OpenIndexFile(const IdxName: string);
begin
	IndexOpen( IdxName );
end;

{LAS: 05-23-00, Added}
{==============================================================================
  CloseIndexFile
 ==============================================================================}
procedure TApolloTable.CloseIndexFile(const IdxName: string);
var
	iOldOrd, iOrd: Integer;
begin
	if iWA > 0 then
  begin
    // If the file does not already exist, nothing to do
    if cs_FileExists( IdxName ) then
    begin
	    iOldOrd := 0;
  	  cs_Select( iWA );

      // If this index is open, close it first
      iOrd := cs_TagArea( PChar( ExtractFileName( IdxName )));
      if iOrd > 0 then
      begin
        // Only set order if not current
        if iOrd <> FIndexOrd then
          iOldOrd := cs_SetOrder( iOrd )
        else
        	FIndexName := '';  // Otherwise, clear out IndexName

        // Close it
        cs_IndexClose;
      end;

      // Clear this index file from the ExtraIndexes list, if it's there
      FExtraIndexes.Delete( FExtraIndexes.IndexOf( IdxName ));

			// Then, update the IndexDefs list
      IndexDefs.Updated := False;
      UpdateIndexDefs;

      // if we switched orders to close it, switch back to original order
      if iOldOrd > 0 then
        cs_SetOrder( iOrd )
      else
      begin
				FCountNeedsRefresh := True;
      	Resync([]);
      end;
		end;
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{LAS: 05-24-00, Added}
{==============================================================================
  FindNearest
 ==============================================================================}
procedure TApolloTable.FindNearest(const KeyValues: array of const);
begin
	if (iWA > 0) then
  begin
		cs_Select( iWA );
		if not FSoftSeek then                           // If SoftSeek was OFF
			cs_SysProp( SDE_SP_SETSOFTSEEK, Pointer(1));  // Turn ON SoftSeek

		FindKey( KeyValues );

		if not FSoftSeek then                           // If SoftSeek was OFF
			cs_SysProp( SDE_SP_SETSOFTSEEK, Pointer(0));  // Turn if back OFF
  end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{LAS: 05-23-00, Added}
{==============================================================================
  DeleteIndex
 ==============================================================================}
procedure TApolloTable.DeleteIndex( const IdxName: string);
var
	iOldOrd, iOrd: Integer;
begin
	if iWA > 0 then
  begin
		// Can't Delete an index if table not opened EXCLUSIVE
		if not Exclusive and (FErrorFunc = -1) then
  		RaiseApollo( ERR_MUST_OPEN_EXCLUSIVE );

    // If the file does not already exist, nothing to do
    if cs_FileExists( IdxName ) then
    begin
	    iOldOrd := 0;
  	  cs_Select( iWA );

      // If this index is open, close it first
      iOrd := cs_TagArea( PChar( ExtractFileName( IdxName )));
      if iOrd > 0 then
      begin
        // Only set order if not current
        if iOrd <> FIndexOrd then
          iOldOrd := cs_SetOrder( iOrd )
        else
        	FIndexName := '';  // Otherwise, clear out IndexName

        // Close it
        cs_IndexClose;
      end;

			// Delete the file
      cs_DeleteFile( IdxName );

      // Clear this index file from the ExtraIndexes list, if it's there
      FExtraIndexes.Delete( FExtraIndexes.IndexOf( IdxName ));

			// Then, update the IndexDefs list
      IndexDefs.Updated := False;
      UpdateIndexDefs;

      // if we switched orders to close it, switch back to original order
      if iOldOrd > 0 then
        cs_SetOrder( iOrd )
      else
      begin
				FCountNeedsRefresh := True;
      	Resync([]);
      end;
		end;
	end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{LAS: 05-23-00, Added}
{==============================================================================
  EditKey
 ==============================================================================}
procedure TApolloTable.EditKey;
begin
	if (iWA = 0) then
		RaiseApollo( ERR_MUST_OPEN_TABLE )
	else if (FIndexName = '') then
		RaiseApollo( ERR_NO_INDEX_ACTIVE );

	FOldRec := PRecInfo(ActiveBuffer + FRecInfoOfs).Bookmark;
	SetState( dsSetKey );
end;

{LAS: 05-23-00, Added}
{==============================================================================
  SetKey
 ==============================================================================}
procedure TApolloTable.SetKey;
begin
	if (iWA = 0) then
		RaiseApollo( ERR_MUST_OPEN_TABLE )
	else if (FIndexName = '') then
		RaiseApollo( ERR_NO_INDEX_ACTIVE );

	FOldRec := PRecInfo(ActiveBuffer + FRecInfoOfs).Bookmark;
	SetState( dsSetKey );
  InternalInitRecord(ActiveBuffer);
end;

{LAS: 05-25-00, Added}
{==============================================================================
  GotoCurrent
 ==============================================================================}
procedure TApolloTable.GotoCurrent( Table: TApolloTable );
begin
	Go( Table.RecNo );
end;

{LAS: 05-23-00, Added}
{==============================================================================
  GotoKey
 ==============================================================================}
function TApolloTable.GotoKey: Boolean;
begin
	if (iWA = 0) then
		RaiseApollo( ERR_MUST_OPEN_TABLE )
	else if (FIndexName = '') then
		RaiseApollo( ERR_NO_INDEX_ACTIVE );

	Result := FindKey( [FieldByName(IndexKeyField).AsString] );
  SetState( dsBrowse );
	if not Result then
    PRecInfo(ActiveBuffer + FRecInfoOfs)^.Bookmark := FOldRec;
  Resync([]);
end;

{LAS: 05-25-00, Added}
{==============================================================================
  GotoNearest
 ==============================================================================}
procedure TApolloTable.GotoNearest;
begin
	if (iWA > 0) then
  begin
		cs_Select( iWA );
		if not FSoftSeek then                           // If SoftSeek was OFF
			cs_SysProp( SDE_SP_SETSOFTSEEK, Pointer(1));  // Turn ON SoftSeek

		GotoKey;

		if not FSoftSeek then                           // If SoftSeek was OFF
			cs_SysProp( SDE_SP_SETSOFTSEEK, Pointer(0));  // Turn if back OFF
  end
  else
		RaiseApollo( ERR_MUST_OPEN_TABLE );
end;

{LAS: 05-23-00, Added}
{==============================================================================
  EditRangeStart
 ==============================================================================}
procedure TApolloTable.EditRangeStart;
begin
	SetRangeStart;
end;

{LAS: 05-23-00, Added}
{==============================================================================
  EditRangeEnd
 ==============================================================================}
procedure TApolloTable.EditRangeEnd;
begin
	SetRangeEnd;
end;

{LAS: 05-23-00, Added}
{==============================================================================
  SetRangeStart
 ==============================================================================}
procedure TApolloTable.SetRangeStart;
begin
	if (iWA = 0) then
		RaiseApollo( ERR_MUST_OPEN_TABLE )
	else if (FIndexName = '') then
		RaiseApollo( ERR_NO_INDEX_ACTIVE );

	FRangeSet := rsStart;
	FOldRec := PRecInfo(ActiveBuffer + FRecInfoOfs).Bookmark;
	SetState( dsSetKey );
end;

{LAS: 05-23-00, Added}
{==============================================================================
  SetRangeEnd
 ==============================================================================}
procedure TApolloTable.SetRangeEnd;
begin
	if (iWA = 0) then
		RaiseApollo( ERR_MUST_OPEN_TABLE )
	else if (FIndexName = '') then
		RaiseApollo( ERR_NO_INDEX_ACTIVE );

	FRangeSet := rsEnd;
	FOldRec := PRecInfo(ActiveBuffer + FRecInfoOfs).Bookmark;
	SetState( dsSetKey );
end;

{LAS: 05-23-00, Added}
{==============================================================================
  ApplyRange
 ==============================================================================}
procedure TApolloTable.ApplyRange;
begin
	if (iWA = 0) then
		RaiseApollo( ERR_MUST_OPEN_TABLE )
	else if (FIndexName = '') then
		RaiseApollo( ERR_NO_INDEX_ACTIVE );

	FRangeSet := rsNone;
  SetState( dsBrowse );
	SetRange( [FRangeStart], [FRangeEnd] );
end;


end.
